#include <stdio.h>
#include <stdlib.h>

#define MAX_HEIGHT 10000
#define MAX_FILENAME 513

void vsmwritebmp_ (char *file, int *flen, int *w, int *h, float *image)
{
  FILE *fp;
  int i, j, k, l, m;
  int filesize, width, height, cwidth;
  int val = 255;
  char filename[MAX_FILENAME];
  const int MaxC=255;
  const int OFFSET = 54;
  char val1, val2, val3, val4;
  char *imageUC;
  int *imageI;

  if(*h > MAX_HEIGHT){
    printf("OutputBMPImage: image height exceeds %d\n",MAX_HEIGHT);
    exit(1);
  }

  if(*flen > MAX_FILENAME-1){
    printf("OutputBMPImage: filename too long. less than %d characters\n",MAX_FILENAME);
    exit(1);
  }

  for(i=0;i< *flen;i++){
    filename[i] = file[i];
  }

  filename[*flen]   ='\0';

  width = *w;  height = *h;
  filesize = (width*3 + width%4)*height+OFFSET;
  imageUC = (char *)malloc(filesize);
  imageI = (int *)malloc((width*height*3)*sizeof(int));

  /* BMP Header*/
  /* ---------File Header --------- */
     imageUC[0] = 'B'; imageUC[1] = 'M'; 
     /* 4 bytes File size*/
     val1 = (char)(val & filesize); val2 = (char)(val & (filesize>>8));
     val3 = (char)(val & (filesize>>16));  val4 = (char)(val&(filesize>>24));

    imageUC[2] = val1;  imageUC[3] = val2;
    imageUC[4] = val3;  imageUC[5] = val4;

    for(i=6;i<=9;i++) imageUC[i] = (char)0; 

    /* offset */
     imageUC[10] = (char)54; imageUC[11] = (char)0;
     imageUC[12] = (char)0; imageUC[13] = (char)0;

     imageUC[14] = (char)40; imageUC[15] = (char)0;
     imageUC[16] = (char)0; imageUC[17] = (char)0;

    /* Width */
     val1 = (char)(val&width); val2 = (char)(val&(width >>8));
     val3 = (char)(val &(width >>16));  val4 = (char)(val&(width>>24));

    imageUC[18] = val1;  imageUC[19] = val2;
    imageUC[20] = val3;  imageUC[21] = val4;

     /* Height */
     val1 = (char)(val&height); val2 = (char)(val&(height>>8));
     val3 = (char)(val&(height>>16));  val4 = (char)(val&(height>>24));

     imageUC[22] = val1;  imageUC[23] = val2;
     imageUC[24] = val3;  imageUC[25] = val4;

     imageUC[26] = (char)1; imageUC[27] = (char)0; 
     imageUC[28] = (char)24; imageUC[29] = (char)0; 
     imageUC[30] = (char)0; imageUC[31] = (char)0; 
     imageUC[32] = (char)0; imageUC[33] = (char)0;

     imageUC[34] = (char)0; imageUC[35] = (char)0;
     imageUC[36] = (char)0; imageUC[37] = (char)0;

     for(i=38;i<=53;i++) imageUC[i] = (char)0;

    cwidth = 3*width + (width%4);
#pragma omp parallel for private(i, j, k, l, m)
    for(j=0;j<height;j++){
      for(i=0;i<width;i++){
         /* real::rgba -> unsigned char :: rgb*/
         m = i+j*width;
         k = m*5;
         l = m *3;
         imageI[l+0] = 256*image[k+0];
         imageI[l+1] = 256*image[k+1];
         imageI[l+2] = 256*image[k+2];
         if(imageI[l+0] > MaxC) imageI[l+0] = MaxC;
         if(imageI[l+1] > MaxC) imageI[l+1] = MaxC;
         if(imageI[l+2] > MaxC) imageI[l+2] = MaxC;
      }
    }

#pragma omp parallel for private(i, j, l, m)
    for(j=0;j<height;j++){
      for(i=0;i<width;i++){
         l = 3*i+(height- j-1)*cwidth+OFFSET;
         m = 3*(i+j*width);
         imageUC[l] = (char)imageI[m+2];/* not RGB but BGR*/
         imageUC[l+1] = (char)imageI[m+1];
         imageUC[l+2] = (char)imageI[m+0];
      }
    }

  if ((fp=fopen(filename, "wb")) == NULL){
    puts("File Open for png failed");
    exit(1);
  }
  fwrite(imageUC, 1, filesize, fp);
  fclose(fp);

  free(imageUC);
  free(imageI);
  return;
}

void vsmwriteppm_ (char *file, int *flen, int *width, int *height, char *limage)
{
  FILE *fp;
  int i, imagesize;
  char filename[MAX_FILENAME];

  if(*height > MAX_HEIGHT){
    printf("OutputBMPImage: image height exceeds %d\n",MAX_HEIGHT);
    exit(1);
  }

  if(*flen > MAX_FILENAME-1){
    printf("OutputBMPImage: filename too long. less than %d characters\n",MAX_FILENAME);
    exit(1);
  }

  for(i=0;i< *flen;i++){
    filename[i] = file[i];
  }

  filename[*flen]   ='\0';
  imagesize = 3*(*width)*(*height);
  if ((fp=fopen(filename, "wb"))==NULL){
    puts("File Open for png failed");
    exit(1);
  }
  fprintf(fp, "P6\n");
  fprintf(fp, "%d %d\n", *width, *height);
  fprintf(fp, "255\n");
  fwrite(limage, sizeof(unsigned char), imagesize, fp);
  fclose(fp);
  return;
}

