module vsmArrows
 use vsmConst
 use vsmFunction
 use vsmData
 use vsmColor
 use vsmCoordinates
 use vsmParam
 use vsmImage
 use vsmSolidDrawer
 use vsmIntpol

 implicit none
 private

 type vismo_ArrowComDat
   real(kind=vdp), dimension(3, 2) :: wd_vertx, ld_vertx
   real(kind=vdp) :: dc
 end type vismo_ArrowComDat

 type(vismo_ArrowComDat),save :: vsmArrowsComDat
 public vsmArrowsPrepare, vsmDrawArrowsIM

 contains

 subroutine vsmArrowsPrepare
!   vsmArrowsComDat%dc = vsmCoordData%MaxRad/vsmCoord%nx(1)
   vsmArrowsComDat%dc = vsmCoordData%MaxRad/100_vdp

   vsmArrowsComDat%wd_vertx(1:3, 1) = -vsmCoordData%MaxRad
   vsmArrowsComDat%wd_vertx(1:3, 2) = vsmCoordData%MaxRad

   vsmArrowsComDat%ld_vertx(1:3, 1) = -vsmCoordData%MaxRad
   vsmArrowsComDat%ld_vertx(1:3, 2) = vsmCoordData%MaxRad
 end subroutine vsmArrowsPrepare

 subroutine vsmDrawArrowsIM(vis_num, cam_num)
   integer, intent(in) :: vis_num, cam_num

     call vsmyyDrawArrowsIM(vis_num, cam_num)

 end subroutine vsmDrawArrowsIM

 subroutine vsmyyDrawArrowsIM(vis_num, cam_num)
   integer, intent(in) :: vis_num, cam_num

   integer :: i, j, x, y, xc, yc
   integer :: nvec
   integer, dimension(3) :: trans_bool
   real(kind=vdp), dimension(3) :: arpos
   real(kind=vdp) :: rad, rad15, leng, trans, dens, dc, sdc
   real(kind=vdp), dimension(3) :: temp_v, pos, spos
   real(kind=vdp), dimension(3, 8) :: trori, trlori
   real(kind=vdp), dimension(3) :: max_wd, min_wd, max_ld, min_ld
   real(kind=vdp), dimension(4) :: sc
   real(kind=vdp), dimension(3, 3) :: matrix, invmatrix
   real(kind=vdp), dimension(3, 8) :: wd_vertx, ld_vertx

   real(kind=vdp), dimension(3) :: dx, vec_val
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp), dimension(3, 8) :: grid_val
   integer, dimension(3) :: ci
   real(kind=vdp), dimension(3) :: pos1, pos2
   logical :: normalize
   integer :: tnuml1

   call vsmLogPrint("draw arrows")

   dc = vsmArrowsComDat%dc
   sdc = dc/15.0_vdp
   wd_vertx(1:3, 1) = vsmArrowsComDat%wd_vertx(1:3,1)
   wd_vertx(1:3, 8) = vsmArrowsComDat%wd_vertx(1:3,2)
!   print *,"wd",vsmArrowsComDat%wd_vertx(1:3,1)
!   print *,"wd",vsmArrowsComDat%wd_vertx(1:3,)

   wd_vertx(2:3, 2) = wd_vertx(2:3, 1)
   wd_vertx(1, 2) = wd_vertx(1, 8)

   wd_vertx(1, 3) = wd_vertx(1, 1)
   wd_vertx(2, 3) = wd_vertx(2, 8)
   wd_vertx(3, 3) = wd_vertx(3, 1)

   wd_vertx(1, 4) = wd_vertx(1, 8)
   wd_vertx(2, 4) = wd_vertx(2, 8)
   wd_vertx(3, 4) = wd_vertx(3, 1)

   wd_vertx(1, 5) = wd_vertx(1, 1)
   wd_vertx(2, 5) = wd_vertx(2, 1)
   wd_vertx(3, 5) = wd_vertx(3, 8)

   wd_vertx(1, 6) = wd_vertx(1, 8)
   wd_vertx(2, 6) = wd_vertx(2, 1)
   wd_vertx(3, 6) = wd_vertx(3, 8)

   wd_vertx(1, 7) = wd_vertx(1, 1)
   wd_vertx(2, 7) = wd_vertx(2, 8)
   wd_vertx(3, 7) = wd_vertx(3, 8)

   ld_vertx(1:3, 1) = vsmArrowsComDat%ld_vertx(1:3,1)
   ld_vertx(1:3, 8) = vsmArrowsComDat%ld_vertx(1:3,2)

   ld_vertx(2:3, 2) = ld_vertx(2:3, 1)
   ld_vertx(1, 2) = ld_vertx(1, 8)

   ld_vertx(1, 3) = ld_vertx(1, 1)
   ld_vertx(2, 3) = ld_vertx(2, 8)
   ld_vertx(3, 3) = ld_vertx(3, 1)

   ld_vertx(1, 4) = ld_vertx(1, 8)
   ld_vertx(2, 4) = ld_vertx(2, 8)
   ld_vertx(3, 4) = ld_vertx(3, 1)

   ld_vertx(1, 5) = ld_vertx(1, 1)
   ld_vertx(2, 5) = ld_vertx(2, 1)
   ld_vertx(3, 5) = ld_vertx(3, 8)

   ld_vertx(1, 6) = ld_vertx(1, 8)
   ld_vertx(2, 6) = ld_vertx(2, 1)
   ld_vertx(3, 6) = ld_vertx(3, 8)

   ld_vertx(1, 7) = ld_vertx(1, 1)
   ld_vertx(2, 7) = ld_vertx(2, 8)
   ld_vertx(3, 7) = ld_vertx(3, 8)

   !if(vsmDbg) then
   !  call vsmLogPrint("dc = ", dc)
   !  call vsmLogPrint("wd 1 = ", wd_vertx(1:3, 1))
   !  call vsmLogPrint("wd 2 = ", wd_vertx(1:3, 2))
   !  call vsmLogPrint("ld 1 = ", ld_vertx(1:3, 1))
   !  call vsmLogPrint("ld 2 = ", ld_vertx(1:3, 2))
   !end if

   trans_bool = 0
   do j=1, vsmVisParam(vis_num)%arrow_num
     sc(1:4) = vsmVisParam(vis_num)%arrowparam(j)%sc(1:4) 
     nvec      = vsmVisParam(vis_num)%arrowparam(j)%nvec
     rad       = vsmVisParam(vis_num)%arrowparam(j)%radius
     leng      = vsmVisParam(vis_num)%arrowparam(j)%length
     dens      = vsmVisParam(vis_num)%arrowparam(j)%density
     rad15 = rad*1.5_vdp

!     if(vsmDbg) then
!       call vsmLogPrint("sc = ", sc)
!       call vsmLogPrint("vec = ", nvec)
!       call vsmLogPrint("rad = ", rad)
!       call vsmLogPrint("leng = ", leng)
!       call vsmLogPrint("dens = ", dens)
!     end if

     call vsmPlaneTrans(sc, trans, trans_bool)

     !call vsmLogPrint("trans =", trans)
     !call vsmLogPrint("trans_bool = ", trans_bool)

     call vsmRotateMatrix(sc, matrix)
     call vsmRotateInvMatrix(sc, invmatrix)

     do i=1, 8
       call vsmTransVertx(wd_vertx(1:3, i), temp_v, trans, trans_bool)
       call vsmMatrixCrossVect(matrix, temp_v, trori(:, i))
       call vsmLogPrint("wd 1 ->", temp_v)
       call vsmLogPrint(" ->", trori(:,i))

       call vsmTransVertx(ld_vertx(1:3, i), temp_v, trans, trans_bool)
       call vsmMatrixCrossVect(matrix, temp_v, trlori(:, i))
       call vsmLogPrint("ld 1 ->", temp_v)
       call vsmLogPrint(" ->", trlori(:,i))
     end do

     max_wd(1:3) = max(trori(1:3,1),trori(1:3,2),trori(1:3,3),trori(1:3,4), &
                       trori(1:3,5),trori(1:3,6),trori(1:3,7), trori(1:3,8)) - sdc

     min_wd(1:3) = min(trori(1:3,1),trori(1:3,2),trori(1:3,3),trori(1:3,4), &
                       trori(1:3,5),trori(1:3,6),trori(1:3,7), trori(1:3,8)) + sdc

     max_ld(1:3) = max(trlori(1:3,1),trlori(1:3,2),trlori(1:3,3),trlori(1:3,4), &
                       trlori(1:3,5),trlori(1:3,6),trlori(1:3,7), trlori(1:3,8))

     min_ld(1:3) = min(trlori(1:3,1),trlori(1:3,2),trlori(1:3,3),trlori(1:3,4), &
                       trlori(1:3,5),trlori(1:3,6),trlori(1:3,7), trlori(1:3,8))

     call vsmLogPrint("max_wd", max_wd)
     call vsmLogPrint("min_wd", min_wd)
     call vsmLogPrint("max_ld", max_ld)
     call vsmLogPrint("min_ld", min_ld)

!     print *, "max_wd", max_wd
!     print *, "min_wd", min_wd
!     print *, "max_ld", max_ld
!     print *, "min_ld", min_ld

     ! the plane never cross local data
     if(min_ld(3) > 0.0_vdp .or. max_ld(3) < 0.0_vdp) cycle

     arpos(2) = min_wd(2) - dc*2.0_vdp; arpos(3) = 0.0_vdp
     yc = int((max_wd(2) - arpos(2))/(dc/dens))
     normalize = .true.
#ifdef OMPNEST
     call vsmSetOMPNest(yc)
     tnuml1 = vsmGetOMPNestL1()
!$omp parallel do private(x,y,xc,temp_v,pos,dx,ci,vx,grid_val,vec_val,pos1,pos2,spos, arpos) num_threads(tnuml1)
     do y = 1, yc !do while(arpos(2) < max_wd(2)) ! y-direc
#else
!$omp parallel do private(x,y,xc,temp_v,pos,dx,ci,vx,grid_val,vec_val,pos1,pos2,spos, arpos)
     do y = 1, yc !do while(arpos(2) < max_wd(2)) ! y-direc
#endif
       arpos(3) = 0.0_vdp
       arpos(2) = min_wd(2) - dc*2.0_vdp + dc/dens*y
       arpos(1) = min_wd(1) - dc*2.0_vdp
       xc = int((max_wd(1) - arpos(1))/(dc/dens))
       do x=1, xc !do while(arpos(1) < max_wd(1)) ! x-direc
         arpos(1) = arpos(1) + dc/dens
         if(arpos(1) < max_ld(1) .and. arpos(1) >= min_ld(1) .and. &
                  arpos(2) < max_ld(2) .and. arpos(2) >= min_ld(2)) then

           call vsmMatrixCrossVect(invmatrix, arpos, temp_v)
           !call vsmLogPrint("arpos  -> temp_v", arpos)
           call vsmTransVertx(temp_v, spos, -trans, trans_bool)
           !call vsmLogPrint("temp_v -> ", spos)
           if(vsmCoordData%YinYang == vsmYin) then
             pos = spos
           else
             pos(1) = -spos(1)
             pos(2) = spos(3)
             pos(3) = spos(2)
           end if
           spos = 0.0_vdp
           call vsmCoordsToAngles(pos, spos)
           if(.not.vsmIsInsidel(spos)) cycle
           !print *,"spos", spos, pos, acos(pos(3) / spos(1)), atan2(pos(2), pos(1))

           call vsmFindCoord(spos, dx, ci)
           call vsmSetFrac(dx, vx)
           call vsmSetVectGridVal(nvec, ci, grid_val)
           call vsmGetVectVal(vx, grid_val, vec_val, normalize)

           call vsmNormVec(vec_val)

           if(vsmCoordData%YinYang == vsmYin) then
             spos = pos
           else
             spos(1) = -pos(1)
             spos(2) = pos(3)
             spos(3) = pos(2)
           end if

           pos1(1:3) = spos(1:3) - vec_val(1:3)*leng/2.0_vdp
           pos2(1:3) = spos(1:3) + vec_val(1:3)*leng/4.0_vdp

           call vsmDrawCylinder(pos1, pos2, rad, cam_num, vis_num, &
                               vsmArAmb, vsmArDif, vsmArSpec, vsmArwShin)

           pos1(1:3) = pos2(1:3)
           pos2(1:3) = pos1(1:3) + vec_val(1:3)*leng/4.0_vdp

           call vsmDrawCone(pos1, pos2, rad15, cam_num, vis_num, vsmArcAmb, &
                              vsmArcDif, vsmArcSpec, vsmArwShin)

         end if
       end do
     end do

   end do

   call vsmLogPrint("draw arrows ends")

 end subroutine vsmyyDrawArrowsIM

end module vsmArrows

