module vsmBSwap
#ifndef CLANG
 use mpi
#endif
 use vsmConst
 use vsmFunction
 use vsmParam
 use vsmImage

 implicit none
 private

 type(vismo_bswap), save, public :: vsmBSwapParam
 
 public vsmInitBSwap, vsmFinalizeBSwap, vsmBinarySwapIMRS, vsmBinarySwapIM, vsmBinarySwap4SolidIM

 contains

 subroutine vsmFinalizeBSwap
  deallocate(vsmBSwapParam%partner, vsmBSwapParam%above)
 end subroutine vsmFinalizeBSwap

 subroutine vsmInitBSwap(rank, psize)
   integer, intent(in) :: rank, psize
   integer :: i
   integer :: allocstate

   call vsmLogPrint("vsmInitBSwap is called")
   call vsmCheckPowerOf2(psize, vsmBSwapParam%numstage, vsmBSwapParam%possible)
   vsmBSwapParam%myrank = rank
   vsmBSwapParam%psize = psize
   vsmBSwapParam%bswap = .false.

   allocate(vsmBSwapParam%partner(vsmBSwapParam%numstage), stat=allocstate)
   if(allocstate /= 0) call vsmError("can't allocate bswap partner")

   allocate(vsmBSwapParam%above(vsmBSwapParam%numstage), stat=allocstate)
   if(allocstate /= 0) call vsmError("can't allocate bswap above")

   do i=1, vsmBSwapParam%numstage
     vsmBSwapParam%partner(i) = -1
     vsmBSwapParam%above(i)= .false.
   end do

   call vsmDvProcess4BS(rank, psize)

   call vsmLogPrint("vsmInitBSwap ends")
 end subroutine vsmInitBSwap

 subroutine vsmDvProcess4BS(rank, psize)
  integer, intent(in) :: rank, psize
  integer :: nstage
  integer :: NumBSProcess
  integer :: prank
  integer :: st, ed, dv
  logical :: a

  nstage = vsmBSwapParam%numstage

  NumBSProcess = 2**nstage
  if(psize == NumBSProcess) vsmBSwapParam%bswap = .true.

  if(rank > NumBSProcess-1) return

  st = 1; ed = NumBSProcess; dv = ed

  do
   if(dv > 1) then
    dv = dv/2
    if(rank+1 < st+dv) then
      prank = rank + dv
      a=.true.
      ed = st+dv-1
    else
      prank = rank - dv
      a=.false.
      st = st+dv
    end if

    vsmBSwapParam%partner(nstage) = prank
    vsmBSwapParam%above(nstage) = a
    call vsmLogPrint("binary swap 4 solid: stage = ", nstage)
    call vsmLogPrint("binary swap 4 solid : myrank = ", rank)
    call vsmLogPrint("binary swap 4 solid : parter rank = ", prank)
    nstage = nstage - 1
    if(nstage == 0) exit
   end if
  end do
 end subroutine vsmDvProcess4BS

 subroutine vsmCheckImageRegionB(startpixel, endpixel)
   integer, intent(out) :: startpixel, endpixel
   integer :: i, numpxls
!   integer :: comm, myrank, psize
!   call vsmGetMPIVars(comm, myrank, psize)

   startpixel = 1
   endpixel   = vsmImageData%pxlsize

   do i=0, vsmImageData%pxlsize-1
     if(vsmImageData%image(i*5+5) < vsmFarAwayd10) then
       startpixel = i
       exit
     end if
   end do

   do i=vsmImageData%pxlsize-1, 0, -1
     if(vsmImageData%image(i*5+5) < vsmFarAwayd10) then
       endpixel = i
       exit
     end if
   end do

   numpxls = endpixel - startpixel + 1
   if(numpxls < 0) then
   end if
 end subroutine vsmCheckImageRegionB

 subroutine vsmBinarySwapIM(solid, cam, vis, tim)
   logical, intent(in) :: solid
   integer, intent(in) :: cam, vis, tim
   integer :: comm, myrank, psize, ierr!, master
#ifndef CLANG
   integer, dimension(mpi_status_size) :: stat
#endif
!   integer :: pixels, pixel5
   integer :: i, st_pixel, ed_pixel, prank
   integer, dimension(2) :: spixl
   integer :: counts, countr, length
   integer :: tag, count, dest, src

!   character(4) :: viskind, camnum, rank 
!   character(7) :: timestep

   call vsmLogPrint("----- Binary Swap IM")
   call vsmGetMPIVars(comm, myrank, psize)
   tag = 0
   st_pixel = 1
   ed_pixel = vsmImageData%rimgsize !vsmImageData%width * vsmImageData%height ! send receive data*5

   call vsmCheckImageL(st_pixel, ed_pixel)
!   do i=vsmBSwapParam%numstage, 1, -1
   do i=1, vsmBSwapParam%numstage, 1 ! 2015/9/27
     call vsmLogPrint("binary swap stage = ", i)
     prank = vsmBSwapParam%partner(i)
     length = ed_pixel - st_pixel +1

     call vsmLogPrint("bswap partner ", prank)
     if(mod(length,2) /=0) call vsmError("bswap: length is not even")

     call vsmClearRImage
     if(vsmBSwapParam%above(i)) then
       countr = length/2
       !counts = countr 
       counts = length - countr

       count = counts*5
#ifndef CLANG
       call mpi_send(vsmImageData%image((st_pixel+countr-1)*5+1), count, mpi_real, prank, tag, comm, ierr)
#else
       call VSMMPISENDFLOAT(vsmImageData%image((st_pixel+countr-1)*5+1), count, prank)
#endif
       count = countr*5
#ifndef CLANG
       call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVFLOAT(vsmImageData%image_r, count, prank)
#endif

       st_pixel = st_pixel
       ed_pixel = st_pixel+countr-1
     else
       counts = length/2
       !countr = counts 
       countr = length - counts

       count = countr*5
#ifndef CLANG 
       call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVFLOAT(vsmImageData%image_r, count, prank)
#endif
       count = counts*5
#ifndef CLANG 
       call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, prank, tag, comm, ierr)
#else
       call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, prank) 
#endif

       ed_pixel = ed_pixel
       st_pixel = st_pixel+counts
     end if

     call vsmLogPrint("stpixel ", st_pixel)
     call vsmLogPrint("edpixel ", ed_pixel)
     if(solid) then
       call vsmLogPrint("solid blending")
       call vsmSolidBlending(st_pixel, ed_pixel)
     else
       call vsmLogPrint("alpha blending")
       call vsmAlphaBlending(st_pixel, ed_pixel)
     end if
     call vsmLogPrint("checkimagel")
     call vsmCheckImageL(st_pixel, ed_pixel)
   end do

   call vsmLogPrint("start gathering")

   if(myrank == 0) then
     do i=1, psize-1
       count = 2
#ifndef CLANG
       call mpi_recv(spixl, count, mpi_integer, i, tag, comm, stat, ierr)
#else
       call VSMMPIRECVINT(spixl, count, i)
#endif
       count = spixl(2)*5
#ifndef CLANG
       call mpi_recv(vsmImageData%image((spixl(1)-1)*5+1), count, mpi_real, i, tag, comm, stat, ierr)
#else
       call VSMMPIRECVFLOAT(vsmImageData%image((spixl(1)-1)*5+1), count, i)
#endif
     end do

   else
     spixl(1) = st_pixel
     spixl(2) = ed_pixel-st_pixel+1
     dest = 0
     count = 2
#ifndef CLANG
     call mpi_send(spixl, count, mpi_integer, dest, tag, comm, ierr)
#else
     call VSMMPISENDINT(spixl, count, dest)
#endif
     count = spixl(2)*5
#ifndef CLANG
     call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, dest, tag, comm, ierr)
#else
     call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, dest)
#endif

   end if

   if(solid) then
     src = 0
     count = vsmImageData%rimgsize*5
#ifndef CLANG
     call mpi_bcast(vsmImageData%image, count, mpi_real, src, comm, ierr)
#else
     call VSMMPIBCASTFLOAT(vsmImageData%image, count, src)
#endif
   end if

   call vsmLogPrint("----- Binary Swap IM ends")
 end subroutine vsmBinarySwapIM

 subroutine vsmBinarySwapIMRS(cam, vis, tim)
   integer, intent(in) :: cam, vis, tim
   integer :: comm, myrank, psize, ierr!, master
#ifndef CLANG
   integer, dimension(mpi_status_size) :: stat
#endif
   integer :: i, st_pixel, ed_pixel, prank
   integer, dimension(2) :: spixl
   integer :: counts, countr, length, numr, nums, soffset!, fstage
   integer :: tag, src, dest, count

!   character(4) :: viskind, camnum, rank 
!   character(7) :: timestep
   tag = 0
!   call vsmLogPrint("----- Binary Swap IMRS")
   call vsmGetMPIVars(comm, myrank, psize)

   !print *,"bswap im rs",myrank
   if(vsmDbg) then
     call vsmLogPrint("----- Binary Swap IMRS")
     call vsmLogPrint("vsmDbg raysegment")
     call vsmBlendRaySegmentF(0)
     call vsmLogPrint("vsmDbg raysegment ends")
     select case(myrank)
       case (0)
         call vsmWriteImage('image0')
       case(1)
         call vsmWriteImage('image1')
       case(2)
         call vsmWriteImage('image2')
       case(3)
         call vsmWriteImage('image3')
       case(4)
         call vsmWriteImage('image4')
       case(5)
         call vsmWriteImage('image5')
       case(6)
         call vsmWriteImage('image6')
       case(7)
         call vsmWriteImage('image7')
       end select
   end if

   st_pixel = 1
   ed_pixel = vsmImageData%rimgsize !vsmImageData%width * vsmImageData%height ! send receive data*5

!   fstage = vsmBSwapParam%numstage
   do i=1, vsmBSwapParam%numstage !do i=1, fstage
     call vsmLogPrint("binary swap stage = ", i)
     prank = vsmBSwapParam%partner(i)
     length = ed_pixel - st_pixel +1

     call vsmLogPrint("bswap partner ", prank)

     call vsmClearRaySegmentR

     call vsmLogPrint("RaySegmentR cleared")
     if(vsmBSwapParam%above(i)) then
       countr = length/2
       counts = length - countr
       nums = counts*vsmRSD*vsmRSBNum*(2**(i-1))
       count = 1
#ifndef CLANG
       call mpi_send(nums, count, mpi_integer, prank, tag, comm, ierr)
#else
       call VSMMPISENDINT(nums, count, prank)
#endif

       if(vsmDbg) then
          call vsmLogPrint("data sent max",vsmRSD*vsmRSBNum*vsmImageData%rimgsize)
          call vsmLogPrint("data sent",nums)
          call vsmLogPrint("data send from",countr*vsmRSBNum*vsmRSD*(2**(i-1))+1)
          call vsmLogPrint("data send to",countr*vsmRSBNum*vsmRSD*(2**(i-1))+nums)
       end if
#ifndef CLANG
       call mpi_send(vsmImageData%rsdata(countr*vsmRSBNum*vsmRSD*(2**(i-1))+1), nums, &
            mpi_integer, prank, tag, comm, ierr)
#else
       call VSMMPISENDINT(vsmImageData%rsdata(countr*vsmRSBNum*vsmRSD*(2**(i-1))+1), nums, prank)
#endif

       call vsmLogPrint("rsdata sent")

      count = 1
#ifndef CLANG
       call mpi_recv(numr, count, mpi_integer, prank, tag, comm, stat, ierr)
       call vsmLogPrint("data receive", numr)
       call mpi_recv(vsmImageData%rsdata_r, numr, mpi_integer, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVINT(numr, count, prank)
       call vsmLogPrint("data receive", numr)
       call VSMMPIRECVINT(vsmImageData%rsdata_r, numr, prank)
#endif

       call vsmLogPrint("data received")
       st_pixel = st_pixel
       ed_pixel = st_pixel+countr-1
       soffset = 1
       if(vsmDbg) then
         call vsmLogPrint("st_pixel", st_pixel)
         call vsmLogPrint("ed_pixel", ed_pixel)
       end if

     else
       counts = length/2
       countr = length - counts
       nums = counts*vsmRSD*vsmRSBNum*(2**(i-1))

       count = 1
#ifndef CLANG
       call mpi_recv(numr, count, mpi_integer, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVINT(numr, count, prank)
#endif
       if(vsmDbg) then
         call vsmLogPrint("data received max",vsmRSD*vsmRSBNum*vsmImageData%rimgsize)
         call vsmLogPrint("data received",numr)
      end if

#ifndef CLANG
       call mpi_recv(vsmImageData%rsdata_r, numr, mpi_integer, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVINT(vsmImageData%rsdata_r, numr, prank)
#endif

       if(vsmDbg) then
         call vsmLogPrint("rs data received",numr)
         call vsmLogPrint("all data received")
       end if

       count = 1
#ifndef CLANG
       call mpi_send(nums, count, mpi_integer, prank, tag, comm, ierr)
       call vsmLogPrint("data send",nums)
       call mpi_send(vsmImageData%rsdata, nums, mpi_integer, prank, tag, comm, ierr)
#else
       call VSMMPISENDINT(nums, count, prank)
       call vsmLogPrint("data send",nums)
       call VSMMPISENDINT(vsmImageData%rsdata, nums, prank)
#endif


       call vsmLogPrint("data sent")

       st_pixel = st_pixel+counts
       ed_pixel = ed_pixel
       soffset = counts+1
       if(vsmDbg) then
         call vsmLogPrint("st_pixel", st_pixel)
         call vsmLogPrint("ed_pixel", ed_pixel)
       end if
     end if
     call vsmLogPrint("check segment ")

     call vsmCheckRaySegment

     call vsmLogPrint("copy segment ")

     call vsmCopyRaySegment(soffset, countr, numr, i)

     call vsmLogPrint("check segment ")
     call vsmCheckRaySegment
     call vsmLogPrint("clear segment ")

     call vsmClearRaySegment
     call vsmLogPrint("Merge segment ")

     call vsmMergeRaySegment(i, countr)

     if(vsmDbg) then
       call vsmLogPrint("stpixel ", st_pixel)
       call vsmLogPrint("edpixel ", ed_pixel)
     end if
     call vsmCheckRaySegment

!     call vsmCheckImageL(st_pixel, ed_pixel)
   end do
   call vsmLogPrint("start blend ray segment")

   call vsmBlendRaySegmentF(vsmBSwapParam%numstage)!(fstage)

   call vsmLogPrint("start gathering")

   if(myrank == 0) then
     do i=1, psize-1
#ifndef CLANG
         count = 2
         call mpi_recv(spixl, count, mpi_integer, i, tag, comm, stat, ierr)
         count = spixl(2)*5
         call mpi_recv(vsmImageData%image((spixl(1)-1)*5+1), count, mpi_real, i, tag, comm, stat, ierr)
#else
         count = 2
         call VSMMPIRECVINT(spixl, count, i)
         count = spixl(2)*5
         call VSMMPIRECVFLOAT(vsmImageData%image((spixl(1)-1)*5+1), count, i)
#endif
     end do
   else
     spixl(1) = st_pixel
     spixl(2) = ed_pixel-st_pixel+1
       dest = 0
#ifndef CLANG
       count = 2
       call mpi_send(spixl, count, mpi_integer, dest, tag, comm, ierr)
       count = spixl(2)*5
       call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, dest, tag, comm, ierr)
#else
       count = 2
       call VSMMPISENDINT(spixl, count, dest)
       count = spixl(2)*5
       call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, dest)
#endif
   end if

   call vsmLogPrint("----- Binary Swap ends")
 end subroutine vsmBinarySwapIMRS

 subroutine vsmBinarySwap4SolidIM(cam, vis, tim)
   integer, intent(in) :: cam, vis, tim
   integer :: comm, myrank, psize, ierr!, master
#ifndef CLANG
   integer, dimension(mpi_status_size) :: stat
#endif
!   integer :: pixels, pixel5
   integer :: i, st_pixel, ed_pixel, prank
   integer :: nbsproc, exproc
   integer, dimension(2) :: spixl
   integer :: counts, countr, length
   integer :: count, tag, src, dest

!   character(4) :: viskind, camnum, rank 
!   character(7) :: timestep
   tag = 0
   call vsmLogPrint("----- Binary Swap for Solid")
   call vsmGetMPIVars(comm, myrank, psize)

   nbsproc = 2**vsmBSwapParam%numstage

   if(psize > nbsproc) then

     exproc = psize - nbsproc

     if(myrank < exproc) then
       prank = myrank + nbsproc
       count = vsmImageData%rimgsize*5
#ifndef CLANG
       call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
#else
       call VSMMPIRECVFLOAT(vsmImageData%image_r, count, prank)
#endif
       call vsmSolidBlending(1, vsmImageData%rimgsize)
     else if(myrank > nbsproc-1) then
       prank = myrank - nbsproc
       count = vsmImageData%rimgsize*5
#ifndef CLANG
       call mpi_send(vsmImageData%image, count, mpi_real, prank, tag, comm, ierr)
#else
       call VSMMPISENDFLOAT(vsmImageData%image, count, prank)
#endif
     end if
   end if

#ifndef CLANG
   call mpi_barrier(comm, ierr)
#else
   call VSMMPIBARRIER()
#endif

   if(myrank < nbsproc) then

     st_pixel = 1
     ed_pixel = vsmImageData%rimgsize !vsmImageData%width * vsmImageData%height ! send receive data*5

     call vsmCheckImageL(st_pixel, ed_pixel)
!     do i=vsmBSwapParam%numstage, 1, -1
     do i=1, vsmBSwapParam%numstage
       call vsmLogPrint("binary swap for solid stage = ", i)
       prank = vsmBSwapParam%partner(i)
       length = ed_pixel - st_pixel +1
       call vsmClearRImage
       if(vsmBSwapParam%above(i)) then
         countr = length/2
         !counts = countr
         counts = length - countr

#ifndef CLANG
         count = counts*5
         call mpi_send(vsmImageData%image((st_pixel+countr-1)*5+1), count, mpi_real, prank, tag, comm, ierr)
         count = countr*5
         call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
#else
         count = counts*5
         call VSMMPISENDFLOAT(vsmImageData%image((st_pixel+countr-1)*5+1), count, prank)
         count = countr*5
         call VSMMPISENDFLOAT(vsmImageData%image_r, count, prank)
#endif
         st_pixel = st_pixel
         ed_pixel = st_pixel+countr-1
       else
         counts = length/2
         !countr = counts 
         countr = length - counts
#ifndef CLANG
         count = countr*5
         call mpi_recv(vsmImageData%image_r, count, mpi_real, prank, tag, comm, stat, ierr)
         count = counts*5
         call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, prank, tag, comm, ierr)
#else
         count = countr*5
         call VSMMPIRECVFLOAT(vsmImageData%image_r, count, prank)
         count = counts*5
         call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, prank)
#endif
         ed_pixel = ed_pixel
         st_pixel = st_pixel+counts
       end if

       call vsmSolidBlending(st_pixel, ed_pixel)
       call vsmCheckImageL(st_pixel, ed_pixel)
     end do

     call vsmLogPrint("start gathering")

     if(myrank == 0) then
       do i=1, nbsproc-1
#ifndef CLANG
         count = 2
         call mpi_recv(spixl, count, mpi_integer, i, tag, comm, stat, ierr)
         count = spixl(2)*5
         call mpi_recv(vsmImageData%image((spixl(1)-1)*5+1), count, mpi_real, i, tag, comm, stat, ierr)
#else
         count = 2
         call VSMMPIRECVINT(spixl, count, i)
         count = spixl(2)*5
         call VSMMPIRECVFLOAT(vsmImageData%image((spixl(1)-1)*5+1), count, i)
#endif
       end do
     else
       spixl(1) = st_pixel
       spixl(2) = ed_pixel-st_pixel+1
       dest = 0
#ifndef CLANG
       count = 2
       call mpi_send(spixl, count, mpi_integer, dest, tag, comm, ierr)
       count = spixl(2)*5
       call mpi_send(vsmImageData%image((st_pixel-1)*5+1), count, mpi_real, dest, tag, comm, ierr)
#else
       count = 2
       call VSMMPISENDINT(spixl, count, dest)
       count = spixl(2)*5
       call VSMMPISENDFLOAT(vsmImageData%image((st_pixel-1)*5+1), count, dest)
#endif
     end if

   end if

   src = 0
   count = vsmImageData%rimgsize*5
#ifndef CLANG
   call mpi_bcast(vsmImageData%image, count, mpi_real, src, comm, ierr)
#else
   call VSMMPIBCASTFLOAT(vsmImageData%image, count, src)
#endif

   !----
   call vsmLogPrint("----- Binary Swap for Solid ends")
 end subroutine vsmBinarySwap4SolidIM

end module vsmBSwap

