module vsmCameraData
 use vsmConst
 use vsmFunction

 implicit none
 private

 type(vismo_camera), dimension(vsmMaxCamera),save, public :: vsmCamera
 integer,save, public :: vsmCameraNum
 logical, save, public :: vsmMultipleCameras
 
 public vsmInitCamera, vsmSetCamera

 contains
 
 ! initialize camera module
 subroutine vsmInitCamera
  integer :: i, j
  vsmCamera(1:vsmMaxCamera)%fovy = 30.0_vdp
  vsmCamera(1:vsmMaxCamera)%near = 1.0_vdp
  vsmCamera(1:vsmMaxCamera)%far  = 1000.0_vdp
  vsmCamera(1:vsmMaxCamera)%pers  = .true.
  vsmCamera(1:vsmMaxCamera)%dp  = -1.0_vdp ! 20190219
  vsmCamera(1:vsmMaxCamera)%width = 0.0_vdp

  do i=1, 3
    vsmCamera(1:vsmMaxCamera)%pos(i) = 0.0_vdp
    vsmCamera(1:vsmMaxCamera)%front(i) = 0.0_vdp
    vsmCamera(1:vsmMaxCamera)%up(i) = 0.0_vdp
    vsmCamera(1:vsmMaxCamera)%right(i) = 0.0_vdp
  end do
  vsmCameraNum = 0
  vsmMultipleCameras = .false.

  vsmCamera(1:vsmMaxCamera)%calculated = .false.
  do i=1, 3
    do j=1, 6
      vsmCamera(1:vsmMaxCamera)%vertex(i, j) = 0.0_vdp
    end do
    do j=1, 3
      vsmCamera(1:vsmMaxCamera)%l(i, j) = 0.0_vdp
    end do
  end do

  do i=1,2
    vsmCamera(1:vsmMaxCamera)%swidth(i) = -1
    vsmCamera(1:vsmMaxCamera)%sheight(i) = -1
    vsmCamera(1:vsmMaxCamera)%dph(i) = -1.0_vdp
  end do
  do i=1, 4
    do j=1, 6
      vsmCamera(1:vsmMaxCamera)%plane(i, j) = 0.0_vdp
    end do
  end do
 end subroutine vsmInitCamera

 subroutine vsmSetCamera(near, far, pos, up, front, pers, fovy, width)
  real(kind=vdp), intent(in) :: near, far, fovy, width
  logical, intent(in) :: pers
  real(kind=vdp), dimension(3), intent(in) :: pos, front, up

  vsmCameraNum = vsmCameraNum + 1
  if(vsmCameraNum > vsmMaxCamera) call vsmError('set vsmCamera :exceed max vsmCamera')

  vsmCamera(vsmCameraNum)%fovy = fovy
  vsmCamera(vsmCameraNum)%pers = pers
  vsmCamera(vsmCameraNum)%width = width

  vsmCamera(vsmCameraNum)%near = near
  vsmCamera(vsmCameraNum)%far  = far
  vsmCamera(vsmCameraNum)%pos(1:3) = pos(1:3)
  vsmCamera(vsmCameraNum)%front(1:3) = front(1:3)
  call vsmNormVec(vsmCamera(vsmCameraNum)%front)

  vsmCamera(vsmCameraNum)%up(1:3) = up(1:3)
  call vsmNormVec(vsmCamera(vsmCameraNum)%up)

  call vsmCrossProd(vsmCamera(vsmCameraNum)%right, vsmCamera(vsmCameraNum)%front, vsmCamera(vsmCameraNum)%up)
  call vsmNormVec(vsmCamera(vsmCameraNum)%right)

  call vsmCrossProd(vsmCamera(vsmCameraNum)%up, vsmCamera(vsmCameraNum)%right, vsmCamera(vsmCameraNum)%front)

 end subroutine vsmSetCamera

end module vsmCameraData

