module vsmColormap
 use vsmConst
 use vsmFunction

 implicit none
 private

 type(vismo_scolormap), dimension(:), allocatable, save, public :: vsmSColormap
 type(vismo_vcolormap), dimension(:), allocatable, save, public :: vsmVColormap
 integer, save, public :: vsmSCMapNum, vsmVCMapNum
 
 public vsmGetAlpha, vsmGetMaxAlpha
 public vsmSetVColorArray, vsmSetSColorArray, vsmInitColMap, vsmGetVColor, vsmGetScolor, vsmFinalizeColMap

 contains

 subroutine vsmInitColmap(scalnum, vectnum)
   integer, intent(in) :: scalnum, vectnum
   integer :: allocstate, i, j
   vsmSCMapNum = scalnum; vsmVCMapNum = vectnum

   if(scalnum > 0) then
     allocate(vsmSColormap(scalnum), stat=allocstate)
     if(allocstate /= 0) call vsmError('cant allocate colormap')
   end if

   if(vectnum > 0) then
     allocate(vsmVColormap(vectnum), stat=allocstate)
     if(allocstate /= 0) call vsmError('cant allocate colormap')
   end if

   do i=1, scalnum
!$omp parallel do private(j)
    do j=1,vsmMaxColmap
      vsmSColormap(i)%value(j) = 0.0_vdp
      vsmSColormap(i)%color(1:4, j) = 0.0
    end do
    vsmSColormap(i)%dv = -1.0_vdp
    vsmSColormap(i)%colnum = 0
    vsmSColormap(i)%set = .false.
    vsmSColormap(i)%eqspace = .false.
   end do

   do i=1, vectnum
!$omp parallel do private(j)
    do j=1,vsmMaxColmap
      vsmVColormap(i)%value(j) = 0.0_vdp
      vsmVColormap(i)%color(1:3, j) = 0.0
    end do
    vsmVColormap(i)%dv = -1.0_vdp
    vsmVColormap(i)%colnum = 0
    vsmVColormap(i)%set = .false.
    vsmVColormap(i)%eqspace = .false.
   end do
 end subroutine vsmInitColmap

 subroutine vsmFinalizeColmap
  if(allocated(vsmSColormap)) deallocate(vsmSColormap)
  if(allocated(vsmVColormap)) deallocate(vsmVColormap)
 end subroutine vsmFinalizeColmap

 subroutine vsmSetSColorArray(nscal, n, val, col, es)
  integer, intent(in) :: n, nscal
  logical, intent(in) :: es
  real(kind=vdp), dimension(vsmMaxColmap), intent(in) :: val
  real, dimension(4, vsmMaxColmap), intent(in) :: col
  integer :: i, j

  if(n > vsmMaxColmap .or. n < 1) call vsmError('exceed bound :: vsmSColormap')
  if(nscal > vsmMaxScalar .or. nscal < 1) call vsmError('scal num :: vsmSColormap')

!$omp parallel do private(i)
  do i=1, n
    call vsmCheckFA01(col(1:4, i))
    vsmSColormap(nscal)%color(1:4, i) = col(1:4, i)
    vsmSColormap(nscal)%value(i) = val(i)
  end do
  vsmSColormap(nscal)%colnum = n
  vsmSColormap(nscal)%set = .true.
  if(es) then
    vsmSColormap(nscal)%dv = (vsmSColormap(nscal)%value(2)-vsmSColormap(nscal)%value(1))
    vsmSColormap(nscal)%eqspace = .true.
  end if
 end subroutine vsmSetSColorArray

 subroutine vsmSetSColorES(nscal)
    integer, intent(in) :: nscal
    real(kind=vdp) :: dval
    if(.not.vsmSColormap(nscal)%set) call vsmError("scal num not set vsmSColorES")
    if(nscal > vsmMaxScalar .or. nscal <= 0) call vsmError("scal num vsmSColorES")
    vsmSColormap(nscal)%eqspace = .true.
    dval = (vsmSColormap(nscal)%value(2)-vsmSColormap(nscal)%value(1))
    vsmSColormap(nscal)%dv = dval
 end subroutine vsmSetSColorES

 subroutine vsmSetVColorArray(nvec, n, val, col, es)
  integer, intent(in) :: n, nvec
  logical, intent(in) :: es
  real(kind=vdp), dimension(vsmMaxColmap), intent(in) :: val
  real, dimension(3, vsmMaxColmap), intent(in) :: col
  integer :: i, j

  if(n > vsmMaxColmap .or. n < 1) call vsmError('exceed bound :: vsmVColormap')
  if(nvec > vsmMaxVector .or. nvec < 1) call vsmError('vect num :: vsmVColormap')

!$omp parallel do private(i)
  do i=1, n
    call vsmCheckFA01(col(1:3, i))
    vsmVColormap(nvec)%color(1:3, i) = col(1:3, i)
    vsmVColormap(nvec)%value(i) = val(i)
  end do

  vsmVColormap(nvec)%colnum = n
  vsmVColormap(nvec)%set = .true.
  if(es) then
    vsmVColormap(nvec)%dv = (vsmVColormap(nvec)%value(2)-vsmVColormap(nvec)%value(1))
    vsmVColormap(nvec)%eqspace = .true.
  end if

 end subroutine vsmSetVColorArray

 subroutine vsmSetVColorES(nvec)
    integer, intent(in) :: nvec
    real(kind=vdp) :: dval
    if(.not.vsmVColormap(nvec)%set) call vsmError("vect num not set vsmSetVColorES")
    if(nvec > vsmMaxVector .or. nvec <= 0) call vsmError("vect num vsmSetVColorES")
    vsmVColormap(nvec)%eqspace = .true.
    dval = (vsmVColormap(nvec)%value(2) - vsmVColormap(nvec)%value(1))
    vsmVColormap(nvec)%dv = dval
 end subroutine vsmSetVColorES

 subroutine vsmGetSColor(val, nscal, col)
  !binary search
  real(kind=vdp), intent(in) :: val
  integer, intent(in) :: nscal
  real, dimension(4), intent(out) :: col
  real(kind=vdp) :: cp
  integer :: first, last, middle, maxc, colpos
  logical :: found
  real :: dc

  if(nscal < 1 .or. nscal > vsmMaxScalar) call vsmError('error get scalar color')
  maxc = vsmSColormap(nscal)%colnum

  if(vsmSColormap(nscal)%eqspace) then
    cp = (val - vsmSColormap(nscal)%value(1))/vsmSColormap(nscal)%dv
    colpos = aint(cp) 

    if(colpos >= maxc - 1) then
      col(1:4) = vsmSColormap(nscal)%color(1:4, maxc)
      return
    end if

    if(colpos <= 0) then
      col(1:4) = vsmSColormap(nscal)%color(1:4, 1)
      return
    end if
    
    dc = cp - colpos
    colpos = colpos + 1
    col(1:4) = (1.0-dc)*vsmSColormap(nscal)%color(1:4, colpos) + dc*vsmSColormap(nscal)%color(1:4, colpos+1)
    return
  end if

  if(val <= vsmSColormap(nscal)%value(1)) then
    col(1:4) = vsmSColormap(nscal)%color(1:4, 1)
    return
  end if

  if(val >= vsmSColormap(nscal)%value(maxc)) then
    col(1:4) = vsmSColormap(nscal)%color(1:4, maxc)
    return
  end if

  first = 1
  last  = maxc
  found = .false.

  do
    if (last- first < 0 .or. found) exit
    middle = (first+last)/2
    if(val < vsmSColormap(nscal)%value(middle)) then
      last = middle
    else if(val >= vsmSColormap(nscal)%value(middle+1)) then
      first = middle
    else
      found = .true.
      dc = real((val - vsmSColormap(nscal)%value(middle))/ &
              (vsmSColormap(nscal)%value(middle+1) - vsmSColormap(nscal)%value(middle)))
      col(1:4) = dc*vsmSColormap(nscal)%color(1:4, middle+1) + &
                       (1.0 - dc)*vsmSColormap(nscal)%color(1:4, middle)
      exit
    end if
  end do
  if(.not.found) call vsmError('cant find color (scalar)')

 end subroutine vsmGetSColor

 function vsmGetMaxAlpha(dat, nscal)
  real(kind=vdp), dimension(8), intent(in) :: dat
  real(kind=vdp) :: mincp, maxcp, val
  integer, intent(in) :: nscal
  real(kind=vdp) :: maxv, minv
  real :: vsmGetMaxAlpha, alpha
  integer, dimension(2) :: colpos

  !binary search
  integer :: first, last, middle, maxc, i
  logical :: found

  if(nscal < 1 .or. nscal > vsmMaxScalar) call vsmError('error get max alpha')

  maxv = maxval(dat)
  minv = minval(dat)
  if(minv > maxv) call vsmError('error get max alpha')

  maxc = vsmSColormap(nscal)%colnum

  if(vsmSColormap(nscal)%eqspace) then
    mincp = (minv - vsmSColormap(nscal)%value(1))/vsmSColormap(nscal)%dv
    colpos(1) = aint(mincp) 

    maxcp = (maxv - vsmSColormap(nscal)%value(1))/vsmSColormap(nscal)%dv
    colpos(2) = aint(maxcp) 

    if(colpos(1) >= maxc - 1) then
       colpos(1) = maxc
    end if

    if(colpos(1) <= 0) then
       colpos(1) = 1
    end if

    if(colpos(2) >= maxc - 1) then
       colpos(2) = maxc
    end if

    if(colpos(2) <= 0) then
       colpos(2) = 1
    end if

    alpha = 0.0
    do i=colpos(1), colpos(2)
       if(alpha < vsmSColormap(nscal)%color(4, i)) alpha = vsmSColormap(nscal)%color(4, i)
    end do
    vsmGetMaxAlpha = alpha
    return
  end if

  do i=1, 2
    if(i == 1) then
      val = minv
    else
      val = maxv
    end if

    if(val <= vsmSColormap(nscal)%value(1)) then
      colpos(i) = 1
      cycle
    end if

    if(val >= vsmSColormap(nscal)%value(maxc)) then
      colpos(i) = maxc
      cycle
    end if

    first = 1
    last  = maxc
    found = .false.

    do
      if (last- first < 0 .or. found) exit
      middle = (first+last)/2
      if(val < vsmSColormap(nscal)%value(middle)) then
        last = middle
      else if(val >= vsmSColormap(nscal)%value(middle+1)) then
        first = middle
      else
        found = .true.
        colpos(i) = middle
        exit
      end if
    end do
    if(.not.found) call vsmError('cant find color (scalar)')
  end do

  alpha = 0.0
  do i=colpos(1), colpos(2)
     if(alpha < vsmSColormap(nscal)%color(4, i)) alpha = vsmSColormap(nscal)%color(4, i)
  end do
  vsmGetMaxAlpha = alpha

 end function vsmGetMaxAlpha

 function vsmGetAlpha(val, nscal)   ! set_alpha
  real(kind=vdp), intent(in) :: val
  real(kind=vdp) :: cp
  integer, intent(in) :: nscal
  real :: vsmGetAlpha
  integer :: colpos

  !binary search
  integer :: first, last, middle, maxc
  logical :: found
  real :: dc

  if(nscal < 1 .or. nscal > vsmMaxScalar) call vsmError('error get scalar color')
  maxc = vsmSColormap(nscal)%colnum

  if(vsmSColormap(nscal)%eqspace) then
    cp = (val - vsmSColormap(nscal)%value(1))/vsmSColormap(nscal)%dv
    colpos = aint(cp) 

    if(colpos >= maxc - 1) then
      vsmGetAlpha = vsmSColormap(nscal)%color(4, maxc)
      return
    end if

    if(colpos <= 0) then
      vsmGetAlpha = vsmSColormap(nscal)%color(4, 1)
      return
    end if
    
    dc = cp - colpos
    colpos = colpos + 1
    vsmGetAlpha = (1.0-dc)*vsmSColormap(nscal)%color(4, colpos) + dc*vsmSColormap(nscal)%color(4, colpos+1)
    return
  end if

  if(val <= vsmSColormap(nscal)%value(1)) then
    vsmGetAlpha = vsmSColormap(nscal)%color(4, 1)
    return
  end if

  if(val >= vsmSColormap(nscal)%value(maxc)) then
    vsmGetAlpha = vsmSColormap(nscal)%color(4, maxc)
    return
  end if

  first = 1
  last  = maxc
  found = .false.

  do
    if (last- first < 0 .or. found) exit
    middle = (first+last)/2
    if(val < vsmSColormap(nscal)%value(middle)) then
      last = middle
    else if(val >= vsmSColormap(nscal)%value(middle+1)) then
      first = middle
    else
      found = .true.
      dc = real((val - vsmSColormap(nscal)%value(middle))/ &
              (vsmSColormap(nscal)%value(middle+1) - vsmSColormap(nscal)%value(middle)))
      vsmGetAlpha = dc*vsmSColormap(nscal)%color(4, middle+1) + &
                       (1.0 - dc)*vsmSColormap(nscal)%color(4, middle)
      exit
    end if
  end do
  if(.not.found) call vsmError('cant find color (scalar)')
 end function vsmGetAlpha

 subroutine vsmGetVColor(val, nvec, col)
  !binary search
  real(kind=vdp), intent(in) :: val
  integer, intent(in) :: nvec
  real(kind=vdp) :: cp
  real, dimension(3), intent(out) :: col
  integer :: first, last, middle, maxc, colpos
  logical :: found
  real :: dc

  if(nvec < 1 .or. nvec > vsmMaxVector) call vsmError('error get vector color')
  maxc = vsmSColormap(nvec)%colnum

  if(vsmVColormap(nvec)%eqspace) then
    cp = (val - vsmVColormap(nvec)%value(1))/vsmVColormap(nvec)%dv
    colpos = aint(cp) 

    if(colpos >= maxc - 1) then
      col(1:3) = vsmVColormap(nvec)%color(1:3, maxc)
      return
    end if

    if(colpos <= 0) then
      col(1:3) = vsmVColormap(nvec)%color(1:3, 1)
      return
    end if
    
    dc = cp - colpos
    colpos = colpos + 1
    col(1:3) = (1.0-dc)*vsmVColormap(nvec)%color(1:3, colpos) + dc*vsmVColormap(nvec)%color(1:3, colpos+1)
    return
  end if

  if(val <= vsmVColormap(nvec)%value(1)) then
    col(1:3) = vsmVColormap(nvec)%color(1:3, 1)
    return
  end if

  if(val >= vsmVColormap(nvec)%value(maxc)) then
    col(1:3) = vsmVColormap(nvec)%color(1:3, maxc)
    return
  end if

  first = 1
  last  = maxc
  found = .false.

  do
    if (last- first < 0 .or. found) exit
      middle = (first+last)/2
    if(val < vsmVColormap(nvec)%value(middle)) then
      last = middle
    else if(val >= vsmVColormap(nvec)%value(middle+1)) then
      first = middle
    else
      found = .true.
      dc = real((val - vsmVColormap(nvec)%value(middle))/ &
              (vsmVColormap(nvec)%value(middle+1) - vsmVColormap(nvec)%value(middle)))
      col(1:3) = dc*vsmVColormap(nvec)%color(1:3, middle+1) + &
                       (1.0 - dc)*vsmVColormap(nvec)%color(1:3, middle)
      exit
    end if
  end do
  if(.not.found) call vsmError('cant find color (vector)')

 end subroutine vsmGetVColor

end module vsmColormap

