module vsmConst
 ! big endian for outputting bmp
 logical, parameter :: bigend = .false.

 !for double precision
 integer, parameter :: vdp = selected_real_kind(14)

 !for 2bytes integer (short), from -32768 to 32767
 integer, parameter :: isp = selected_int_kind(4)

 integer, parameter :: int_max = 2147483647

 !debug
 logical, parameter :: vsmDbg = .false.

 !debug
 logical, parameter :: vsmReInitMode = .false.

 !pi
 real(kind=vdp) :: pi = 3.1415926535897932384_vdp

 !for OpenMP
 integer, parameter :: vsmMaxOMP = 512
 integer, parameter :: vsmMaxOMP4Solid = 23

 !for file open
! from  vsmUnitNum to vsmUnitMaxNum may be used
! About 10 Unit Numbers are needed
 integer, parameter :: vsmUnitNum =  35
 integer, parameter :: vsmUnitMaxNum = 99

 ! image
 character(5), parameter :: vsmImageFile = 'image' ! imagev[vis]c[camera]t[time].ppm
 integer, parameter :: vsmImgKind = 2 ! 0:BMP, 1:PPM, 2:PNG

 ! config change record
 character(16), parameter :: vsmConfigChangeFile = 'ConfigChange.vsm'

 character(15), parameter :: vsmDataInfoFile = 'vsmDataInfo.vsm'

 !for camera
 integer, parameter :: vsmMaxCamera = 10

 !for light
 integer, parameter :: vsmMaxLight = 4

 !image
 real(kind=vdp), parameter :: vsmFaraway = 1.0D+20
 real(kind=vdp), parameter :: vsmFarawayd10 = 1.0D+19

 integer, parameter :: vsmPbLen = 512
 !integer, parameter :: vsmPbLen = 10000
 integer, parameter :: vsmPackConst = 900

 !isosurface
 integer, parameter :: vsmMaxIsosurf = 5 ! less than 20
 !slice
 integer, parameter :: vsmMaxSlice = 5
 !stream line
 integer, parameter :: vsmMaxSeed = 50
 integer, parameter :: vsmMaxSlineSteps = 40000
 real(kind=vdp), parameter :: vsmIntegStride = 0.1_vdp ! min grid size * vsmIntegStride = Runge-Kutta integration interval
 !arrow : plane of arrows
 integer, parameter :: vsmMaxArrows = 5
 !volume rendering
 integer, parameter :: vsmMaxVolren = 2
 integer, parameter :: vsmMaxRayCount = 1000000
 real, parameter :: vsmErtValue = 0.999999
 real(kind=vdp), parameter :: vsmRayStrideVR = 0.5_vdp ! min grid size * vsmRayStride = sampling interval
 real(kind=vdp), parameter :: vsmRayStride   = 0.2_vdp ! min grid size * vsmRayStride = sampling interval

 integer, parameter :: vsmMaxColmap = 256

 integer, parameter :: vsmMaxVisKind = 15

 ! data
 integer, parameter :: vsmMaxScalar = 5
 integer, parameter :: vsmMaxVector = 5
 integer, parameter :: vsmMaxPtcl   = 2

 ! Octree
 integer, parameter :: vsmMaxOctLevel = 3
 integer, parameter :: vsmMaxOctNum   = 1 + 8 + 8*8 + 8*8*8
 integer, dimension(0:vsmMaxOctLevel) :: vsmOctNum= (/1, 1+8, 1+8+8*8, 1+8+8*8+8*8*8/)
 real, parameter :: vsmSkipAlpha = 0.0001

 !Transformation
 type vismo_transdata
   real(kind=vdp), dimension(3,3) :: matrix
   real(kind=vdp), dimension(3,3) :: invmatrix
   real(kind=vdp), dimension(3)   :: pos
 end type vismo_transdata

 ! For Vectorization :: SX-ACE, etc...
 integer, parameter :: vsmVecLen = 256
 integer, parameter :: vsmNoVecLen = 4
 integer, parameter :: vsmVecMaxOpenMP = 4
 integer, parameter :: vsmMaxNumArrows = vsmVecLen*8+1 ! max number of arrows per an image for vector machine
 integer, parameter :: vsmVecMaxOMP = 4
 integer, parameter :: vsmOMPVecLen = vsmVecLen*vsmVecMaxOpenMP
 
 ! YinYang
 real(kind=vdp), parameter :: vsmYinMinT =  0.25_vdp*3.1415926535897932384_vdp
 real(kind=vdp), parameter :: vsmYinMaxT = 0.75_vdp*3.1415926535897932384_vdp
 real(kind=vdp), parameter :: vsmYinMinP = -0.75_vdp*3.1415926535897932384_vdp
 real(kind=vdp), parameter :: vsmYinMaxP = 0.75_vdp*3.1415926535897932384_vdp
 integer, parameter :: vsmYin = 0
 integer, parameter :: vsmYang = 1

 !yyFrame
 integer, parameter :: vsmyyFrameRNum = 6
 integer, parameter :: vsmyyFrameLongNum = 28
 integer, parameter :: vsmyyFrameLatNum = 10

 integer, parameter :: vsmSpFrameLongNum = 36
 integer, parameter :: vsmSpFrameLatNum = 18

 integer, parameter :: vsmRSBNum = 6 ! more than 6
 integer, parameter :: vsmRSD = 3 ! more than 3

 ! structures
 !image
  type vismo_image
   integer :: width, height
   real, dimension(:), allocatable :: image
   real, dimension(:), allocatable :: image_r
   real(kind=vdp), dimension(3) :: topleft
   real(kind=vdp) :: dp
   real, dimension(3) :: bgcolor
   integer :: pxlsize, rimgsize

   integer, dimension(:), allocatable :: rsdata    ! pixn, in, out, color
   integer, dimension(:), allocatable :: rsdata_r
   character, dimension(:), allocatable :: imageUC
   logical :: BGC
  end type vismo_image

 ! data
 type vismo_scaldata
#ifndef FLOAT
  real(kind=vdp), dimension(:,:,:), pointer :: scal
#else
  real, dimension(:,:,:), pointer :: scal
#endif
  real(kind=vdp) :: minvalue, maxvalue
 end type vismo_scaldata

 type vismo_vectdata
#ifndef FLOAT
  real(kind=vdp), dimension(:,:,:), pointer :: vect_x
  real(kind=vdp), dimension(:,:,:), pointer :: vect_y
  real(kind=vdp), dimension(:,:,:), pointer :: vect_z
#else
  real, dimension(:,:,:), pointer :: vect_x
  real, dimension(:,:,:), pointer :: vect_y
  real, dimension(:,:,:), pointer :: vect_z
#endif
  real(kind=vdp) :: minvalue, maxvalue
 end type vismo_vectdata

 type vismo_ptdata
#ifndef FLOAT
  real(kind=vdp), dimension(:), pointer :: x
  real(kind=vdp), dimension(:), pointer :: y
  real(kind=vdp), dimension(:), pointer :: z
#else
  real, dimension(:), pointer :: x
  real, dimension(:), pointer :: y
  real, dimension(:), pointer :: z
#endif
  integer, dimension(:), pointer :: q ! q(m) == 0 : partcile(m) is drawn, q(m) /= 0: not drawn
end type vismo_ptdata

 type vismo_datainfo
   integer :: NumOfScal
   integer :: NumOfVect
   integer :: NumOfPtcl
 end type vismo_datainfo

 ! Coord
  !whole data
 type vismo_coord
  integer, dimension(3) :: nx ! number of toftal points
  real(kind=vdp), dimension(3) :: origin  !origin
  real(kind=vdp), dimension(3) :: cend    !max
  real(kind=vdp), dimension(3) :: dx
  real(kind=vdp), dimension(:), allocatable :: coordx, coordy, coordz
  logical :: uniform
  real(kind=vdp), dimension(3, 8) :: wdvertex
  integer, dimension(2, 12) :: lvnum
 end type vismo_coord

 !local data (distributed data)
 type vismo_localcoord
  integer, dimension(3) :: nx  ! local number of total points (with overlapped points)
  integer, dimension(3) :: lx0 ! local real origin's number
  integer, dimension(3) :: lnx ! local real number of points (without overlapped points)
  real(kind=vdp), dimension(3) :: lorigin  !local origin (without overlapped points)
  real(kind=vdp), dimension(3) :: lcend    !local max (without overlapped points)
  real(kind=vdp), dimension(3) :: origin  !local origin
  real(kind=vdp), dimension(3) :: cend    !local max
  real(kind=vdp), dimension(3) :: dx
  real(kind=vdp), dimension(:), allocatable :: coordx, coordy, coordz
  logical :: uniform
 end type vismo_localcoord

type vismo_coorddata
  integer :: YinYang
  logical :: YYZMode
  real(kind=vdp) :: MinRad, MaxRad, dr
end type vismo_coorddata

! Visualization Parameters
! scalar data visualization
 type vismo_isodata
   real(kind=vdp) :: iso_level
   integer :: nscal
   real :: alpha
   real, dimension(3) :: amb, dif, spec
   real :: shin
 end type vismo_isodata

 type vismo_slicedata
   real(kind=vdp), dimension(4) :: sc !a, b, c, d ! f(x,y,z) = ax+by+cz+d or radius
   integer :: nscal
   integer :: type ! 0:plane, 1:sphere 
   real :: alpha, amb, dif, spec, shin
 end type vismo_slicedata

 type vismo_volrendata
   integer :: nscal
   real :: amb, dif, spec, shin
 end type vismo_volrendata

 ! vector data visualization
 type vismo_strlinedata
   real(kind=vdp), dimension(3, vsmMaxSeed) :: seed
   integer :: num_seed
   integer, dimension(vsmMaxSeed) :: steps_remained
   integer :: nvec
   real, dimension(3) :: amb, dif, spec ! for monocolor lines
   real :: shin
   real(kind=vdp) :: radius
   logical :: monocolor
 end type vismo_strlinedata

 type vismo_arrowdata
   integer :: nvec
   real(kind=vdp), dimension(4) :: sc !a, b, c, d ! f(x,y,z) = ax+by+cz+d 
   real(kind=vdp) :: density
   real(kind=vdp) :: radius, length
 end type vismo_arrowdata

  ! Those are the common variables for any vis_kinds
 type vismo_interval
   real(kind=vdp) :: ray_interval ! for Ray-Casting
   real(kind=vdp) :: dh           ! for StreamLines' Runge-Kutta integration
 end type vismo_interval

 ! Those are the common variables for mpi

 ! Frame data
 type vismo_framedata
   logical :: insphere
   logical :: inspframe, outspframe
   logical :: inyyframe, outyyframe
   logical :: egyyframe
 end type vismo_framedata

 type vismo_mpi
   integer :: comm   ! communicator
   integer :: myrank ! rank
   integer :: psize
   integer :: omp_tn ! OpenMP threads
   integer :: omp_tnl1, omp_tnl2 ! OpenMP threads , level1, level2
 end type vismo_mpi

 ! visualization parameters 
 type vismo_visparameters
   ! isosurface
   type(vismo_isodata), dimension(vsmMaxIsosurf) :: isoparam
   integer :: iso_num
   logical :: iso_solid

   ! color slice
   type(vismo_slicedata), dimension(vsmMaxSlice) :: sliceparam
   integer :: slice_num
   logical :: slice_solid

   ! volume rendering
   type(vismo_volrendata), dimension(vsmMaxScalar) :: volparam
   integer :: volren_num

   ! stream line
   type(vismo_strlinedata), dimension(vsmMaxVector) :: slineparam
   integer :: sline_num

   ! arrows
   type(vismo_arrowdata), dimension(vsmMaxArrows) :: arrowparam
   integer :: arrow_num

   !lights
   integer :: light_num
   logical, dimension(vsmMaxLight) :: light_used

   !depthcue
   logical :: depthq

   type(vismo_framedata) :: frameparam
 end type vismo_visparameters


!Light
 type vismo_light
  real, dimension(3) :: pos
  real, dimension(3) :: direc

  real, dimension(3) :: anpos
  real, dimension(3) :: andirec

  real, dimension(3) :: amb, dif, spec
  logical :: para !parallel light or point light
 end type vismo_light

!Depth Cue
 type vismo_depthparam
  integer :: depth_kind ! 1:liner, 2:exp(-x), 3:exp(-x**2)
  real(kind=vdp), dimension(2) :: dist
 end type vismo_depthparam

! ColorMap
 type vismo_scolormap
  real(kind=vdp), dimension(vsmMaxColmap) :: value
  real(kind=vdp) :: dv
  real, dimension(4, vsmMaxColmap) :: color
  integer :: colnum
  logical :: set, eqspace
 end type vismo_scolormap

 type vismo_vcolormap
  real(kind=vdp), dimension(vsmMaxColmap) :: value
  real(kind=vdp) :: dv
  real, dimension(3, vsmMaxColmap) :: color
  integer :: colnum
  logical :: set, eqspace
 end type vismo_vcolormap

! Camera
 ! pos, front, up are required
 type vismo_camera
  real(kind=vdp), dimension(3) :: pos
  real(kind=vdp), dimension(3) :: front, up, right
  real(kind=vdp) :: fovy
  real(kind=vdp) :: near, far
  real(kind=vdp) :: width
  logical :: pers
  real(kind=vdp) :: dp ! 20200219
  !---Iso PC Data
  logical :: calculated
  real(kind=vdp), dimension(3, 6) :: vertex
  real(kind=vdp), dimension(3, 3) :: l
  real(kind=vdp), dimension(4, 6) :: plane
  integer, dimension(2) :: swidth, sheight
  real(kind=vdp), dimension(2) :: dph
 end type vismo_camera

! Binary Swap
 type vismo_bswap
   logical :: possible, bswap
   integer :: numstage
   integer :: myrank
   integer :: psize
   integer, dimension(:), allocatable :: partner
   logical, dimension(:), allocatable :: above
 end type vismo_bswap


 end module vsmConst

