module vsmDepthQ
 use vsmConst
 use vsmFunction
 use vsmImage

 implicit none
 private
 type(vismo_depthparam), save, public :: vsmDepthParam
 
 public vsmDepthQFactor, vsmSetDepthParam, vsmDepthCueing, vsmDepthCueingImage

 contains

 subroutine vsmInitDepthParam
  vsmDepthParam%depth_kind = -1
  vsmDepthParam%dist(1:2) = 0.0_vdp
 end subroutine vsmInitDepthParam

 subroutine vsmSetDepthParam(dkind, dist)
  integer, intent(in) :: dkind
  real(kind=vdp), dimension(2), intent(in) :: dist

  if(dkind < 1 .or. dkind > 3) call vsmError("depth kind is bizarre")
  vsmDepthParam%depth_kind = dkind
  vsmDepthParam%dist(1:2) = dist(1:2)
 end subroutine vsmSetDepthParam

 function vsmDepthQFactor(dist)
  real(kind=vdp), intent(in) :: dist
  real :: vsmDepthQFactor
  real(kind=vdp) :: c
  real :: tmp

  select case (vsmDepthParam%depth_kind)
    case (1)
     !linear
     tmp = real((vsmDepthParam%dist(2) - dist) / (vsmDepthParam%dist(2) - vsmDepthParam%dist(1)))
     tmp = vsmScalef01(tmp)
    case (2)
     c = dist*vsmDepthParam%dist(1)
     tmp = real(exp(-c))
    case (3)
     c = dist*dist*vsmDepthParam%dist(1)*vsmDepthParam%dist(1)
     tmp = real(exp(-c))
    case default
     call vsmError("strange depth cue kind")
  end select
  vsmDepthQFactor = vsmScaleF01(tmp)
 end function vsmDepthQFactor

 subroutine vsmDepthCueing(dist, lumi)
  real(kind=vdp), intent(in) :: dist
  real, dimension(3), intent(inout) :: lumi
  real(kind=vdp) :: c
  real :: tmp, tmp2

  select case (vsmDepthParam%depth_kind)
    case (1)
     !linear
     tmp = real((vsmDepthParam%dist(2) - dist) / (vsmDepthParam%dist(2) - vsmDepthParam%dist(1))); 
    case (2)
     c = dist*vsmDepthParam%dist(1)
     tmp = real(exp(-c))
    case (3)
     c = dist*dist*vsmDepthParam%dist(1)*vsmDepthParam%dist(1)
     tmp = real(exp(-c))
    case default
     call vsmError("strange depth cue kind")
  end select
  tmp2 = vsmScaleF01(tmp)

  if(vsmImageData%BGC) then
    lumi(1:3) = tmp2*lumi(1:3) + (1.0-tmp2)*vsmImageData%bgcolor(1:3)
  else
    lumi(1:3) = tmp2*lumi(1:3)
  end if

 end subroutine vsmDepthCueing

 subroutine vsmDepthCueingImage
  integer :: i, j
  real(kind=vdp) :: c
  real :: tmp, dist
  real, dimension(3) :: lumi
  integer :: wdh

  if(vsmDepthParam%depth_kind < 1 .or. vsmDepthParam%depth_kind > 3) return

!$omp parallel do private(i,j,wdh,dist, lumi, tmp, c) schedule(dynamic, 1)
  do j=1, vsmImageData%height
    do i=1, vsmImageData%width
      wdh = (i-1 + (j-1)*vsmImageData%width)*5
      dist = vsmImageData%image(wdh+5)
      lumi(1:3) = vsmImageData%image(wdh+1:wdh+3)

      select case (vsmDepthParam%depth_kind)
        case (1)
        !linear
         tmp = real((vsmDepthParam%dist(2) - dist) / (vsmDepthParam%dist(2) - vsmDepthParam%dist(1))); 
        case (2)
         c = dist*vsmDepthParam%dist(1)
         tmp = real(exp(-c))
        case (3)
         c = dist*dist*vsmDepthParam%dist(1)*vsmDepthParam%dist(1)
         tmp = real(exp(-c))
        case default
         call vsmError("strange depth cue kind")
      end select

      if(tmp > 1.0) then
        tmp = 1.0
      else if(tmp < 0.0) then
        tmp = 0.0
      end if

      if(vsmImageData%BGC) then
        lumi(1:3) = tmp*lumi(1:3) + (1.0-tmp)*vsmImageData%bgcolor(1:3)
      else
        lumi(1:3) = tmp*lumi(1:3)
      end if
      vsmImageData%image(wdh+1:wdh+3) = lumi(1:3)
    end do
  end do
 end subroutine vsmDepthCueingImage

end module vsmDepthQ

