module vsmFrame
 use vsmConst
 use vsmFunction
 use vsmParam
 use vsmImage
 use vsmColor
 use vsmSolidDrawer
 use vsmCameraData
 use vsmCoordinates
! use vsmDepthQ
 use vsmLightData

 implicit none
 private

 logical :: bon ! blend on
 integer, save :: framewidth
 public vsmDrawinnerSphere, vsmDrawSphereFramePara, vsmDrawYinYangFrameAPara, vsmDrawYinYangFrameBPara
 public vsmInitFrame, vsmSetFrameWidth, vsmGetFrameWidth
 contains

 subroutine vsmInitFrame
   framewidth = 1
 end subroutine vsmInitFrame
 
 subroutine vsmGetFrameWidth(width)
   integer, intent(out) :: width
   width = framewidth
 end subroutine vsmGetFrameWidth
 
 subroutine vsmSetFrameWidth(width)
   integer, intent(in) :: width
   if(width <= 0) then
      call vsmLogPrint("frame width should be positive ", width)
      return
   end if
   framewidth = width
 end subroutine vsmSetFrameWidth

 function vsmCheckFPix(pix)
   integer, dimension(2), intent(in) :: pix
   logical :: vsmCheckFPix

   if(pix(1) < 1 .or. pix(1) > vsmImageData%width .or. &
        pix(2) < 1 .or. pix(2) > vsmImageData%height) then
     vsmCheckFPix = .false.
   else
     vsmCheckFPix = .true.
   end if
 end function vsmCheckFPix

 subroutine vsmDrawInnerSphere(cam_num, vis_num)
   real(kind=vdp) :: radius
   integer, intent(in) :: cam_num, vis_num

   integer :: i,j
   integer, dimension(2) :: px
   real(kind=vdp), dimension(3) :: epos, topleft, cpos
   real(kind=vdp), dimension(3) :: img_right, img_down
   real(kind=vdp), dimension(3) :: pixp, direc
!   real(kind=vdp) :: A, B, C, D, t, t1, t2, dist
   real(kind=vdp) :: B, C, D, t, t1, t2, dist
   real(kind=vdp), dimension(3) :: cr_sph, fpos ! cross point 
   real, dimension(3) :: lumi, normal
!   real :: dq_factor

   radius = vsmCoordData%MinRad
   !call vsmLogPrint("-- raycast sphere")
   ! translation :: the center of the sphere moves to the origin
   cpos(1:3) = vsmCamera(cam_num)%pos(1:3)
   epos(1:3) = cpos(1:3)
   topleft(1:3) = vsmImageData%topleft(1:3)
   img_right(1:3) = vsmCamera(cam_num)%right(1:3)
   img_down(1:3) = -vsmCamera(cam_num)%up(1:3)

!$omp parallel do private(i,j,pixp,direc,b,c,d,t,cr_sph,t1,t2,normal,fpos,dist,lumi,px)
   do j = 1, vsmImageData%height
     do i = 1, vsmImageData%width
       pixp(1:3) = topleft(1:3) + (i-1)*vsmImageData%dp*img_right(1:3) + &
                   (j-1)*vsmImageData%dp*img_down(1:3)
    !   call vsmLogPrint("realpos:pixel ", pixp)
       if(vsmCamera(cam_num)%pers) then
         direc(1:3) = pixp(1:3) - epos(1:3)
       else
         direc(1:3) = vsmCamera(cam_num)%front(1:3)
       end if
       call vsmNormVec(direc)
       !A = 1.0_vdp  ! dot_product(direc, direc)
       B = dot_product(direc, pixp)
       C = dot_product(pixp, pixp) - radius**2
       ! determinant
       D = B**2 - C
       if(D < 0.0_vdp) then
         cycle
       else if(D == 0.0_vdp) then
         t = -B
         if(t < 0.0_vdp) cycle
           cr_sph(1:3) = pixp(1:3) + direc(1:3)*t
       else if(D > 0.0_vdp) then
         t1 = (-B+sqrt(D))
         t2 = (-B-sqrt(D))
         if(t1 >= 0.0_vdp .and. t2 >= 0.0_vdp) then
           t = min(t1,t2)
         else if(t1 < 0.0_vdp .and. t2 >= 0.0_vdp) then
           t = t2
         else if(t1 >= 0.0_vdp .and. t2 < 0.0_vdp) then
           t = t1
         else 
           cycle
         end if
         cr_sph(1:3) = pixp(1:3) + direc(1:3)*t
       end if

       normal(1:3) = real(cr_sph(1:3))
       call vsmNormVecF(normal)
       fpos(1:3) = cr_sph(1:3)

       if(vsmCamera(cam_num)%pers) then
           if(vsmCoordData%YinYang == vsmYin) then
              call vsmCalcColorv(vsmFSAmb, vsmFSDif, vsmFSspec, vsmFSshin, cpos, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            else
              call vsmCalcColorvAn(vsmFSamb, vsmFSdif, vsmFSspec, vsmFSshin, cpos, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            end if
            dist = sqrt(dot_product(cr_sph-epos, cr_sph-epos))
        else
           if(vsmCoordData%YinYang == vsmYin) then
              call vsmCalcColorvOrtho(vsmFSamb, vsmFSdif, vsmFSspec, vsmFSshin, direc, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            else
              call vsmCalcColorvOrthoAn(vsmFSamb, vsmFSdif, vsmFSspec, vsmFSshin, direc, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            end if
            dist = t + vsmCamera(cam_num)%near
        end if
        
!       if(vsmVisParam(vis_num)%depthq) then
!         dq_factor = vsmDepthQFactor(dist)
!         lumi(1:3) = dq_factor*lumi(1:3)
!       end if

       if(lumi(1) < 0.0 .or. lumi(2) < 0.0 .or. lumi(3) < 0.0) print *, "raycast sphere"
       px(1) = i; px(2) = j
       call vsmPaintPixel3(px, lumi, dist)
     end do
   end do 

 end subroutine vsmDrawInnerSphere

subroutine vsmDrawSphereFrame(cam_num, blend, inout)
   integer, intent(in) :: cam_num
   logical, intent(in) :: blend
   integer, intent(in) :: inout
   real(kind=vdp), dimension(3, 2) :: pos
   real(kind=vdp) :: da, radius, dda, phi, theta, theta1, theta2, phi1, phi2, r1, r2, z
   integer :: i, j, nl!, k, nt, np

   call vsmLogPrint("draw inner sphere")
   bon = blend
   da = 2.0_vdp*pi/vsmspFrameLongNum

   if(inout == 0) then
     radius = vsmCoordData%MinRad !- vsmCoordData%dr
   else
     radius = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp ! for safety
   end if
   !print *,"minrad = ", radius

   nl = 180+1
   dda = pi/nl
   do j=1, vsmSpFrameLongNum
     phi = (j-1)*da
     do i=1, nl
        theta1 = (i-1)*dda
        theta2 = i*dda
        if(theta1 >= pi) theta1 = pi
        if(theta2 >= pi) theta2 = pi
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
     end do
   end do

   dda = 2.0_vdp*pi/nl
   da = pi/vsmSpFrameLatNum
   do j=1, vsmSpFrameLatNum
     theta = (j-1)*da
     nl = 360+1
     if(j==1 .or. theta >= pi) cycle
     z = radius*cos(theta)
     do i=1, nl
        phi1 = (i-1)*dda
        phi2 = i*dda
        if(phi1 >= 2.0_vdp*pi) phi1 = 2.0_vdp*pi
        if(phi2 >= 2.0_vdp*pi) phi2 = 2.0_vdp*pi
        r1 = radius * sin(theta)

        pos(1,1) = r1*cos(phi1)
        pos(2,1) = r1*sin(phi1)
        pos(3,1) = z

        pos(1,2) = r1*cos(phi2)
        pos(2,2) = r1*sin(phi2)
        pos(3,2) = z
        call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
     end do
   end do
 end subroutine vsmDrawSphereFrame

subroutine vsmDrawSphereFramePara(cam_num, blend, inout, rank, psize)
   integer, intent(in) :: cam_num
   logical, intent(in) :: blend
   integer, intent(in) :: inout, rank, psize
   real(kind=vdp), dimension(3, 2) :: pos
   real(kind=vdp) :: da, radius, dda, phi, theta, theta1, theta2, phi1, phi2, r1, r2, z
   integer :: i, j, nl!, k, nt, np
   integer :: st, ed, chnk

   call vsmLogPrint("draw inner sphere, psize=", psize)
   call vsmLogPrint("draw inner sphere, rank=", rank)

   bon = blend

   if(inout == 0) then
     radius = vsmCoordData%MinRad !- vsmCoordData%dr
   else
     radius = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp
   end if

   da = 2.0_vdp*pi/vsmspFrameLongNum
   nl = 180+1
   if(nl < psize) then
     nl = psize + 1
   end if
   dda = pi/nl
   chnk = nl/psize
   st = chnk*rank + 1
   ed = chnk*(rank+1)
   if(rank == psize-1) ed = nl

!   call vsmLogPrint("draw inner sphere, st=", st)
!   call vsmLogPrint("draw inner sphere, ed=", ed)

   do j=1, vsmSpFrameLongNum
     phi = (j-1)*da
     do i=st, ed
        theta1 = (i-1)*dda
        theta2 = i*dda
        if(theta1 >= pi) theta1 = pi
        if(theta2 >= pi) theta2 = pi
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
     end do
   end do

   da = pi/vsmSpFrameLatNum
   nl = 360+1
   if(nl < psize) then
     nl = psize + 1
   end if
   dda = 2.0_vdp*pi/nl
   chnk = nl/psize
   st = chnk*rank + 1
   ed = chnk*(rank+1)
   if(rank == psize-1) ed = nl

!   call vsmLogPrint("draw inner sphere, st=", st)
!   call vsmLogPrint("draw inner sphere, ed=", ed)

   do j=1, vsmSpFrameLatNum
     theta = (j-1)*da
     if(j==1 .or. theta >= pi) cycle
     z = radius*cos(theta)
     do i=st, ed
        phi1 = (i-1)*dda
        phi2 = i*dda
        if(phi1 >= 2.0_vdp*pi) phi1 = 2.0_vdp*pi
        if(phi2 >= 2.0_vdp*pi) phi2 = 2.0_vdp*pi
        r1 = radius * sin(theta)

        pos(1,1) = r1*cos(phi1)
        pos(2,1) = r1*sin(phi1)
        pos(3,1) = z

        pos(1,2) = r1*cos(phi2)
        pos(2,2) = r1*sin(phi2)
        pos(3,2) = z
        call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
     end do
   end do
 end subroutine vsmDrawSphereFramePara

subroutine vsmDrawYinYangFrameA(cam_num, blend, inout, yy)
   integer, intent(in) :: cam_num
   logical, intent(in) :: blend
   integer, intent(in) :: inout, yy ! yin:yy=0, yng:yy=1. yinyang = 2
   real(kind=vdp), dimension(3, 2) :: pos
   real(kind=vdp), dimension(3, 2) :: apos
   real(kind=vdp) :: da, radius, dda, phi, theta, theta1, theta2, phi1, phi2, r1, r2, z!,rad1,rad2
   integer :: i, j, nl

   call vsmLogPrint("draw inner sphere")
   bon = blend
   da = 270.0_vdp/(vsmyyFrameLongNum-1)
   da = vsmDeg2Rad(da)
!   dl = vsmImageData%dp

   if(inout == 0) radius = vsmCoordData%MinRad !- vsmCoordData%dr
   if(inout == 1) radius = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp !- vsmCoordData%dr

   nl = 90+1
   dda = pi/nl/2.0_vdp
!   longlat = 0
   do j=1, vsmyyFrameLongNum
     phi = (j-1)*da - pi/4.0_vdp*3.0_vdp
!     print *,"phi = ", phi, da
     do i=1, nl
        theta1 = (i-1)*dda + pi/4.0_vdp 
        theta2 = i*dda + pi/4.0_vdp
        if(theta1 >= pi/4.0_vdp*3.0_vdp) theta1 = pi/4.0_vdp*3.0_vdp
        if(theta2 >= pi/4.0_vdp*3.0_vdp) theta2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy==2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)

        if(yy==1 .or. yy==2) then 
           apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
           apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
           call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
   end do

   nl = 270+1
   dda = pi/nl*1.5_vdp
   da = pi/2.0_vdp/(vsmyyFrameLatNum-1)
!   longlat = 1
   do j=1, vsmyyFrameLatNum
     theta = (j-1)*da+pi/4.0_vdp
     nl = 270+1
     if(theta >= pi/4.0_vdp*3.0_vdp) theta = pi/4.0_vdp*3.0_vdp
     z = radius*cos(theta)
     do i=1, nl
!        print *,"i=", i
        phi1 = (i-1)*dda - pi/4.0_vdp*3.0_vdp
        phi2 = i*dda - pi/4.0_vdp*3.0_vdp
        if(phi1 >= pi/4.0_vdp*3.0_vdp) phi1 = pi/4.0_vdp*3.0_vdp
        if(phi2 >= pi/4.0_vdp*3.0_vdp) phi2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta)
!        r2 = radius*sin(theta)
        pos(1,1) = r1*cos(phi1)
        pos(2,1) = r1*sin(phi1)
        pos(3,1) = z

        pos(1,2) = r1*cos(phi2)
        pos(2,2) = r1*sin(phi2)
        pos(3,2) = z
        if(yy==0 .or. yy==2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)

        if(yy==1 .or. yy==2) then 
           apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
           apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
           call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
   end do

  end subroutine vsmDrawYinYangFrameA

subroutine vsmDrawYinYangFrameAPara(cam_num, blend, inout, yy, rank, psize)
   integer, intent(in) :: cam_num
   logical, intent(in) :: blend
   integer, intent(in) :: inout, yy ! yin:yy=0, yng:yy=1. yinyang = 2
   integer, intent(in) :: rank, psize
   real(kind=vdp), dimension(3, 2) :: pos
   real(kind=vdp), dimension(3, 2) :: apos
   real(kind=vdp) :: da, radius, dda, phi, theta, theta1, theta2, phi1, phi2, r1, r2, z!,rad1,rad2
   integer :: i, j, nl,st,ed,chnk

   call vsmLogPrint("draw inner sphere")
   bon = blend
   da = 270.0_vdp/(vsmyyFrameLongNum-1)
   da = vsmDeg2Rad(da)

   if(inout == 0) radius = vsmCoordData%MinRad !- vsmCoordData%dr
   if(inout == 1) radius = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp !- vsmCoordData%dr

   nl = 90+1
   if(nl < psize) then
     nl = psize + 1
   end if
   dda = pi/nl/2.0_vdp
   chnk = nl/psize
   st = chnk*rank + 1
   ed = chnk*(rank+1)
   if(rank == psize-1) ed = nl
   do j=1, vsmyyFrameLongNum
     phi = (j-1)*da - pi/4.0_vdp*3.0_vdp
!     print *,"phi = ", phi, da
     do i=st, ed
        theta1 = (i-1)*dda + pi/4.0_vdp 
        theta2 = i*dda + pi/4.0_vdp
        if(theta1 >= pi/4.0_vdp*3.0_vdp) theta1 = pi/4.0_vdp*3.0_vdp
        if(theta2 >= pi/4.0_vdp*3.0_vdp) theta2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy==2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)

        if(yy==1 .or. yy==2) then 
           apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
           apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
           call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
   end do

   nl = 270+1
   if(nl < psize) then
     nl = psize + 1
   end if
   dda = pi/nl*1.5_vdp
   chnk = nl/psize
   st = chnk*rank + 1
   ed = chnk*(rank+1)
   if(rank == psize-1) ed = nl
   da = pi/2.0_vdp/(vsmyyFrameLatNum-1)
   do j=1, vsmyyFrameLatNum
     theta = (j-1)*da+pi/4.0_vdp
     nl = 270+1
     if(theta >= pi/4.0_vdp*3.0_vdp) theta = pi/4.0_vdp*3.0_vdp
     z = radius*cos(theta)
     do i=st, ed
!        print *,"i=", i
        phi1 = (i-1)*dda - pi/4.0_vdp*3.0_vdp
        phi2 = i*dda - pi/4.0_vdp*3.0_vdp
        if(phi1 >= pi/4.0_vdp*3.0_vdp) phi1 = pi/4.0_vdp*3.0_vdp
        if(phi2 >= pi/4.0_vdp*3.0_vdp) phi2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta)
!        r2 = radius*sin(theta)
        pos(1,1) = r1*cos(phi1)
        pos(2,1) = r1*sin(phi1)
        pos(3,1) = z

        pos(1,2) = r1*cos(phi2)
        pos(2,2) = r1*sin(phi2)
        pos(3,2) = z
        if(yy==0 .or. yy==2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy==2) then 
           apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
           apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
           call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
   end do
  end subroutine vsmDrawYinYangFrameAPara

subroutine vsmDrawYinYangFrameB(cam_num, blend, yy)
   integer, intent(in) :: cam_num
   logical, intent(in) :: blend
   integer, intent(in) :: yy
   real(kind=vdp), dimension(3, 2) :: pos
   real(kind=vdp), dimension(3, 2) :: apos
   real(kind=vdp) :: da, radius, dda, phi, theta, theta1, theta2, phi1, phi2, r1, r2, z,rad1,rad2
   integer :: i, j, nl

   call vsmLogPrint("draw inner sphere")
   bon = blend
   da = 1.5_vdp*pi/(vsmyyFrameLongNum-1)

   nl = 50
   dda = (vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp - vsmCoordData%MinRad)/dble(nl)
   do j=1, vsmyyFrameLongNum
     phi = (j-1)*da - pi/4.0_vdp*3.0_vdp
     do i=1, nl
        theta = pi/4.0_vdp
        r1 = (i-1)*dda + vsmCoordData%MinRad
        r2 = i*dda + vsmCoordData%MinRad
        if(r1 >= vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp) r1 = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp
        if(r2 >= vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp) r2 = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp
        rad1 = r1 * sin(theta)
        rad2 = r2 * sin(theta)
        pos(1,1) = rad1*cos(phi)
        pos(2,1) = rad1*sin(phi)
        pos(3,1) = r1*cos(theta)

        pos(1,2) = rad2*cos(phi)
        pos(2,2) = rad2*sin(phi)
        pos(3,2) = r2*cos(theta)
        if(yy==0 .or. yy==2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if

        theta = pi/4.0_vdp*3.0_vdp

        pos(1,1) = rad1*cos(phi)
        pos(2,1) = rad1*sin(phi)
        pos(3,1) = r1*cos(theta)

        pos(1,2) = rad2*cos(phi)
        pos(2,2) = rad2*sin(phi)
        pos(3,2) = r2*cos(theta)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
 end do

   da = 0.5_vdp*pi/(vsmyyFrameLatNum-1)
   do j=1, vsmyyFrameLatNum
     theta = (j-1)*da+pi/4.0_vdp
     if(theta >= pi/4.0_vdp*3.0_vdp) theta = pi/4.0_vdp*3.0_vdp
     z = radius*cos(theta)
     do i=1, nl
        phi =  - pi/4.0_vdp*3.0_vdp
        r1 = (i-1)*dda + vsmCoordData%MinRad
        r2 = i*dda + vsmCoordData%MinRad
        if(r1 >= vsmCoordData%MaxRad) r1 = vsmCoordData%MaxRad
        if(r2 >= vsmCoordData%MaxRad) r2 = vsmCoordData%MaxRad
        rad1 = r1 * sin(theta)
        rad2 = r2 * sin(theta)
        pos(1,1) = rad1*cos(phi)
        pos(2,1) = rad1*sin(phi)
        pos(3,1) = r1*cos(theta)

        pos(1,2) = rad2*cos(phi)
        pos(2,2) = rad2*sin(phi)
        pos(3,2) = r2*cos(theta)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if

        phi = pi/4.0_vdp*3.0_vdp
        pos(1,1) = rad1*cos(phi)
        pos(2,1) = rad1*sin(phi)
        pos(3,1) = r1*cos(theta)

        pos(1,2) = rad2*cos(phi)
        pos(2,2) = rad2*sin(phi)
        pos(3,2) = r2*cos(theta)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
 end do

!     print *,"phi = ", phi, da
   da = (vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp - vsmCoordData%MinRad)/(vsmyyFrameRNum-1)
   nl = 50+1
   dda = pi/nl/2.0_vdp
   do j=1, vsmyyFrameRNum
     radius = vsmCoordData%MinRad + da*(j-1)
     if(radius >= vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp) radius = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp
     do i=1, nl
        theta1 = (i-1)*dda + pi/4.0_vdp 
        theta2 = i*dda + pi/4.0_vdp
        if(theta1 >= pi/4.0_vdp*3.0_vdp) theta1 = pi/4.0_vdp*3.0_vdp
        if(theta2 >= pi/4.0_vdp*3.0_vdp) theta2 = pi/4.0_vdp*3.0_vdp
        phi =  - pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if

        phi =  pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do

   nl = 90+1
   dda = pi/nl/2.0_vdp
     do i=1, nl
        phi = - pi/4.0_vdp*3.0_vdp
        theta1 = (i-1)*dda + pi/4.0_vdp 
        theta2 = i*dda + pi/4.0_vdp
        if(theta1 >= pi/4.0_vdp*3.0_vdp) theta1 = pi/4.0_vdp*3.0_vdp
        if(theta2 >= pi/4.0_vdp*3.0_vdp) theta2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy == 2)  call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2)  then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if

        phi =  pi/4.0_vdp*3.0_vdp

        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do

   nl = 270+1
   dda = pi/nl*1.5_vdp
   nl = 270+1
    theta = pi/4.0_vdp
     z = radius*cos(theta)
     do i=1, nl
        phi1 = (i-1)*dda - pi/4.0_vdp*3.0_vdp
        phi2 = i*dda - pi/4.0_vdp*3.0_vdp
        if(phi1 >= pi/4.0_vdp*3.0_vdp) phi1 = pi/4.0_vdp*3.0_vdp
        if(phi2 >= pi/4.0_vdp*3.0_vdp) phi2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta)
        pos(1,1) = r1*cos(phi1)
        pos(2,1) = r1*sin(phi1)
        pos(3,1) = z

        pos(1,2) = r1*cos(phi2)
        pos(2,2) = r1*sin(phi2)
        pos(3,2) = z
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do

    theta = pi/4.0_vdp*3.0_vdp
     z = radius*cos(theta)
     do i=1, nl
!        print *,"i=", i
        phi1 = (i-1)*dda - pi/4.0_vdp*3.0_vdp
        phi2 = i*dda - pi/4.0_vdp*3.0_vdp
        if(phi1 >= pi/4.0_vdp*3.0_vdp) phi1 = pi/4.0_vdp*3.0_vdp
        if(phi2 >= pi/4.0_vdp*3.0_vdp) phi2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta)
!        r2 = radius*sin(theta)
        pos(1,1) = r1*cos(phi1)
        pos(2,1) = r1*sin(phi1)
        pos(3,1) = z

        pos(1,2) = r1*cos(phi2)
        pos(2,2) = r1*sin(phi2)
        pos(3,2) = z
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)

        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
   end do

  end subroutine vsmDrawYinYangFrameB

subroutine vsmDrawYinYangFrameBPara(cam_num, blend, yy, rank, psize)
   integer, intent(in) :: cam_num
   logical, intent(in) :: blend
   integer, intent(in) :: yy, rank, psize
   real(kind=vdp), dimension(3, 2) :: pos
   real(kind=vdp), dimension(3, 2) :: apos
   real(kind=vdp) :: da, radius, dda, phi, theta, theta1, theta2, phi1, phi2, r1, r2, z,rad1,rad2
   integer :: i, j, nl, st,ed,chnk

   call vsmLogPrint("draw inner sphere")
   bon = blend
   da = 1.5_vdp*pi/(vsmyyFrameLongNum-1)

   nl = 50
   if(nl < psize) then
     nl = psize + 1
   end if
   dda = (vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp - vsmCoordData%MinRad)/dble(nl)
   chnk = nl/psize
   st = chnk*rank + 1
   ed = chnk*(rank+1)
   if(rank == psize-1) ed = nl
   do j=1, vsmyyFrameLongNum
     phi = (j-1)*da - pi/4.0_vdp*3.0_vdp
     do i=st, ed
        theta = pi/4.0_vdp
        r1 = (i-1)*dda + vsmCoordData%MinRad
        r2 = i*dda + vsmCoordData%MinRad
        if(r1 >= vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp) r1 = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp
        if(r2 >= vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp) r2 = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp
        rad1 = r1 * sin(theta)
        rad2 = r2 * sin(theta)
        pos(1,1) = rad1*cos(phi)
        pos(2,1) = rad1*sin(phi)
        pos(3,1) = r1*cos(theta)

        pos(1,2) = rad2*cos(phi)
        pos(2,2) = rad2*sin(phi)
        pos(3,2) = r2*cos(theta)
        if(yy==0 .or. yy==2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if

        theta = pi/4.0_vdp*3.0_vdp

        pos(1,1) = rad1*cos(phi)
        pos(2,1) = rad1*sin(phi)
        pos(3,1) = r1*cos(theta)

        pos(1,2) = rad2*cos(phi)
        pos(2,2) = rad2*sin(phi)
        pos(3,2) = r2*cos(theta)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
 end do

   da = 0.5_vdp*pi/(vsmyyFrameLatNum-1)
   do j=1, vsmyyFrameLatNum
     theta = (j-1)*da+pi/4.0_vdp
     if(theta >= pi/4.0_vdp*3.0_vdp) theta = pi/4.0_vdp*3.0_vdp
     z = radius*cos(theta)
     do i=st, ed
        phi =  - pi/4.0_vdp*3.0_vdp
        r1 = (i-1)*dda + vsmCoordData%MinRad
        r2 = i*dda + vsmCoordData%MinRad
        if(r1 >= vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp) r1 = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp
        if(r2 >= vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp) r2 = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp
        rad1 = r1 * sin(theta)
        rad2 = r2 * sin(theta)
        pos(1,1) = rad1*cos(phi)
        pos(2,1) = rad1*sin(phi)
        pos(3,1) = r1*cos(theta)

        pos(1,2) = rad2*cos(phi)
        pos(2,2) = rad2*sin(phi)
        pos(3,2) = r2*cos(theta)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if

        phi = pi/4.0_vdp*3.0_vdp
        pos(1,1) = rad1*cos(phi)
        pos(2,1) = rad1*sin(phi)
        pos(3,1) = r1*cos(theta)

        pos(1,2) = rad2*cos(phi)
        pos(2,2) = rad2*sin(phi)
        pos(3,2) = r2*cos(theta)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
 end do

!     print *,"phi = ", phi, da
   da = (vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp - vsmCoordData%MinRad)/(vsmyyFrameRNum-1)
   nl = 50+1
   if(nl < psize) then
     nl = psize + 1
   end if
   dda = pi/nl/2.0_vdp
   chnk = nl/psize
   st = chnk*rank + 1
   ed = chnk*(rank+1)
   if(rank == psize-1) ed = nl
   do j=1, vsmyyFrameRNum
     radius = vsmCoordData%MinRad + da*(j-1)
     if(radius >= vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp) radius = vsmCoordData%MaxRad + vsmCoordData%dr*0.5_vdp
     do i=st, ed
        theta1 = (i-1)*dda + pi/4.0_vdp 
        theta2 = i*dda + pi/4.0_vdp
        if(theta1 >= pi/4.0_vdp*3.0_vdp) theta1 = pi/4.0_vdp*3.0_vdp
        if(theta2 >= pi/4.0_vdp*3.0_vdp) theta2 = pi/4.0_vdp*3.0_vdp
        phi =  - pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if

        phi =  pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do

   nl = 90+1
   if(nl < psize) then
     nl = psize + 1
   end if
   dda = pi/nl/2.0_vdp
   chnk = nl/psize
   st = chnk*rank + 1
   ed = chnk*(rank+1)
   if(rank == psize-1) ed = nl
     do i=st, ed
        phi = - pi/4.0_vdp*3.0_vdp
        theta1 = (i-1)*dda + pi/4.0_vdp 
        theta2 = i*dda + pi/4.0_vdp
        if(theta1 >= pi/4.0_vdp*3.0_vdp) theta1 = pi/4.0_vdp*3.0_vdp
        if(theta2 >= pi/4.0_vdp*3.0_vdp) theta2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta1)
        r2 = radius*sin(theta2)
        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy == 2)  call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2)  then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if

        phi =  pi/4.0_vdp*3.0_vdp

        pos(1,1) = r1*cos(phi)
        pos(2,1) = r1*sin(phi)
        pos(3,1) = radius*cos(theta1)

        pos(1,2) = r2*cos(phi)
        pos(2,2) = r2*sin(phi)
        pos(3,2) = radius*cos(theta2)
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do

   nl = 270+1
   if(nl < psize) then
     nl = psize + 1
   end if
   dda = pi/nl*1.5_vdp
   chnk = nl/psize
   st = chnk*rank + 1
   ed = chnk*(rank+1)
   if(rank == psize-1) ed = nl

    theta = pi/4.0_vdp
     z = radius*cos(theta)
     do i=st, ed
        phi1 = (i-1)*dda - pi/4.0_vdp*3.0_vdp
        phi2 = i*dda - pi/4.0_vdp*3.0_vdp
        if(phi1 >= pi/4.0_vdp*3.0_vdp) phi1 = pi/4.0_vdp*3.0_vdp
        if(phi2 >= pi/4.0_vdp*3.0_vdp) phi2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta)
        pos(1,1) = r1*cos(phi1)
        pos(2,1) = r1*sin(phi1)
        pos(3,1) = z

        pos(1,2) = r1*cos(phi2)
        pos(2,2) = r1*sin(phi2)
        pos(3,2) = z
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)
        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do

    theta = pi/4.0_vdp*3.0_vdp
     z = radius*cos(theta)
     do i=st, ed
!        print *,"i=", i
        phi1 = (i-1)*dda - pi/4.0_vdp*3.0_vdp
        phi2 = i*dda - pi/4.0_vdp*3.0_vdp
        if(phi1 >= pi/4.0_vdp*3.0_vdp) phi1 = pi/4.0_vdp*3.0_vdp
        if(phi2 >= pi/4.0_vdp*3.0_vdp) phi2 = pi/4.0_vdp*3.0_vdp
        r1 = radius * sin(theta)
!        r2 = radius*sin(theta)
        pos(1,1) = r1*cos(phi1)
        pos(2,1) = r1*sin(phi1)
        pos(3,1) = z

        pos(1,2) = r1*cos(phi2)
        pos(2,2) = r1*sin(phi2)
        pos(3,2) = z
        if(yy==0 .or. yy == 2) call vsmDrawSphereLine(pos(:,1), pos(:,2), cam_num)

        if(yy==1 .or. yy == 2) then
          apos(1,1) = -pos(1,1); apos(2,1) = pos(3,1); apos(3,1) = pos(2,1)
          apos(1,2) = -pos(1,2); apos(2,2) = pos(3,2); apos(3,2) = pos(2,2)
          call vsmDrawSphereLine(apos(:,1), apos(:,2), cam_num)
        end if
     end do
   end do

  end subroutine vsmDrawYinYangFrameBPara


 subroutine vsmDrawSphereLine(pos1, pos2, cam_num)
   real(kind=vdp), dimension(3), intent(in) :: pos1, pos2
!   real(kind=vdp), intent(in) :: phiz
   integer, intent(in) :: cam_num
   logical :: cross1, cross2
   integer, dimension(2) :: pix1, pix2
!   integer, intent(in) :: longlat  ! longat = 0 (long), longlat=1(lat)
   real(kind=vdp) :: dist1, dist2

   ! pos1 -> pos2
   call vsmPProjection4f(pos1, cam_num, cross1, pix1, dist1)
   call vsmPProjection4f(pos2, cam_num, cross2, pix2, dist2)
   !print *,"dist1, dist2", dist1, dist2
   if(cross1 .and. cross2) then
     call vsmConnectSphere(pix1, pix2, cam_num, dist1, dist2)
   else if(.not.cross1 .and. cross2) then
     call vsmLProjection4f(pos1, pos2, cam_num, cross1, pix1, dist1)
     if(.not.cross1) call vsmError("frame drawing 1")
     call vsmConnectSphere(pix1, pix2, cam_num, dist1, dist2)
   else if(cross1 .and. .not.cross2) then
     call vsmLProjection4f(pos2, pos1, cam_num, cross2, pix2, dist2)
     if(.not.cross2) call vsmError("frame drawing 2")
     call vsmConnectSphere(pix1, pix2, cam_num, dist1, dist2)
   end if

 end subroutine vsmDrawSphereLine

 subroutine vsmConnectSphere(pix1, pix2, cam_num, dist1, dist2)
   integer, dimension(2), intent(in) :: pix1, pix2
   integer, intent(in) :: cam_num
   real(kind=vdp), intent(in) :: dist1, dist2
   integer :: xs, xb, ys, yb ! small, big
   integer, dimension(2) :: pix, ppx
   integer :: i, fs, bpx
   real(kind=vdp) :: A, dy, dx
   real(kind=vdp) :: distance, ds, mindist
   integer :: x, y

   fs = framewidth; bpx = (fs-1)/2

   if(fs==1) then
   if(pix1(1) == pix2(1)) then

     ys = min(pix1(2), pix2(2)); yb = max(pix1(2), pix2(2))
!     pix(1) = pix1(1)
     if(pix1(2) < pix2(2)) then
       ds = dist2-dist1
       mindist = dist1
     else
       ds = dist1 - dist2
       mindist = dist2
     end if
     dy = yb - ys
     if(dy == 0) dy = 1
!$omp parallel do private(i, distance, pix)
     do i=1, yb-ys+1
       pix(1) = pix1(1)
       pix(2) = ys+i-1
       distance = mindist + ds/dy*(i-1)
       if(.not.vsmCheckFPix(pix)) cycle

!         if(bon) then
!           call vsmPaintPixelBlend(pix, vsmFrameColor, distance)
!         else
           call vsmPaintPixel3(pix, vsmFrameColor, distance)
!         end if
     end do

   else if(pix1(2) == pix2(2)) then

     xs = min(pix1(1), pix2(1)); xb = max(pix1(1), pix2(1))
!     pix(2) = pix1(2)
     if(pix1(1) < pix2(1)) then
       ds = dist2 - dist1
       mindist = dist1
     else
       ds = dist1 - dist2
       mindist = dist2
     end if
     dx = xb - xs
     if(dx == 0) dx = 1
!$omp parallel do private(i, distance, pix)
     do i=1, xb-xs+1
       pix(1) = xs+i-1; pix(2) = pix1(2)
       if(.not.vsmCheckFPix(pix)) cycle
       distance = mindist + ds/dx*(i-1)

!         if(bon) then
!           call vsmPaintPixelBlend(pix, vsmFrameColor, distance)
!         else
           call vsmPaintPixel3(pix, vsmFrameColor, distance)
!         end if
     end do

   else 

     A = dble((pix2(2)-pix1(2)))/dble((pix2(1)-pix1(1)))

     if(abs(A) <= 1.0_vdp) then

       if(pix1(1) < pix2(1)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
         ds = dist2 - dist1
         mindist = dist1
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
         ds = dist1 - dist2
         mindist = dist2
       end if
       dx = xb - xs
       if(dx == 0) dx =1
!$omp parallel do private(i, pix, distance)
       do i=1, xb-xs+1
         pix(1) = xs+i-1
         pix(2) = nint(A*(pix(1)-xs) + ys)
         distance = mindist + ds/dx*(i-1)
         if(.not.vsmCheckFPix(pix)) cycle
!           if(bon) then
!             call vsmPaintPixelBlend(pix, vsmFrameColor, distance)
!           else
             call vsmPaintPixel3(pix, vsmFrameColor, distance)
!           end if
       end do

     else

       if(pix1(2) < pix2(2)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
         ds = dist2 - dist1
         mindist = dist1
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
         ds = dist1 - dist2
         mindist = dist2
       end if
       dy = yb - ys
       if(dy == 0) dy = 1
!$omp parallel do private(i, pix, distance)
       do i=1, yb-ys+1
         pix(2) = ys+i-1
         pix(1) = nint((pix(2)-ys)/A + xs)
         distance = mindist + ds/dy*(i-1)
         if(.not.vsmCheckFPix(pix)) cycle
!           if(bon) then
!             call vsmPaintPixelBlend(pix, vsmFrameColor, distance)
!           else
             call vsmPaintPixel3(pix, vsmFrameColor, distance)
!           end if
       end do

     end if

   end if

   else ! FRAME WIDTH > 1
   fs = fs -1
   if(pix1(1) == pix2(1)) then

     ys = min(pix1(2), pix2(2)); yb = max(pix1(2), pix2(2))
!     pix(1) = pix1(1)
     if(pix1(2) < pix2(2)) then
       ds = dist2-dist1
       mindist = dist1
     else
       ds = dist1 - dist2
       mindist = dist2
     end if
     dy = yb - ys
     if(dy == 0) dy = 1
     do i=1, yb-ys+1
       pix(1) = pix1(1)
       pix(2) = ys+i-1
       distance = mindist + ds/dy*(i-1)
       if(.not.vsmCheckFPix(pix)) cycle
         do y=pix(2) - bpx, pix(2)-bpx + fs
           do x = pix(1) - bpx, pix(1) - bpx + fs
             ppx(1) = x; ppx(2) = y
             if(.not.vsmCheckFPix(ppx)) cycle
             call vsmPaintPixel3(ppx, vsmFrameColor, distance)
           end do
         end do
     end do

   else if(pix1(2) == pix2(2)) then

     xs = min(pix1(1), pix2(1)); xb = max(pix1(1), pix2(1))
!     pix(2) = pix1(2)
     if(pix1(1) < pix2(1)) then
       ds = dist2 - dist1
       mindist = dist1
     else
       ds = dist1 - dist2
       mindist = dist2
     end if
     dx = xb - xs
     if(dx == 0) dx = 1
     do i=1, xb-xs+1
       pix(1) = xs+i-1; pix(2) = pix1(2)
       if(.not.vsmCheckFPix(pix)) cycle
       distance = mindist + ds/dx*(i-1)
         do y=pix(2) - bpx, pix(2)-bpx + fs
           do x = pix(1) - bpx, pix(1) - bpx + fs
             ppx(1) = x; ppx(2) = y
             if(.not.vsmCheckFPix(ppx)) cycle
             call vsmPaintPixel3(ppx, vsmFrameColor, distance)
           end do
         end do
     end do

   else 

     A = dble((pix2(2)-pix1(2)))/dble((pix2(1)-pix1(1)))

     if(abs(A) <= 1.0_vdp) then

       if(pix1(1) < pix2(1)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
         ds = dist2 - dist1
         mindist = dist1
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
         ds = dist1 - dist2
         mindist = dist2
       end if
       dx = xb - xs
       if(dx == 0) dx =1
       do i=1, xb-xs+1
         pix(1) = xs+i-1
         pix(2) = nint(A*(pix(1)-xs) + ys)
         distance = mindist + ds/dx*(i-1)
         if(.not.vsmCheckFPix(pix)) cycle
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixel3(ppx, vsmFrameColor, distance)
             end do
           end do
       end do

     else

       if(pix1(2) < pix2(2)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
         ds = dist2 - dist1
         mindist = dist1
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
         ds = dist1 - dist2
         mindist = dist2
       end if
       dy = yb - ys
       if(dy == 0) dy = 1
       do i=1, yb-ys+1
         pix(2) = ys+i-1
         pix(1) = nint((pix(2)-ys)/A + xs)
         distance = mindist + ds/dy*(i-1)
         if(.not.vsmCheckFPix(pix)) cycle
           do y=pix(2) - bpx, pix(2)-bpx + fs
             do x = pix(1) - bpx, pix(1) - bpx + fs
               ppx(1) = x; ppx(2) = y
               if(.not.vsmCheckFPix(ppx)) cycle
               call vsmPaintPixel3(ppx, vsmFrameColor, distance)
             end do
           end do
       end do

     end if

   end if

   end if
 end subroutine vsmConnectSphere

 subroutine vsmMeasureDistSphere(pix, longlat, phiz, cam_num, dist, draw)
   real(kind=vdp), intent(in) :: phiz
   integer, dimension(2), intent(in) :: pix
   integer, intent(in) :: longlat, cam_num
   real(kind=vdp), intent(out) :: dist
   logical, intent(out) :: draw
   real(kind=vdp), dimension(3) :: epos, ppos, direc
   real(kind=vdp) :: a, b, c

   dist = vsmFarAway
   draw = .false.

   epos(1:3) = vsmCamera(cam_num)%pos(1:3)
   ppos(1:3) = vsmImageData%topleft(1:3) + vsmImageData%dp*((pix(1)-1)*vsmCamera(cam_num)%right(1:3) &
                    - (pix(2)-1)*vsmCamera(cam_num)%up(1:3))

   if(vsmCamera(cam_num)%pers) then
     direc(1:3) = ppos(1:3) - epos(1:3)
   else
     direc(1:3) = vsmCamera(cam_num)%front(1:3)
   end if

   call vsmNormVec(direc)
   if(longlat == 0) then
     a = -sin(phiz); b=cos(phiz)
     c = a*direc(1)+b*direc(2) 
     if(c == 0.0_vdp) then
        draw = .false.
     else
        draw = .true.
        if(vsmCamera(cam_num)%pers) then
          dist = -(epos(1)*a+epos(2)*b)/c
        else
          dist = -(ppos(1)*a+ppos(2)*b)/c
          dist = dist + vsmCamera(cam_num)%near
        end if
    end if
   else
     if(direc(3) == 0.0_vdp) then
        draw = .false.
     else
        draw = .true.
        if(vsmCamera(cam_num)%pers) then
          dist = (phiz - epos(3))/direc(3)
        else
          dist =  (phiz - ppos(3))/direc(3)
          dist = dist + vsmCamera(cam_num)%near
        end if
     end if
   end if
 end subroutine vsmMeasureDistSphere

 subroutine vsmDrawFrame(cam_num, blend)
   integer, intent(in) :: cam_num
   logical, intent(in) :: blend
   real(kind=vdp), dimension(3, 8) :: pos
   real(kind=vdp) :: dx
   integer :: n

   bon = blend
   dx = vsmIntval%ray_interval
   pos(1:3, 1) = vsmCoord%origin(1:3) - dx
   pos(1:3, 8) = vsmCoord%cend(1:3) + dx

   pos(1,2) = vsmCoord%cend(1) + dx; pos(2,2) = pos(2, 1); pos(3,2) = pos(3, 1)
   pos(1,3) = pos(1, 1); pos(2,3) = vsmCoord%cend(2)+dx; pos(3,3) = pos(3, 1)
   pos(1,4) = vsmCoord%cend(1) + dx; pos(2,4) = vsmCoord%cend(2) + dx; pos(3,4) = pos(3, 1)
   pos(1:2, 5) = pos(1:2, 1); pos(3, 5) = vsmCoord%cend(3)+dx
   pos(1,6) = vsmCoord%cend(1)+dx; pos(2,6) = pos(2, 1); pos(3,6) = vsmCoord%cend(3)+dx
   pos(1,7) = pos(1, 1); pos(2,7) = vsmCoord%cend(2)+dx; pos(3,7) = vsmCoord%cend(3)+dx

   ! pos1 -> pos2
   n=1
   call vsmDrawLine(pos(:, 1), pos(:, 2), cam_num, n)

   ! pos1 -> pos3
   n=2
   call vsmDrawLine(pos(:, 1), pos(:, 3), cam_num, n)

   ! pos2 -> pos4
   n=3
   call vsmDrawLine(pos(:, 2), pos(:, 4), cam_num, n)

   ! pos3 -> pos4
   n=4
   call vsmDrawLine(pos(:, 3), pos(:, 4), cam_num, n)

   ! pos5 -> pos6
   n=5
   call vsmDrawLine(pos(:, 5), pos(:, 6), cam_num, n)

   ! pos5 -> pos7
   n=6
   call vsmDrawLine(pos(:, 5), pos(:, 7), cam_num, n)

   ! pos6 -> pos8
   n=7
   call vsmDrawLine(pos(:, 6), pos(:, 8), cam_num, n)

   ! pos7 -> pos8
   n=8
   call vsmDrawLine(pos(:, 7), pos(:, 8), cam_num, n)

   ! pos1 -> pos5
   n=9
   call vsmDrawLine(pos(:, 1), pos(:, 5), cam_num, n)

   ! pos2 -> pos6
   n=10
   call vsmDrawLine(pos(:, 2), pos(:, 6), cam_num, n)

   ! pos3 -> pos7
   n=11
   call vsmDrawLine(pos(:, 3), pos(:, 7), cam_num, n)

   ! pos4 -> pos8
   n=12
   call vsmDrawLine(pos(:, 4), pos(:, 8), cam_num, n)

 end subroutine vsmDrawFrame

 subroutine vsmDrawLine(pos1, pos2, cam_num, lineNum)
   real(kind=vdp), dimension(3), intent(in) :: pos1, pos2
   integer, intent(in) :: cam_num, lineNum
   logical :: cross1, cross2
   integer, dimension(2) :: pix1, pix2

   ! pos1 -> pos2
   call vsmPProjection4f(pos1, cam_num, cross1, pix1)
   call vsmPProjection4f(pos2, cam_num, cross2, pix2)
   if(cross1 .and. cross2) then
     call vsmConnect(pix1, pix2, lineNum, pos1, cam_num)
!   else if(.not.cross1 .and. cross2) then
!     call vsmLProjection4f(pos1, pos2, cam_num, cross1, pix1)
!     if(.not.cross1) call vsmError("frame drawing 1")
!     call vsmConnect(pix1, pix2, lineNum, pos2, cam_num)
!   else if(cross1 .and. .not.cross2) then
!     call vsmLProjection4f(pos2, pos1, cam_num, cross2, pix2)
!     if(.not.cross2) call vsmError("frame drawing 2")
!     call vsmConnect(pix1, pix2, lineNum, pos1, cam_num)
   end if

 end subroutine vsmDrawLine

 subroutine vsmConnect(pix1, pix2, lineNum, pos, cam_num)
   integer, dimension(2), intent(in) :: pix1, pix2
   real(kind=vdp), dimension(3), intent(in) :: pos
   integer, intent(in) :: lineNum, cam_num
   integer :: xs, xb, ys, yb ! small, big
   integer, dimension(2) :: pix
   integer :: i
   real(kind=vdp) :: A
   logical :: draw
   real(kind=vdp) :: distance

   if(pix1(1) == pix2(1)) then

     ys = min(pix1(2), pix2(2)); yb = max(pix1(2), pix2(2))
     pix(1) = pix1(1)
!$omp parallel do private(i, distance, draw) firstprivate(pix)
     do i=1, yb-ys+1
       pix(2) = ys+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)

       if(draw) then
         if(bon) then
           call vsmPaintPixelBlend(pix, vsmFrameColor, distance)
         else
           call vsmPaintPixel3(pix, vsmFrameColor, distance)
         end if
       end if
     end do

   else if(pix1(2) == pix2(2)) then

     xs = min(pix1(1), pix2(1)); xb = max(pix1(1), pix2(1))
     pix(2) = pix1(2)
!$omp parallel do private(i, distance, draw) firstprivate(pix)
     do i=1, xb-xs+1
       pix(1) = xs+i-1
       if(.not.vsmCheckFPix(pix)) cycle
       call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
       if(draw) then
         if(bon) then
           call vsmPaintPixelBlend(pix, vsmFrameColor, distance)
         else
           call vsmPaintPixel3(pix, vsmFrameColor, distance)
         end if
       end if
     end do

   else 

     A = dble((pix2(2)-pix1(2)))/dble((pix2(1)-pix1(1)))

     if(abs(A) <= 1.0_vdp) then

       if(pix1(1) < pix2(1)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

!$omp parallel do private(i, pix, distance, draw)
       do i=1, xb-xs+1
         pix(1) = xs+i-1
         pix(2) = nint(A*(pix(1)-xs) + ys)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
           if(bon) then
             call vsmPaintPixelBlend(pix, vsmFrameColor, distance)
           else
             call vsmPaintPixel3(pix, vsmFrameColor, distance)
           end if
         end if
       end do

     else

       if(pix1(2) < pix2(2)) then
         xs = pix1(1); ys = pix1(2)
         xb = pix2(1); yb = pix2(2)
       else
         xs = pix2(1); ys = pix2(2)
         xb = pix1(1); yb = pix1(2)
       end if

!$omp parallel do private(i, pix, distance, draw)
       do i=1, yb-ys+1
         pix(2) = ys+i-1
         pix(1) = nint((pix(2)-ys)/A + xs)
         if(.not.vsmCheckFPix(pix)) cycle
         call vsmMeasureDist(pix, lineNum, pos, cam_num, distance, draw)
         if(draw) then
           if(bon) then
             call vsmPaintPixelBlend(pix, vsmFrameColor, distance)
           else
             call vsmPaintPixel3(pix, vsmFrameColor, distance)
           end if
         end if
       end do

     end if

   end if

 end subroutine vsmConnect

 subroutine vsmMeasureDist(pix, num, pos, cam_num, dist, draw)
   real(kind=vdp), dimension(3), intent(in) :: pos
   integer, dimension(2), intent(in) :: pix
   integer, intent(in) :: num, cam_num
   real(kind=vdp), intent(out) :: dist
   logical, intent(out) :: draw
   real(kind=vdp), dimension(3) :: epos, ppos, direc

   dist = vsmFarAway
   draw = .false.

   if(num < 1 .or. num > 12) call vsmError("frame num is bad")

   epos(1:3) = vsmCamera(cam_num)%pos(1:3)
   ppos(1:3) = vsmImageData%topleft(1:3) + vsmImageData%dp*((pix(1)-1)*vsmCamera(cam_num)%right(1:3) &
                    - (pix(2)-1)*vsmCamera(cam_num)%up(1:3))

   if(num==1 .or. num==5 .or. num==4 .or. num==8) then
     direc(1:3) = ppos(1:3) - epos(1:3)
     call vsmNormVec(direc)
     if(direc(2) /= 0.0_vdp) then
       dist = (pos(2) - epos(2))/ direc(2)
       draw = .true.
     else if(direc(3) /= 0.0_vdp) then
       dist = (pos(3) - epos(3))/ direc(3)
       draw = .true.
     else
       draw = .false.
     end if
     return
   end if

   if(num==2 .or. num==3 .or. num==6 .or. num==7) then
     direc(1:3) = ppos(1:3) - epos(1:3)
     call vsmNormVec(direc)
     if(direc(1) /= 0.0_vdp) then
       dist = (pos(1) - epos(1))/ direc(1)
       draw = .true.
     else if(direc(3) /= 0.0_vdp) then
       dist = (pos(3) - epos(3))/ direc(3)
       draw = .true.
     else
       draw = .false.
     end if
     return
   end if

   if(num==9 .or. num==10 .or. num==11 .or. num==12) then
     direc(1:3) = ppos(1:3) - epos(1:3)
     call vsmNormVec(direc)
     if(direc(1) /= 0.0_vdp) then
       dist = (pos(1) - epos(1))/ direc(1)
       draw = .true.
     else if(direc(2) /= 0.0_vdp) then
       dist = (pos(2) - epos(2))/ direc(2)
       draw = .true.
     else
       draw = .false.
     end if
     return
   end if

 end subroutine vsmMeasureDist

 subroutine vsmPProjection4f(pos, cam_num, cross, pix, dist)
   real(kind=vdp), dimension(3), intent(in) :: pos
   integer, intent(in) :: cam_num
   logical, intent(inout) :: cross
   integer, dimension(2), intent(out) :: pix
   real(kind=vdp), intent(inout), optional :: dist

   real(kind=vdp), dimension(3) :: eyepos
   real(kind=vdp), dimension(3) :: norm, plane_c
   real(kind=vdp), dimension(3) :: direc
   real(kind=vdp) :: A, C, B, ip1, ip2
   real(kind=vdp), dimension(3) :: cross_pos
   real(kind=vdp), dimension(3) :: img_topleft, img_right, img_down

!   call vsmLogPrint("PProjection4f ", pos)

   norm(1:3) = vsmCamera(cam_num)%front(1:3)
   eyepos(1:3) = vsmCamera(cam_num)%pos(1:3)

   plane_c(1:3) = eyepos(1:3) + vsmCamera(cam_num)%near * norm(1:3)

   if(present(dist)) dist = 0.0_vdp

   direc = pos - plane_c
   A = dot_product(norm, direc)
   if(A <= 0.0_vdp) then
       cross = .false.
       return
   end if

   if(vsmCamera(cam_num)%pers) then
     !Perspective Projection
     direc(1:3) = pos(1:3) - eyepos(1:3)
     call vsmNormVec(direc)

     C = dot_product(norm, plane_c - eyepos)
     B = dot_product(norm, direc)

     if(B == 0.0_vdp) then
       cross = .false.
       return
     end if

     A = C/B

     if(A < 0.0_vdp) then
       cross = .false.
       return
     end if

     cross_pos(1:3) = A*direc(1:3) + eyepos(1:3)
     direc = pos - eyepos
     if(present(dist)) dist = sqrt(direc(1)*direc(1)+direc(2)*direc(2)+direc(3)*direc(3))

   else
     !Parallel Projection
     direc(1:3) = -norm(1:3)

     C = dot_product(norm, plane_c - pos)
     !B = -1.0_vdp

     A = -C

     if(A < 0.0_vdp) then
       cross = .false.
       return
     end if

     cross_pos(1:3) = A*direc(1:3) + pos(1:3)
     if(present(dist)) dist = A + vsmCamera(cam_num)%near
   end if

   img_topleft(1:3) = vsmImageData%topleft(1:3)
   img_right(1:3)   = vsmCamera(cam_num)%right(1:3)
   img_down(1:3)    = -vsmCamera(cam_num)%up(1:3)

   direc(1:3) = cross_pos(1:3) - img_topleft(1:3)

   ip1 = dot_product(direc, img_right)
   ip2 = dot_product(direc, img_down)
!   call vsmLogPrint("ip1 = ", ip1)
!   call vsmLogPrint("ip2 = ", ip2)
   pix(1) = nint(ip1/vsmImageData%dp)
   pix(2) = nint(ip2/vsmImageData%dp)

   cross = .true.
 end subroutine vsmPProjection4f

 subroutine vsmLProjection4f(pos1, pos2, cam_num, cross, pix, dist)
   ! pos1 is below the plane and pos2 is above the plane
   real(kind=vdp), dimension(3), intent(in) :: pos1, pos2
   integer, intent(in) :: cam_num
   logical, intent(inout) :: cross
   integer, dimension(2), intent(out) :: pix
   real(kind=vdp), intent(inout), optional :: dist

   real(kind=vdp), dimension(3) :: norm,eyepos,  plane_c
   real(kind=vdp), dimension(3) :: direc
   real(kind=vdp) :: A, C, B, ip1, ip2
   real(kind=vdp), dimension(3) :: cross_pos
   real(kind=vdp), dimension(3) :: img_topleft, img_right, img_down

   norm(1:3) = vsmCamera(cam_num)%front(1:3)
   eyepos(1:3) = vsmCamera(cam_num)%pos(1:3)

   plane_c(1:3) = eyepos(1:3) + vsmCamera(cam_num)%near * norm(1:3)

   if(present(dist)) dist = 0.0_vdp
   !Perspective Projection & Orthogonal
   direc(1:3) = pos2(1:3) - pos1(1:3)
   call vsmNormVec(direc)

   C = dot_product(norm, plane_c - pos1)
   B = dot_product(norm, direc)

   if(B == 0.0_vdp) then
     cross = .false.
     return
   end if

   A = C/B

   if(A < 0.0_vdp) then
     cross = .false.
     return
   end if

   cross_pos(1:3) = A*direc(1:3) + pos1(1:3)

   if(present(dist)) then 
   if(vsmCamera(cam_num)%pers) then
     direc = cross_pos - vsmCamera(cam_num)%pos
     dist = sqrt(direc(1)*direc(1) + direc(2)*direc(2) + direc(3)*direc(3))
   else
     dist = vsmCamera(cam_num)%near
   end if
   end if

   img_topleft(1:3) = vsmImageData%topleft(1:3)
   img_right(1:3)   = vsmCamera(cam_num)%right(1:3)
   img_down(1:3)    = -vsmCamera(cam_num)%up(1:3)

   direc(1:3) = cross_pos(1:3) - img_topleft(1:3)

   ip1 = dot_product(direc, img_right)
   ip2 = dot_product(direc, img_down)

   pix(1) = nint(ip1/vsmImageData%dp)
   pix(2) = nint(ip2/vsmImageData%dp)

   cross = .true.
 end subroutine vsmLProjection4f


 end module vsmFrame

