! general functions for visualization
module vsmFunction
#ifndef CLANG
 use mpi
#endif

 use vsmConst
!$ use omp_lib

 implicit none
 private
 integer, save :: vsmUnitNumLog
 
interface vsmLogPrint
  module procedure vsmLogPrintda, vsmLogPrintla, vsmLogPrintl, vsmLogPrintia, vsmLogPrinti, vsmLogPrintfa, &
    vsmLogPrintf, vsmLogPrintd, vsmLogPrintt
end interface vsmLogPrint

 public vsmError, vsmDeg2Rad, vsmScaleF01, vsmScaleD01, vsmSearchUnitNum, vsmUnPack1sp23sp, vsmPack3sp21sp
 public vsmAtoF, vsmAtoI, vsmAtoD, vsmInt2Asc9, vsmInt2Asc5, vsmInt2Asc4
 public vsmTransVertx, vsmSetZeroF, vsmRotateMatrix, vsmRotateInvmatrix, vsmPlaneTrans, vsmNormVecF, vsmNOrmVec
 public vsmMatrixCrossVectf, vsmMatrixCrossVect, vsmCrossProd, vsmCheckPowerof2, vsmCheckFa01, vsmCheckF01
 public vsmLogStart, vsmLogPrint, vsmLogFlush, vsmLogEnd
 public vsmUnPack1int23sp, vsmPack3sp21int
 
#ifdef NO_ISNAN
public  isnan
#endif


 contains

 function vsmFuncDistanced2d(a, b)
   real(kind=vdp), dimension(3), intent(in) :: a, b
   real(kind=vdp) :: vsmFuncDistanced2d
   vsmFuncDistanced2d = sqrt(dot_product(a-b, a-b))
 end function vsmFuncDistanced2d


 subroutine vsmCheckPowerOf2(psize, stage, poss)
   integer, intent(in) :: psize
   integer, intent(out) :: stage
   logical, intent(out) :: poss
   integer :: p, n

   call vsmLogPrint("vsmCheckPowerOf2 is called")
   p=1; n=0; stage = 1
   do
     p = p * 2
     n = n + 1

     if(p == psize) then
       stage = n
       poss = .true.
       exit
     end if

     if(p > psize) then
       stage = n-1
       poss = .false.
       exit
     end if
   end do

   if(vsmDbg) then
     call vsmLogPrint("stage = ", stage)
     call vsmLogPrint("possibility = ", poss)
     call vsmLogPrint("vsmCheckPowerOf2 ends")
   end if
 end subroutine vsmCheckPowerOf2

 subroutine vsmPlaneTrans(sc, trans, bool)
   integer, dimension(3), intent(out) :: bool
   real(kind=vdp), intent(out) :: trans
   real(kind=vdp), dimension(4), intent(in) :: sc

   bool = 0
   if(sc(1) /= 0.0_vdp) then
     bool(1) = 1
     trans = sc(4) / sc(1)
   else if(sc(2) /= 0.0_vdp) then
     bool(2) = 1
     trans = sc(4) / sc(2)
   else if(sc(3) /= 0.0_vdp) then
     bool(3) = 1
     trans = sc(4) / sc(3)
   else
     call vsmError("arrow eq is strange")
   end if
 end subroutine vsmPlaneTrans

 subroutine vsmTransVertx(pos1, pos2, trans, bool)
   integer, dimension(3), intent(in) :: bool
   real(kind=vdp), intent(in) :: trans
   real(kind=vdp), dimension(3), intent(in) :: pos1
   real(kind=vdp), dimension(3), intent(out) :: pos2

   integer :: i

   do i=1, 3
     if(bool(i) == 0) then
       pos2(i) = pos1(i)
     else
       pos2(i) = pos1(i) + trans
     end if
   end do

 end subroutine vsmTransVertx

 subroutine vsmMatrixCrossVect(matrix, in_vect, out_vect)
   real(kind=vdp), dimension(3,3), intent(in) :: matrix
   real(kind=vdp), dimension(3), intent(in)   :: in_vect
   real(kind=vdp), dimension(3), intent(out)  :: out_vect
   ! (1,1) (2,1) (3,1)   (1)
   ! (1,2) (2,2) (3,2) x (2)
   ! (1,3) (2,3) (3,3)   (3)
   out_vect(1:3) = matrix(1,1:3)*in_vect(1) + matrix(2,1:3)*in_vect(2) + matrix(3,1:3)*in_vect(3)
 end subroutine vsmMatrixCrossVect

 subroutine vsmMatrixCrossVectf(matrix, in_vect, out_vect)
   real(kind=vdp), dimension(3,3), intent(in) :: matrix
   real, dimension(3), intent(in)   :: in_vect
   real, dimension(3), intent(out)  :: out_vect
   ! (1,1) (2,1) (3,1)   (1)
   ! (1,2) (2,2) (3,2) x (2)
   ! (1,3) (2,3) (3,3)   (3)
   out_vect(1:3) = matrix(1,1:3)*in_vect(1) + matrix(2,1:3)*in_vect(2)  + matrix(3,1:3)*in_vect(3)
 end subroutine vsmMatrixCrossVectf

 subroutine vsmTransVertxV(pos1, pos2, trans, bool, v)
   integer, dimension(3), intent(in) :: bool
   integer, intent(in) :: v
   real(kind=vdp), intent(in) :: trans
   real(kind=vdp), dimension(3, v), intent(in) :: pos1
   real(kind=vdp), dimension(3, v), intent(out) :: pos2
   integer :: i, j

   do j=1, 3
     do i=1, v
       if(bool(j) == 0) then
         pos2(j, i) = pos1(j, i)
       else
         pos2(j, i) = pos1(j, i) + trans
       end if
     end do
   end do
 end subroutine vsmTransVertxV

 subroutine vsmMatrixCrossVectV(matrix, in_vect, out_vect, v)
   real(kind=vdp), dimension(3,3), intent(in) :: matrix
   integer, intent(in) :: v
   real(kind=vdp), dimension(3, v), intent(in)   :: in_vect
   real(kind=vdp), dimension(3, v), intent(out)  :: out_vect
   integer :: i
   ! (1,1) (2,1) (3,1)   (1)
   ! (1,2) (2,2) (3,2) x (2)
   ! (1,3) (2,3) (3,3)   (3)
   do i=1, v
     out_vect(1, i) = matrix(1,1)*in_vect(1, i) + matrix(2,1)*in_vect(2, i) + matrix(3,1)*in_vect(3, i)
     out_vect(2, i) = matrix(1,2)*in_vect(1, i) + matrix(2,2)*in_vect(2, i) + matrix(3,2)*in_vect(3, i)
     out_vect(3, i) = matrix(1,3)*in_vect(1, i) + matrix(2,3)*in_vect(2, i) + matrix(3,3)*in_vect(3, i)
   end do
 end subroutine vsmMatrixCrossVectV

 subroutine vsmTranslate(pos1, pos2, pos1a, pos2a)
   !translate pos1 to origin
   real(kind=vdp), dimension(3), intent(in) :: pos1, pos2
   real(kind=vdp), dimension(3), intent(out) :: pos1a, pos2a

   pos1a(1:3) = 0.0_vdp
   pos2a(1:3) = pos2(1:3) - pos1(1:3)
 end subroutine vsmTranslate

 subroutine vsmRotateMatrix(pos, matrix)
   !get a matrix (0,0,0 to x,y,z) to (0,0,0 to 0,0,z)
   ! (1,1) (2,1) (3,1)
   ! (1,2) (2,2) (3,2)
   ! (1,3) (2,3) (3,3)
   ! assuming |pos| /= 0
   real(kind=vdp), dimension(:), intent(in) :: pos
   real(kind=vdp), dimension(3,3), intent(out) :: matrix

   real(kind=vdp) :: cos_phi, sin_phi, cos_theta, sin_theta
   real(kind=vdp) :: len_xy, len_xyz

   if(pos(1) == 0.0_vdp .and. pos(2) == 0.0_vdp) then
     ! already parallel to z
     matrix(1,1) = 1.0_vdp; matrix(2,1) = 0.0_vdp; matrix(3,1) = 0.0_vdp;  
     matrix(1,2) = 0.0_vdp; matrix(2,2) = 1.0_vdp; matrix(3,2) = 0.0_vdp; 
     matrix(1,3) = 0.0_vdp; matrix(2,3) = 0.0_vdp; matrix(3,3) = 1.0_vdp; 
   else
     len_xy = sqrt(dot_product(pos(1:2), pos(1:2)))
     cos_phi = pos(1)/len_xy;  sin_phi = pos(2)/len_xy
     !if(pos(2)<0.0_vdp) sin_phi=-sqrt(1.0_vdp-cos_phi*cos_phi)
     !if(pos(2)>=0.0_vdp) sin_phi=sqrt(1.0_vdp-cos_phi*cos_phi)

     len_xyz = sqrt(dot_product(pos(1:3), pos(1:3)))
     cos_theta = pos(3)/len_xyz; sin_theta = len_xy/len_xyz
     !sin_theta = sqrt(1.0_Vdp-cos_theta*cos_theta)

     !matrix(1,1) = cos_theta*sin_phi; matrix(2,1) = cos_theta*sin_phi; matrix(3,1) = -sin_theta;  
     matrix(1,1) = cos_theta*cos_phi; matrix(2,1) = cos_theta*sin_phi; matrix(3,1) = -sin_theta;  
     matrix(1,2) = -sin_phi;          matrix(2,2) = cos_phi;           matrix(3,2) = 0.0_vdp; 
     matrix(1,3) = sin_theta*cos_phi; matrix(2,3) = sin_theta*sin_phi; matrix(3,3) = cos_theta; 
   end if
 end subroutine vsmRotateMatrix

 subroutine vsmRotateInvMatrix(pos, matrix)
   !get a matrix (0,0,0 to 0,0,z) to (0,0,0 to x,y,z)
   ! (1,1) (2,1) (3,1)
   ! (1,2) (2,2) (3,2)
   ! (1,3) (2,3) (3,3)
   ! assuming |pos| /= 0
   real(kind=vdp), dimension(:), intent(in) :: pos
   real(kind=vdp), dimension(3,3), intent(out) :: matrix

   real(kind=vdp) :: cos_phi, sin_phi, cos_theta, sin_theta
   real(kind=vdp) :: len_xy, len_xyz

   if(pos(1) == 0.0_vdp .and. pos(2) == 0.0_vdp) then
     ! already parallel to z
     matrix(1,1) = 1.0_vdp; matrix(2,1) = 0.0_vdp; matrix(3,1) = 0.0_vdp;  
     matrix(1,2) = 0.0_vdp; matrix(2,2) = 1.0_vdp; matrix(3,2) = 0.0_vdp; 
     matrix(1,3) = 0.0_vdp; matrix(2,3) = 0.0_vdp; matrix(3,3) = 1.0_vdp; 
   else
     len_xy = sqrt(dot_product(pos(1:2), pos(1:2)))
     cos_phi = pos(1)/len_xy;  sin_phi = pos(2)/len_xy
     len_xyz = sqrt(dot_product(pos(1:3), pos(1:3)))
     cos_theta = pos(3)/len_xyz; sin_theta = len_xy/len_xyz

     !matrix(1,1) = cos_theta*cos_phi; matrix(2,1) = sin_phi; matrix(3,1) = sin_theta*cos_phi;  
     matrix(1,1) = cos_theta*cos_phi; matrix(2,1) = -sin_phi; matrix(3,1) = sin_theta*cos_phi;  
     matrix(1,2) = cos_theta*sin_phi; matrix(2,2) = cos_phi; matrix(3,2) = sin_theta*sin_phi; 
     matrix(1,3) = -sin_theta;        matrix(2,3) = 0.0_vdp; matrix(3,3) = cos_theta; 
   end if
 end subroutine vsmRotateInvMatrix

 !log
 subroutine vsmLogStart(filename)
   character(*), intent(in) :: filename
   integer :: iostatus
   if(.not.vsmDbg) return
   vsmUnitNumLog = vsmSearchUnitNum("Log")
   open(vsmUnitNumLog, file=trim(filename), action='write', iostat=iostatus)
   if(iostatus /= 0) then 
     call vsmError('cant open logfile')
   end if
 end subroutine vsmLogStart

 subroutine vsmLogFlush
   if(.not.vsmDbg) return
   call flush(vsmUnitNumLog)
 end subroutine vsmLogFlush

 subroutine vsmLogEnd
   if(.not.vsmDbg) return
   !call flush(vsmUnitNumLog)
   close(vsmUnitNumLog)
 end subroutine vsmLogEnd

 subroutine vsmLogPrintt(mes)
  character(*), intent(in) :: mes
  if(.not.vsmDbg) return
!$omp critical (vsm_ocv_Log)
  write(vsmUnitNumLog, '(2A)') 'vsmlog : ', mes
!$omp end critical (vsm_ocv_Log)
 call vsmLogFlush
 end subroutine vsmLogPrintt

 subroutine vsmLogPrinti(mes, i)
  character(*), intent(in) :: mes
  integer, intent(in) :: i
  if(.not.vsmDbg) return
!$omp critical (vsm_ocv_Log)
  write(vsmUnitNumLog, '(2A)', advance='no') 'vsmlog : ', mes
  write(vsmUnitNumLog, '(1I10)') i
!$omp end critical (vsm_ocv_Log)
 call vsmLogFlush
 end subroutine vsmLogPrinti

 subroutine vsmLogPrintia(mes, ia)
  character(*), intent(in) :: mes
  integer, dimension(:), intent(in) :: ia
  integer :: i
  if(.not.vsmDbg) return
!$omp critical (vsm_ocv_Log)
  write(vsmUnitNumLog, '(2A)', advance='no') 'vsmlog : ', mes
  do i=1, size(ia)
    write(vsmUnitNumLog, '(1I10,1X)', advance='no') ia(i)
  end do
  write(vsmUnitNumLog, *) ''
!$omp end critical (vsm_ocv_Log)
  call vsmLogFlush
 end subroutine vsmLogPrintia

 subroutine vsmLogPrintf(mes, f)
  character(*), intent(in) :: mes
  real, intent(in) :: f
  if(.not.vsmDbg) return
!$omp critical (vsm_ocv_Log)
  write(vsmUnitNumLog, '(2A)', advance='no') 'vsmlog : ', mes
  write(vsmUnitNumLog, '(1E10.3)') f
!$omp end critical (vsm_ocv_Log)
 call vsmLogFlush
 end subroutine vsmLogPrintf

 subroutine vsmLogPrintfa(mes, fa)
  character(*), intent(in) :: mes
  real, dimension(:), intent(in) :: fa
  integer :: i
  if(.not.vsmDbg) return
!$omp critical (vsm_ocv_Log)
  write(vsmUnitNumLog, '(2A)', advance='no') 'vsmlog : ', mes
  do i=1, size(fa)
    write(vsmUnitNumLog, '(1E10.3,1X)', advance='no') fa(i)
  end do
  write(vsmUnitNumLog, *) ''
!$omp end critical (vsm_ocv_Log)
 call vsmLogFlush
 end subroutine vsmLogPrintfa

 subroutine vsmLogPrintd(mes, d)
  character(*), intent(in) :: mes
  real(kind=vdp), intent(in) :: d
  if(.not.vsmDbg) return
!$omp critical (vsm_ocv_Log)
  write(vsmUnitNumLog, '(2A)', advance='no') 'vsmlog : ', mes
!  write(vsmUnitNumLog, '(1E10.3)') d
  write(vsmUnitNumLog, '(1F10.6)') d
!$omp end critical (vsm_ocv_Log)
 call vsmLogFlush
 end subroutine vsmLogPrintd

 subroutine vsmLogPrintda(mes, da)
  character(*), intent(in) :: mes
  real(kind=vdp), dimension(:), intent(in) :: da
  integer :: i

  if(.not.vsmDbg) return
!$omp critical (vsm_ocv_Log)
  write(vsmUnitNumLog, '(2A)', advance='no') 'vsmlog : ', mes
  do i=1, size(da)
!    write(vsmUnitNumLog, '(1E10.3,1X)', advance='no') da(i)
    write(vsmUnitNumLog, '(1F10.6,1X)', advance='no') da(i)
  end do
  write(vsmUnitNumLog, *) ''
!$omp end critical (vsm_ocv_Log)
 call vsmLogFlush
 end subroutine vsmLogPrintda

 subroutine vsmLogPrintl(mes, l)
  character(*), intent(in) :: mes
  logical, intent(in) :: l
  if(.not.vsmDbg) return
!$omp critical (vsm_ocv_Log)
  write(vsmUnitNumLog, '(2A)', advance='no') 'vsmlog : ', mes
  write(vsmUnitNumLog, '(1L2)') l
!$omp end critical (vsm_ocv_Log)
 call vsmLogFlush
 end subroutine vsmLogPrintl

 subroutine vsmLogPrintla(mes, la)
  character(*), intent(in) :: mes
  logical, dimension(:), intent(in) :: la
  integer :: i
  if(.not.vsmDbg) return
!$omp critical (vsm_ocv_Log)
  write(vsmUnitNumLog, '(2A)', advance='no') 'vsmlog : ', mes
  do i=1, size(la)
    write(vsmUnitNumLog, '(1L2,1X)', advance='no') la(i)
  end do
  write(vsmUnitNumLog, *) ''
!$omp end critical (vsm_ocv_Log)
 call vsmLogFlush
 end subroutine vsmLogPrintla

 !normalize a vector
 subroutine vsmNormVec(vec)
  real(kind=vdp), dimension(3), intent(inout) :: vec
  real(kind=vdp) :: temp

  temp = sqrt(dot_product(vec, vec))
  if(temp > 0.0_vdp) vec(1:3) = vec(1:3)/temp
 end subroutine vsmNormVec

 subroutine vsmNormVecF(vec)
  real, dimension(3), intent(inout) :: vec
  real :: temp

  temp = sqrt(dot_product(vec, vec))
  if(temp > 0.0) vec(1:3) = vec(1:3)/temp
 end subroutine vsmNormVecF

 subroutine vsmSetZero(vec)
   real(kind=vdp), dimension(:), intent(out) :: vec
   integer :: n

   n = size(vec)
   vec(1:n) = 0.0_vdp
 end subroutine vsmSetZero

 subroutine vsmSetZeroF(vec)
   real, dimension(:), intent(out) :: vec
   integer :: n

   n = size(vec)
   vec(1:n) = 0.0
 end subroutine vsmSetZeroF

 !calculate cross product of vectors
 subroutine vsmCrossProd(vec, vec1, vec2)
  real(kind=vdp), dimension(3), intent(out) :: vec
  real(kind=vdp), dimension(3), intent(in) :: vec1
  real(kind=vdp), dimension(3), intent(in) :: vec2

  vec(1) = vec1(2)*vec2(3) - vec1(3)*vec2(2)
  vec(2) = vec1(3)*vec2(1) - vec1(1)*vec2(3)
  vec(3) = vec1(1)*vec2(2) - vec1(2)*vec2(1)
 end subroutine vsmCrossProd

 subroutine vsmTransvec(a, o, oa)
   real(kind=vdp), dimension(3), intent(in):: a, o
   real(kind=vdp), dimension(3), intent(out):: oa

   oa(1:3) = a(1:3) - o(1:3)
 end subroutine

 function vsmDistance(vec1, vec2)
    real(kind=vdp), dimension(3), intent(in):: vec1
    real(kind=vdp), dimension(3), intent(in):: vec2
    real(kind=vdp) :: vsmDistance
    
    real(kind=vdp), dimension(3) :: v21
    v21 = vec2 - vec1
    vsmDistance = sqrt(dot_product(v21(1:3),v21(1:3)))

 end function vsmDistance

 !degree to radian
 function vsmDeg2Rad(deg)
  real(kind=vdp), intent(in):: deg
  real(kind=vdp) :: vsmDeg2Rad
  vsmDeg2Rad = pi/180.0_vdp*deg
 end function vsmDeg2Rad

 !an error occurs
 !program displays a message and stops
 subroutine vsmError(mes, comm)
  character(*), intent(in) :: mes
  integer, intent(in), optional :: comm
  integer :: ecord, ierr

  write(*, '(2A)') 'VISMO ERROR : ', mes
  write(*, '(A)') 'stop this program'
  !call vsmLogFlush
  call vsmLogEnd
#ifndef CLANG
  if(present(comm)) call mpi_abort(comm, ecord, ierr)
#else
  if(present(comm)) call VSMMPIABORT(ecord)
#endif
  stop
 end subroutine vsmError

 function vsmAtoI(a)
   integer :: vsmAtoI
   character(*), intent(in) :: a
   read(a, *) vsmAtoI
 end function vsmAtoI

 function vsmAtoF(a)
   real :: vsmAtoF
   character(*), intent(in) :: a
   read(a, *) vsmAtoF
 end function vsmAtoF

 function vsmAtoD(a)
   real(kind = vdp) :: vsmAtoD
   character(*), intent(in) :: a
   read(a, *) vsmAtoD
 end function vsmAtoD

 function vsmItoA(i)
  character(25) :: vsmItoA
  integer, intent(in) :: i
  character(25) :: a
  write (a, *) i
  vsmItoA = a
 end function vsmItoA

 subroutine vsmCheckFA01(var)
   real, dimension(:), intent(in) :: var
   integer :: i
   do i=1, size(var)
    if(var(i) > 1.0 .or. var(i) < 0.0) then
      call vsmLogPrint('there are variables with unproper values')
    end if
   end do
 end subroutine vsmCheckFA01

 subroutine vsmCheckF01(var)
   real, intent(in) :: var
    if(var > 1.0 .or. var < 0.0) then
      call vsmLogPrint('there are variables with unproper values')
    end if
 end subroutine vsmCheckF01

 function vsmScaleF01(val)
  real, intent(in) :: val
  real :: vsmScaleF01

  vsmScaleF01 = val
  if(vsmScaleF01 < 0.0) then 
   vsmScaleF01 = 0.0
  else if(vsmScaleF01 > 1.0) then 
   vsmScaleF01 = 1.0
  end if
 end function vsmScaleF01
 
 function vsmScaleD01(val)
  real(kind=vdp), intent(in) :: val
  real(kind=vdp) :: vsmScaleD01

  vsmScaleD01 = val
  if(vsmScaleD01 < 0.0_vdp) then 
   vsmScaleD01 = 0.0_vdp
  else if(vsmScaleD01 > 1.0_vdp) then 
   vsmScaleD01 = 1.0_vdp
  end if
 end function vsmScaleD01

 function vsmScaleFA01(val)
  real, dimension(:), intent(in) :: val
  real, dimension(size(val)) :: vsmScaleFA01
  integer :: i

  vsmScaleFA01 = val
  do i=1, size(val)
    if(vsmScaleFA01(i) < 0.0) then 
      vsmScaleFA01(i) = 0.0
    else if(vsmScaleFA01(i) > 1.0) then 
      vsmScaleFA01(i) = 1.0
    end if
  end do
 end function vsmScaleFA01

 function vsmSearchUnitNum(mes)
   character(*), intent(in) :: mes
   integer :: vsmSearchUnitNum
   integer :: i
   logical :: used

   vsmSearchUnitNum = -1
   do i = vsmUnitNum, vsmUnitMaxNum
      inquire(unit=i, opened=used)
      if(used .neqv. .true.) then
         vsmSearchUnitNum = i
         exit
      end if
   end do

   if(vsmSearchUnitNum == -1) then
      call vsmError("can't find any unused unit numbers:"//mes)
   end if
 end function vsmSearchUnitNum

 subroutine vsmInt2Asc4(num, chara)
   ! integer to ascii with 0 in the left
   integer, intent(in) :: num
   character(4) :: chara
   write(chara, '(I4.4)') num
 end subroutine vsmInt2Asc4

 subroutine vsmInt2Asc5(num, chara)
   ! integer to ascii with 0 in the left
   integer, intent(in) :: num
   character(5) :: chara
   write(chara, '(I5.5)') num
 end subroutine vsmInt2Asc5

 subroutine vsmInt2Asc7(num, chara)
   ! integer to ascii with 0 in the left
   integer, intent(in) :: num
   character(7) :: chara
   write(chara, '(I7.7)') num
 end subroutine vsmInt2Asc7

 subroutine vsmInt2Asc9(num, chara)
   ! integer to ascii with 0 in the left
   integer, intent(in) :: num
   character(9) :: chara
   write(chara, '(I9.9)') num
 end subroutine vsmInt2Asc9

 subroutine vsmPack3SP21SP(color, sp)
   real, dimension(3), intent(in) :: color
   real, intent(out) :: sp
   integer :: red, green, blue, PixelColor

   !call vsmLogPrint("rgb=",color)

   green = int(color(2)*vsmPackConst)
   blue  = int(color(3)*vsmPackConst)

   if(color(1) >= 1.0) then
     red = vsmPackConst
   else
     red = int(color(1)*vsmPackConst)
   end if

   if(color(2) >= 1.0) then
     green = vsmPackConst
   else
     green = int(color(2)*vsmPackConst)
   end if

   if(color(3) >= 1.0) then
     blue = vsmPackConst
   else
     blue = int(color(3)*vsmPackConst)
   end if

   if(red < 0) then
     call vsmError("negative value:vsmPack3SP2SP red")
   end if

   if(green < 0) then
     call vsmError("negative value:vsmPack3SP2SP green")
   end if

   if(blue < 0) then
     call vsmError("negative value:vsmPack3SP2SP blue")
   end if

   PixelColor = red + 1000*green + 1000000*blue
   if(PixelColor > 1000000000) then
     call vsmLogPrint("Pack:PixelColor=",PixelColor)
     call vsmError("big value:pack PixelColor")
   end if
#ifndef CTRANS
   sp = transfer(PixelColor, 1.0)
#else
   call VSMTRANSFER4I2F(PixelColor, sp)
#endif
 end subroutine vsmPack3SP21SP


 subroutine vsmPack3SP21INT(color, intcol)
   real, dimension(3), intent(in) :: color
   integer, intent(out) :: intcol
   integer :: red, green, blue, PixelColor

   !call vsmLogPrint("rgb=",color)

   green = int(color(2)*vsmPackConst)
   blue  = int(color(3)*vsmPackConst)

   if(color(1) >= 1.0) then
     red = vsmPackConst
   else
     red = int(color(1)*vsmPackConst)
   end if

   if(color(2) >= 1.0) then
     green = vsmPackConst
   else
     green = int(color(2)*vsmPackConst)
   end if

   if(color(3) >= 1.0) then
     blue = vsmPackConst
   else
     blue = int(color(3)*vsmPackConst)
   end if

   if(red < 0) then
     call vsmError("negative value:vsmPack3SP2SP red")
   end if

   if(green < 0) then
     call vsmError("negative value:vsmPack3SP2SP green")
   end if

   if(blue < 0) then
     call vsmError("negative value:vsmPack3SP2SP blue")
   end if

   PixelColor = red + 1000*green + 1000000*blue
   if(PixelColor > 1000000000) then
     call vsmLogPrint("Pack:PixelColor=",PixelColor)
     call vsmError("big value:pack PixelColor")
   end if
   
   intcol = PixelColor
 end subroutine vsmPack3SP21INT

 subroutine vsmUnpack1SP23SP(color, sp)
   real, dimension(3), intent(out) :: color
   real, intent(in) :: sp
   integer :: red, green, blue, PixelColor
   integer :: MaxConst
   real :: fsp

#ifndef CTRANS
   PixelColor = transfer(sp, 1)
#else
   fsp = sp
   call VSMTRANSFER4F2I(fsp, PixelColor)
#endif
   if(PixelColor > 1000000000) then
     call vsmLogPrint("Unpack:PixelColor=",PixelColor)
     call vsmLogPrint("Unpack:sp=",sp)
     call vsmError("big value:unpack PixelColor")
   end if

   blue  = PixelColor/1000000
   green = (PixelColor - blue*1000000)/1000
   red   =  PixelColor - blue*1000000 - green*1000

   MaxConst = vsmPackConst*1.1
   if(red < 0 .or. red > MaxConst) then
     call vsmLogPrint("red=",red)
     call vsmError("negative or very big value:vsmUnpack3SP2SP red")
   end if

   if(green < 0 .or. green > MaxConst) then
     call vsmLogPrint("green=",green)
     call vsmError("negative or very big value:vsmUnpack3SP2SP green")
   end if

   if(blue < 0 .or. blue > MaxConst) then
     call vsmLogPrint("blue=",blue)
     call vsmError("negative or very big value:vsmUnpack3SP2SP blue")
   end if

   color(1) = real(red)/vsmPackConst
   color(2) = real(green)/vsmPackConst
   color(3) = real(blue)/vsmPackConst

 end subroutine vsmUnpack1SP23SP

subroutine vsmUnpack1INT23SP(color, intcol)
   real, dimension(3), intent(out) :: color
   integer, intent(in) :: intcol
   integer :: red, green, blue, PixelColor
   integer :: MaxConst

   PixelColor = intcol
   if(PixelColor > 1000000000) then
     call vsmLogPrint("Unpackint:PixelColor=",PixelColor)
!     print *,"",pixelcolor
     call vsmError("big value:unpackint PixelColor")
   end if

   blue  = PixelColor/1000000
   green = (PixelColor - blue*1000000)/1000
   red   =  PixelColor - blue*1000000 - green*1000

   MaxConst = vsmPackConst*1.1
   if(red < 0 .or. red > MaxConst) then
     call vsmLogPrint("red=",red)
     call vsmError("negative or very big value:vsmUnpack3SP2SP red")
   end if

   if(green < 0 .or. green > MaxConst) then
     call vsmLogPrint("green=",green)
     call vsmError("negative or very big value:vsmUnpack3SP2SP green")
   end if

   if(blue < 0 .or. blue > MaxConst) then
     call vsmLogPrint("blue=",blue)
     call vsmError("negative or very big value:vsmUnpack3SP2SP blue")
   end if

   color(1) = real(red)/vsmPackConst
   color(2) = real(green)/vsmPackConst
   color(3) = real(blue)/vsmPackConst

 end subroutine vsmUnpack1INT23SP

#ifdef NO_ISNAN
 logical function isnan(v)
   real(kind=vdp) :: v
   isnan = .false.
   if(v .ne. v) then
       isnan = .true.
   endif
 end function isnan
#endif

end module vsmFunction

