module vsmSolidDrawer
 use vsmConst
 use vsmFunction
 use vsmCameraData
 use vsmImage
 use vsmLightData
 use vsmDepthQ
 use vsmParam
 use vsmCoordinates
!$ use omp_lib
 implicit none
 private

! type(vismo_SolidParticle), private, save:: SolidPoint

 integer,save :: SolidPointNum
 real(kind=vdp), dimension(:, :), allocatable,save ::  SolidPointcoord ! (:, 3)
 real, dimension(:, :), allocatable,save ::  SolidPointnorm ! (:, 3)
 real, dimension(:, :), allocatable,save ::  SolidPointcolor ! (:, 3)
 real(kind=vdp), dimension(:), allocatable,save :: SolidPointdepth
 integer, dimension(:), allocatable,save :: SolidPointpxlpos
 integer, dimension(:), allocatable,save :: SolidPointid

 public vsmDrawCylinder, vsmDrawSphere, vsmDrawCone

 contains

! subroutine vsmAllocSolidParticle(Particle, Num)
 subroutine vsmAllocSolidParticle(Num)
!   type(vismo_SolidParticle) :: Particle
   integer, intent(in) :: Num
   integer :: n, allocstate
   
   n = Num
 !  Particle%Num = n
   SolidPointNum = n
    if(mod(n, 2) == 0) then
     n = n + 1
   end if
   
!   allocate(Particle%coord(n, 3), Particle%norm(n, 3), Particle%color(n, 3), stat = allocstate)
   allocate(SolidPointcoord(n, 3), SolidPointnorm(n, 3), SolidPointcolor(n, 3), stat = allocstate)
   if(allocstate /= 0) then
     call vsmLogPrint("allocstate",allocstate)
     call vsmError("can't allocate Particle coord")
   end if

!   allocate(Particle%depth(n), Particle%pxlpos(n), Particle%id(n), stat = allocstate)
   allocate(SolidPointdepth(n), SolidPointpxlpos(n), SolidPointid(n), stat = allocstate)
   if(allocstate /= 0) then
     call vsmLogPrint("allocstate",allocstate)
     call vsmError("can't allocate Particle depth")
   end if

  end subroutine vsmAllocSolidParticle

!  subroutine vsmDeallocSolidParticle(Particle)
  subroutine vsmDeallocSolidParticle
!   type(vismo_SolidParticle) :: Particle
!   if(allocated(Particle%coord)) deallocate(Particle%coord)
!   if(allocated(Particle%norm)) deallocate(Particle%norm)
!   if(allocated(Particle%color)) deallocate(Particle%color)
!   if(allocated(Particle%depth)) deallocate(Particle%depth)
!   if(allocated(Particle%pxlpos)) deallocate(Particle%pxlpos)
!   if(allocated(SolidPointid)) deallocate(SolidPointid)
   if(allocated(SolidPointcoord)) deallocate(SolidPointcoord)
   if(allocated(SolidPointnorm)) deallocate(SolidPointnorm)
   if(allocated(SolidPointcolor)) deallocate(SolidPointcolor)
   if(allocated(SolidPointdepth)) deallocate(SolidPointdepth)
   if(allocated(SolidPointpxlpos)) deallocate(SolidPointpxlpos)
   if(allocated(SolidPointid)) deallocate(SolidPointid)
  end subroutine vsmDeallocSolidParticle

 subroutine vsmInitSolidParticle(totalnum)
!   integer :: i
   integer, intent(in) :: totalnum
   call vsmLogPrint("total particle", totalnum)
   !print *,"solid total num", totalnum
!   call vsmAllocSolidParticle(SolidPoint, totalnum)
   call vsmAllocSolidParticle(totalnum)
   call vsmLogPrint("alloc success")
   call vsmClearSolidData
   call vsmLogPrint("clear data")
 end subroutine vsmInitSolidParticle

 subroutine vsmFinalizeSolidParticle
!  call vsmDeallocSolidParticle(SolidPoint)
  call vsmDeallocSolidParticle
 end subroutine vsmFinalizeSolidParticle

 subroutine vsmClearSolidData
   integer :: i
!$omp parallel do private(i) schedule(static)
   do i=1, SolidPointNum
     SolidPointcoord(i, 1) = vsmFarAway
     SolidPointcoord(i, 2) = vsmFarAway
     SolidPointcoord(i, 3) = vsmFarAway
     SolidPointnorm(i, 1) = 0.0_vdp; SolidPointnorm(i, 2) = 0.0_vdp; SolidPointnorm(i, 3) = 0.0_vdp
     SolidPointcolor(i, 1) = 0.0; SolidPointcolor(i, 2) = 0.0; SolidPointcolor(i, 3) = 0.0
     SolidPointdepth(i) = vsmFarAway
     SolidPointpxlpos(i) = -1
     SolidPointid(i) = -2
   end do
 end subroutine vsmClearSolidData

 function vsmCheckDraw(ba_pixel)
  logical :: vsmCheckDraw
  integer, dimension(2,4), intent(in) :: ba_pixel

  if(ba_pixel(1,1) > vsmImageData%width) then
    vsmCheckDraw = .false.
    return
  end if

  if(ba_pixel(1,2) < 1) then
    vsmCheckDraw = .false.
    return
  end if

  if(ba_pixel(2,1) > vsmImageData%height) then
    vsmCheckDraw = .false.
    return
  end if

  if(ba_pixel(2,4) < 1) then ! 2,3 -> 2,4 2013/8/30
    vsmCheckDraw = .false.
    return
  end if

  vsmCheckDraw = .true.

 end function vsmCheckDraw

 subroutine vsmDrawSphere(pos, radius, cam_num, vis_num, amb, dif, spec, shin)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), intent(in) :: radius
   integer, intent(in) :: cam_num, vis_num
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin

   real(kind=vdp), dimension(3, 8) :: bv_verts
   integer, dimension(2,20) :: pix
   integer, dimension(2,4) :: ba_pix
   integer :: i
   logical, dimension(20) :: cross
   logical, dimension(8) :: above ! 12->8 2013/8/30
   real(kind=vdp), dimension(3, 12) :: cross_points
   integer :: num_cp
   logical :: draw = .false.

   cross = .false.
   num_cp = 0
 !  call vsmLogPrint("-- vsmDrawSphere")
 !  call vsmLogPrint("calc bvolume sphere")
   call vsmCalcBvolumeSphere(pos, radius, bv_verts)

 !  call vsmLogPrint("pos: ", pos(:)); call vsmLogPrint("radius: ", radius)

!   call vsmLogPrint("bv verts1: ", bv_verts(:,1))
!   call vsmLogPrint("bv verts2: ", bv_verts(:,2))
!   call vsmLogPrint("bv verts3: ", bv_verts(:,3))
!   call vsmLogPrint("bv verts4: ", bv_verts(:,4))
!   call vsmLogPrint("bv verts5: ", bv_verts(:,5))
!   call vsmLogPrint("bv verts6: ", bv_verts(:,6))
!   call vsmLogPrint("bv verts7: ", bv_verts(:,7))
!   call vsmLogPrint("bv verts8: ", bv_verts(:,8))

  ! call vsmLogPrint("check bv")
   call vsmCheckBv(bv_verts, cam_num, radius, cross_points, num_cp, above)

   !if(vsmDbg .and. num_cp > 0) then
  !   call vsmLogPrint("cross points1: ", cross_points(:,1))
  !   call vsmLogPrint("cross points2: ", cross_points(:,2))
   !  call vsmLogPrint("cross points3: ", cross_points(:,3))
  !   call vsmLogPrint("cross points4: ", cross_points(:,4))
  !   call vsmLogPrint("cross points5: ", cross_points(:,5))
  !   call vsmLogPrint("cross points6: ", cross_points(:,6))
  !   call vsmLogPrint("cross points7: ", cross_points(:,7))
  !   call vsmLogPrint("cross points8: ", cross_points(:,8))
  !   call vsmLogPrint("cross points9: ", cross_points(:,9))
  !   call vsmLogPrint("cross points10: ", cross_points(:,10))
  !   call vsmLogPrint("cross points11: ", cross_points(:,11))
  !   call vsmLogPrint("cross points12: ", cross_points(:,12))
   !end if
  ! call vsmLogPrint("point projection 1")
     do i=1, 8
       if(.not.above(i)) cycle
       call vsmPointProjection(bv_verts(1:3, i), cam_num, cross(i), pix(1:2, i))
       draw = .true.
    !if(vsmDbg .and. cross(i)) then
    !  call vsmLogPrint("cross true:", i)
    !  call vsmLogPrint("bvverts:",bv_verts(:, i))
    !  call vsmLogPrint("pix:",pix(:,i))
    !end if
     end do

 !  call vsmLogPrint("point projection 2")
     do i=9, 8+num_cp
      call vsmPointProjection(cross_points(1:3, i-8), cam_num, cross(i), pix(1:2, i))
      draw = .true.
    !if(vsmDbg .and. cross(i)) then
     ! call vsmLogPrint("cross true:")
     ! call vsmLogPrint("crosspoints:",cross_points(:, i-8))
     ! call vsmLogPrint("pix:",pix(:,i))
    !end if
     end do

   if(draw) then
     call vsmBoundarea(pix, ba_pix, cross)
#ifdef OPNEST
     call vsmRaycastSphereOMP(pos, radius, ba_pix, cam_num, vis_num, amb, dif, spec, shin)
#else
     call vsmRaycastSphere(pos, radius, ba_pix, cam_num, vis_num, amb, dif, spec, shin)
#endif
   end if
 end subroutine vsmDrawSphere

 subroutine vsmDrawCone(pos1, pos2, radius, cam_num, vis_num, amb, dif, spec, shin)
   real(kind=vdp), dimension(3), intent(in) :: pos1, pos2 ! pos1:bottom, pos2:top
   real(kind=vdp), intent(in) :: radius
   integer, intent(in) :: cam_num, vis_num
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin

   real(kind=vdp), dimension(3) :: norm
   real(kind=vdp) :: height

   real(kind=vdp), dimension(3, 8) :: bv_verts
   integer, dimension(2, 20) :: pix
   integer, dimension(2, 4) :: ba_pix
   integer :: i
   logical, dimension(20) :: cross
   logical, dimension(8) :: above ! 12->8 2013/8/30
   real(kind=vdp), dimension(3, 12) :: cross_points
   integer :: num_cp
   logical :: draw = .false.
   type(vismo_transdata) :: tr_data

   cross = .false.
   num_cp = 0
   norm(1:3) = pos2(1:3) - pos1(1:3)
   height = sqrt(dot_product(norm, norm))

   if(height <= abs(vsmImageData%dp)*0.1_vdp) return

   call vsmNormVec(norm)
   call vsmRotateMatrix(norm, tr_data%matrix)
   call vsmRotateInvMatrix(norm, tr_data%invmatrix)
   tr_data%pos(1:3) = pos1(1:3)
   call vsmCalcBvolumeCylinder(height, radius, bv_verts)

   call vsmCheckBv(bv_verts, cam_num, radius, cross_points, num_cp, above, tr_data)

     do i=1, 8
      if(.not.above(i)) cycle
      call vsmPointProjection(bv_verts(1:3, i), cam_num, cross(i), pix(1:2, i), tr_data)
      draw = .true.
     end do

     do i=9, 8+num_cp
      call vsmPointProjection(cross_points(1:3, i-8), cam_num, cross(i), pix(1:2, i), tr_data)
      draw = .true.
     end do

   if(draw) then
     call vsmBoundarea(pix, ba_pix, cross)
#ifdef OMPNEST
     call vsmRaycastConeOMP(tr_data, radius, height, ba_pix, cam_num, vis_num, amb, dif, spec, shin)
#else
     call vsmRaycastCone(tr_data, radius, height, ba_pix, cam_num, vis_num, amb, dif, spec, shin)
#endif
   end if
 end subroutine vsmDrawCone

 subroutine vsmDrawCylinder(pos1, pos2, radius, cam_num, vis_num, amb, dif, spec, shin)
   real(kind=vdp), dimension(3), intent(in) :: pos1, pos2 ! pos1:bottom, pos2:top
   real(kind=vdp), intent(in) :: radius
   integer, intent(in) :: cam_num, vis_num
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin

   real(kind=vdp), dimension(3) :: norm
   real(kind=vdp) :: height

   real(kind=vdp), dimension(3, 8) :: bv_verts
   integer, dimension(2, 20) :: pix
   integer, dimension(2, 4) :: ba_pix
   integer :: i
   logical, dimension(20) :: cross
   logical, dimension(8) :: above ! 12->8 2013/8/30
   real(kind=vdp), dimension(3, 12) :: cross_points
   integer :: num_cp
   logical :: draw = .false.
   !real(kind=vdp), dimension(3, 3) :: matrix
   type(vismo_transdata) :: tr_data

   cross = .false.
   num_cp = 0
   norm(1:3) = pos2(1:3) - pos1(1:3)
   height = sqrt(dot_product(norm, norm))

   call vsmNormVec(norm)
   call vsmRotateMatrix(norm, tr_data%matrix)
   call vsmRotateInvMatrix(norm, tr_data%invmatrix)
   tr_data%pos(1:3) = pos1(1:3)

   call vsmCalcBvolumeCylinder(height, radius, bv_verts)
   call vsmCheckBv(bv_verts, cam_num, radius, cross_points, num_cp, above, tr_data)

     do i=1, 8
      if(above(i)) then
        call vsmPointProjection(bv_verts(1:3, i), cam_num, cross(i), pix(1:2, i), tr_data)
        draw = .true.
      end if
     end do

     do i=9, 8+num_cp
      call vsmPointProjection(cross_points(1:3, i-8), cam_num, cross(i), pix(1:2, i), tr_data)
      draw = .true.
     end do

   if(draw) then
     call vsmBoundarea(pix, ba_pix, cross)
#ifdef OMPNEST
     call vsmRaycastCylinderOMP(tr_data, radius, height, ba_pix, cam_num, vis_num, amb, dif, spec, shin)
#else
     call vsmRaycastCylinder(tr_data, radius, height, ba_pix, cam_num, vis_num, amb, dif, spec, shin)
#endif
   end if
 end subroutine vsmDrawCylinder
 
 subroutine vsmRaycastSphere(pos, radius, ba_pixel, cam_num, vis_num, amb, dif, spec, shin)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), intent(in) :: radius
   integer, dimension(2,4), intent(in) :: ba_pixel
   integer, intent(in) :: cam_num, vis_num
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin

   integer :: i,j
!   integer, dimension(2) :: px
   real(kind=vdp), dimension(3) :: epos, topleft, cpos
   real(kind=vdp), dimension(3) :: img_right, img_down
   real(kind=vdp), dimension(3) :: pixp, direc
!   real(kind=vdp) :: A, B, C, D, t, t1, t2, dist
   real(kind=vdp) :: B, C, D, t, t1, t2, dist
   real(kind=vdp), dimension(3) :: cr_sph, fpos ! cross point 
   real, dimension(3) :: lumi, normal
!   real :: dq_factor
   real, dimension(4, vsmPbLen) :: picbuf
   integer :: pc, ps, pe, wlen ! pixel count, pixel start, pixel end

   if(.not.vsmCheckDraw(ba_pixel)) then
     return
   end if

   !call vsmLogPrint("-- raycast sphere")
   ! translation :: the center of the sphere moves to the origin
   cpos(1:3) = vsmCamera(cam_num)%pos(1:3)
   epos(1:3)      = cpos(1:3) - pos(1:3)
   topleft(1:3)   = vsmImageData%topleft(1:3) - pos(1:3)
   img_right(1:3) = vsmCamera(cam_num)%right(1:3)
   img_down(1:3)  = -vsmCamera(cam_num)%up(1:3)

   !if(vsmDbg) then
   !  call vsmLogPrint("bapixel(1,1): ", ba_pixel(1,1))
   !  call vsmLogPrint("bapixel(1,2): ", ba_pixel(1,2))
   !  call vsmLogPrint("bapixel(2,1): ", ba_pixel(2,1))
   !  call vsmLogPrint("bapixel(2,3): ", ba_pixel(2,3))
   !end if

   pc = 0; ps = 1; pe = 0
   wlen = ba_pixel(1,2) - ba_pixel(1,1) + 1
   picbuf(1:4, 1:vsmPbLen) = vsmFarAway*1.1_vdp
   do j = ba_pixel(2,1), ba_pixel(2, 4)
     do i = ba_pixel(1,1), ba_pixel(1,2)
       pc=pc+1
       if(pc == vsmPbLen+1) then
         call vsmPaintPixel3ms(picbuf, ps, vsmPBLen, ba_pixel(1,1), ba_pixel(2,1), wlen)
         ps=ps+vsmPbLen; pc=1; pe=0
         picbuf(1:4, 1:vsmPbLen) = vsmFarAway*1.1_vdp
       end if

       pixp(1:3) = topleft(1:3) + (i-1)*vsmImageData%dp*img_right(1:3) + &
                   (j-1)*vsmImageData%dp*img_down(1:3)
    !   call vsmLogPrint("realpos:pixel ", pixp)
       if(vsmCamera(cam_num)%pers) then
         direc(1:3) = pixp(1:3) - epos(1:3)
       else
         direc(1:3) = vsmCamera(cam_num)%front(1:3)
       end if
       call vsmNormVec(direc)
       !A = 1.0_vdp  ! dot_product(direc, direc)
       B = dot_product(direc, pixp)
       C = dot_product(pixp, pixp) - radius**2
       ! determinant
       D = B**2 - C
       if(D < 0.0_vdp) then
         cycle
       else if(D == 0.0_vdp) then
         t = -B
         if(t < 0.0_vdp) cycle
           cr_sph(1:3) = pixp(1:3) + direc(1:3)*t
       else if(D > 0.0_vdp) then
         t1 = (-B+sqrt(D))
         t2 = (-B-sqrt(D))
         if(t1 >= 0.0_vdp .and. t2 >= 0.0_vdp) then
           t = min(t1,t2)
         else if(t1 < 0.0_vdp .and. t2 >= 0.0_vdp) then
           t = t2
         else if(t1 >= 0.0_vdp .and. t2 < 0.0_vdp) then
           t = t1
         else 
           cycle
         end if
         cr_sph(1:3) = pixp(1:3) + direc(1:3)*t
       end if

       normal(1:3) = real(cr_sph(1:3))
       call vsmNormVecF(normal)
       fpos(1:3) = cr_sph(1:3)+pos(1:3)

       if(vsmCamera(cam_num)%pers) then
           if(vsmCoordData%YinYang == vsmYin) then
              call vsmCalcColorv(amb, dif, spec, shin, cpos, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            else
              call vsmCalcColorvAn(amb, dif, spec, shin, cpos, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            end if
            dist = sqrt(dot_product(cr_sph-epos, cr_sph-epos))
        else
           if(vsmCoordData%YinYang == vsmYin) then
              call vsmCalcColorvOrtho(amb, dif, spec, shin, direc, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            else
              call vsmCalcColorvOrthoAn(amb, dif, spec, shin, direc, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            end if
            dist = t + vsmCamera(cam_num)%near
        end if

!       if(vsmVisParam(vis_num)%depthq) then
!         dq_factor = vsmDepthQFactor(dist)
!         lumi(1:3) = dq_factor*lumi(1:3)
!       end if

!       px(1) = i; px(2) = j
!       call vsmPaintPixel3s(px, lumi, dist)
        picbuf(1:3,pc) = lumi
        picbuf(4,pc) = real(dist)

       !if(i < 1 .or. i > vsmImageData%width .or. j < 1 .or. j > vsmImageData%height) then
       !   print *, "raycast sphere pixel position", i, j
       !end if
       if(lumi(1) < 0.0 .or. lumi(2) < 0.0 .or. lumi(3) < 0.0) print *, "raycast sphere"

       !call vsmLogPrint("lumi:", lumi)
       !call vsmLogPrint("dist:", dist)
     end do
   end do 
   if(pc >= 1) then
      pe=pc
      call vsmPaintPixel3ms(picbuf, ps, pe, ba_pixel(1,1), ba_pixel(2,1), wlen)
    end if

   !call vsmLogPrint("-- raycast sphere ends")
 end subroutine vsmRaycastSphere

 subroutine vsmRaycastSphereOMP(pos, radius, ba_pixel, cam_num, vis_num, amb, dif, spec, shin)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), intent(in) :: radius
   integer, dimension(2,4), intent(in) :: ba_pixel
   integer, intent(in) :: cam_num, vis_num
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin

   integer :: i,j,k,l
   real(kind=vdp), dimension(3) :: epos, topleft, cpos
   real(kind=vdp), dimension(3) :: img_right, img_down
   real(kind=vdp), dimension(3) :: pixp, direc
   real(kind=vdp) :: B, C, D, t, t1, t2, dist
   real(kind=vdp), dimension(3) :: cr_sph, fpos ! cross point 
   real, dimension(3) :: lumi, normal
   real, dimension(4, vsmPbLen) :: picbuf
   integer :: ps, wlen
   integer :: comm, myrank, psize, ompnum, snum
   integer :: totallen, loop, res, lsize, offset

   if(.not.vsmCheckDraw(ba_pixel)) then
     return
   end if
   call vsmGetMPIVars(comm, myrank, psize, ompnum)
   snum = ompnum/vsmMaxOMP4Solid
   if(snum < 2) snum = 1

   !call vsmLogPrint("-- raycast sphere")
   ! translation :: the center of the sphere moves to the origin
   cpos(1:3) = vsmCamera(cam_num)%pos(1:3)
   epos(1:3)      = cpos(1:3) - pos(1:3)
   topleft(1:3)   = vsmImageData%topleft(1:3) - pos(1:3)
   img_right(1:3) = vsmCamera(cam_num)%right(1:3)
   img_down(1:3)  = -vsmCamera(cam_num)%up(1:3)

   ps = 1
   wlen = ba_pixel(1,2) - ba_pixel(1,1) + 1

!$omp parallel do private(i) num_threads(snum)
   do i=1, vsmPblen
     picbuf(1:4, i) = vsmFarAway*1.1_vdp
   end do

   totallen = (ba_pixel(2,1) - ba_pixel(2, 4) )* (ba_pixel(1,1) - ba_pixel(1,4))
   loop = totallen/vsmPbLen
   res = mod(totallen,vsmPbLen)

!   do j = ba_pixel(2,1), ba_pixel(2, 4)
!     do i = ba_pixel(1,1), ba_pixel(1,2)
   do k=1, loop+1
     if(k==loop+1) then
       lsize = res
     else
       lsize = vsmPblen
     end if
     offset = (k-1)*vsmPblen
!$omp parallel do private(l,i,j,pixp,direc,b,c,d,t1,t2,cr_sph,t,normal,fpos,lumi,dist) num_threads(snum)
     do l=1, lsize
       j = (l+offset-1)/wlen+ba_pixel(2,1)
       i = mod(l+offset-1, wlen)+ba_pixel(1,1)
       pixp(1:3) = topleft(1:3) + (i-1)*vsmImageData%dp*img_right(1:3) + &
                   (j-1)*vsmImageData%dp*img_down(1:3)
    !   call vsmLogPrint("realpos:pixel ", pixp)
       if(vsmCamera(cam_num)%pers) then
         direc(1:3) = pixp(1:3) - epos(1:3)
       else
         direc(1:3) = vsmCamera(cam_num)%front(1:3)
       end if
       call vsmNormVec(direc)
       !A = 1.0_vdp  ! dot_product(direc, direc)
       B = dot_product(direc, pixp)
       C = dot_product(pixp, pixp) - radius**2
       ! determinant
       D = B**2 - C
       if(D < 0.0_vdp) then
         cycle
       else if(D == 0.0_vdp) then
         t = -B
         if(t < 0.0_vdp) cycle
           cr_sph(1:3) = pixp(1:3) + direc(1:3)*t
       else if(D > 0.0_vdp) then
         t1 = (-B+sqrt(D))
         t2 = (-B-sqrt(D))
         if(t1 >= 0.0_vdp .and. t2 >= 0.0_vdp) then
           t = min(t1,t2)
         else if(t1 < 0.0_vdp .and. t2 >= 0.0_vdp) then
           t = t2
         else if(t1 >= 0.0_vdp .and. t2 < 0.0_vdp) then
           t = t1
         else 
           cycle
         end if
         cr_sph(1:3) = pixp(1:3) + direc(1:3)*t
       end if

       normal(1:3) = real(cr_sph(1:3))
       call vsmNormVecF(normal)
       fpos(1:3) = cr_sph(1:3)+pos(1:3)

       if(vsmCamera(cam_num)%pers) then
           if(vsmCoordData%YinYang == vsmYin) then
              call vsmCalcColorv(amb, dif, spec, shin, cpos, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            else
              call vsmCalcColorvAn(amb, dif, spec, shin, cpos, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            end if
            dist = sqrt(dot_product(cr_sph-epos, cr_sph-epos))
        else
           if(vsmCoordData%YinYang == vsmYin) then
              call vsmCalcColorvOrtho(amb, dif, spec, shin, direc, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            else
              call vsmCalcColorvOrthoAn(amb, dif, spec, shin, direc, &
                          fpos, normal, vsmVisParam(vis_num)%light_used, lumi)
            end if
            dist = t + vsmCamera(cam_num)%near
        end if

        picbuf(1:3,l) = lumi
        picbuf(4,l) = real(dist)

       if(lumi(1) < 0.0 .or. lumi(2) < 0.0 .or. lumi(3) < 0.0) print *, "raycast sphere"
     end do
     call vsmPaintPixel3ms(picbuf, ps, lsize, ba_pixel(1,1), ba_pixel(2,1), wlen)
     ps=ps+vsmPbLen
!$omp parallel do private(i) num_threads(snum)
     do i=1, vsmPblen
       picbuf(1:4, i) = vsmFarAway*1.1_vdp
     end do
   end do 
 end subroutine vsmRaycastSphereOMP

 subroutine vsmRaycastCylinder(transdat, radius, height, ba_pixel, cam_num, vis_num, amb, dif, spec, shin)
   type(vismo_transdata), intent(in) :: transdat
   real(kind=vdp), intent(in) :: radius, height
   integer, dimension(2,4), intent(in) :: ba_pixel
   integer, intent(in) :: cam_num, vis_num
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin

   integer :: i, j
   real(kind=vdp), dimension(3) :: epos, topleft, front
   real(kind=vdp), dimension(3) :: img_right, img_down
   real(kind=vdp), dimension(3) :: pixp, direc, tpos
   real(kind=vdp) :: A, B, D, t, t1, t2, dist
   real(kind=vdp), dimension(3) :: cr_cyl,cr_cyl1,cr_cyl2 ! cross point 
   real, dimension(3) :: lumi, tnorm
   real, dimension(4, vsmPbLen) :: picbuf
   integer :: pc, ps, pe, wlen

   !call vsmLogPrint("raycast cylinder:")

   if(.not.vsmCheckDraw(ba_pixel)) then
     return
   end if

   ! translation and rotation:: the center of the bottom of cylinder moves to the origin
   tpos(1:3) = vsmCamera(cam_num)%pos(1:3) - transdat%pos(1:3)
   call vsmMatrixCrossVect(transdat%matrix, tpos, epos)
   tpos(1:3) = vsmImageData%topleft(1:3) - transdat%pos(1:3)
   call vsmMatrixCrossVect(transdat%matrix, tpos, topleft)
   call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%right, img_right)
   tpos(1:3) = -vsmCamera(cam_num)%up
   call vsmMatrixCrossVect(transdat%matrix, tpos, img_down)

   if(.not.vsmCamera(cam_num)%pers) then
     call  vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%front, front)
   end if

   pc = 0; ps = 1; pe = 0
   wlen = ba_pixel(1,2) - ba_pixel(1,1) + 1
   picbuf(1:4, 1:vsmPbLen) = vsmFarAway*1.1_vdp
   do j=ba_pixel(2,1), ba_pixel(2, 4)
     do i=ba_pixel(1,1), ba_pixel(1,4)
        pc=pc+1
       if(pc == vsmPbLen+1) then
         call vsmPaintPixel3ms(picbuf, ps, vsmPBLen, ba_pixel(1,1), ba_pixel(2,1), wlen)
         ps=ps+vsmPbLen; pc=1; pe=0
         picbuf(1:4, 1:vsmPbLen) = vsmFarAway*1.1_vdp
       end if

       pixp(1:3) = topleft(1:3) + (i-1)*vsmImageData%dp*img_right(1:3) + (j-1)*vsmImageData%dp*img_down(1:3)

       if(vsmCamera(cam_num)%pers) then
         direc(1:3) = pixp(1:3) - epos(1:3)
       else
         direc(1:3) = front(1:3)
       end if

       call vsmNormVec(direc)
       A = direc(1)*direc(1) + direc(2)*direc(2)
       if(A == 0.0_vdp) cycle

       B = direc(1)*pixp(1) + direc(2)*pixp(2)
       D = B*B-A*(pixp(1)*pixp(1)+pixp(2)*pixp(2)-radius*radius)
       if(D < 0.0_vdp) cycle

       t1 = (-B+sqrt(D))/A
       t2 = (-B-sqrt(D))/A

       if(t1 < 0.0_vdp .and. t2 < 0.0_vdp) then
         cycle
       else if(t1 >= 0.0_vdp .and. t2 >= 0.0_vdp) then
         cr_cyl1(1:3) = pixp(1:3) + direc(1:3)*t1
         cr_cyl2(1:3) = pixp(1:3) + direc(1:3)*t2

         if(cr_cyl1(3) >= 0.0_vdp .and. cr_cyl1(3) <= height) then

           if(cr_cyl2(3) >= 0.0_vdp .and. cr_cyl2(3) <= height) then
             if(t1 > t2) then
               t = t2
               cr_cyl(1:3) = cr_cyl2(1:3)
             else
               t = t1
               cr_cyl(1:3) = cr_cyl1(1:3)
             end if
           else
               t = t1
               cr_cyl(1:3) = cr_cyl1(1:3)
           end if

         else if(cr_cyl2(3) >= 0.0_vdp .and. cr_cyl2(3) <= height) then
             t = t2
             cr_cyl(1:3) = cr_cyl2(1:3)
           else
             cycle
         end if

       else if(t1 >= 0.0_vdp .and. t2 < 0.0_vdp) then
         cr_cyl1(1:3) = pixp(1:3) + direc(1:3)*t1
         if(cr_cyl1(3) >= 0.0_vdp .and. cr_cyl1(3) <= height) then
           t = t1
           cr_cyl(1:3) = cr_cyl1(1:3)
         else
           cycle
         end if

       else if(t1 < 0.0_vdp .and. t2 >= 0.0_vdp) then
         cr_cyl2(1:3) = pixp(1:3) + direc(1:3)*t2
         if(cr_cyl2(3) >= 0.0_vdp .and. cr_cyl2(3) <= height) then
           t = t2
           cr_cyl(1:3) = cr_cyl2(1:3)
         else
           cycle
         end if
       end if

       tnorm(1:2) = real(cr_cyl(1:2))
       tnorm(3) = 0.0
       call vsmNormVecF(tnorm)

       if(vsmCamera(cam_num)%pers) then
         if(vsmCoordData%YinYang == vsmYin) then
           call vsmCalcColorv(amb, dif, spec, shin, epos, &
                          cr_cyl, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         else
           call vsmCalcColorvAn(amb, dif, spec, shin, epos, &
                          cr_cyl, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         end if
         dist = sqrt(dot_product(cr_cyl-epos, cr_cyl-epos))
       else
         if(vsmCoordData%YinYang == vsmYin) then
              call vsmCalcColorvOrtho(amb, dif, spec, shin, direc, &
                          cr_cyl, tnorm, vsmVisParam(vis_num)%light_used, lumi)
         else
              call vsmCalcColorvOrthoAn(amb, dif, spec, shin, direc, &
                          cr_cyl, tnorm, vsmVisParam(vis_num)%light_used, lumi)
         end if
         dist = t + vsmCamera(cam_num)%near
       end if

        picbuf(1:3,pc) = lumi(1:3)
        picbuf(4,pc) = real(dist)

       if(lumi(1) < 0.0 .or. lumi(2) < 0.0 .or. lumi(3) < 0.0) print *, "raycast cylinder"

     end do
   end do 

   if(pc >= 1) then
      pe=pc
      call vsmPaintPixel3ms(picbuf, ps, pe, ba_pixel(1,1), ba_pixel(2,1), wlen)
    end if
 end subroutine vsmRaycastCylinder

 subroutine vsmRaycastCylinderOMP(transdat, radius, height, ba_pixel, cam_num, vis_num, amb, dif, spec, shin)
   type(vismo_transdata), intent(in) :: transdat
   real(kind=vdp), intent(in) :: radius, height
   integer, dimension(2, 4), intent(in) :: ba_pixel
   integer, intent(in) :: cam_num, vis_num
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin

   integer :: i, j, k, l
   real(kind=vdp), dimension(3) :: epos, topleft, front
   real(kind=vdp), dimension(3) :: img_right, img_down
   real(kind=vdp), dimension(3) :: pixp, direc, tpos
   real(kind=vdp) :: A, B, D, t, t1, t2, dist
   real(kind=vdp), dimension(3) :: cr_cyl,cr_cyl1,cr_cyl2 ! cross point 
   real, dimension(3) :: lumi, tnorm
   real, dimension(4, vsmPbLen) :: picbuf
   integer :: ps, wlen
   integer :: comm, myrank, psize, ompnum, snum
   integer :: totallen, loop, res, lsize, offset

   !call vsmLogPrint("raycast cylinder:")

   if(.not.vsmCheckDraw(ba_pixel)) then
     return
   end if
   call vsmGetMPIVars(comm, myrank, psize, ompnum)
   snum = ompnum/vsmMaxOMP4Solid
   if(snum < 2) snum = 1
   ! translation and rotation:: the center of the bottom of cylinder moves to the origin
   tpos(1:3) = vsmCamera(cam_num)%pos(1:3) - transdat%pos(1:3)
   call vsmMatrixCrossVect(transdat%matrix, tpos, epos)
   tpos(1:3) = vsmImageData%topleft(1:3) - transdat%pos(1:3)
   call vsmMatrixCrossVect(transdat%matrix, tpos, topleft)
   call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%right, img_right)
   tpos(1:3) = -vsmCamera(cam_num)%up
   call vsmMatrixCrossVect(transdat%matrix, tpos, img_down)

   if(.not.vsmCamera(cam_num)%pers) then
     call  vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%front, front)
   end if

   ps = 1
   wlen = ba_pixel(1,2) - ba_pixel(1,1) + 1

!$omp parallel do private(i) num_threads(snum)
   do i=1, vsmPblen
     picbuf(1:4, i) = vsmFarAway*1.1_vdp
   end do

   totallen = (ba_pixel(2,1) - ba_pixel(2, 4) )* (ba_pixel(1,1) - ba_pixel(1,4))
   loop = totallen/vsmPbLen
   res = mod(totallen,vsmPbLen)

!   do j=ba_pixel(2,1), ba_pixel(2, 4)
!     do i=ba_pixel(1,1), ba_pixel(1,4)
   do k=1, loop+1
     if(k==loop+1) then
       lsize = res
     else
       lsize = vsmPblen
     end if
     offset = (k-1)*vsmPblen
!$omp parallel do private(l,i,j,pixp,direc,a,b,d,t1,t2,cr_cyl1,cr_cyl2,t,cr_cyl,tnorm,lumi,dist) num_threads(snum)
     do l=1, lsize
       j = (l+offset-1)/wlen+ba_pixel(2,1)
       i = mod(l+offset-1, wlen)+ba_pixel(1,1)

       pixp(1:3) = topleft(1:3) + (i-1)*vsmImageData%dp*img_right(1:3) + (j-1)*vsmImageData%dp*img_down(1:3)
       if(vsmCamera(cam_num)%pers) then
         direc(1:3) = pixp(1:3) - epos(1:3)
       else
         direc(1:3) = front(1:3)
       end if

       call vsmNormVec(direc)
       A = direc(1)*direc(1) + direc(2)*direc(2)
       if(A == 0.0_vdp) cycle

       B = direc(1)*pixp(1) + direc(2)*pixp(2)
       D = B*B-A*(pixp(1)*pixp(1)+pixp(2)*pixp(2)-radius*radius)
       if(D < 0.0_vdp) cycle

       t1 = (-B+sqrt(D))/A
       t2 = (-B-sqrt(D))/A

       if(t1 < 0.0_vdp .and. t2 < 0.0_vdp) then
         cycle
       else if(t1 >= 0.0_vdp .and. t2 >= 0.0_vdp) then
         cr_cyl1(1:3) = pixp(1:3) + direc(1:3)*t1
         cr_cyl2(1:3) = pixp(1:3) + direc(1:3)*t2

         if(cr_cyl1(3) >= 0.0_vdp .and. cr_cyl1(3) <= height) then

           if(cr_cyl2(3) >= 0.0_vdp .and. cr_cyl2(3) <= height) then
             if(t1 > t2) then
               t = t2
               cr_cyl(1:3) = cr_cyl2(1:3)
             else
               t = t1
               cr_cyl(1:3) = cr_cyl1(1:3)
             end if
           else
               t = t1
               cr_cyl(1:3) = cr_cyl1(1:3)
           end if

         else if(cr_cyl2(3) >= 0.0_vdp .and. cr_cyl2(3) <= height) then
             t = t2
             cr_cyl(1:3) = cr_cyl2(1:3)
           else
             cycle
         end if

       else if(t1 >= 0.0_vdp .and. t2 < 0.0_vdp) then
         cr_cyl1(1:3) = pixp(1:3) + direc(1:3)*t1
         if(cr_cyl1(3) >= 0.0_vdp .and. cr_cyl1(3) <= height) then
           t = t1
           cr_cyl(1:3) = cr_cyl1(1:3)
         else
           cycle
         end if

       else if(t1 < 0.0_vdp .and. t2 >= 0.0_vdp) then
         cr_cyl2(1:3) = pixp(1:3) + direc(1:3)*t2
         if(cr_cyl2(3) >= 0.0_vdp .and. cr_cyl2(3) <= height) then
           t = t2
           cr_cyl(1:3) = cr_cyl2(1:3)
         else
           cycle
         end if
       end if

       tnorm(1:2) = real(cr_cyl(1:2))
       tnorm(3) = 0.0
       call vsmNormVecF(tnorm)

       if(vsmCamera(cam_num)%pers) then
         if(vsmCoordData%YinYang == vsmYin) then
           call vsmCalcColorv(amb, dif, spec, shin, epos, &
                          cr_cyl, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         else
           call vsmCalcColorvAn(amb, dif, spec, shin, epos, &
                          cr_cyl, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         end if
         dist = sqrt(dot_product(cr_cyl-epos, cr_cyl-epos))
       else
         if(vsmCoordData%YinYang == vsmYin) then
              call vsmCalcColorvOrtho(amb, dif, spec, shin, direc, &
                          cr_cyl, tnorm, vsmVisParam(vis_num)%light_used, lumi)
         else
              call vsmCalcColorvOrthoAn(amb, dif, spec, shin, direc, &
                          cr_cyl, tnorm, vsmVisParam(vis_num)%light_used, lumi)
         end if
         dist = t + vsmCamera(cam_num)%near
       end if

        picbuf(1:3,l) = lumi(1:3)
        picbuf(4,l) = real(dist)

       if(lumi(1) < 0.0 .or. lumi(2) < 0.0 .or. lumi(3) < 0.0) print *, "raycast cylinder"
     end do

     call vsmPaintPixel3ms(picbuf, ps, lsize, ba_pixel(1,1), ba_pixel(2,1), wlen)
     ps=ps+vsmPbLen
!$omp parallel do private(i) num_threads(snum)
     do i=1, vsmPblen
       picbuf(1:4, i) = vsmFarAway*1.1_vdp
     end do
   end do 

 end subroutine vsmRaycastCylinderOMP

 subroutine vsmRaycastCone(transdat, radius, height, ba_pixel, cam_num, vis_num, amb, dif, spec, shin)
   type(vismo_transdata), intent(in) :: transdat
   real(kind=vdp), intent(in) :: radius, height
   integer, dimension(2,4), intent(in) :: ba_pixel
   integer, intent(in) :: cam_num, vis_num
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin
   integer :: i, j
!   integer, dimension(2) :: px
   real(kind=vdp), dimension(3) :: epos, topleft, front
   real(kind=vdp), dimension(3) :: img_right, img_down
   real(kind=vdp), dimension(3) :: pixp, direc, tpos

   real(kind=vdp) :: A, B, C, D, t, t1, t2, dist
   real(kind=vdp), dimension(3) :: cr_cone, cr_cone1, cr_cone2 ! cross point 
   real, dimension(3) :: lumi, tnorm
!   real :: dq_factor
   real, dimension(4, vsmPbLen) :: picbuf
   integer :: pc, ps, pe, wlen ! pixel count, pixel start, pixel end
   integer :: cone_state

   if(.not.vsmCheckDraw(ba_pixel)) then
     return
   end if

   ! translation and rotation:: the center of the bottom of cylinder moves to the origin
   tpos(1:3) = vsmCamera(cam_num)%pos(1:3) - transdat%pos(1:3)
   call vsmMatrixCrossVect(transdat%matrix, tpos, epos)

   tpos(1:3) = vsmImageData%topleft(1:3) - transdat%pos(1:3)
   call vsmMatrixCrossVect(transdat%matrix, tpos, topleft)

   call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%right, img_right)

   tpos(1:3) = -vsmCamera(cam_num)%up
   call vsmMatrixCrossVect(transdat%matrix, tpos, img_down)

   if(.not.vsmCamera(cam_num)%pers) then
     call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%front, front)
     call vsmNormVec(front)
   end if

   pc = 0; ps = 1; pe = 0
   wlen = ba_pixel(1,2) - ba_pixel(1,1) + 1
   picbuf(1:4, 1:vsmPbLen) = vsmFarAway*1.1_vdp

   do j=ba_pixel(2,1), ba_pixel(2, 4)
     do i=ba_pixel(1,1), ba_pixel(1,4)
       pc=pc+1
       if(pc == vsmPbLen+1) then
         call vsmPaintPixel3ms(picbuf, ps, vsmPBLen, ba_pixel(1,1), ba_pixel(2,1), wlen)
         ps=ps+vsmPbLen; pc=1; pe=0
         picbuf(1:4, 1:vsmPbLen) = vsmFarAway*1.1_vdp
       end if

       pixp(1:3) = topleft(1:3) + (i-1)*vsmImageData%dp*img_right(1:3) + (j-1)*vsmImageData%dp*img_down(1:3)

       if(vsmCamera(cam_num)%pers) then
         direc(1:3) = pixp(1:3) - epos(1:3)
         call vsmNormVec(direc)
       else
         direc(1:3) = front(1:3)
       end if

       A = direc(1)*direc(1) + direc(2)*direc(2) - radius*radius*direc(3)*direc(3)/(height*height)

       if(A == 0.0_vdp) cycle

       B = direc(1)*pixp(1)+direc(2)*pixp(2)+radius*radius*direc(3)*(-pixp(3)/(height*height)+1.0_vdp/height)
       C = pixp(1)*pixp(1)+pixp(2)*pixp(2)+&
            radius*radius*(-1.0_vdp-pixp(3)*pixp(3)/(height*height)+2.0_vdp*pixp(3)/height)
       D = B*B-A*C

       if(D < 0.0_vdp) cycle

       t1 = (-B+sqrt(D))/A
       t2 = (-B-sqrt(D))/A

!       cr_cone = vsmFarAway
       cone_state = -1
       cr_cone1(1:3) = pixp(1:3) + direc(1:3)*t1
       cr_cone2(1:3) = pixp(1:3) + direc(1:3)*t2
       if(t1 < 0.0_vdp .and. t2 < 0.0_vdp) then
           cone_state = 0
!          cycle
       else if(t1 >= 0.0_vdp .and. t2 >= 0.0_vdp) then
!          cr_cone1(1:3) = pixp(1:3) + direc(1:3)*t1
!          cr_cone2(1:3) = pixp(1:3) + direc(1:3)*t2
          if(cr_cone1(3) >= 0.0_vdp .and. cr_cone1(3) <= height) then
            if(cr_cone2(3) >= 0.0_vdp .and. cr_cone2(3) <= height) then
               if(t1 > t2) then
!                 t = t2
!                 cr_cone(1:3) = cr_cone2(1:3)
                 cone_state = 2
               else
!                 t = t1
!                 cr_cone(1:3) = cr_cone1(1:3)
                 cone_state = 1
               end if
            else
!                t = t1
!                cr_cone(1:3) = cr_cone1(1:3)
                cone_state = 1
            end if
         else if(cr_cone2(3) >= 0.0_vdp .and. cr_cone2(3) <= height) then
!             t = t2
!             cr_cone(1:3) = cr_cone2(1:3)
             cone_state = 2
        else
!             cycle
             cone_state = 0
        end if
       else if(t1 >= 0.0_vdp .and. t2 < 0.0_vdp) then
         cr_cone1(1:3) = pixp(1:3) + direc(1:3)*t1
         if(cr_cone1(3) >= 0.0_vdp .and. cr_cone1(3) <= height) then
!           t = t1
!           cr_cone(1:3) = cr_cone1(1:3)
           cone_state = 1
         else
!           cycle
            cone_state = 0
         end if
       else if(t1 < 0.0_vdp .and. t2 >= 0.0_vdp) then
         cr_cone2(1:3) = pixp(1:3) + direc(1:3)*t2
         if(cr_cone2(3) >= 0.0_vdp .and. cr_cone2(3) <= height) then
!           t = t2
!           cr_cone(1:3) = cr_cone2(1:3)
           cone_state = 2
         else
!           cycle
           cone_state = 0
         end if
       end if

!       if(cr_cone(1) > vsmFarAway*0.1 .or. cr_cone(2) > vsmFarAway*0.1 .or. cr_cone(3) > vsmFarAway*0.1) then
!          print *,cr_cone
!          call vsmError("cr_cone undefined")
!       end if

       select case (cone_state)
           case (0)
              cycle
           case (1)
              t = t1
              cr_cone(1:3) = cr_cone1(1:3)
           case (2)
              t = t2
              cr_cone(1:3) = cr_cone2(1:3)
           case (-1)
              print *,"rad, height, cam, vis, pc",radius, height,cam_num,vis_num,pc
              print *,"amb, dif, spec",amb,dif, spec
              print *,"right dwon front",img_right, img_down, front
              print *,"transdat pos", transdat%pos
              print *,"transdat matrix", transdat%matrix
              print *,"pixp, direc, tpos",pixp,direc,tpos
              print *,"A, B,C,D",A,B,C,D
              print *,"cone1, cone2",cr_cone1, cr_cone2
              call vsmError("cr_cone undefined")
       end select
       tnorm(1:2) = real(cr_cone(1:2))
       A = tnorm(1)*tnorm(1)+tnorm(2)*tnorm(2) 
       if(A /= 0.0_vdp) then
         tnorm(1:2) = real(tnorm(1:2) / sqrt(A))
         tnorm(3) = real(radius/height)
         call vsmNormVecF(tnorm)
       else
         tnorm(1:2) = 0.0
         tnorm(3) = 1.0
       end if

       if(vsmCamera(cam_num)%pers) then
         if(vsmCoordData%YinYang == vsmYin) then
           call vsmCalcColorv(amb, dif, spec, shin, epos, &
                          cr_cone, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         else
           call vsmCalcColorvAn(amb, dif, spec, shin, epos, &
                          cr_cone, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         end if
         dist = sqrt(dot_product(cr_cone-epos, cr_cone-epos))
       else
         if(vsmCoordData%YinYang == vsmYin) then
           call vsmCalcColorvOrtho(amb, dif, spec, shin, direc, &
                          cr_cone, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         else
           call vsmCalcColorvOrthoAn(amb, dif, spec, shin, direc, &
                          cr_cone, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         end if
         dist = t + vsmCamera(cam_num)%near
       end if

!       if(vsmVisParam(vis_num)%depthq) then
!         dq_factor = vsmDepthQFactor(dist)
!         lumi(1:3) = dq_factor*lumi(1:3)
!       end if

!       px(1) = i; px(2) = j
!       call vsmPaintPixel3s(px, lumi, dist)
        picbuf(1:3,pc) = lumi
        picbuf(4,pc) = real(dist)

       !if(i < 1 .or. i > vsmImageData%width .or. j < 1 .or. j > vsmImageData%height) then
       !   print *, "raycast cone pixel position", i, j
       !end if
       if(lumi(1) < 0.0 .or. lumi(2) < 0.0 .or. lumi(3) < 0.0) print *, "raycast cone"
     end do
   end do

   if(pc >= 1) then
      pe=pc
      call vsmPaintPixel3ms(picbuf, ps, pe, ba_pixel(1,1), ba_pixel(2,1), wlen)
    end if

 end subroutine vsmRaycastCone

 subroutine vsmRaycastConeOMP(transdat, radius, height, ba_pixel, cam_num, vis_num, amb, dif, spec, shin)
   type(vismo_transdata), intent(in) :: transdat
   real(kind=vdp), intent(in) :: radius, height
   integer, dimension(2,4), intent(in) :: ba_pixel
   integer, intent(in) :: cam_num, vis_num
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin
   integer :: i, j,k,l
   real(kind=vdp), dimension(3) :: epos, topleft, front
   real(kind=vdp), dimension(3) :: img_right, img_down
   real(kind=vdp), dimension(3) :: pixp, direc, tpos

   real(kind=vdp) :: A, B, C, D, t, t1, t2, dist
   real(kind=vdp), dimension(3) :: cr_cone, cr_cone1, cr_cone2 ! cross point 
   real, dimension(3) :: lumi, tnorm
   real, dimension(4, vsmPbLen) :: picbuf
   integer :: ps, wlen
   integer :: cone_state
   integer :: comm, myrank, psize, ompnum, snum
   integer :: totallen, loop, res, lsize, offset

   if(.not.vsmCheckDraw(ba_pixel)) then
     return
   end if

   call vsmGetMPIVars(comm, myrank, psize, ompnum)
   snum = ompnum/vsmMaxOMP4Solid
   if(snum < 2) snum = 1
   ! translation and rotation:: the center of the bottom of cylinder moves to the origin
   tpos(1:3) = vsmCamera(cam_num)%pos(1:3) - transdat%pos(1:3)
   call vsmMatrixCrossVect(transdat%matrix, tpos, epos)

   tpos(1:3) = vsmImageData%topleft(1:3) - transdat%pos(1:3)
   call vsmMatrixCrossVect(transdat%matrix, tpos, topleft)

   call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%right, img_right)

   tpos(1:3) = -vsmCamera(cam_num)%up
   call vsmMatrixCrossVect(transdat%matrix, tpos, img_down)

   if(.not.vsmCamera(cam_num)%pers) then
     call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%front, front)
     call vsmNormVec(front)
   end if

   ps = 1
   wlen = ba_pixel(1,2) - ba_pixel(1,1) + 1

!$omp parallel do private(i) num_threads(snum)
   do i=1, vsmPblen
     picbuf(1:4, i) = vsmFarAway*1.1_vdp
   end do

   totallen = (ba_pixel(2,1) - ba_pixel(2, 4) )* (ba_pixel(1,1) - ba_pixel(1,4))
   loop = totallen/vsmPbLen
   res = mod(totallen,vsmPbLen)

!   do j=ba_pixel(2,1), ba_pixel(2, 4)
!     do i=ba_pixel(1,1), ba_pixel(1,4)
   do k=1, loop+1
     if(k==loop+1) then
       lsize = res
     else
       lsize = vsmPblen
     end if
     offset = (k-1)*vsmPblen
!$omp parallel do private(l,i,j,pixp,direc,a,b,c,d,cone_state,t1,t2,cr_cone1,cr_cone2,t,cr_cone,tnorm,lumi,dist) num_threads(snum)
     do l=1, lsize
       j = (l+offset-1)/wlen+ba_pixel(2,1)
       i = mod(l+offset-1, wlen)+ba_pixel(1,1)

       pixp(1:3) = topleft(1:3) + (i-1)*vsmImageData%dp*img_right(1:3) + (j-1)*vsmImageData%dp*img_down(1:3)

       if(vsmCamera(cam_num)%pers) then
         direc(1:3) = pixp(1:3) - epos(1:3)
         call vsmNormVec(direc)
       else
         direc(1:3) = front(1:3)
       end if

       A = direc(1)*direc(1) + direc(2)*direc(2) - radius*radius*direc(3)*direc(3)/(height*height)

       if(A == 0.0_vdp) cycle

       B = direc(1)*pixp(1)+direc(2)*pixp(2)+radius*radius*direc(3)*(-pixp(3)/(height*height)+1.0_vdp/height)
       C = pixp(1)*pixp(1)+pixp(2)*pixp(2)+&
            radius*radius*(-1.0_vdp-pixp(3)*pixp(3)/(height*height)+2.0_vdp*pixp(3)/height)
       D = B*B-A*C

       if(D < 0.0_vdp) cycle

       t1 = (-B+sqrt(D))/A
       t2 = (-B-sqrt(D))/A

!       cr_cone = vsmFarAway
       cone_state = -1
       cr_cone1(1:3) = pixp(1:3) + direc(1:3)*t1
       cr_cone2(1:3) = pixp(1:3) + direc(1:3)*t2
       if(t1 < 0.0_vdp .and. t2 < 0.0_vdp) then
           cone_state = 0
!          cycle
       else if(t1 >= 0.0_vdp .and. t2 >= 0.0_vdp) then
!          cr_cone1(1:3) = pixp(1:3) + direc(1:3)*t1
!          cr_cone2(1:3) = pixp(1:3) + direc(1:3)*t2
          if(cr_cone1(3) >= 0.0_vdp .and. cr_cone1(3) <= height) then
            if(cr_cone2(3) >= 0.0_vdp .and. cr_cone2(3) <= height) then
               if(t1 > t2) then
!                 t = t2
!                 cr_cone(1:3) = cr_cone2(1:3)
                 cone_state = 2
               else
!                 t = t1
!                 cr_cone(1:3) = cr_cone1(1:3)
                 cone_state = 1
               end if
            else
!                t = t1
!                cr_cone(1:3) = cr_cone1(1:3)
                cone_state = 1
            end if
         else if(cr_cone2(3) >= 0.0_vdp .and. cr_cone2(3) <= height) then
!             t = t2
!             cr_cone(1:3) = cr_cone2(1:3)
             cone_state = 2
        else
!             cycle
             cone_state = 0
        end if
       else if(t1 >= 0.0_vdp .and. t2 < 0.0_vdp) then
         cr_cone1(1:3) = pixp(1:3) + direc(1:3)*t1
         if(cr_cone1(3) >= 0.0_vdp .and. cr_cone1(3) <= height) then
!           t = t1
!           cr_cone(1:3) = cr_cone1(1:3)
           cone_state = 1
         else
!           cycle
            cone_state = 0
         end if
       else if(t1 < 0.0_vdp .and. t2 >= 0.0_vdp) then
         cr_cone2(1:3) = pixp(1:3) + direc(1:3)*t2
         if(cr_cone2(3) >= 0.0_vdp .and. cr_cone2(3) <= height) then
!           t = t2
!           cr_cone(1:3) = cr_cone2(1:3)
           cone_state = 2
         else
!           cycle
           cone_state = 0
         end if
       end if

       select case (cone_state)
           case (0)
              cycle
           case (1)
              t = t1
              cr_cone(1:3) = cr_cone1(1:3)
           case (2)
              t = t2
              cr_cone(1:3) = cr_cone2(1:3)
           case (-1)
              print *,"rad, height, cam, vis, pc",radius, height,cam_num,vis_num
              print *,"amb, dif, spec",amb,dif, spec
              print *,"right dwon front",img_right, img_down, front
              print *,"transdat pos", transdat%pos
              print *,"transdat matrix", transdat%matrix
              print *,"pixp, direc, tpos",pixp,direc,tpos
              print *,"A, B,C,D",A,B,C,D
              print *,"cone1, cone2",cr_cone1, cr_cone2
              call vsmError("cr_cone undefined")
       end select
       tnorm(1:2) = real(cr_cone(1:2))
       A = tnorm(1)*tnorm(1)+tnorm(2)*tnorm(2) 
       if(A /= 0.0_vdp) then
         tnorm(1:2) = real(tnorm(1:2) / sqrt(A))
         tnorm(3) = real(radius/height)
         call vsmNormVecF(tnorm)
       else
         tnorm(1:2) = 0.0
         tnorm(3) = 1.0
       end if

       if(vsmCamera(cam_num)%pers) then
         if(vsmCoordData%YinYang == vsmYin) then
           call vsmCalcColorv(amb, dif, spec, shin, epos, &
                          cr_cone, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         else
           call vsmCalcColorvAn(amb, dif, spec, shin, epos, &
                          cr_cone, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         end if
         dist = sqrt(dot_product(cr_cone-epos, cr_cone-epos))
       else
         if(vsmCoordData%YinYang == vsmYin) then
           call vsmCalcColorvOrtho(amb, dif, spec, shin, direc, &
                          cr_cone, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         else
           call vsmCalcColorvOrthoAn(amb, dif, spec, shin, direc, &
                          cr_cone, tnorm, vsmVisParam(vis_num)%light_used, lumi, transdat)
         end if
         dist = t + vsmCamera(cam_num)%near
       end if

        picbuf(1:3,l) = lumi
        picbuf(4,l) = real(dist)

       if(lumi(1) < 0.0 .or. lumi(2) < 0.0 .or. lumi(3) < 0.0) print *, "raycast cone"
     end do
     call vsmPaintPixel3ms(picbuf, ps, lsize, ba_pixel(1,1), ba_pixel(2,1), wlen)
     ps=ps+vsmPbLen
!$omp parallel do private(i) num_threads(snum)
     do i=1, vsmPbLen
       picbuf(1:4, i) = vsmFarAway*1.1_vdp
     end do
   end do

 end subroutine vsmRaycastConeOMP

 subroutine vsmBoundarea(pixel, ba_pixel, cross)
   integer, dimension(2,20), intent(in) :: pixel
   integer, dimension(2,4), intent(out) :: ba_pixel
   logical, dimension(20), intent(in) :: cross

   integer :: min_x, max_x, min_y, max_y
   integer :: i

!   if(vsmDbg) then
!     call vsmLogPrint("---vsmBoundarea")
!     call vsmLogPrint("pixel(1,:):",pixel(1,:))
!     call vsmLogPrint("pixel(2,:):",pixel(2,:))
!   end if

!   if (vsmDbg) then
!     do i=1, 20
!       if(cross(i)) then 
!         call vsmLogPrint("true: i=", i)
!       else
!         call vsmLogPrint("false: i=", i)
!       end if
!     end do
!   end if
   min_x = vsmImageData%width+1
   max_x = 0
   min_y = vsmImageData%height+1
   max_y = 0
   
   do i=1, 20
      if(.not. cross(i)) cycle
      if(min_x > pixel(1,i)) min_x = pixel(1,i)
      if(max_x < pixel(1,i)) max_x = pixel(1,i)
      if(min_y > pixel(2,i)) min_y = pixel(2,i)
      if(max_y < pixel(2,i)) max_y = pixel(2,i)
   end do
!   min_x = minval(pixel(1,:), cross) - 1 ! for secure
!   max_x = maxval(pixel(1,:), cross) + 1 ! for secure

!   min_y = minval(pixel(2,:), cross) - 1 ! for secure
!   max_y = maxval(pixel(2,:), cross) + 1 ! for secure

   if(min_x < 1) then
     min_x = 1
   else if(min_x == vsmImageData%width) then
     min_x = vsmImageData%width
   end if
   if(max_x > vsmImageData%width) then
     max_x = vsmImageData%width
   else if(max_x == 0) then
     max_x = 1
   end if

   if(min_y < 1) then
     min_y = 1
   else if(min_y == vsmImageData%height + 1) then
     min_y = vsmImageData%height
   end if
   if(max_y > vsmImageData%height) then
     max_y = vsmImageData%height
   else if(max_y == 0) then
     max_y = 1
   end if

   !if(vsmDbg) then
    ! call vsmLogPrint("min x:",min_x)
    ! call vsmLogPrint("min x:",pixel(1,:))
    ! call vsmLogPrint("min y:",min_y)
    ! call vsmLogPrint("max x:",max_x)
    ! call vsmLogPrint("max y:",max_y)
   !end if

   ba_pixel(1,1) = min_x;  ba_pixel(2,1) = min_y
   ba_pixel(1,2) = max_x;  ba_pixel(2,2) = min_y
   ba_pixel(1,3) = min_x;  ba_pixel(2,3) = max_y
   ba_pixel(1,4) = max_x;  ba_pixel(2,4) = max_y
 end subroutine vsmBoundarea

 subroutine vsmCheckBv(verts, cam_num, rad, cross_points, num_cp, above, transdat)
   real(kind=vdp), dimension(3, 8), intent(in) :: verts
   real(kind=vdp), intent(in) :: rad
   integer, intent(in) :: cam_num
   integer, intent(out) :: num_cp
   real(kind=vdp), dimension(3, 12), intent(out) :: cross_points
   logical, dimension(8), intent(out) :: above
   type(vismo_transdata), optional, intent(in) :: transdat

   integer :: i, num_up
   real(kind=vdp), dimension(3) :: norm, plane_c
   real(kind=vdp) :: C, val, t
   real(kind=vdp), dimension(3) :: temp_vec, temp_vec2

   !call vsmLogPrint("---- check bv")
   if(.not.present(transdat)) then
     norm(1:3) = vsmCamera(cam_num)%front(1:3)
     plane_c(1:3) = vsmCamera(cam_num)%pos(1:3) + vsmCamera(cam_num)%near * norm(1:3)
   else
     call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%front, norm)
     call vsmNormVec(norm)
     temp_vec2(1:3) = vsmCamera(cam_num)%pos(1:3)-transdat%pos(1:3)
     call vsmMatrixCrossVect(transdat%matrix, temp_vec2, temp_vec)
     plane_c(1:3) = temp_vec(1:3) + vsmCamera(cam_num)%near * norm(1:3)
   end if
   
  ! call vsmLogPrint("Norm", norm)
 !  call vsmLogPrint("cneter of plane", plane_c)

   C = -dot_product(norm, plane_c)

 !  call vsmLogPrint("norm*plane_c = ", C)

   num_up = 0
   do i=1, 8
     val = dot_product(verts(1:3,i), norm) + C
     if(val >= 0.0_vdp) then
       num_up = num_up + 1
       above(i) = .true.
     else
       above(i) = .false.
     end if
   end do
 !  call vsmLogPrint("num_up = ", num_up)

   num_cp = 0
   if(num_up /= 0 .and. num_up /= 8) then
     ! X direction x 4 patterns
     if(norm(1) /= 0.0_vdp) then
      ! there are cross points
      !t(1) = (dot_product(norm, verts(1:3, 1) + C)/norm(1)
      !t(2) = (dot_product(norm, verts(1:3, 3) + C)/norm(1)
      !t(3) = (dot_product(norm, verts(1:3, 5) + C)/norm(1)
      !t(4) = (dot_product(norm, verts(1:3, 7) + C)/norm(1)
      do i=1, 4
        t = (dot_product(norm, verts(1:3, 2*(i-1)+1)) + C)/norm(1)
        if(t >= 0.0_vdp .and. t <= 2.0_vdp*rad) then
          num_cp = num_cp + 1
          cross_points(1, num_cp) = verts(1, 2*(i-1)+1) + t
          cross_points(2, num_cp) = verts(2, 2*(i-1)+1)
          cross_points(3, num_cp) = verts(3, 2*(i-1)+1)
        end if
      end do
     end if

     ! Y direction x 4 patterns
     if(norm(2) /= 0.0_vdp) then
      ! there are cross points
      !t(5) = (dot_product(norm, verts(1:3, 1) + C)/norm(2)
      !t(6) = (dot_product(norm, verts(1:3, 2) + C)/norm(2)
      do i=5, 6
        t = (dot_product(norm, verts(1:3, i-4)) + C)/norm(2)
        if(t >= 0.0_vdp .and. t <= 2.0_vdp*rad) then
          num_cp = num_cp + 1
          cross_points(1, num_cp) = verts(1, i-4)
          cross_points(2, num_cp) = verts(2, i-4) + t
          cross_points(3, num_cp) = verts(3, i-4)
        end if
      end do
      !t(7) = (dot_product(norm, verts(1:3, 5) + C)/norm(2)
      !t(8) = (dot_product(norm, verts(1:3, 6) + C)/norm(2)
      do i=7, 8
        t = (dot_product(norm, verts(1:3, i-2)) + C)/norm(2)
        if(t >= 0.0_vdp .and. t <= 2.0_vdp*rad) then
          num_cp = num_cp + 1
          cross_points(1, num_cp) = verts(1, i-2)
          cross_points(2, num_cp) = verts(2, i-2) + t
          cross_points(3, num_cp) = verts(3, i-2)
        end if
      end do
     end if

     ! Z direction x 4 patterns
     if(norm(3) /= 0.0_vdp) then
      ! there are cross points
      !t(9)  = (dot_product(norm, verts(1:3, 1) + C)/norm(3)
      !t(10) = (dot_product(norm, verts(1:3, 2) + C)/norm(3)
      !t(11) = (dot_product(norm, verts(1:3, 3) + C)/norm(3)
      !t(12) = (dot_product(norm, verts(1:3, 4) + C)/norm(3)
      do i=9, 12
        t = (dot_product(norm, verts(1:3, i-8)) + C)/norm(3)
        if(t >= 0.0_vdp .and. t <= 2.0_vdp*rad) then
          num_cp = num_cp + 1
          cross_points(1, num_cp) = verts(1, i-8)
          cross_points(2, num_cp) = verts(2, i-8)
          cross_points(3, num_cp) = verts(3, i-8) + t
        end if
      end do
     end if
   end if

 end subroutine vsmCheckBv

 subroutine vsmCheckBa(verts, cam_num, rad, cross_points, num_cp, above, transdat)
   real(kind=vdp), dimension(3, 4), intent(in) :: verts ! already translated and rotated
   real(kind=vdp), intent(in) :: rad
   integer, intent(in) :: cam_num
   integer, intent(out) :: num_cp
   real(kind=vdp), dimension(3, 4), intent(out) :: cross_points
   logical, dimension(4), intent(out) :: above
   type(vismo_transdata), optional, intent(in) :: transdat

   integer :: i, num_up
   real(kind=vdp), dimension(3) :: norm, plane_c
   real(kind=vdp) :: C, val, t
   real(kind=vdp), dimension(3) :: temp_vec, temp_vec2

   if(.not.present(transdat)) then
     norm(1:3) = vsmCamera(cam_num)%front(1:3)
   else
     call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%front(1:3), norm)
   end if

   if(.not.present(transdat)) then
     plane_c(1:3) = vsmCamera(cam_num)%pos(1:3) + vsmCamera(cam_num)%near * norm(1:3)
   else
     temp_vec2 = vsmCamera(cam_num)%pos(1:3)-transdat%pos(1:3)

     call vsmMatrixCrossVect(transdat%matrix, temp_vec2, temp_vec)
     plane_c(1:3) = temp_vec(1:3) + vsmCamera(cam_num)%near * norm(1:3)
   end if
   C = -dot_product(norm, plane_c)

   num_up = 0
   do i=1, 4
     val = dot_product(verts(1:3,i), norm) + C
     if(val >= 0.0_vdp) then
       num_up = num_up + 1
       above(i) = .true.
     else
       above(i) = .false.
     end if
   end do

   num_cp = 0
   if(num_up /= 0 .and. num_up /= 4) then
     ! X direction x 2 patterns
     if(norm(1) /= 0.0_vdp) then
      ! there are cross points
      !t(1) = (dot_product(norm, verts(1:3, 1) + C)/norm(1)
      !t(2) = (dot_product(norm, verts(1:3, 3) + C)/norm(1)
      do i=1, 2
        t = (dot_product(norm, verts(1:3, 2*(i-1)+1)) + C)/norm(1)
        if(t >= 0.0_vdp .and. t <= 2.0_vdp*rad) then
          num_cp = num_cp + 1
          cross_points(1, num_cp) = verts(1, 2*(i-1)+1) + t
          cross_points(2, num_cp) = verts(2, 2*(i-1)+1)
          cross_points(3, num_cp) = verts(3, 2*(i-1)+1)
        end if
      end do
     end if

     ! Y direction x 2 patterns
     if(norm(2) /= 0.0_vdp) then
      ! there are cross points
      !t(3) = (dot_product(norm, verts(1:3, 1) + C)/norm(2)
      !t(4) = (dot_product(norm, verts(1:3, 2) + C)/norm(2)
      do i=1, 2
        t = (dot_product(norm, verts(1:3, i)) + C)/norm(2)
        if(t >= 0.0_vdp .and. t <= 2.0_vdp*rad) then
          num_cp = num_cp + 1
          cross_points(1, num_cp) = verts(1, i)
          cross_points(2, num_cp) = verts(2, i) + t
          cross_points(3, num_cp) = verts(3, i)
        end if
      end do
     end if
   end if

 end subroutine vsmCheckBa

 subroutine vsmPointProjection(pos, cam_num, cross, pix, transdat)
   real(kind=vdp), dimension(3), intent(in) :: pos ! already translated and rotated
   integer, intent(in) :: cam_num
   logical, intent(out) :: cross
   integer, dimension(2), intent(inout) :: pix
   type(vismo_transdata), intent(in), optional :: transdat

   real(kind=vdp), dimension(3) :: eyepos
   real(kind=vdp), dimension(3) :: norm, plane_c
   real(kind=vdp), dimension(3) :: direc
   real(kind=vdp) :: A, C, B, ip1, ip2
   real(kind=vdp), dimension(3) :: cross_pos, tpos
   real(kind=vdp), dimension(3) :: img_topleft, img_right, img_down
   !real(kind=vdp), dimension(3) :: temp_vec, temp_vec2
   real(kind=vdp), dimension(3) :: temp_vec2

   !call vsmLogPrint("--- point projection")

   if(.not. present(transdat)) then
     norm(1:3) = vsmCamera(cam_num)%front(1:3)
     eyepos(1:3) = vsmCamera(cam_num)%pos(1:3)
   else
     call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%front, norm)
     call vsmNormVec(norm)
     temp_vec2(1:3) = vsmCamera(cam_num)%pos(1:3) - transdat%pos(1:3)
     call vsmMatrixCrossVect(transdat%matrix, temp_vec2, eyepos)
   end if

   plane_c(1:3) = eyepos(1:3) + vsmCamera(cam_num)%near * norm(1:3)

   if(vsmCamera(cam_num)%pers) then
     !Perspective Projection
     direc(1:3) = pos(1:3) - eyepos(1:3)
     call vsmNormVec(direc)

     C = dot_product(norm, plane_c - eyepos)
     B = dot_product(norm, direc)

     if(B == 0.0_vdp) then
       cross = .false.
       return
     end if

     A = C/B

     if(A < 0.0_vdp) then
       cross = .false.
       return
     end if

     cross_pos(1:3) = A*direc(1:3) + eyepos(1:3)

   else
     !Parallel Projection
     direc(1:3) = -norm(1:3)

     C = dot_product(norm, plane_c - pos)
     !B = -1.0_vdp

     A = -C

     if(A < 0.0_vdp) then
       cross = .false.
       return
     end if

     cross_pos(1:3) = A*direc(1:3) + pos(1:3)
   end if

 !  call vsmLogPrint("cross point on image", cross_pos)

   if(.not. present(transdat)) then
     img_topleft(1:3) = vsmImageData%topleft(1:3)
     img_right(1:3)   = vsmCamera(cam_num)%right(1:3)
     img_down(1:3)    = -vsmCamera(cam_num)%up(1:3)
   else
     temp_vec2(1:3) = vsmImageData%topleft(1:3) - transdat%pos(1:3)
     call vsmMatrixCrossVect(transdat%matrix, temp_vec2, img_topleft)
     call vsmMatrixCrossVect(transdat%matrix, vsmCamera(cam_num)%right, img_right)
     tpos(1:3) = -vsmCamera(cam_num)%up
     call vsmMatrixCrossVect(transdat%matrix, tpos, img_down)
     call vsmNormVec(img_right)
     call vsmNormVec(img_down)
     !print *,"cam_right:", dot_product(vsmCamera(cam_num)%right, vsmCamera(cam_num)%right)
     !print *,"img_right:", sqrt(dot_product(img_right, img_right))
     !print *,"img_down:", sqrt(dot_product(img_down, img_down))
   end if

   !call vsmLogPrint("image topleft", img_topleft)
   !call vsmLogPrint("image right", img_right)
   !call vsmLogPrint("image down", img_down)

   direc(1:3) = cross_pos(1:3) - img_topleft(1:3)

   !call vsmLogPrint("topleft to crosspoint on image", direc)
   ip1 = dot_product(direc, img_right)
   ip2 = dot_product(direc, img_down)

   pix(1) = nint(ip1/vsmImageData%dp)
   pix(2) = nint(ip2/vsmImageData%dp)

  ! if(pix(1) < 1 .or. pix(1) > image_width .or. pix(2) < 1 .or. pix(2) > image_height) then
  !   cross = .false.
  !   return
  ! end if

   cross = .true.
   !call vsmLogPrint("point projection ends")
 end subroutine vsmPointProjection

 subroutine vsmCalcBvolumeSphere(pos, radius, vertx)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), intent(in) :: radius
   real(kind=vdp), dimension(3, 8), intent(out) :: vertx

   !vertex 1
   vertx(1,1) = pos(1)-radius; vertx(2,1) = pos(2)-radius; vertx(3,1) = pos(3)-radius
   !vertex 2
   vertx(1,2) = pos(1)+radius; vertx(2,2) = pos(2)-radius; vertx(3,2) = pos(3)-radius
   !vertex 3
   vertx(1,3) = pos(1)-radius; vertx(2,3) = pos(2)+radius; vertx(3,3) = pos(3)-radius
   !vertex 4
   vertx(1,4) = pos(1)+radius; vertx(2,4) = pos(2)+radius; vertx(3,4) = pos(3)-radius

   !vertex 5
   vertx(1,5) = pos(1)-radius; vertx(2,5) = pos(2)-radius; vertx(3,5) = pos(3)+radius
   !vertex 6
   vertx(1,6) = pos(1)+radius; vertx(2,6) = pos(2)-radius; vertx(3,6) = pos(3)+radius
   !vertex 7
   vertx(1,7) = pos(1)-radius; vertx(2,7) = pos(2)+radius; vertx(3,7) = pos(3)+radius
   !vertex 8
   vertx(1,8) = pos(1)+radius; vertx(2,8) = pos(2)+radius; vertx(3,8) = pos(3)+radius

 end subroutine vsmCalcBvolumeSphere

 subroutine vsmCalcBvolumeCylinder(height, radius, vertx)
   real(kind=vdp), intent(in) :: height, radius
   real(kind=vdp), dimension(3, 8), intent(out) :: vertx

   !vertex 1
   vertx(1,1) = -radius; vertx(2,1) = -radius; vertx(3,1) = 0.0_vdp
   !vertex 2
   vertx(1,2) =  radius; vertx(2,2) = -radius; vertx(3,2) = 0.0_vdp
   !vertex 3
   vertx(1,3) = -radius; vertx(2,3) =  radius; vertx(3,3) = 0.0_vdp 
   !vertex 4
   vertx(1,4) =  radius; vertx(2,4) =  radius; vertx(3,4) = 0.0_vdp 

   !vertex 5
   vertx(1,5) =  -radius; vertx(2,5) = -radius; vertx(3,5) = height
   !vertex 6
   vertx(1,6) =  radius;  vertx(2,6) = -radius; vertx(3,6) = height
   !vertex 7
   vertx(1,7) = -radius;  vertx(2,7) = radius;  vertx(3,7) = height
   !vertex 8
   vertx(1,8) =  radius;  vertx(2,8) = radius;  vertx(3,8) = height

 end subroutine vsmCalcBvolumeCylinder

end module vsmSolidDrawer
