module vsmStreamLine
 use vsmConst
 use vsmFunction
 use vsmData
 use vsmParam
 use vsmColormap
 use vsmCoordinates
 use vsmIntpol
 use vsmSolidDrawer
#ifndef CLANG
 use mpi
#endif

 implicit none
 private
 
 public vsmDrawStreamlinesIM

 contains

 subroutine vsmyyRungeKutta4(x, nvec, h, inside, abs_val, first)
   integer, intent(in) :: nvec
   real(kind=vdp), dimension(3), intent(inout) :: x
   real(kind=vdp), intent(in) :: h
   real(kind=vdp), intent(out), optional :: abs_val
   logical, intent(out) :: inside
   logical, intent(inout) :: first

   real(kind=vdp), dimension(3) :: xt, k1, k2, k3, k4, spos
   real(kind=vdp), dimension(3) :: dx, vec_val
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp), dimension(3, 8) :: grid_val
   integer, dimension(3) :: ci
   logical :: normalize

   normalize = .true.
   ! calculating k1
   xt(1:3) = x(1:3)
   !call vsmLogPrint("Runge:b4 insidel")
   call vsmCoordsToAngles(xt, spos)
   if(first) then
      first = .false.
      if(vsmIsAnotherSet(xt, spos)) then
        inside = .false.
      else
        inside = vsmIsInsidel(spos)
      end if
   else
      inside = vsmIsInsidelWO(spos)
   end if
   if(.not.inside) return
   !call vsmLogPrint("Runge:after insidel")

   call vsmFindCoord(spos, dx, ci)
   call vsmSetFrac(dx, vx)
   call vsmSetVectGridVal(nvec, ci, grid_val)
   call vsmGetVectVal(vx, grid_val, vec_val, normalize)
   if(vsmCoordData%YinYang == vsmYin) then
     k1(1:3) = vec_val(1:3)*h
   else
     k1(1) = -vec_val(1)*h
     k1(2) = vec_val(3)*h
     k1(3) = vec_val(2)*h
   end if

   ! calculating k2
   xt(1:3) = x(1:3) + k1(1:3)/2.0_vdp

   !print *,"Runge:b4 insidel 2", xt
   call vsmCoordsToAngles(xt, spos)
   inside = vsmIsInsidelWO(spos)
   if(.not.inside) then
     if(vsmCoordData%YYZMode) then
       if(.not.vsmIsInsideSphere(xt)) x(1:3) = xt(1:3)
     else
       if(.not.vsmIsBetweenSpheres(xt)) x(1:3) = xt(1:3)
     end if
    return
   end if
   !print *,"Runge:after insidel 2"

   call vsmFindCoord(spos, dx, ci)
   call vsmSetFrac(dx, vx)
   call vsmSetVectGridVal(nvec, ci, grid_val)
   call vsmGetVectVal(vx, grid_val, vec_val, normalize)
   if(vsmCoordData%YinYang == vsmYin) then
     k2(1:3) = vec_val(1:3)*h
   else
     k2(1) = -vec_val(1)*h
     k2(2) = vec_val(3)*h
     k2(3) = vec_val(2)*h
   end if

   ! calculating k3
   xt(1:3) = x(1:3) + k2(1:3)/2.0_vdp

   call vsmCoordsToAngles(xt, spos)
   inside = vsmIsInsidelWO(spos)
   if(.not.inside) then
     if(vsmCoordData%YYZMode) then
       if(.not.vsmIsInsideSphere(xt)) x(1:3) = xt(1:3)
     else
       if(.not.vsmIsBetweenSpheres(xt)) x(1:3) = xt(1:3)
     end if
    return
   end if

   call vsmFindCoord(spos, dx, ci)
   call vsmSetFrac(dx, vx)
   call vsmSetVectGridVal(nvec, ci, grid_val)
   call vsmGetVectVal(vx, grid_val, vec_val, normalize)
   if(vsmCoordData%YinYang == vsmYin) then
     k3(1:3) = vec_val(1:3)*h
   else
     k3(1) = -vec_val(1)*h
     k3(2) = vec_val(3)*h
     k3(3) = vec_val(2)*h
   end if
   
   ! calculating k4
   xt(1:3) = x(1:3) + k3(1:3)

   call vsmCoordsToAngles(xt, spos)
   inside = vsmIsInsidelWO(spos)
   if(.not.inside) then
     if(vsmCoordData%YYZMode) then
       if(.not.vsmIsInsideSphere(xt)) x(1:3) = xt(1:3)
     else
       if(.not.vsmIsBetweenSpheres(xt)) x(1:3) = xt(1:3)
     end if
     return
   end if

   call vsmFindCoord(spos, dx, ci)
   call vsmSetFrac(dx, vx)
   call vsmSetVectGridVal(nvec, ci, grid_val)
   call vsmGetVectVal(vx, grid_val, vec_val, normalize)
   if(vsmCoordData%YinYang == vsmYin) then
     k4(1:3) = vec_val(1:3)*h
   else
     k4(1) = -vec_val(1)*h
     k4(2) = vec_val(3)*h
     k4(3) = vec_val(2)*h
   end if

   x(1:3) = x(1:3) + (k1(1:3) + 2.0_vdp*k2(1:3) + 2.0_vdp*k3(1:3)+k4(1:3))/6.0_vdp

   call vsmCoordsToAngles(x, spos)
   inside = vsmIsInsidelWO(spos)
   if(.not.inside) return
 
   if(present(abs_val)) then
     call vsmFindCoord(spos, dx, ci)
     call vsmSetFrac(dx, vx)
     call vsmSetVectGridVal(nvec, ci, grid_val)
     call vsmGetVectVal(vx, grid_val, vec_val)
     !abs_val = sqrt(vec_val(1)*vec_val(1)+vec_val(2)*vec_val(2)+vec_val(3)*vec_val(3))
     abs_val = sqrt(dot_product(vec_val, vec_val))!2013/8/30
   end if
   inside = .true.
 end subroutine vsmyyRungeKutta4

 subroutine vsmzRungeKutta4(x, nvec, h, inside, abs_val, first)
   integer, intent(in) :: nvec
   real(kind=vdp), dimension(3), intent(inout) :: x
   real(kind=vdp), intent(in) :: h
   real(kind=vdp), intent(out), optional :: abs_val
   logical, intent(out) :: inside
   logical, intent(inout) :: first

   real(kind=vdp), dimension(3) :: xt, k1, k2, k3, k4
   real(kind=vdp), dimension(3) :: dx, vec_val
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp), dimension(3, 8) :: grid_val
   integer, dimension(3) :: ci
   logical :: normalize, in1, in2

   normalize = .true.
   ! calculating k1
   xt(1:3) = x(1:3)
   call vsmLogPrint("Runge:b4 insidel")
   if(first) then
!      if(vsmIsInsidel(xt) .and. vsmIsInsideInnerSphere4z(xt)) then
      in1 = vsmIsInsidel(xt); in2 = vsmIsInsideInnerSphere4z0dr(xt)
      if(in1 .and. in2) then
        inside = .true.
      else
        inside = .false.
!        call vsmLogPrint("first insidel ", vsmIsInsidel(xt))
!        call vsmLogPrint("first IsInsideInnerSphere ", vsmIsInsideInnerSphere4z(xt))
      end if
      first = .false.
   else
!      if(vsmIsInsidelWO(xt) .and. vsmIsInsideInnerSphere4z(xt)) then
      in1 = vsmIsInsidelWO(xt); in2 = vsmIsInsideInnerSphere4z1dr(xt)
      if(in1 .and. in2) then
        inside = .true.
      else
        inside = .false.
      end if
   end if
   if(.not.inside) return
   call vsmLogPrint("Runge:after insidel")

   call vsmFindCoord(xt, dx, ci)
   call vsmSetFrac(dx, vx)
   call vsmSetVectGridVal(nvec, ci, grid_val)
   call vsmGetVectVal(vx, grid_val, vec_val, normalize)
   k1(1:3) = vec_val(1:3)*h

   ! calculating k2
   xt(1:3) = x(1:3) + k1(1:3)/2.0_vdp

   !print *,"Runge:b4 insidel 2", xt
!   if(vsmIsInsidelWO(xt) .and. vsmIsInsideInnerSphere4z(xt)) then
   in1 = vsmIsInsidelWO(xt); in2 = vsmIsInsideInnerSphere4z1dr(xt)
   if(in1 .and. in2) then
     inside = .true.
   else
     inside = .false.
   end if
   if(.not.inside) then
    if(.not.vsmIsInsideSphere(xt)) x(1:3) = xt(1:3)
    return
   end if
   !print *,"Runge:after insidel 2"

   call vsmFindCoord(xt, dx, ci)
   call vsmSetFrac(dx, vx)
   call vsmSetVectGridVal(nvec, ci, grid_val)
   call vsmGetVectVal(vx, grid_val, vec_val, normalize)
   k2(1:3) = vec_val(1:3)*h

   ! calculating k3
   xt(1:3) = x(1:3) + k2(1:3)/2.0_vdp

!   if(vsmIsInsidelWO(xt) .and. vsmIsInsideInnerSphere4z(xt)) then
   in1 = vsmIsInsidelWO(xt); in2 = vsmIsInsideInnerSphere4z1dr(xt)
   if(in1 .and. in2) then
     inside = .true.
   else
     inside = .false.
   end if
   if(.not.inside) then
    if(.not.vsmIsInsideSphere(xt)) x(1:3) = xt(1:3)
    return
   end if

   call vsmFindCoord(xt, dx, ci)
   call vsmSetFrac(dx, vx)
   call vsmSetVectGridVal(nvec, ci, grid_val)
   call vsmGetVectVal(vx, grid_val, vec_val, normalize)
   k3(1:3) = vec_val(1:3)*h
   
   ! calculating k4
   xt(1:3) = x(1:3) + k3(1:3)

!   if(vsmIsInsidelWO(xt) .and. vsmIsInsideInnerSphere4z(xt)) then
      in1 = vsmIsInsidelWO(xt); in2 = vsmIsInsideInnerSphere4z1dr(xt)
    if(in1 .and. in2) then
     inside = .true.
   else
     inside = .false.
   end if
   if(.not.inside) then
    if(.not.vsmIsInsideSphere(xt)) x(1:3) = xt(1:3)
    return
   end if

   call vsmFindCoord(xt, dx, ci)
   call vsmSetFrac(dx, vx)
   call vsmSetVectGridVal(nvec, ci, grid_val)
   call vsmGetVectVal(vx, grid_val, vec_val, normalize)
   k4(1:3) = vec_val(1:3)*h

   x(1:3) = x(1:3) + (k1(1:3) + 2.0_vdp*k2(1:3) + 2.0_vdp*k3(1:3)+k4(1:3))/6.0_vdp

!   if(vsmIsInsidelWO(x) .and. vsmIsInsideInnerSphere4z(x)) then
   in1 = vsmIsInsidelWO(xt); in2 = vsmIsInsideInnerSphere4z1dr(xt)
   if(in1 .and. in2) then
     inside = .true.
   else
     inside = .false.
   end if
   if(.not.inside) return
 
   if(present(abs_val)) then
     call vsmFindCoord(x, dx, ci)
     call vsmSetFrac(dx, vx)
     call vsmSetVectGridVal(nvec, ci, grid_val)
     call vsmGetVectVal(vx, grid_val, vec_val)
     !abs_val = sqrt(vec_val(1)*vec_val(1)+vec_val(2)*vec_val(2)+vec_val(3)*vec_val(3))
     abs_val = sqrt(dot_product(vec_val, vec_val))!2013/8/30
   end if
   inside = .true.
 end subroutine vsmzRungeKutta4


 subroutine vsmDrawStreamLinesIM(vis_num, cam_num, dp)
   integer, intent(in) :: vis_num, cam_num
   real(kind=vdp), intent(in) :: dp
     call vsmyyDrawStreamLinesIM(vis_num, cam_num, dp)
 end subroutine vsmDrawStreamLinesIM

 subroutine vsmyyDrawStreamLinesIM(vis_num, cam_num, dp)
   integer, intent(in) :: vis_num, cam_num
   real(kind=vdp), intent(in) :: dp
   integer :: comm, myrank
   integer :: i, j, k, ierr, l
   integer :: vec_loop !, step_loop
   real(kind=vdp), dimension(3) :: x, xp, spos, xx, xxp
   real(kind=vdp) :: h, rad, absval, dc
   integer :: nvec, nseed, st_rem, intval
   logical :: calc_suc
   real, dimension(3) :: col, amb, dif, spec
   real :: shin

   real(kind=vdp), dimension(3, vsmMaxSeed) :: pos, pos1
   integer, dimension(vsmMaxSeed) :: st_remained, st_remained1
   logical, dimension(vsmMaxSeed) :: first
   integer, parameter :: maxSlineLoop = 100
   integer :: sum_remain, m
   integer :: count
   integer :: tnuml1
   integer, dimension(vsmMaxSeed) :: pc_rec

   call vsmGetMPIVars(comm, myrank)

   vec_loop = vsmVisParam(vis_num)%sline_num
   !call vsmLogPrint("vecloop:", vec_loop)
   h = vsmIntval%dh
   !call vsmLogPrint("dh:", h)
   if(dp > 0.0_vdp) then
      dc = dp
   else
      dc = -dp
   end if

   intval = int(dc/h)
   if(intval < 1) intval = 1

   pc_rec=0
   do j=1, vec_loop

     call vsmLogPrint("vis_num:", vis_num)
     call vsmLogPrint("vec loop j:", j)
     nvec  = vsmVisParam(vis_num)%slineparam(j)%nvec
     nseed = vsmVisParam(vis_num)%slineparam(j)%num_seed
     rad   = vsmVisParam(vis_num)%slineparam(j)%radius

     if(vsmVisParam(vis_num)%slineparam(j)%monocolor) then
        amb(1:3)  = vsmVisParam(vis_num)%slineparam(j)%amb(1:3)
        dif(1:3)  = vsmVisParam(vis_num)%slineparam(j)%dif(1:3)
        spec(1:3) = vsmVisParam(vis_num)%slineparam(j)%spec(1:3)
     end if
     ! common for color and mono
     shin = vsmVisParam(vis_num)%slineparam(j)%shin

     ! initialization
     !call vsmLogPrint("seed initialization loop ")
     do k=1, nseed
       first(k) = .true.
       if(vsmCoordData%YinYang == vsmYin) then
         pos(1:3, k)    =  vsmVisParam(vis_num)%slineparam(j)%seed(1:3, k)
       else
         pos(1, k)  =  -vsmVisParam(vis_num)%slineparam(j)%seed(1, k)
         pos(2, k)  =  vsmVisParam(vis_num)%slineparam(j)%seed(3, k)
         pos(3, k)  =  vsmVisParam(vis_num)%slineparam(j)%seed(2, k)
       end if

       st_remained(k) =  vsmVisParam(vis_num)%slineparam(j)%steps_remained(k)
       !call vsmLogPrint("seed initialization loop pos:", pos(:, k))
       !call vsmLogPrint("seed initialization loop remained:", st_remained(k))
     end do
#ifdef OMPNEST
     call vsmSetOMPNest(nseed)
     tnuml1 = vsmGetOMPNestL1()

     do l=1, maxSlineLoop ! -----  MPI  ------! max barrier number is fixed for safety
!$omp parallel do private(i,k,x,st_rem,xp,calc_suc,absval,col,spos,xx,xxp) firstprivate(amb,dif,spec,shin) &
!$omp num_threads(tnuml1)
      do i=1, nseed
#else
     do l=1, maxSlineLoop ! -----  MPI  ------! max barrier number is fixed for safety
!$omp parallel do private(i,k,x,st_rem,xp,calc_suc,absval,col,spos,xx,xxp) firstprivate(amb,dif,spec,shin)
      do i=1, nseed
#endif
       !call vsmLogPrint("seed loop i:", i)
       x(1:3) = pos(1:3, i)
       st_rem = st_remained(i) 

       do k=1, st_remained(i)
         !call vsmLogPrint("pos = ", x)
!         if(myrank == 0) print *,"0:st_rem(i):", st_remained(1:4), k
!         if(myrank == 1) print *,"1:st_rem(i):", st_remained(1:4), k

         if(k==1) then
           xp(1:3) = x(1:3)
           if(.not.vsmCoordData%YYZMode) then
              if( .not.vsmIsBetweenSpheres(x)) then
                !st_remained(i) = 0
                st_rem = 0
                call vsmLogPrint("outside whole data, yymode remaind = 0: ", i)
                call vsmLogPrint("outside whole data, yymode", vsmCoordData%YYZMode)
                exit
              end if
           else if(vsmCoordData%YYZMode) then
              if(.not.vsmIsInsideSphere(x)) then
                !st_remained(i) = 0
                st_rem = 0
                call vsmLogPrint("outside whole data, yyz mode remaind = 0: ", i)
                call vsmLogPrint("outside whole data, yymode", vsmCoordData%YYZMode)
                exit
              end if
           else
              call vsmError("yyzmode is strange")
           end if

           if( (.not.vsmIsBetweenSpheres(x)  .and. .not. vsmCoordData%YYZMode).or. &
                (.not.vsmIsInsideSphere(x) .and. vsmCoordData%YYZMode)) then
             !st_remained(i) = 0
             st_rem = 0
             call vsmLogPrint("outside whole data, remaind = 0: ", i)
             exit
           end if

           call vsmCoordsToAngles(pos(1:3, i), spos)
           if(.not.vsmIsInsidelWO(spos)) then
!              print *,"insidel pos:", myrank, pos(:,i)
              pos(1:3, i) = vsmFaraway
              if(vsmDbg) then
                call vsmLogPrint("outside local data, pos will be faraway: ", pos(1:3, i))
                call vsmLogPrint("outside local data: ", vsmLocalCoord%lorigin)
                call vsmLogPrint("outside local data: ", vsmLocalCoord%lcend)
              end if
              exit
           end if
         end if

         call vsmyyRungeKutta4(x, nvec, h, calc_suc, absval, first(i))

         if(calc_suc .eqv. .false.) then
           !st_remained(i) = st_rem
           call vsmLogPrint("calc failed : ", i)
!           call vsmCoordsToAngles(x, spos)
           if( (.not.vsmIsBetweenSpheres(x)  .and. .not. vsmCoordData%YYZMode).or. &
                (.not.vsmIsInsideSphere(x) .and. vsmCoordData%YYZMode)) then
             st_remained(i) = 0
           end if
           if(k==1) then
              pos(1:3, i) = vsmFaraway
           else
              pos(1:3, i) = x(1:3)
           end if
             exit
         end if

         !if(myrank == 3 .and. i==1) print *,"RungeKutta st_rem", st_rem

         if(.not.vsmVisParam(vis_num)%slineparam(j)%monocolor) then
            call vsmGetVColor(absval, nvec, col)
            amb(1:3)  = col(1:3) * vsmVisParam(vis_num)%slineparam(j)%amb(1)
            dif(1:3)  = col(1:3) * vsmVisParam(vis_num)%slineparam(j)%dif(1)
            spec(1:3) = col(1:3) * vsmVisParam(vis_num)%slineparam(j)%spec(1)
         end if

         if(mod((vsmMaxSlineSteps - st_rem), intval) == 0) then
           if(vsmCoordData%YinYang == vsmYin) then
             xx =  x
             xxp = xp
           else
             xx(1) =  -x(1); xx(2) =  x(3); xx(3) =  x(2)
             xxp(1) =  -xp(1); xxp(2) =  xp(3); xxp(3) =  xp(2)
           end if
           call vsmDrawSphere(xx, rad, cam_num, vis_num, amb, dif, spec, shin)
           call vsmDrawCylinder(xxp, xx, rad, cam_num, vis_num, amb, dif, spec, shin)
           xp(1:3) = x(1:3)
         end if

         st_rem = st_rem - 1

         if(st_rem == 0) then
           if(vsmCoordData%YinYang == vsmYin) then
             xx =  x
             xxp = xp
           else
             xx(1) =  -x(1); xx(2) =  x(3); xx(3) =  x(2)
             xxp(1) =  -xp(1); xxp(2) =  xp(3); xxp(3) =  xp(2)
           end if
           call vsmDrawCylinder(xxp, xx, rad, cam_num, vis_num, amb, dif, spec, shin)
           exit
         end if

         if(absval == 0.0_vdp) then
           st_rem = 0
           exit
         end if
       end do
     st_remained(i) = st_rem
     pos(1:3, i) = x(1:3)
     end do

!     if(myrank >= 0) print *,"st_rem(i):", myrank, st_remained(1)
!     if(myrank >= 0) print *,"pos:", myrank, pos(:,1)
!     call vsmCoordsToAngles(pos(1:3, 1), spos)
!     if(myrank >= 0) print *,"spos:", myrank, spos(1), spos(2)*180.0_vdp/pi, spos(3)*180.0_vdp/pi
     !if(myrank == 1) print *,"1:st_rem(i):", st_remained(1:4), pos(:,1:4)o
!     print *,"mpi sync"
    ! MPI sync
     !if(nseed < 1) call vsmError("nseed is less than 1")
     !call vsmLogPrint("nseed ", nseed)
     !call vsmLogPrint("st remain ", st_remained(1:nseed))

     if(cam_num==1) call vsmLogPrint("call mpi_allreduce: remained", st_remained(1:3))
#ifndef CLANG
     call mpi_allreduce(st_remained, st_remained1, nseed, mpi_integer, mpi_min, comm, ierr)
#else
     call vsmmpiallreduceintmin(st_remained, st_remained1, nseed)
#endif
     if(cam_num==1) then
       if(vsmDbg) then
         call vsmLogPrint("after: call mpi_allreduce: remained", st_remained1(1:3))
         call vsmLogPrint("remained pos1", pos(:,1))
         call vsmLogPrint("remained pos2", pos(:,2))
         call vsmLogPrint("remained pos3", pos(:,3))
      end if
     end if

     do m=1, nseed
       if(vsmCoordData%YinYang == vsmYin) then
         pos1(1:3, m) =  pos(1:3, m)
       else
         pos1(1, m)  =  -pos(1,m)
         pos1(2, m)  =  pos(3, m)
         pos1(3, m)  =  pos(2, m)
       end if
     end do

     do m=1, nseed
        if(st_remained1(m) < st_remained(m)) then
           pos1(1:3, m) = vsmFarAway
        end if
     end do

     if(cam_num==1) then
       if(vsmDbg) then
         call vsmLogPrint("remained pos1", pos(:,1))
         call vsmLogPrint("remained pos2", pos(:,2))
         call vsmLogPrint("remained pos3", pos(:,3))
       end if
     end if

     count =3*nseed
#ifndef CLANG
     call mpi_allreduce(pos1, pos, count, mpi_double_precision, mpi_min, comm, ierr)
#else
     call vsmmpiallreducedblemin(pos1, pos, count)
#endif

     if(cam_num==1) then
       if(vsmDbg) then
         call vsmLogPrint("after: second call mpi_allreduce remained")
         call vsmLogPrint("remained pos1", pos1(:,1))
         call vsmLogPrint("remained pos2", pos1(:,2))
         call vsmLogPrint("remained pos3", pos1(:,3))
      end if
     end if

     call vsmLogPrint("call mpi_allreduce ends")

     if(vsmCoordData%YinYang == vsmYang) then
       do m=1, nseed
         pos1(1, m)  =  -pos(1,m)
         pos1(2, m)  =  pos(3, m)
         pos1(3, m)  =  pos(2, m)
       end do
     pos(1:3, 1:nseed) = pos1(1:3, 1:nseed)
     end if

!     pos(1:3, 1:nseed) = pos1(1:3, 1:nseed)
     st_remained(1:nseed) = st_remained1(1:nseed)
     !call vsmLogPrint("sum stremained+nseed = ",sum(st_remained, nseed))
     if(vsmDbg) then
       call vsmLogPrint("pos1", pos(:,1)); call vsmLogPrint("remain 1",st_remained(1))
       call vsmLogPrint("pos2", pos(:,2)); call vsmLogPrint("remain 2",st_remained(2))
       call vsmLogPrint("pos3", pos(:,3)); call vsmLogPrint("remain 3",st_remained(3))
     end if

     sum_remain = 0
     do m=1, nseed
       sum_remain = sum_remain + st_remained(m)
     end do
!     print *,"sum remain", sum_remain
     if(sum_remain == 0) then
       exit
     end if

     end do !---- MPI ----!
     if(l==maxSlineLoop) call vsmLogPrint("Stream Lines could't finish safely")
   end do
  if(cam_num == 1) then
    if(vsmDbg) then
      call vsmLogPrint("vis: ",vis_num)
      call vsmLogPrint("pc_rec1", pc_rec(1))
      call vsmLogPrint("pc_rec2", pc_rec(2))
      call vsmLogPrint("pc_rec3", pc_rec(3))
    end if
  end if
  !call vsmMergeBuf
 end subroutine vsmyyDrawStreamLinesIM

end module vsmStreamLine

