module vsmParam
 use vsmConst
 use vsmFunction

 implicit none
 private
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 !---- declaration of variables ----
 !
 type(vismo_visparameters), dimension(vsmMaxVisKind), save, public :: vsmVisParam
 integer, save, public :: vsmNumViskind
 type(vismo_interval), save, public :: vsmIntval
 type(vismo_mpi), save, public :: vsmMPIVars
 !
 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 
 public vsmGetCurrentIsonum, vsmIsAboveSlice, vsmIsAboveIsoL
 public vsmAddSline, vsmAddVisualization, vsmCalcIntValr, vsmCalcIntvalu, vsmegyyFrameon, vsmGetIsoAlpha
 public vsmGetIsoColor, vsmGetMPIVars, vsmGetSliceAlpha, vsmGetSliceColor, vsmGetVolColor, vsmInitVisParam
 public vsmInspFrameOn, vsmInSphereOn, vsmInyyFrameOn, vsmLightOn, vsmOutSpFrameOn, vsmOutyyFrameOn
 public vsmSetArrowParam, vsmSetDepthq, vsmSetIsoParam, vsmSetSeedParam, vsmSetSliceParam
 public vsmSetSlineParam, vsmSetVolren
 public vsmSetOMPNest, vsmGetOMPNestL1, vsmGetOMPNestL2

 contains

 subroutine vsmInitVisParam(comm, rank, psize, ompth)
   integer, intent(in) :: comm, rank, psize, ompth
   ! after mpi_init
   call vsmInitIsoparam
   call vsmInitSliceparam
   call vsmInitVolparam
   call vsmInitSlineparam
   call vsmInitArrowparam
   call vsmInitLight
   call vsmInitDepthq
   call vsmInitFrame
   vsmNumViskind = 0
   vsmMPIVars%comm   = comm
   vsmMPIVars%myrank = rank
   vsmMPIVars%psize  = psize
   vsmMPIVars%omp_tn = ompth
   vsmMPIVars%omp_tnl1 = ompth
   vsmMPIVars%omp_tnl2 = 1
 end subroutine vsmInitVisParam

 subroutine vsmAddVisualization
   vsmNumViskind = vsmNumViskind + 1
   if(vsmNumViskind > vsmMaxVisKind) call vsmError('add visualization : exceed vsmMaxVisKind')
 end subroutine vsmAddVisualization

 function vsmGetCurrentIsonum()
   integer :: vsmGetCurrentIsonum
   vsmGetCurrentIsonum = vsmVisParam(vsmNumViskind)%iso_num
 end function vsmGetCurrentIsonum

 ! subroutine for lights
 subroutine vsmInitLight
  integer :: i
  vsmVisParam(1:vsmMaxVisKind)%light_num = 0
  do i=1, vsmMaxVisKind
    vsmVisParam(i)%light_used(1:vsmMaxLight) = .false.
  end do
 end subroutine vsmInitLight

 subroutine vsmLightOn(nlight)
   integer, intent(in) :: nlight
   integer :: light_num

   if(nlight > vsmMaxLight .or. nlight < 1) call vsmError('num of lights')
   light_num = vsmVisParam(vsmNumViskind)%light_num + 1
   vsmVisParam(vsmNumViskind)%light_num = light_num
   vsmVisParam(vsmNumViskind)%light_used(nlight) = .true.
 end subroutine vsmLightOn

 ! subroutines for depthcue
 subroutine vsmInitDepthq
   vsmVisParam(1:vsmMaxVisKind)%depthq = .false.
 end subroutine vsmInitDepthq

 subroutine vsmSetDepthQ
   vsmVisParam(vsmNumViskind)%depthq = .true.
 end subroutine vsmSetDepthQ

 subroutine vsmInitFrame
  vsmVisParam(1:vsmMaxVisKind)%frameparam%insphere = .false.
  vsmVisParam(1:vsmMaxVisKind)%frameparam%inspframe = .false.
  vsmVisParam(1:vsmMaxVisKind)%frameparam%outspframe = .false.
  vsmVisParam(1:vsmMaxVisKind)%frameparam%inyyframe = .false.
  vsmVisParam(1:vsmMaxVisKind)%frameparam%outyyframe = .false.
  vsmVisParam(1:vsmMaxVisKind)%frameparam%egyyframe = .false.
 end subroutine vsmInitFrame

 subroutine vsmInSphereOn
   vsmVisParam(vsmNumViskind)%frameparam%insphere = .true.
 end subroutine vsmInSphereOn
 
 subroutine vsmInSpFrameOn
   vsmVisParam(vsmNumViskind)%frameparam%inspframe = .true.
 end subroutine vsmInSpFrameOn

 subroutine vsmOutSpFrameOn
   vsmVisParam(vsmNumViskind)%frameparam%outspframe = .true.
 end subroutine vsmOutSpFrameOn

 subroutine vsmInyyFrameOn
   vsmVisParam(vsmNumViskind)%frameparam%inyyframe = .true.
 end subroutine vsmInyyFrameOn

 subroutine vsmOutyyFrameOn
   vsmVisParam(vsmNumViskind)%frameparam%outyyframe = .true.
 end subroutine vsmOutyyFrameOn

 subroutine vsmegyyFrameOn
   vsmVisParam(vsmNumViskind)%frameparam%egyyframe = .true.
 end subroutine vsmegyyFrameOn
 
 ! subroutines for isosurface
 subroutine vsmInitIsoparam
   integer :: i, k

   do k=1, vsmMaxVisKind
     do i=1, vsmMaxIsosurf
       vsmVisParam(k)%isoparam(i)%iso_level = 0.0_vdp
       vsmVisParam(k)%isoparam(i)%alpha = 0.0
       vsmVisParam(k)%isoparam(i)%nscal = -1
       vsmVisParam(k)%isoparam(i)%amb(1:3)  = 0.0
       vsmVisParam(k)%isoparam(i)%dif(1:3)  = 0.0
       vsmVisParam(k)%isoparam(i)%spec(1:3) = 0.0
       vsmVisParam(k)%isoparam(i)%shin = 0.0
     end do
     vsmVisParam(k)%iso_num = 0
     vsmVisParam(k)%iso_solid = .true.
   end do
 end subroutine vsmInitIsoparam

 function vsmCheckIso(vnum, niso)
   integer, intent(in) :: vnum, niso
   logical :: vsmCheckIso

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckIso = .false.
     return
   end if

   if(niso < 1 .or. niso > vsmVisParam(vnum)%iso_num) then
     vsmCheckIso = .false.
     return
   end if
   vsmCheckIso = .true.
 end function vsmCheckIso

 function vsmCheckIso2(vnum, niso, lpse)
   integer, intent(in) :: vnum
   integer, dimension(vsmVecLen+1), intent(in) :: niso
   integer, dimension(2), intent(in) :: lpse
   logical :: vsmCheckIso2
   integer :: i, err

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckIso2 = .false.
     return
   end if

   err = 0
   do i=lpse(1), lpse(2)
      if(niso(i) < 1 .or. niso(i) > vsmVisParam(vnum)%iso_num) then
         err = err + 1
      end if 
   end do

   if(err > 0) then
     vsmCheckIso2 = .false.
     return
   end if

   vsmCheckIso2 = .true.
 end function vsmCheckIso2

 subroutine vsmSetIsoParam(level, nscal, alpha, amb, dif, spec, shin)
   integer, intent(in) :: nscal
   real(kind=vdp), intent(in) :: level
   real, intent(in) :: alpha, shin
   real, dimension(3), intent(in) :: amb, dif, spec
   integer :: iso_num

   if(nscal > vsmMaxScalar .or. nscal < 1 .or. vsmNumViskind > vsmMaxVisKind) then
     call vsmError('vsmSetIsoParam')
   end if
   iso_num = vsmVisParam(vsmNumViskind)%iso_num + 1

   if(iso_num > vsmMaxIsosurf) call vsmError('set_isparam : exceed vsmMaxIsosurf')
   vsmVisParam(vsmNumViskind)%iso_num = iso_num
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%iso_level = level
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%alpha = alpha
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%nscal = nscal

   call vsmCheckFA01(amb); call vsmCheckFA01(dif); call vsmCheckFA01(spec)
   call vsmCheckF01(alpha)
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%amb(1:3)  = amb(1:3)
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%dif(1:3)  = dif(1:3)
   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%spec(1:3) = spec(1:3)

   vsmVisParam(vsmNumViskind)%isoparam(iso_num)%shin = shin
 end subroutine vsmSetIsoParam

subroutine vsmChangeIsoLevel(visnum, isonum, level)
   integer, intent(in) :: visnum, isonum
   real(kind=vdp), intent(in) :: level

   if(visnum > vsmNumViskind .or. visnum < 1) then
     call vsmError('vsmChangeIsoLevel :: visnum is strange')   
   end if

   if(isonum > vsmVisParam(visnum)%iso_num .or. isonum < 1) then
     call vsmError('vsmChangeIsoLevel :: isonum is strange')   
   end if

   vsmVisParam(visnum)%isoparam(isonum)%iso_level = level
   if(vsmDbg) then
     call vsmLogPrint("change iso level visnum =", visnum)
     call vsmLogPrint("change iso level isonum =", isonum)
     call vsmLogPrint("change iso level to ", level)
   end if
 end subroutine vsmChangeIsoLevel

 subroutine vsmGetIsoColor(visnum, niso, alpha, amb, dif, spec, shin)
   integer, intent(in) :: visnum, niso
   real, intent(out) :: alpha, shin
   real, dimension(3), intent(out) :: amb, dif, spec

   if(.not.vsmCheckIso(visnum, niso)) call vsmError("error: vsmGetIsoColor")
   alpha = vsmVisParam(visnum)%isoparam(niso)%alpha
   amb(1:3) = vsmVisParam(visnum)%isoparam(niso)%amb(1:3)
   dif(1:3) = vsmVisParam(visnum)%isoparam(niso)%dif(1:3)
   spec(1:3) = vsmVisParam(visnum)%isoparam(niso)%spec(1:3)
   shin = vsmVisParam(visnum)%isoparam(niso)%shin
 end subroutine vsmGetIsoColor

 function vsmIsAboveIsoL(visnum, niso, s)
   integer, intent(in) :: niso, visnum
   real(kind=vdp), intent(in) :: s
   real(kind=vdp) :: vsmIsAboveIsoL

   if(.not.vsmCheckIso(visnum, niso)) call vsmError("error: vsmIsAboveIsoL")
   vsmIsAboveIsoL = vsmVisParam(visnum)%isoparam(niso)%iso_level - s
 end function vsmIsAboveIsoL

 subroutine vsmGetIsoAlpha(visnum, niso, alpha)
   integer, intent(in) :: visnum, niso
   real, intent(out) :: alpha

   if(.not.vsmCheckIso(visnum, niso)) call vsmError("error: vsmGetIsoAlpha")
   alpha = vsmVisParam(visnum)%isoparam(niso)%alpha
 end subroutine vsmGetIsoAlpha

 ! subroutines and functions for color slice
 subroutine vsmInitSliceparam
   integer :: i, k

   do k=1, vsmMaxVisKind
     do i=1, vsmMaxSlice
       vsmVisParam(k)%sliceparam(i)%sc(1:4) = 0.0_vdp
       vsmVisParam(k)%sliceparam(i)%alpha = 0.0
       vsmVisParam(k)%sliceparam(i)%nscal = -1
       vsmVisParam(k)%sliceparam(i)%type = 0
       vsmVisParam(k)%sliceparam(i)%amb = 0.0
       vsmVisParam(k)%sliceparam(i)%dif = 0.0
       vsmVisParam(k)%sliceparam(i)%spec = 0.0
       vsmVisParam(k)%sliceparam(i)%shin = 0.0
     end do
     vsmVisParam(k)%slice_num = 0
     vsmVisParam(k)%slice_solid = .true.
   end do
 end subroutine vsmInitSliceparam

 function vsmCheckSlice(vnum, nsli)
   integer, intent(in) :: vnum, nsli
   logical :: vsmCheckSlice

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckSlice = .false.
     return
   end if

   if(nsli < 1 .or. nsli > vsmVisParam(vnum)%slice_num) then
     vsmCheckSlice = .false.
     return
   end if
   vsmCheckSlice = .true.
 end function vsmCheckSlice

 subroutine vsmSetSliceParam(coef, nscal, alpha, amb, dif, spec, shin, type)
   real(kind=vdp), dimension(4), intent(in) :: coef
   real(kind=vdp), dimension(4) :: tco
   real(kind=vdp) :: temp
   integer, intent(in) :: nscal, type
   real, intent(in) :: alpha, amb, dif, spec, shin
   integer :: slice_num

   if(nscal > vsmMaxScalar .or. vsmNumViskind > vsmMaxVisKind) call vsmError('vsmSetSliceParam')
   slice_num = vsmVisParam(vsmNumViskind)%slice_num + 1
   if(slice_num > vsmMaxSlice) call vsmError('vsmSetSliceParam : exceed vsmMaxSlice')
   vsmVisParam(vsmNumViskind)%slice_num = slice_num
   if(type == 1) then
      vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%sc(1:4) = coef(1:4)
   else if(type == 0) then
      tco(1:4) = coef(1:4)
      temp = sqrt(tco(1)*tco(1)+tco(2)*tco(2)+tco(3)*tco(3))
      vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%sc(1:4) = coef(1:4)/temp
   end if

   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%nscal = nscal
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%type = type

   call vsmCheckF01(alpha)
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%alpha = alpha
   call vsmCheckF01(amb); call vsmCheckF01(dif); call vsmCheckF01(spec); 
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%amb  = amb
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%dif  = dif
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%spec = spec
   vsmVisParam(vsmNumViskind)%sliceparam(slice_num)%shin = shin
 end subroutine vsmSetSliceParam

 function vsmIsAboveSlice(visnum, nslice, x)
   integer, intent(in) :: nslice, visnum
   real(kind=vdp), dimension(3), intent(in) :: x
   real(kind=vdp) :: vsmIsAboveSlice
   integer :: i

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmIsAboveSlice")
   vsmIsAboveSlice = vsmVisParam(visnum)%sliceparam(nslice)%sc(4)
   do i=1, 3
     vsmIsAboveSlice = vsmIsAboveSlice + vsmVisParam(visnum)%sliceparam(nslice)%sc(i) * x(i)
   end do
 end function vsmIsAboveSlice

 subroutine vsmGetSliceColor(visnum, nslice, alpha, amb, dif, spec, shin)
   integer, intent(in) :: nslice, visnum
   real, intent(out) :: alpha, amb, dif, spec, shin

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmGetSliceColor")
   alpha = vsmVisParam(visnum)%sliceparam(nslice)%alpha
   amb   = vsmVisParam(visnum)%sliceparam(nslice)%amb
   dif   = vsmVisParam(visnum)%sliceparam(nslice)%dif
   spec  = vsmVisParam(visnum)%sliceparam(nslice)%spec
   shin  = vsmVisParam(visnum)%sliceparam(nslice)%shin
 end subroutine vsmGetSliceColor

 subroutine vsmGetSliceAlpha(visnum, nslice, alpha)
   integer, intent(in) :: nslice, visnum
   real, intent(out) :: alpha

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmGetSliceAlpha")
   alpha = vsmVisParam(visnum)%sliceparam(nslice)%alpha
 end subroutine vsmGetSliceAlpha

 subroutine vsmGetSliceNorm(visnum, nslice, norm)
   integer, intent(in) :: nslice, visnum
   real(kind=vdp), dimension(3), intent(out) :: norm

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmGetSliceNorm")
   norm(1:3) = vsmVisParam(visnum)%sliceparam(nslice)%sc(1:3)
   !call vsmNormVec(norm)
 end subroutine vsmGetSliceNorm

 subroutine vsmGetSliceNormf(visnum, nslice, norm)
   integer, intent(in) :: nslice, visnum
   real, dimension(3), intent(out) :: norm

   if(.not.vsmCheckSlice(visnum, nslice)) call vsmError("error: vsmGetSliceNormf")
   norm(1:3) = real(vsmVisParam(visnum)%sliceparam(nslice)%sc(1:3))
   call vsmNormVecF(norm)
 end subroutine vsmGetSliceNormf

 ! subroutines for volume rendering
 subroutine vsmInitVolparam
  integer :: i, k

  do k=1, vsmMaxVisKind
    do i=1, vsmMaxScalar
      vsmVisParam(k)%volparam(i)%nscal = -1
      vsmVisParam(k)%volparam(i)%amb  = 0.0
      vsmVisParam(k)%volparam(i)%dif  = 0.0
      vsmVisParam(k)%volparam(i)%spec = 0.0
      vsmVisParam(k)%volparam(i)%shin = 0.0
    end do
    vsmVisParam(k)%volren_num = 0
  end do
 end subroutine vsmInitVolparam

 function vsmCheckVolren(vnum, nvol)
   integer, intent(in) :: vnum, nvol
   logical :: vsmCheckVolren

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckVolren = .false.
     return
   end if

   if(nvol < 1 .or. nvol > vsmVisParam(vnum)%volren_num) then
     vsmCheckVolren = .false.
     return
   end if
   vsmCheckVolren = .true.
 end function vsmCheckVolren

 subroutine vsmSetVolren(nscal, amb, dif, spec, shin)
   integer, intent(in) :: nscal
   real, intent(in) :: amb, dif, spec, shin
   integer :: volren_num

   if(nscal > vsmMaxScalar .or. nscal < 1) call vsmError('vsmSetVolren')
   volren_num = vsmVisParam(vsmNumViskind)%volren_num + 1
   if(volren_num > vsmMaxVolren) call vsmError('vsmSetVolren : exceed vsmMaxVolren')
   vsmVisParam(vsmNumViskind)%volren_num = volren_num
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%nscal = nscal

   call vsmCheckF01(amb); call vsmCheckF01(dif); call vsmCheckF01(spec); 
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%amb  = amb
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%dif  = dif
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%spec = spec
   vsmVisParam(vsmNumViskind)%volparam(volren_num)%shin = shin
 end subroutine vsmSetVolren

 subroutine vsmGetVolColor(visnum, nvol, amb, dif, spec, shin)
   integer, intent(in) :: nvol, visnum
   real, intent(out) :: amb, dif, spec, shin

   if(.not.vsmCheckVolren(visnum, nvol)) call vsmError("error: vsmGetVolColor")
   amb   = vsmVisParam(visnum)%volparam(nvol)%amb
   dif   = vsmVisParam(visnum)%volparam(nvol)%dif
   spec  = vsmVisParam(visnum)%volparam(nvol)%spec
   shin  = vsmVisParam(visnum)%volparam(nvol)%shin
 end subroutine vsmGetVolColor

 ! common subroutines for ray casting and StreamLines
 ! vsmCalcIntvalU or vsmCalcIntvalR must be called when setting coordinates
 subroutine vsmCalcIntvalU(dx)
   real(kind=vdp), dimension(3), intent(in) :: dx
   real(kind=vdp) :: mind

   call vsmLogPrint("calcintvalu is called")
   mind = min(dx(1), dx(2), dx(3))
   vsmIntval%ray_interval = mind*vsmRayStride
   vsmIntval%dh = mind*vsmIntegStride
   if(vsmDbg) then
     call vsmLogPrint("ray_interval:", vsmIntval%ray_interval)
     call vsmLogPrint("dh:", vsmIntval%dh)
   end if
!   print *,"ray_intval, dh",vsmIntval%ray_interval,vsmIntval%dh
 end subroutine vsmCalcIntvalU

 subroutine vsmCalcIntvalR(coordx, coordy, coordz, nx)
   integer, dimension(3), intent(in) :: nx
   real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
   real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
   real(kind=vdp), dimension(nx(3)), intent(in) :: coordz
   real(kind=vdp) :: mind, d
   integer :: i

   mind = coordx(nx(1)) - coordx(1)
   do i=1, nx(1)-1
     d = coordx(i+1)-coordx(i)
     if( d < mind) then
        mind = d
     end if
   end do

   do i=1, nx(2)-1
     d = coordy(i+1)-coordy(i)
     if( d < mind) then
        mind = d
     end if
   end do

   do i=1, nx(3)-1
     d = coordz(i+1)-coordz(i)
     if( d < mind) then
        mind = d
     end if
   end do

   vsmIntval%ray_interval = mind*vsmRayStride
   vsmIntval%dh = mind*vsmIntegStride
 end subroutine vsmCalcIntvalR

 ! subrtouines for stream lines
 subroutine vsmInitSlineparam
   integer :: i, k

   do k=1, vsmMaxVisKind
     do i=1, vsmMaxVector !partially vectrize
       vsmVisParam(k)%slineparam(i)%seed(1, 1:vsmMaxSeed) = 0.0_vdp
       vsmVisParam(k)%slineparam(i)%seed(2, 1:vsmMaxSeed) = 0.0_vdp
       vsmVisParam(k)%slineparam(i)%seed(3, 1:vsmMaxSeed) = 0.0_vdp
       vsmVisParam(k)%slineparam(i)%num_seed = 0
       vsmVisParam(k)%slineparam(i)%steps_remained(1:vsmMaxSeed) = vsmMaxSlineSteps
       vsmVisParam(k)%slineparam(i)%nvec = -1
       vsmVisParam(k)%slineparam(i)%amb(1)  = 0.0
       vsmVisParam(k)%slineparam(i)%amb(2)  = 0.0
       vsmVisParam(k)%slineparam(i)%amb(3)  = 0.0
       vsmVisParam(k)%slineparam(i)%dif(1)  = 0.0
       vsmVisParam(k)%slineparam(i)%dif(2)  = 0.0
       vsmVisParam(k)%slineparam(i)%dif(3)  = 0.0
       vsmVisParam(k)%slineparam(i)%spec(1) = 0.0
       vsmVisParam(k)%slineparam(i)%spec(2) = 0.0
       vsmVisParam(k)%slineparam(i)%spec(3) = 0.0
       vsmVisParam(k)%slineparam(i)%shin = 0.0
       vsmVisParam(k)%slineparam(i)%radius = 1.0_vdp
       vsmVisParam(k)%slineparam(i)%monocolor = .false.
     end do
     vsmVisParam(k)%sline_num = 0
   end do
 end subroutine vsmInitSlineparam

 function vsmCheckSline(vnum, nsline)
   integer, intent(in) :: vnum, nsline
   logical :: vsmCheckSline

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckSline = .false.
     return
   end if

   if(nsline < 1 .or. nsline > vsmVisParam(vnum)%sline_num) then
     vsmCheckSline = .false.
     return
   end if
   vsmCheckSline = .true.
 end function vsmCheckSline

 subroutine vsmAddSline
   integer :: sline_num

   sline_num = vsmVisParam(vsmNumViskind)%sline_num + 1
   if(sline_num > vsmMaxVector) call vsmError('vsmAddSline : exceed max_sline')
   vsmVisParam(vsmNumViskind)%sline_num = sline_num
 end subroutine vsmAddSline

 subroutine vsmSetSeedParam(x)
   real(kind=vdp), dimension(3), intent(in) :: x
   integer :: sline_num, num_seed

   sline_num = vsmVisParam(vsmNumViskind)%sline_num
   num_seed = vsmVisParam(vsmNumViskind)%slineparam(sline_num)%num_seed + 1
   if(num_seed > vsmMaxSeed) call vsmError('set seedparam : exceed vsmMaxSeed')
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%num_seed = num_seed
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%seed(1:3, num_seed) = x(1:3)
 end subroutine vsmSetSeedParam

 subroutine vsmSetSlineParam(nvec, mono, amb, dif, spec, shin, radius)
   integer, intent(in) :: nvec
   logical, intent(in) :: mono
   real, dimension(3), intent(in) :: amb, dif, spec
   real, intent(in) :: shin
   real(kind=vdp), intent(in) :: radius
   integer :: sline_num

   if(nvec > vsmMaxVector .or. nvec < 1) call vsmError('vsmSetSlineParam :: nvec')
   sline_num = vsmVisParam(vsmNumViskind)%sline_num
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%nvec = nvec
   call vsmCheckFA01(amb); call vsmCheckFA01(dif); call vsmCheckFA01(spec)
     if(.not.mono) then
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%amb(1:3) = amb(1:3)
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%dif(1:3) = dif(1:3)
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%spec(1:3) = spec(1:3)
     else
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%amb(1:3) = amb(1)
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%dif(1:3) = dif(1)
       vsmVisParam(vsmNumViskind)%slineparam(sline_num)%spec(1:3) = spec(1)
     end if
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%shin = shin
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%radius = radius
   vsmVisParam(vsmNumViskind)%slineparam(sline_num)%monocolor = mono
 end subroutine vsmSetSlineParam

subroutine vsmGetSlineColor(visnum, nsline, amb, dif, spec, shin)
   integer, intent(in) :: nsline, visnum
   real, intent(out) :: amb, dif, spec, shin

   if(.not.vsmCheckSline(visnum, nsline)) call vsmError("error: vsmGetSlineColor")
   amb   = vsmVisParam(visnum)%slineparam(nsline)%amb(1)
   dif   = vsmVisParam(visnum)%slineparam(nsline)%dif(1)
   spec  = vsmVisParam(visnum)%slineparam(nsline)%spec(1)
   shin  = vsmVisParam(visnum)%slineparam(nsline)%shin
 end subroutine vsmGetSlineColor

 subroutine vsmGetMonoLineColor(visnum, nsline, amb, dif, spec, shin)
   integer, intent(in) :: nsline, visnum
   real, dimension(3), intent(out) :: amb, dif, spec
   real, intent(out) :: shin

   if(.not.vsmCheckSline(visnum, nsline)) call vsmError("error: vsmGetMonoLineColor")
   amb(1:3)   = vsmVisParam(visnum)%slineparam(nsline)%amb(1:3)
   dif(1:3)   = vsmVisParam(visnum)%slineparam(nsline)%dif(1:3)
   spec(1:3)  = vsmVisParam(visnum)%slineparam(nsline)%spec(1:3)
   shin  = vsmVisParam(visnum)%slineparam(nsline)%shin
 end subroutine vsmGetMonoLineColor

 ! subroutines for arrows
  subroutine vsmInitArrowparam
   integer :: i, k

   do k=1, vsmMaxVisKind
     do i=1, vsmMaxArrows
       vsmVisParam(k)%arrowparam(i)%sc(1:4) = 0.0_vdp
       vsmVisParam(k)%arrowparam(i)%nvec = -1
       vsmVisParam(k)%arrowparam(i)%density = 1.0
       vsmVisParam(k)%arrowparam(i)%radius = 1.0
       vsmVisParam(k)%arrowparam(i)%length = 1.0
     end do
     vsmVisParam(k)%arrow_num = 0
   end do
 end subroutine vsmInitArrowparam

 function vsmCheckArrow(vnum, narrow)
   integer, intent(in) :: vnum, narrow
   logical :: vsmCheckArrow

   if(vnum > vsmMaxVisKind .or. vnum < 1) then
     vsmCheckArrow = .false.
     return
   end if

   if(narrow < 1 .or. narrow > vsmVisParam(vnum)%arrow_num) then
     vsmCheckArrow = .false.
     return
   end if
   vsmCheckArrow = .true.
 end function vsmCheckArrow

 subroutine vsmSetArrowParam(coef, nvec, density, rad, length)
   real(kind=vdp), dimension(4), intent(in) :: coef
   integer, intent(in) :: nvec
   real(kind=vdp), intent(in) :: density, rad, length
   integer :: arrow_num

   arrow_num = vsmVisParam(vsmNumViskind)%arrow_num + 1
   if(nvec > vsmMaxVector .or. arrow_num > vsmMaxArrows) call vsmError('vsmSetArrowParam')
   vsmVisParam(vsmNumViskind)%arrow_num = arrow_num
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%sc(1:4) = coef(1:4)
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%nvec = nvec
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%density = density
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%radius = rad
   vsmVisParam(vsmNumViskind)%arrowparam(arrow_num)%length = length
 end subroutine vsmSetArrowParam

 subroutine vsmGetArPlane(visnum, nslice, eq)
  real(kind=vdp), dimension(4), intent(out) :: eq
  integer, intent(in) :: nslice, visnum

  if(.not.vsmCheckArrow(visnum, nslice)) call vsmError("error: vsmGetArPlane")
  eq(1:4) = vsmVisParam(visnum)%arrowparam(nslice)%sc(1:4)
 end subroutine vsmGetArPlane

 subroutine vsmGetArParam(visnum, nslice, den, rad, length)
  real(kind=vdp), intent(out) :: den, rad, length
  integer, intent(in) :: nslice, visnum

  if(.not.vsmCheckArrow(visnum, nslice)) call vsmError("error: vsmGetArParam")
  den = vsmVisParam(visnum)%arrowparam(nslice)%density
  rad = vsmVisParam(visnum)%arrowparam(nslice)%radius
  length = vsmVisParam(visnum)%arrowparam(nslice)%length
 end subroutine vsmGetArParam

 subroutine vsmGetMPIVars(comm, rank, psize, ompn)
   integer, intent(out) :: comm
   integer, intent(out), optional :: rank
   integer, intent(out), optional :: psize
   integer, intent(out), optional :: ompn

   comm = vsmMPIVars%comm
   if(present(rank))  rank  = vsmMPIVars%myrank
   if(present(psize)) psize = vsmMPIVars%psize
   if(present(ompn))  ompn  = vsmMPIVars%omp_tn
 end subroutine vsmGetMPIVars

 subroutine vsmSetOMPNest(num)
   integer, intent(in) :: num
   integer :: tnuml1, tnuml2
   tnuml1 = min(vsmMaxOMP4Solid, vsmMPIVars%omp_tn, num)
   if(tnuml1 == 0) tnuml1 = 1
   tnuml2 = vsmMPIVars%omp_tn/tnuml1
   if(tnuml2 < 2) tnuml2 = 1
   vsmMPIVars%omp_tnl1=tnuml1
   vsmMPIVars%omp_tnl2=tnuml2
 end subroutine vsmSetOMPNest
 
 function vsmGetOMPNestL1()
   integer :: vsmGetOMPNestL1
   vsmGetOMPNestL1 = vsmMPIVars%omp_tnl1
 end function vsmGetOMPNestL1

 function vsmGetOMPNestL2()
   integer :: vsmGetOMPNestL2
   vsmGetOMPNestL2 = vsmMPIVars%omp_tnl2
 end function vsmGetOMPNestL2

end module vsmParam

