module vsmCoordinates
 use vsmConst
 use vsmFunction
 use vsmParam

 implicit none
 private

 type(vismo_coord), save, public :: vsmCoord
 type(vismo_localcoord), save, public :: vsmLocalCoord
 type(vismo_coorddata), save, public :: vsmCoordData
 
 public vsmRayDist, vsmCheckLCoord, vsmIsInsidelWO, vsmIsInsidelWOE, vsmyyRayDistInside, vsmIsAnotherSet, vsmIsInsideSphere
 public vsmIsInsideInnerSphere4z0dr, vsmIsInsidel, vsmIsInsideInnerSphere4z1dr, vsmIsBetweenSpheres
 public vsmRayDistInside, vsmIsInsideInnerSphere4Z2dr, vsmyyRayDistSphere, vsmyyRayDist
 public vsmAnglesToCoords, vsmCoordsToAngles, vsmFinalizeCoord, vsmFindCoord, vsmInitCoord, vsmPrepareCoord
 public vsmSetlrect, vsmSetLUniform, vsmSetRect, vsmSetUniform, vsmToYangCoord, vsmyyRayDistSphere4RSlice

 contains

 subroutine vsmYYZCoordMode(yyz)
   logical, intent(in) :: yyz
   vsmCoordData%YYZMode = yyz ! false:YinYang, true:YinYangZhong (in the near future)
   call vsmLogPrint("YYZMode = ", yyz)
 end subroutine vsmYYZCoordMode

 subroutine vsmInitCoord(uni, yyz, rad, YYZMode)
  logical, intent(in) :: uni
  integer, intent(in) :: yyz
  real(kind=vdp), dimension(3), intent(in) :: rad
  logical, intent(in) :: YYZMode
  integer :: i

  call vsmYYZCoordMode(.false.)
  vsmCoord%uniform = uni
  vsmLocalCoord%uniform = uni

  if(uni .neqv. .true.) call vsmError("yy vismo accepts only unicoord")

  if(yyz == vsmYin .or. yyz == vsmYang) then
     vsmCoordData%YinYang = yyz
  else
     call vsmError("vsmInitCoord yinyangzhong variable is strange. vsmYin or vsmYang")
  end if

  do i=1, 3
    vsmCoord%nx(i)     = 0
    vsmCoord%origin(i) = 0.0_vdp
    vsmCoord%cend(i)   = 0.0_vdp
    vsmCoord%dx(i)     = 0.0_vdp

    vsmLocalCoord%nx(i)      = 0
    vsmLocalCoord%lx0(i)     = 0
    vsmLocalCoord%lnx(i)     = 0
    vsmLocalCoord%lorigin(i) = 0.0_vdp
    vsmLocalCoord%lcend(i)   = 0.0_vdp
    vsmLocalCoord%origin(i)  = 0.0_vdp
    vsmLocalCoord%cend(i)    = 0.0_vdp
    vsmLocalCoord%dx(i)      = 0.0_vdp
  end do
  vsmCoord%wdvertex = vsmFarAway
  vsmCoord%lvnum = -1
  
  vsmCoordData%MinRad = rad(1)
  vsmCoordData%MaxRad = rad(2)
  vsmCoordData%dr = rad(3)

 end subroutine vsmInitCoord

 subroutine vsmPrepareCoord

 ! nothing to do

 end subroutine vsmPrepareCoord

 function vsmIsAnotherSet(pos, spos)
   logical :: vsmIsAnotherSet
   real(kind=vdp), dimension(3), intent(in) :: pos, spos
   real(kind=vdp), dimension(3) :: apos, aspos

   if(vsmCoordData%YinYang == vsmYin) then

     if(spos(2) <= vsmYinMaxT .and. spos(2) >= vsmYinMinT&
        .and. spos(3) <= vsmYinMaxP .and. spos(3) >= vsmYinMinP) then
        vsmIsAnotherSet = .false.
     else
        vsmIsAnotherSet = .true.
     end if

   else if(vsmCoordData%YinYang == vsmYang) then
     vsmIsAnotherSet = .true.
     apos(1) = -pos(1); apos(2) = pos(3); apos(3) = pos(2)
     call vsmCoordsToAngles(apos, aspos)

     if(aspos(2) <= vsmYinMaxT .and. aspos(2) >= vsmYinMinT&
        .and. aspos(3) <= vsmYinMaxP .and. aspos(3) >= vsmYinMinP) then
        vsmIsAnotherSet = .true.
     else
        vsmIsAnotherSet = .false.
     end if

   end if

 end function vsmIsAnotherSet

 subroutine vsmSetUniform(nx, origin, dx, radon)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  logical, optional, intent(in) :: radon
  real(kind=vdp), dimension(3) :: dr
  logical :: radian
!  integer :: i

  vsmCoord%nx(1:3) = nx(1:3)
  radian = .false.
  if(present(radon)) then
    if(radon) radian = .true.
  end if

  if(vsmCoordData%YinYang == vsmYin .or. vsmCoordData%YinYang == vsmYang) then
    if(radian) then
      vsmCoord%origin(1:3) = origin(1:3)
    else
      vsmCoord%origin(1) = origin(1)
      vsmCoord%origin(2) = vsmDeg2Rad(origin(2))
      vsmCoord%origin(3) = vsmDeg2Rad(origin(3))
    end if
  else
    call vsmError("YinYangZhong value is strange: setuniform")
  end if

  if(vsmCoordData%YinYang == vsmYin .or. vsmCoordData%YinYang == vsmYang) then
    if(radian) then
      vsmCoord%dx(1:3) = dx(1:3)
    else
      vsmCoord%dx(1) = dx(1)
      vsmCoord%dx(2) = vsmDeg2Rad(dx(2))
      vsmCoord%dx(3) = vsmDeg2Rad(dx(3))
    end if
  else
    call vsmError("YinYangZhong value is strange: setuniform")
  end if

  vsmCoord%cend(1:3) = vsmCoord%origin(1:3)+(nx(1:3)-1)*vsmCoord%dx(1:3)

  if(vsmDbg) then
    call vsmLogPrint("origin:", vsmCoord%origin)
    call vsmLogPrint("dx:", vsmCoord%dx)
    call vsmLogPrint("end:", vsmCoord%cend)
    call vsmLogPrint("nx:", vsmCoord%nx)
  end if

  if(vsmCoordData%YinYang == vsmYin .or. vsmCoordData%YinYang == vsmYang) then
    if(radian) then
      dr(1:3) = dx(1:3)
    else
      dr(1) = dx(1)
      dr(2) = vsmDeg2Rad(dx(2))
      dr(3) = vsmDeg2Rad(dx(3))
    end if
    dr(2) = dr(2) * vsmCoordData%MinRad; dr(3) = dr(3) * vsmCoordData%MinRad/sqrt(2.0_vdp) ! *sin(45.0 deg) 
  else
    call vsmError("YinYangZhong value is strange: setuniform")
  end if
  call vsmCalcIntvalU(dr)
 end subroutine vsmSetUniform

 subroutine vsmSetLUniform(nx, lx0, lnx, origin, dx, radon)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  logical, optional, intent(in) :: radon
  logical :: radian
!  integer :: i

  vsmLocalCoord%nx(1:3)  = nx(1:3)
  vsmLocalCoord%lx0(1:3) = lx0(1:3)
  vsmLocalCoord%lnx(1:3) = lnx(1:3)

  radian = .false.
  if( present(radon) ) then
    if ( ( radon) ) radian = .true.
  end if

  if(vsmCoordData%YinYang == vsmYin .or. vsmCoordData%YinYang == vsmYang) then
    if(radian) then
      vsmLocalCoord%dx(1:3)  = dx(1:3)
    else
      vsmLocalCoord%dx(1)  = dx(1)
      vsmLocalCoord%dx(2)  = vsmDeg2Rad(dx(2))
      vsmLocalCoord%dx(3)  = vsmDeg2Rad(dx(3))
    end if
  else
    call vsmError("YinYangZhong value is strange: setluniform")
  end if

  if(vsmCoordData%YinYang == vsmYin .or. vsmCoordData%YinYang == vsmYang) then
    if(radian) then
      vsmLocalCoord%lorigin(1:3) = origin(1:3)+(lx0(1:3)-1)*dx(1:3) !without overlap
    else
      vsmLocalCoord%lorigin(1) = origin(1)+(lx0(1)-1)*dx(1) !without overlap
      vsmLocalCoord%lorigin(2) = vsmDeg2Rad(origin(2))+(lx0(2)-1)*vsmLocalCoord%dx(2) !without overlap
      vsmLocalCoord%lorigin(3) = vsmDeg2Rad(origin(3))+(lx0(3)-1)*vsmLocalCoord%dx(3) !without overlap
    end if
  else
    call vsmError("YinYangZhong value is strange: setluniform")
  end if

  vsmLocalCoord%lcend(1:3)   = vsmLocalCoord%lorigin(1:3)+(lnx(1:3)-1)*vsmLocalCoord%dx(1:3)!without overlap

  if(vsmCoordData%YinYang == vsmYin .or. vsmCoordData%YinYang == vsmYang) then
    if(radian) then
      vsmLocalCoord%origin(1:3)  = origin(1:3) !with overlap
    else
      vsmLocalCoord%origin(1)  = origin(1) !with overlap
      vsmLocalCoord%origin(2)  = vsmDeg2Rad(origin(2)) !with overlap
      vsmLocalCoord%origin(3)  = vsmDeg2Rad(origin(3)) !with overlap
    end if
  else
    call vsmError("YinYangZhong value is strange: setluniform")
  end if

  vsmLocalCoord%cend(1:3) = vsmLocalCoord%origin(1:3)+(nx(1:3)-1)*vsmLocalCoord%dx(1:3) !with overlap

  if(vsmDbg) then
    call vsmLogPrint("local origin:", vsmLocalCoord%lorigin)
    call vsmLogPrint("local end:", vsmLocalCoord%lcend)
    call vsmLogPrint("local origin WO:", vsmLocalCoord%origin)
    call vsmLogPrint("local end WO:", vsmLocalCoord%cend)
    call vsmLogPrint("nx:", vsmLocalCoord%nx)
  end if
 end subroutine vsmSetLUniform

 subroutine vsmSetRect(nx, coordx, coordy, coordz)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
  real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
  real(kind=vdp), dimension(nx(3)), intent(in) :: coordz
  integer :: allocstate

  vsmCoord%nx(1:3) = nx(1:3)
  vsmCoord%origin(1) = coordx(1)
  vsmCoord%origin(2) = coordy(1)
  vsmCoord%origin(3) = coordz(1)
  vsmCoord%cend(1) = coordx(nx(1))
  vsmCoord%cend(2) = coordy(nx(2))
  vsmCoord%cend(3) = coordz(nx(3))

  allocate(vsmCoord%coordx(nx(1)), vsmCoord%coordy(nx(2)), &
             vsmCoord%coordz(nx(3)), stat=allocstate)
  if(allocstate /= 0) call vsmError('cant allocate coordinates')

  vsmCoord%coordx(1:nx(1)) = coordx(1:nx(1))
  vsmCoord%coordy(1:nx(2)) = coordy(1:nx(2))
  vsmCoord%coordz(1:nx(3)) = coordz(1:nx(3))
  call vsmCalcIntvalR(coordx, coordy, coordz, nx)
 end subroutine vsmSetRect

 subroutine vsmSetLRect(nx, lx0, lnx, coordx, coordy, coordz)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
  real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
  real(kind=vdp), dimension(nx(3)), intent(in) :: coordz
  integer :: allocstate, i

  do i=1, 3
    vsmLocalCoord%nx(i) = nx(i)
    vsmLocalCoord%lx0(i) = lx0(i)
    vsmLocalCoord%lnx(i) = lnx(i)
  end do

  vsmLocalCoord%lorigin(1) = coordx(lx0(1))
  vsmLocalCoord%lorigin(2) = coordy(lx0(2))
  vsmLocalCoord%lorigin(3) = coordz(lx0(3))
  vsmLocalCoord%lcend(1)   = coordx(lx0(1)+lnx(1)-1)
  vsmLocalCoord%lcend(2)   = coordy(lx0(2)+lnx(2)-1)
  vsmLocalCoord%lcend(3)   = coordz(lx0(3)+lnx(3)-1)

  allocate(vsmLocalCoord%coordx(nx(1)), vsmLocalCoord%coordy(nx(2)), &
             vsmLocalCoord%coordz(nx(3)), stat=allocstate)
  if(allocstate /= 0) call vsmError('cant allocate coordinates')

  vsmLocalCoord%coordx(1:nx(1)) = coordx(1:nx(1))
  vsmLocalCoord%coordy(1:nx(2)) = coordy(1:nx(2))
  vsmLocalCoord%coordz(1:nx(3)) = coordz(1:nx(3))
 end subroutine vsmSetLRect

 subroutine vsmFindCoord(x, ratio, coordpos)
  real(kind=vdp), dimension(3), intent(in) :: x
  real(kind=vdp), dimension(3), intent(inout) :: ratio
  integer, dimension(3), intent(inout) :: coordpos

  if(vsmCoord%uniform) then
    call vsmFindUnicoord(x, ratio, coordpos)
  else
    call vsmError("yyvismo accepts only uni coord")
    call vsmFindReccoord(x, ratio, coordpos)
  end if

 end subroutine vsmFindCoord

 subroutine vsmFindUnicoord(x, ratio, coordpos, lnum)
  real(kind=vdp), dimension(3), intent(in) :: x
  real(kind=vdp), dimension(3), intent(inout) :: ratio
  integer, dimension(3), intent(inout) :: coordpos
  integer, optional :: lnum
  real(kind=vdp), dimension(3) :: rx

  if(x(1) < vsmLocalCoord%origin(1) .or. x(1) > vsmLocalCoord%cend(1)) then
    print *,"x origin end ", x(1),vsmLocalCoord%origin(1),vsmLocalCoord%cend(1)
    print *,"x lorigin end ", x(1),vsmLocalCoord%lorigin(1),vsmLocalCoord%lcend(1)
    if(present(lnum)) print *,"loop=", lnum
    call vsmError('out of local data x')
  end if

  if(x(2) < vsmLocalCoord%origin(2) .or. x(2) > vsmLocalCoord%cend(2)) then
    print *,"y origin end ", x(2),vsmLocalCoord%origin(2),vsmLocalCoord%cend(2)
    print *,"y lorigin end ", x(2),vsmLocalCoord%lorigin(2),vsmLocalCoord%lcend(2)
    if(present(lnum)) print *,"loop=", lnum
    call vsmError('out of local data y')
  end if

  if(x(3) < vsmLocalCoord%origin(3) .or. x(3) > vsmLocalCoord%cend(3)) then
    print *,"z origin end ", x(3),vsmLocalCoord%origin(3),vsmLocalCoord%cend(3)
    print *,"z lorigin end ", x(3),vsmLocalCoord%lorigin(3),vsmLocalCoord%lcend(3)
    if(present(lnum)) print *,"loop=", lnum
    call vsmError('out of local data z')
  end if

  rx(1:3) = (x(1:3) - vsmLocalCoord%origin(1:3))/vsmLocalCoord%dx(1:3)
  coordpos(1:3) = aint(rx(1:3)) + 1

  if(.not.vsmCheckLCoord(coordpos(1),coordpos(2),coordpos(3))) then 
      call vsmLogPrint("local coord:", x)
      print *,x,vsmLocalCoord%origin(1:3)
      call vsmLogPrint("local coord int:", coordpos)
      print *,coordpos
      call vsmError("Error :: local coord is strange :: vsmFindUnicoord")
  end if

  if(coordpos(1) == vsmLocalCoord%nx(1)) then
    coordpos(1) = vsmLocalCoord%nx(1)-1
    ratio(1) = 1.0_vdp
  else
    ratio(1) = rx(1) - aint(rx(1))
  end if

  if(coordpos(2) == vsmLocalCoord%nx(2)) then
    coordpos(2) = vsmLocalCoord%nx(2)-1
    ratio(2) = 1.0_vdp
  else
    ratio(2) = rx(2) - aint(rx(2))
  end if

  if(coordpos(3) == vsmLocalCoord%nx(3)) then
    coordpos(3) = vsmLocalCoord%nx(3)-1
    ratio(3) = 1.0_vdp
  else
    ratio(3) = rx(3) - aint(rx(3))
  end if

!  if(coordpos(1) == vsmLocalCoord%nx(1) .or. coordpos(2) == vsmLocalCoord%nx(2) &
!  	.or. coordpos(3) == vsmLocalCoord%nx(3)) then
!    call vsmLogPrint("x:",x(1:3))
!    call vsmLogPrint("ratio:",ratio(1:3))
!    call vsmLogPrint("local coord int:", coordpos)
!    call vsmError("Error :: local coord is strange :: vsmFindUnicoord 2")
!  end if
 end subroutine vsmFindUnicoord

 subroutine vsmFindReccoord(x, ratio, coordpos)
  !binary search
  real(kind=vdp), dimension(3), intent(in) :: x
  real(kind=vdp), dimension(3), intent(inout) :: ratio
  integer, dimension(3), intent(inout) :: coordpos
  integer :: lower, upper, middle
  logical :: found
  
  found = .false.
  lower = vsmLocalCoord%lx0(1)
  upper = vsmLocalCoord%lx0(1)+vsmLocalCoord%lnx(1)-1
  if(x(1) < vsmLocalCoord%coordx(1) .or. x(1) > vsmLocalCoord%coordx(upper)) then
    call vsmError('out of local data x')
  end if
  found = .false.

  if(x(1)==vsmLocalCoord%coordx(upper)) then
    coordpos(1) = vsmLocalCoord%lx0(1)+vsmLocalCoord%lnx(1)-2
    found = .true.
    ratio(1) = 1.0_vdp
  else
    do
      if (upper- lower < 0) exit
      middle = (lower+upper)/2
      if(x(1) < vsmLocalCoord%coordx(middle)) then
        upper = middle
      else if(x(1) >= vsmLocalCoord%coordx(middle+1)) then
        lower = middle
      else
        found = .true.
        coordpos(1) = middle
        exit
      end if
    end do
    ratio(1) = (x(1) - vsmLocalCoord%coordx(coordpos(1)))/(vsmLocalCoord%coordx(coordpos(1)+1) &
      - vsmLocalCoord%coordx(coordpos(1)))
  end if

  if(.not.found) call vsmError('cant find coord x')
 
  lower = vsmLocalCoord%lx0(2)
  upper = vsmLocalCoord%lx0(2)+vsmLocalCoord%lnx(2)-1
  if(x(2) < vsmLocalCoord%coordy(1) .or. x(2) > vsmLocalCoord%coordy(upper)) then
    call vsmError('out of local data y')
  end if
  found = .false.

  if(x(1)==vsmLocalCoord%coordx(upper)) then
    coordpos(2) = vsmLocalCoord%lx0(2)+vsmLocalCoord%lnx(2)-2
    found = .true.
    ratio(2) = 1.0_vdp
  else
    do
      if (upper- lower < 0) exit
      middle = (lower+upper)/2
      if(x(2) < vsmLocalCoord%coordy(middle)) then
        upper = middle
      else if(x(2) >= vsmLocalCoord%coordy(middle+1)) then
        lower = middle
      else
        found = .true.
        coordpos(2) = middle
        exit
      end if
    end do
    ratio(2) = (x(2) - vsmLocalCoord%coordy(coordpos(2)))/(vsmLocalCoord%coordy(coordpos(2)+1) &
      - vsmLocalCoord%coordy(coordpos(2)))
  end if

  if(.not.found) call vsmError('cant find coord y')

  lower = vsmLocalCoord%lx0(3)
  upper = vsmLocalCoord%lx0(3)+vsmLocalCoord%lnx(3)-1
  if(x(3) < vsmLocalCoord%coordz(1) .or. x(3) > vsmLocalCoord%coordz(upper)) then
    call vsmError('out of local data z')
  end if
  found = .false.
  if(x(1)==vsmLocalCoord%coordx(upper)) then
    coordpos(3) = vsmLocalCoord%lx0(3)+vsmLocalCoord%lnx(3)-2
    found = .true.
    ratio(3) = 1.0_vdp
  else
    do
      if (upper- lower < 0) exit
      middle = (lower+upper)/2
      if(x(3) < vsmLocalCoord%coordz(middle)) then
        upper = middle
      else if(x(3) >= vsmLocalCoord%coordz(middle+1)) then
        lower = middle
      else
        found = .true.
        coordpos(3) = middle
        exit
      end if
    end do
    ratio(3) = (x(3) - vsmLocalCoord%coordz(coordpos(3)))/(vsmLocalCoord%coordz(coordpos(3)+1) &
       - vsmLocalCoord%coordz(coordpos(3)))
  end if
  if(.not.found) call vsmError('cant find coord z')

  if(.not.vsmCheckLCoord(coordpos(1),coordpos(2),coordpos(3))) then
    call vsmError("Error :: localcoord is strange:vsmFindReccoord")
  end if

 end subroutine vsmFindReccoord

function vsmyyRayDistSphere(pos, nvec)
  real(kind=vdp), dimension(3), intent(in) :: pos, nvec
  real(kind=vdp) :: dist, vsmyyRayDistSphere
  real(kind=vdp) :: det, xdn, xdx, ndn
  real(kind=vdp) :: sqrad, sqminrad, sqmaxrad

  sqrad = pos(1) * pos(1) + pos(2) * pos(2) + pos(3) * pos(3)
  sqminrad = vsmCoordData%MinRad * vsmCoordData%MinRad
  sqmaxrad = vsmCoordData%MaxRad * vsmCoordData%MaxRad

  if(sqrad <= sqmaxrad .and. sqrad >= sqminrad) then
     ! inside the data
     dist = 0.0_vdp
   else if(sqrad < sqminrad) then
  ! inside inner sphere
     xdn = (pos(1) * nvec(1) + pos(2) * nvec(2) + pos(3) * nvec(3))
     ndn = (nvec(1) * nvec(1) + nvec(2) * nvec(2) + nvec(3) * nvec(3))
     xdx = sqrad

     det = xdn * xdn - ndn * (xdx - sqminrad)

     if(det < 0.0_vdp) then
       call vsmError("pos is in the inner sphere but no cross point")
     end if
     dist = (-xdn + sqrt(det)) / ndn
     if(dist < 0.0_vdp) call vsmError("a is negative!!")

  else  !outside the data

    xdn = (pos(1) * nvec(1) + pos(2) * nvec(2) + pos(3) * nvec(3))
    ndn = (nvec(1) * nvec(1) + nvec(2) * nvec(2) + nvec(3) * nvec(3))
    xdx = sqrad

    det = xdn * xdn - ndn * (xdx - sqmaxrad)

    if(det < 0.0_vdp) then
       dist =vsmFarAway
    else
       dist = (-xdn - sqrt(det)) / ndn

        if(dist <= 0.0_vdp) then
           dist = vsmFarAway
       end if
    end if
  end if
  vsmyyRayDistSphere = dist
end function vsmyyRayDistSphere

subroutine vsmyyRayDistSphere4Rslice(pos, nvec, rad, dist)
  real(kind=vdp), dimension(3), intent(in) :: pos, nvec
  real(kind=vdp), intent(in) :: rad
  real(kind=vdp), dimension(2), intent(out) :: dist
  real(kind=vdp) :: det, xdn, xdx, ndn
  real(kind=vdp) :: sqrad, rad2

  sqrad = pos(1) * pos(1) + pos(2) * pos(2) + pos(3) * pos(3)
  rad2 = rad*rad

  xdn = (pos(1) * nvec(1) + pos(2) * nvec(2) + pos(3) * nvec(3))
  ndn = (nvec(1) * nvec(1) + nvec(2) * nvec(2) + nvec(3) * nvec(3))
  xdx = sqrad

  det = xdn * xdn - ndn * (xdx - rad2)

  if(det < 0.0_vdp) then
     dist = -1.0_vdp
  else
     dist(1) = (-xdn - sqrt(det)) / ndn
     dist(2) = (-xdn + sqrt(det)) / ndn
  end if

end subroutine vsmyyRayDistSphere4Rslice

 subroutine vsmToYangCoord(x, yanx)
   real(kind=vdp), dimension(3), intent(in) :: x
   real(kind=vdp), dimension(3), intent(inout) :: yanx

   yanx(1) = -x(1)
   yanx(2) = x(3)
   yanx(3) = x(2)
 end subroutine vsmToYangCoord

 subroutine vsmCoordsToAngles(x, spos)
   real(kind=vdp), dimension(3), intent(in) :: x
   real(kind=vdp), dimension(3), intent(inout) :: spos

   !-PI < phi < PI
!   real(kind=vdp) :: x_yy, y_yy, z_yy
   real(kind=vdp) :: r_sin_tht_sq, r_sin_tht, ratio!, r

!    x_yy = x(1)
!    y_yy = x(2)
!    z_yy = x(3)

!    r_sin_tht_sq = x_yy * x_yy + y_yy * y_yy
    r_sin_tht_sq = x(1) * x(1) + x(2) * x(2)

    spos(1) = sqrt(r_sin_tht_sq + x(3) * x(3));
    r_sin_tht = sqrt(r_sin_tht_sq)

    if(spos(1) == 0.0_vdp) then
     ! almost on the origin (0,0,0)
       spos(2) = pi / 2.0_vdp
       spos(3) = 0.0_vdp
    else if(r_sin_tht == 0.0_vdp) then
      if(x(3) >= 0.0_vdp) then!      if(z_yy >= 0.0_vdp) then
         spos(2) = 0.0_vdp
         spos(3) = 0.0_vdp
      else 
         spos(2) = pi
         spos(3) = 0.0_vdp
      end if
   else
      ratio = x(3) / spos(1)!ratio = z_yy / spos(1)
      if(ratio >= 1.0_vdp) then
        spos(2) = 0.0_vdp
      else if(ratio <= -1.0_vdp) then
        spos(2) = pi
      else
        spos(2) = acos(ratio)
      end if
      if(x(1)==0.0_vdp) then
         if(x(2) >= 0.0) then
           spos(3) = pi/2.0_vdp
         else
           spos(3) = -pi/2.0_vdp
         end if
      else
        spos(3) = atan2(x(2), x(1))!spos(3) = atan2(y_yy, x_yy)
      end if
   end if
!   print *, "c2a", x,spos
 end  subroutine vsmCoordsToAngles

 subroutine vsmAnglesToCoords(spos, x)
   real(kind=vdp), dimension(3), intent(out) :: x
   real(kind=vdp), dimension(3), intent(in) :: spos

!   YinYang = vsmCoordData%YinYang

   x(1) = spos(1)* sin(spos(2)) * cos(spos(3))
   x(2) = spos(1)* sin(spos(2)) * sin(spos(3))
   x(3) = spos(1)* cos(spos(2))

 end  subroutine vsmAnglesToCoords

 function vsmIsInside(pos, origin, cend)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), dimension(3), intent(in) :: origin, cend
   logical :: vsmIsInside

   if(pos(1) >= origin(1) .and. pos(1) < cend(1) .and. &
      pos(2) >= origin(2) .and. pos(2) < cend(2) .and. &
      pos(3) >= origin(3) .and. pos(3) < cend(3)) then
     vsmIsInside = .true.
   else
     vsmIsInside = .false.
   end if
 end function vsmIsInside

 function vsmIsInsided(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   logical :: vsmIsInsided

   if(pos(1) >= vsmCoord%origin(1) .and. pos(1) < vsmCoord%cend(1) .and. &
      pos(2) >= vsmCoord%origin(2) .and. pos(2) < vsmCoord%cend(2) .and. &
      pos(3) >= vsmCoord%origin(3) .and. pos(3) < vsmCoord%cend(3)) then
     vsmIsInsided = .true.
   else
     vsmIsInsided = .false.
   end if
 end function vsmIsInsided

 function vsmIsInsideSphere(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp) :: radius2
   logical :: vsmIsInsidesphere

   radius2 = pos(1)*pos(1) + pos(2)*pos(2) + pos(3)*pos(3)!dot_product(pos, pos)
   if(radius2 < vsmCoordData%MaxRad*vsmCoordData%MaxRad) then
     vsmIsInsidesphere = .true.
   else
     vsmIsInsidesphere = .false.
   end if
 end function vsmIsInsideSphere

 function vsmIsInsideInnerSphere4Z0dr(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp) :: radius2, MinR
   logical :: vsmIsInsideinnersphere4z0dr

   radius2 = pos(1)*pos(1) + pos(2)*pos(2) + pos(3)*pos(3)!dot_product(pos, pos)
   MinR = vsmCoordData%MinRad ! for safty

   if(radius2 < MinR*MinR) then
     vsmIsInsideInnersphere4z0dr = .true.
   else
     vsmIsInsideinnersphere4z0dr = .false.
   end if

 end function vsmIsInsideInnerSphere4z0dr

 function vsmIsInsideInnerSphere4Z1dr(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp) :: radius2, MinR
   logical :: vsmIsInsideinnersphere4z1dr

   radius2 = pos(1)*pos(1) + pos(2)*pos(2) + pos(3)*pos(3)!dot_product(pos, pos)
   MinR = vsmCoordData%MinRad+vsmCoordData%dr

   if(radius2 < MinR*MinR) then
     vsmIsInsideInnersphere4z1dr = .true.
   else
     vsmIsInsideinnersphere4z1dr = .false.
   end if

 end function vsmIsInsideInnerSphere4z1dr

 function vsmIsInsideInnerSphere4Z2dr(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp) :: radius2, MinR
   logical :: vsmIsInsideinnersphere4z2dr

   radius2 = pos(1)*pos(1) + pos(2)*pos(2) + pos(3)*pos(3)!dot_product(pos, pos)
   MinR = vsmCoordData%MinRad + dble(2)*vsmCoordData%dr 

   if(radius2 < MinR*MinR) then
     vsmIsInsideInnersphere4z2dr = .true.
   else
     vsmIsInsideinnersphere4z2dr = .false.
   end if

 end function vsmIsInsideInnerSphere4z2dr

 function vsmIsInsideInnerSphere(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp) :: radius2
   logical :: vsmIsInsideinnersphere

   radius2 = pos(1)*pos(1) + pos(2)*pos(2) + pos(3)*pos(3)!dot_product(pos, pos)
   if(radius2 < vsmCoordData%MinRad*vsmCoordData%MinRad) then
     vsmIsInsideInnersphere = .true.
   else
     vsmIsInsideinnersphere = .false.
   end if
 end function vsmIsInsideInnerSphere

 function vsmIsBetweenSpheres(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp) :: radius2
   logical :: vsmIsBetweenspheres

   radius2 = pos(1)*pos(1) + pos(2)*pos(2) + pos(3)*pos(3)!dot_product(pos, pos)
   if(radius2 <= vsmCoordData%MaxRad*vsmCoordData%MaxRad &
        .and. radius2 >= vsmCoordData%MinRad*vsmCoordData%MinRad) then
     vsmIsBetweenspheres = .true.
   else
     vsmIsBetweenspheres = .false.
   end if
 end function vsmIsBetweenSpheres

 function vsmIsInsidel(pos)
   real(kind=vdp), dimension(3), intent(in) :: pos
   logical :: vsmIsInsidel

   if(pos(1) >= vsmLocalCoord%lorigin(1) .and. pos(1) < vsmLocalCoord%lcend(1) .and. &
      pos(2) >= vsmLocalCoord%lorigin(2) .and. pos(2) < vsmLocalCoord%lcend(2) .and. &
      pos(3) >= vsmLocalCoord%lorigin(3) .and. pos(3) < vsmLocalCoord%lcend(3)) then
     vsmIsInsidel = .true.
!     print *,"isinsidel",pos,vsmLocalCoord%lorigin(1),vsmLocalCoord%cend(1)
   else
     vsmIsInsidel = .false.
   end if
 end function vsmIsInsidel

 function vsmIsInsidelWO(pos) ! With Overlap
   real(kind=vdp), dimension(3), intent(in) :: pos
   logical :: vsmIsInsidelWO

   if(pos(1) >= vsmLocalCoord%origin(1) .and. pos(1) < vsmLocalCoord%cend(1) .and. &
      pos(2) >= vsmLocalCoord%origin(2) .and. pos(2) < vsmLocalCoord%cend(2) .and. &
      pos(3) >= vsmLocalCoord%origin(3) .and. pos(3) < vsmLocalCoord%cend(3)) then
     vsmIsInsidelWO = .true.
   else
     vsmIsInsidelWO = .false.
   end if
 end function vsmIsInsidelWO

 function vsmIsInsidelWOE(pos) ! With Overlap and Equal
   real(kind=vdp), dimension(3), intent(in) :: pos
   logical :: vsmIsInsidelWOE

   if(pos(1) >= vsmLocalCoord%origin(1) .and. pos(1) <= vsmLocalCoord%cend(1) .and. &
      pos(2) >= vsmLocalCoord%origin(2) .and. pos(2) <= vsmLocalCoord%cend(2) .and. &
      pos(3) >= vsmLocalCoord%origin(3) .and. pos(3) <= vsmLocalCoord%cend(3)) then
     vsmIsInsidelWOE = .true.
   else
     vsmIsInsidelWOE = .false.
   end if
 end function vsmIsInsidelWOE

 subroutine vsmFinalizeCoord
  integer :: allocstate
  if(vsmCoord%uniform) return
  deallocate(vsmCoord%coordx, vsmCoord%coordy, vsmCoord%coordz, stat=allocstate)
  deallocate(vsmLocalCoord%coordx, vsmLocalCoord%coordy, vsmLocalCoord%coordz, stat=allocstate)
 end subroutine vsmFinalizeCoord

 function vsmCheckLCoord(i, j, k)
   integer, intent(in) :: i, j, k
   logical :: vsmCheckLCoord

   if(i<1 .or. i>vsmLocalCoord%nx(1)) then
     vsmCheckLCoord = .false.
     call vsmLogPrint("local x:", i)
     return
   end if

   if(j<1 .or. j>vsmLocalCoord%nx(2)) then
     vsmCheckLCoord = .false.
     call vsmLogPrint("local y:", j)
     return
   end if

   if(k<1 .or. k>vsmLocalCoord%nx(3)) then
     vsmCheckLCoord = .false.
     call vsmLogPrint("local z:", k)
     return
   end if

   vsmCheckLCoord = .true.
 end function vsmCheckLCoord

 function vsmIsXz(posx, posz, origin, cend)
   real(kind=vdp), intent(in) :: posx, posz
   real(kind=vdp), dimension(3), intent(in) :: origin, cend
   logical :: vsmIsXz

   if(posx >= origin(1) .and. posx <= cend(1) .and. &
      posz >= origin(3) .and. posz <= cend(3)) then
     vsmIsXz = .true.
   else
     vsmIsXz = .false.
   end if
 end function vsmIsXz

 function vsmIsXy(posx, posy, origin, cend)
   real(kind=vdp), intent(in) :: posx, posy
   real(kind=vdp), dimension(3), intent(in) :: origin, cend
   logical :: vsmIsXy

   if(posx >= origin(1) .and. posx <= cend(1) .and. &
      posy >= origin(2) .and. posy <= cend(2)) then
     vsmIsXy = .true.
   else
     vsmIsXy = .false.
   end if
 end function vsmIsXy

 function vsmIsYz(posy, posz, origin, cend)
   real(kind=vdp), intent(in) :: posy, posz
   real(kind=vdp), dimension(3), intent(in) :: origin, cend
   logical :: vsmIsYz

   if(posy >= origin(2) .and. posy <= cend(2) .and. &
      posz >= origin(3) .and. posz <= cend(3)) then
     vsmIsYz = .true.
   else
     vsmIsYz = .false.
   end if
 end function vsmIsYz

subroutine vsmDisR(pos, nvec, rad1, rad2, cp)
  real(kind=vdp), dimension(3), intent(in) :: pos, nvec
  real(kind=vdp), intent(in) :: rad1, rad2
  real(kind=vdp), dimension(4), intent(out) :: cp
  real(kind=vdp) :: det, xdn, xdx, ndn
  real(kind=vdp) :: sqrad, sqminrad, sqmaxrad

  sqrad = pos(1) * pos(1) + pos(2) * pos(2) + pos(3) * pos(3)
  sqminrad = rad1*rad1
  sqmaxrad = rad2*rad2
   
  xdn = pos(1) * nvec(1) + pos(2) * nvec(2) + pos(3) * nvec(3)
  ndn = nvec(1) * nvec(1) + nvec(2) * nvec(2) + nvec(3) * nvec(3)
  xdx = sqrad

  !inner sphere
  det = xdn * xdn - ndn * (xdx - sqminrad)
  if(det < 0.0_vdp) then
    cp(1:2) = vsmFaraway
  else
    cp(1) = (-xdn + sqrt(det)) / ndn
    if(cp(1) <= 0.0_vdp) then
      cp(1) = vsmFarAway
    end if

    cp(2) = (-xdn - sqrt(det)) / ndn
    if(cp(2) <= 0.0_vdp) then
      cp(2) = vsmFarAway
    end if
  end if

  !outside the data
  det = xdn * xdn - ndn * (xdx - sqmaxrad)
  if(det < 0.0_vdp) then
    cp(3:4) =vsmFarAway
  else
    cp(3) = (-xdn - sqrt(det)) / ndn
    if(cp(3) <= 0.0_vdp) then
      cp(3) = vsmFarAway
    end if

    cp(4) = (-xdn + sqrt(det)) / ndn
    if(cp(4) <= 0.0_vdp) then
      cp(4) = vsmFarAway
    end if
  end if
end subroutine vsmDisR

subroutine vsmDisT(pos, nvec, theta1, theta2, cp)
  real(kind=vdp), dimension(3), intent(in) :: pos, nvec
  real(kind=vdp), intent(in) :: theta1, theta2
  real(kind=vdp), dimension(4), intent(out) :: cp
  real(kind=vdp) :: det, tan2
  real(kind=vdp) :: A, B, C!, dtan
  real(kind=vdp) :: z

  !dtan = pi/2.0_vdp - (pi/2.0_vdp)*0.9999999999999_vdp
  !if(theta1 >= pi/2.0_vdp - dtan .and. theta1 <= pi/2.0_vdp + dtan) then
  if(theta1 == pi/2.0_vdp) then
    if(nvec(3) == 0.0_vdp) then
      cp(1:2) = vsmFaraway
    else
      cp(1:2) = -pos(3)/nvec(3)
    end if
    if(cp(1) < 0.0_vdp) cp(1:2) = vsmFaraway
  else
    tan2 = tan(theta1)*tan(theta1)
    A = nvec(1)*nvec(1) + nvec(2)*nvec(2) - nvec(3)*nvec(3)*tan2
    B = pos(1)* nvec(1) + pos(2)* nvec(2) - pos(3)* nvec(3)*tan2
    C = pos(1)* pos(1) + pos(2)* pos(2) - pos(3)* pos(3)*tan2

    det = B*B - A*C

    if(det < 0.0_vdp) then
      cp(1) = vsmFaraway
      cp(2) = vsmFaraway
    else
      cp(1) = (-B + sqrt(det)) / A
      cp(2) = (-B - sqrt(det)) / A

      if(cp(1) <= 0.0_vdp) then
        cp(1) = vsmFarAway
      else
        z = pos(3) + cp(1)*nvec(3)
        if((z > 0.0_vdp .and. theta1 > pi/2.0_vdp) .or. (z < 0.0_vdp .and. theta1 < pi/2.0_vdp)) then
          cp(1) = vsmFarAway
        end if
      end if

      if(cp(2) <= 0.0_vdp) then
        cp(2) = vsmFarAway
      else
        z = pos(3) + cp(2)*nvec(3)
        if((z > 0.0_vdp .and. theta1 > pi/2.0_vdp) .or. (z < 0.0_vdp .and. theta1 < pi/2.0_vdp)) then
          cp(2) = vsmFarAway
        end if
      end if
    end if
  end if

  if(theta2 == pi/2.0_vdp) then
    if(nvec(3) == 0.0_vdp) then
      cp(3:4) = vsmFaraway
    else
      cp(3:4) = -pos(3)/nvec(3)
    end if
    if(cp(3) < 0.0_vdp) cp(3:4) = vsmFaraway
  else
    tan2 = tan(theta2)*tan(theta2)
    A = nvec(1)*nvec(1) + nvec(2)*nvec(2) - nvec(3)*nvec(3)*tan2
    B = pos(1) * nvec(1) + pos(2) * nvec(2) - pos(3) * nvec(3)*tan2
    C = pos(1) * pos(1) + pos(2) * pos(2) - pos(3) * pos(3)*tan2

    det = B*B - A*C

    if(det < 0.0_vdp) then
      cp(3:4) = vsmFaraway
    else
      cp(3) = (-B + sqrt(det)) / A
      cp(4) = (-B - sqrt(det)) / A
      if(cp(3) <= 0.0_vdp) then
        cp(3) = vsmFarAway
      else
        z = pos(3) + cp(3)*nvec(3)
        if((z > 0.0_vdp .and. theta2 > pi/2.0_vdp) .or. (z < 0.0_vdp .and. theta2 < pi/2.0_vdp)) then
          cp(3) = vsmFarAway
        end if
      end if

      if(cp(4) <= 0.0_vdp) then
        cp(4) = vsmFarAway
      else
        z = pos(3) + cp(4)*nvec(3)
        if((z > 0.0_vdp .and. theta2 > pi/2.0_vdp) .or. (z < 0.0_vdp .and. theta2 < pi/2.0_vdp)) then
          cp(4) = vsmFarAway
        end if
      end if

    end if
  end if
end subroutine vsmDisT

subroutine vsmDisP(pos, nvec, phi1, phi2, cp)
  real(kind=vdp), dimension(3), intent(in) :: pos, nvec
  real(kind=vdp), intent(in) :: phi1, phi2
  real(kind=vdp), dimension(2), intent(out) :: cp
  real(kind=vdp) :: det!, tan2, cp1, cp2
!  real(kind=vdp) :: A, B, C, dtan
  real(kind=vdp), dimension(2) :: x

  det = nvec(2)*cos(phi1) - nvec(1)*sin(phi1)
  if(det == 0.0_vdp) then
    cp(1) = vsmFaraway
  else
    cp(1) = -(pos(2)*cos(phi1) - pos(1)*sin(phi1))/det
    if(cp(1) < 0.0_vdp) then
      cp(1) = vsmFaraway
    else
      x(1:2) = pos(1:2) + cp(1)*nvec(1:2)
      if(phi1 >= pi/2.0_vdp) then
        if(x(1) > 0.0_vdp .or. x(2) < 0.0_vdp) then
          cp(1) = vsmFaraway
        end if
      else if(phi1 >= 0.0_vdp)then
        if(x(1) <= 0.0_vdp .or. x(2) < 0.0_vdp) then
          cp(1) = vsmFaraway
        end if
      else if(phi1 <= -pi/2.0_vdp)then
        if(x(1) > 0.0_vdp .or. x(2) >= 0.0_vdp) then
          cp(1) = vsmFaraway
        end if
      else
        if(x(1) <= 0.0_vdp .or. x(2) >=- 0.0_vdp) then
          cp(1) = vsmFaraway
        end if
      end if
    end if
  end if

  det = nvec(2)*cos(phi2) - nvec(1)*sin(phi2)
  if(det == 0.0_vdp) then
    cp(2) = vsmFaraway
  else
    cp(2) = -(pos(2)*cos(phi2) - pos(1)*sin(phi2))/det
    if(cp(2) < 0.0_vdp) then
      cp(2) = vsmFaraway
    else
      x(1:2) = pos(1:2) + cp(2)*nvec(1:2)
      if(phi2 >= pi/2.0_vdp) then
        if(x(1) > 0.0_vdp .or. x(2) < 0.0_vdp) then
          cp(2) = vsmFaraway
        end if
      else if(phi2 >= 0.0_vdp)then
        if(x(1) <= 0.0_vdp .or. x(2) < 0.0_vdp) then
          cp(2) = vsmFaraway
        end if
      else if(phi2 <= -pi/2.0_vdp)then
        if(x(1) > 0.0_vdp .or. x(2) >= 0.0_vdp) then
          cp(2) = vsmFaraway
        end if
      else
        if(x(1) <= 0.0_vdp .or. x(2) >=- 0.0_vdp) then
          cp(2) = vsmFaraway
        end if
      end if
     end if
  end if
end subroutine vsmDisP

 function vsmyyRayDist(opos, norm, origin, cend, flag)
   real(kind=vdp), dimension(3), intent(in) :: opos, norm, origin, cend
   logical, intent(in), optional :: flag
   real(kind=vdp), dimension(3) :: pos, nvec, tpos, rtp
   real(kind=vdp), dimension(10) :: ray
   real(kind=vdp) :: vsmyyRayDist
   integer :: i

   pos(1) = opos(1)
   pos(2) = opos(2)
   pos(3) = opos(3)

   nvec(1) = norm(1)
   nvec(2) = norm(2)
   nvec(3) = norm(3)

   call vsmCoordsToAngles(pos, rtp)
   if(present(flag)) then
      print *,"yyray pos",pos
      print *,"yyray vec",nvec
      print *,"yyray rtp",rtp(1),rtp(2)*180.0_vdp/pi,rtp(3)*180.0_vdp/pi
   end if
   if(vsmIsInside(rtp, origin, cend)) then
      vsmyyRayDist = 0.0_vdp
!      print *,"",rtp, origin,cend
   else
      ray = vsmFaraway
      call vsmDisR(pos, nvec, origin(1), cend(1), ray(1:4))
      do i=1, 4
        if(ray(i) < vsmFarAwayd10) then
           tpos(1:3) = ray(i)*nvec(1:3) + pos(1:3)
           call vsmCoordsToAngles(tpos, rtp)
           if(present(flag)) print *,"yyraydist pos r",tpos,ray(i)
           if(present(flag)) print *,"yyraydist rtp r",rtp(1),rtp(2)*180.0_vdp/pi,rtp(3)*180.0_vdp/pi,ray(i)

           if(.not.vsmIsYz(rtp(2), rtp(3), origin, cend)) ray(i) = vsmFaraway
        end if
      end do

      call vsmDisT(pos, nvec, origin(2), cend(2), ray(5:8))
      do i=5, 8
        if(ray(i) < vsmFarAwayd10) then
           tpos(1:3) = ray(i)*nvec(1:3) + pos(1:3)
           call vsmCoordsToAngles(tpos, rtp)
           if(present(flag)) print *,"yyraydist pos t",tpos,ray(i)
           if(present(flag)) print *,"yyraydist rtp t",rtp(1),rtp(2)*180.0_vdp/pi,rtp(3)*180.0_vdp/pi,ray(i)
           if(.not.vsmIsXz(rtp(1), rtp(3), origin, cend)) ray(i) = vsmFaraway
        end if
      end do

      call vsmDisP(pos, nvec, origin(3), cend(3), ray(9:10))
      do i=9, 10
        if(ray(i) < vsmFarAwayd10) then
           tpos(1:3) = ray(i)*nvec(1:3) + pos(1:3)
           call vsmCoordsToAngles(tpos, rtp)
           if(present(flag)) print *,"yyraydist pos p",tpos,ray(i)
           if(present(flag)) print *,"yyraydist rtp p",rtp(1),rtp(2)*180.0_vdp/pi,rtp(3)*180.0_vdp/pi,ray(i)
           if(.not.vsmIsXy(rtp(1), rtp(2), origin, cend)) ray(i) = vsmFaraway
        end if
      end do

      where(ray < 0.0_vdp)
        ray = vsmFaraway
      end where

      vsmyyRayDist = minval(ray)
      if(vsmyyRayDist < 0.0_vdp) call vsmLogPrint("ray_dist is negative!")
   end if

 end function vsmyyRayDist
 
  function vsmyyRayDistInside(pos, spos, norm, origin, cend)
   real(kind=vdp), dimension(3), intent(in) :: pos, spos, norm, origin, cend
   real(kind=vdp), dimension(3) :: tpos, rtp!, mincp
   real(kind=vdp), dimension(10) :: ray
   real(kind=vdp) :: vsmyyRayDistInside
   integer :: i

   if(.not. vsmIsInside(spos, origin, cend)) then
      vsmyyRayDistInside = 0.0_vdp
      return
   end if

   call vsmDisR(pos, norm, origin(1), cend(1), ray(1:4))
   !print *,"rad", ray(1:4), origin, cend
    do i=1, 4
      if(ray(i) < vsmFarAwayd10) then
        tpos(1:3) = ray(i)*norm(1:3) + pos(1:3)
        call vsmCoordsToAngles(tpos, rtp)
        if(.not.vsmIsYz(rtp(2), rtp(3), origin, cend)) ray(i) = vsmFaraway
      end if
    end do

!   mincp(1) = minval(ray(1:4))

    call vsmDisT(pos, norm, origin(2), cend(2), ray(5:8))
    do i=5, 8
      if(ray(i) < vsmFarAwayd10) then
        tpos(1:3) = ray(i)*norm(1:3) + pos(1:3)
        call vsmCoordsToAngles(tpos, rtp)
        if(.not.vsmIsXz(rtp(1), rtp(3), origin, cend)) ray(i) = vsmFaraway
      end if
    end do

!   mincp(2) = minval(ray(5:8))

    call vsmDisP(pos, norm, origin(3), cend(3), ray(9:10))
    do i=9, 10
      if(ray(i) < vsmFarAwayd10) then
        tpos(1:3) = ray(i)*norm(1:3) + pos(1:3)
        call vsmCoordsToAngles(tpos, rtp)
        if(.not.vsmIsXy(rtp(1), rtp(2), origin, cend)) ray(i) = vsmFaraway
      end if
    end do
!    print *,"",ray

!   mincp(3) = minval(ray(9:10))

    where(ray < 0.0_vdp)
      ray = vsmFaraway
    end where
!   print *,"ray 1:10", ray
    vsmyyRayDistInside = minval(ray)
!    if(mincp(2) < mincp(1) .and. mincp(2) < mincp(3))vsmyyRayDistInside = 0.0_vdp
!    if(mincp(3) < mincp(1) .and. mincp(3) < mincp(2))vsmyyRayDistInside = 0.0_vdp
!    if(mincp(1) < mincp(2) .and. mincp(1) < mincp(3))vsmyyRayDistInside = 0.0_vdp

    if(vsmyyRayDistInside > vsmFarAwayd10) then
       call vsmLogPrint("ray_dist is too big. The Ray tip may be on the borders")
       vsmyyRayDistInside = 0.0_vdp
    end if
    if(vsmyyRayDistInside < 0.0_vdp) call vsmLogPrint("ray_dist is negative!")

!    if(mincp(2) < mincp(1) .and.mincp(2) < mincp(3))vsmyyRayDistInside = 0.0_vdp

 end function vsmyyRayDistInside

 function vsmRayDist(pos, norm, origin, cend)
   real(kind=vdp), dimension(3), intent(in) :: pos, norm, origin, cend
   real(kind=vdp), dimension(6) :: ray
   real(kind=vdp) :: vsmRayDist, p1, p2

   if(vsmIsInside (pos, origin, cend)) then
      vsmRayDist = 0.0_vdp
   else
      if(abs(norm(1)) /= 0.0_vdp) then
        ray(1) = (-pos(1)+origin(1))/norm(1)
        ray(2) = (-pos(1)+cend(1))/norm(1)
        p1 = pos(2)+ray(1)*norm(2); p2 = pos(3)+ray(1)*norm(3)
        if(.not.vsmIsYz(p1, p2, origin, cend)) ray(1) = vsmFaraway
        p1 = pos(2)+ray(2)*norm(2); p2 = pos(3)+ray(2)*norm(3) 
        if(.not.vsmIsYz(p1, p2, origin, cend)) ray(2) = vsmFaraway
      else
        ray(1:2) = vsmFaraway
      end if

      if(abs(norm(2)) /= 0.0_vdp) then
        ray(3) = (-pos(2)+origin(2))/norm(2)
        ray(4) = (-pos(2)+cend(2))/norm(2)
        p1 = pos(1)+ray(3)*norm(1); p2 = pos(3)+ray(3)*norm(3)
        if(.not.vsmIsXz(p1, p2, origin, cend)) ray(3) = vsmFaraway
        p1 = pos(1)+ray(4)*norm(1); p2 = pos(3)+ray(4)*norm(3)
        if(.not.vsmIsXz(p1, p2, origin, cend)) ray(4) = vsmFaraway
      else
        ray(3:4) = vsmFaraway
      end if

      if(abs(norm(3)) /= 0.0_vdp) then
        ray(5) = (-pos(3)+origin(3))/norm(3)
        ray(6) = (-pos(3)+cend(3))/norm(3)
        p1 = pos(1)+ray(5)*norm(1); p2 = pos(2)+ray(5)*norm(2)
        if(.not.vsmIsXy(p1, p2, origin, cend)) ray(5) = vsmFaraway
        p1 = pos(1)+ray(6)*norm(1) ; p2 = pos(2)+ray(6)*norm(2)
        if(.not.vsmIsXy(p1, p2, origin, cend)) ray(6) = vsmFaraway
      else
        ray(5:6) = vsmFaraway
      end if

      where(ray < 0.0_vdp)
        ray = vsmFaraway
      end where

      vsmRayDist = minval(ray)
      if(vsmRayDist < 0.0_vdp) call vsmLogPrint("ray_dist is negative!")

   end if

 end function vsmRayDist

 function vsmRayDistInside(pos, norm, origin, cend)
   real(kind=vdp), dimension(3), intent(in) :: pos, norm, origin, cend
   real(kind=vdp), dimension(6) :: ray
   real(kind=vdp) :: vsmRayDistInside, p1, p2

   if(.not.vsmIsInside (pos, origin, cend)) then
      call vsmLogPrint("pos ", pos)
      call vsmLogPrint("origin ", origin)
      call vsmLogPrint("cend ", cend)
      call vsmError("pos is not inside data")
   end if

   if(abs(norm(1)) /= 0.0_vdp) then
     ray(1) = (-pos(1)+origin(1))/norm(1)
     ray(2) = (-pos(1)+cend(1))/norm(1)
     p1 = pos(2)+ray(1)*norm(2); p2 = pos(3)+ray(1)*norm(3)
     if(.not.vsmIsYz(p1, p2, origin, cend)) ray(1) = vsmFaraway
     p1 = pos(2)+ray(2)*norm(2); p2 = pos(3)+ray(2)*norm(3) 
     if(.not.vsmIsYz(p1, p2, origin, cend)) ray(2) = vsmFaraway
   else
     ray(1:2) = vsmFaraway
   end if

   if(abs(norm(2)) /= 0.0_vdp) then
     ray(3) = (-pos(2)+origin(2))/norm(2)
     ray(4) = (-pos(2)+cend(2))/norm(2)
     p1 = pos(1)+ray(3)*norm(1); p2 = pos(3)+ray(3)*norm(3)
     if(.not.vsmIsXz(p1, p2, origin, cend)) ray(3) = vsmFaraway
     p1 = pos(1)+ray(4)*norm(1); p2 = pos(3)+ray(4)*norm(3)
     if(.not.vsmIsXz(p1, p2, origin, cend)) ray(4) = vsmFaraway
   else
     ray(3:4) = vsmFaraway
   end if

   if(abs(norm(3)) /= 0.0_vdp) then
     ray(5) = (-pos(3)+origin(3))/norm(3)
     ray(6) = (-pos(3)+cend(3))/norm(3)
     p1 = pos(1)+ray(5)*norm(1); p2 = pos(2)+ray(5)*norm(2)
     if(.not.vsmIsXy(p1, p2, origin, cend)) ray(5) = vsmFaraway
     p1 = pos(1)+ray(6)*norm(1) ; p2 = pos(2)+ray(6)*norm(2)
     if(.not.vsmIsXy(p1, p2, origin, cend)) ray(6) = vsmFaraway
   else
     ray(5:6) = vsmFaraway
   end if

   where(ray < 0.0_vdp)
     ray = vsmFaraway
   end where

   vsmRayDistInside = minval(ray)
   if(vsmRayDistInside < 0.0_vdp) call vsmLogPrint("ray_dist(inside) is negative!")

 end function vsmRayDistInside

end module vsmCoordinates

