module vsmData
 use vsmConst
 use vsmFunction
 use vsmColormap

 implicit none
 private

 type(vismo_scaldata), dimension(vsmMaxScalar), save, public :: vsmScalData
 type(vismo_vectdata), dimension(vsmMaxVector), save,public :: vsmVectData
 type(vismo_ptdata), dimension(vsmMaxPtcl), save,public :: vsmPtcData
 type(vismo_datainfo), save, public :: vsmDataInfo
 
 public vsmAddPtclData, vsmAddScalData, vsmAddVectData, vsmInitData

 contains

 subroutine vsmInitData
  vsmDataInfo%NumOfScal = 0
  vsmDataInfo%NumOfVect = 0
  vsmDataInfo%NumOfPtcl = 0
 end subroutine vsmInitData

 subroutine vsmAddScalData(scalar, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOAT
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: scalar
#else
   real, dimension(nx,ny,nz), target, intent(in) :: scalar
#endif
   integer :: NumOfScal

   NumOfScal = vsmDataInfo%NumOfScal + 1
   if(NumOfScal > vsmMaxScalar) call vsmError('too many scalar data')
   vsmScalData(NumOfScal)%scal => scalar
   vsmDataInfo%NumOfScal = NumOfScal
 end subroutine vsmAddScalData

 subroutine vsmAddVectData(vect_x, vect_y, vect_z, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOAT
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#else
   real, dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#endif
   integer :: NumOfVect

   NumOfVect = vsmDataInfo%NumOfVect + 1
   if(NumOfVect > vsmMaxVector) call vsmError('too many vector data')
   vsmVectData(NumOfVect)%vect_x => vect_x
   vsmVectData(NumOfVect)%vect_y => vect_y
   vsmVectData(NumOfVect)%vect_z => vect_z
   vsmDataInfo%NumOfVect = NumOfVect
 end subroutine vsmAddVectData

 subroutine vsmAddPtclData(p_x, p_y, p_z, q, n)
   integer, intent(in) :: n
#ifndef FLOAT
   real(kind=vdp), dimension(n), target, intent(in) :: p_x, p_y, p_z
#else
   real, dimension(n), target, intent(in) :: p_x, p_y, p_z
#endif
   integer, dimension(n), target, intent(in) :: q
   integer :: NumOfPtcl

   NumOfPtcl = vsmDataInfo%NumOfPtcl + 1
   if(NumOfPtcl > vsmMaxPtcl) call vsmError('too many particle data')
   vsmPtcData(NumOfPtcl)%x => p_x
   vsmPtcData(NumOfPtcl)%y => p_y
   vsmPtcData(NumOfPtcl)%z => p_z
   vsmPtcData(NumOfPtcl)%q => q
   vsmDataInfo%NumOfPtcl = NumOfPtcl
 end subroutine vsmAddPtclData

end module vsmData

