module vsmImage
 use vsmConst
 use vsmFunction
 use vsmCoordinates
 use vsmCameraData
 use vsmParam
!$ use omp_lib

 implicit none
 private

  type(vismo_image), save, public :: vsmImageData

  public vsmGetZBuf, vsmAlphaBlending, vsmBlendRaySegmentf, vsmCalcDP, vsmCheckImage, vsmCheckImageL
  public vsmCheckRaySegment, vsmClearImage4Solid, vsmClearImage, vsmClearPixel, vsmClearRaySegment, vsmClearRaySegmentr
  public vsmClearRImage, vsmCopyRaySegment, vsmFinalizeImage, vsmGetPixelColor4, vsmInitImage, vsmInitScreen
  public vsmMergebuf4Solid, vsmMergeRaySegment, vsmPaintPixel3, vsmPaintPixel3MS, vsmPaintPixelBlend, vsmRaySegment
  public vsmSetImageSize, vsmSolidBlending, vsmWriteImage
  public vsmInitBGColor, vsmSetBGColor, vsmBGBlending, vsmPaintPixel4R

 contains

 !initialize image buffer
 subroutine vsmRaySegment(in, out, color, pixn)
   real(kind=vdp), intent(in) :: in, out
   real, dimension(4), intent(inout) :: color
   integer, dimension(2), intent(in) :: pixn
   integer :: px, pos, intin, intout, i, intein
   real :: fin

   if(color(4) < 1.0/256) then
     color(1:4) = 0.0
     return
   end if

   px = pixn(1)-1 + (pixn(2)-1)*vsmImageData%width
   if(px >= vsmImageData%width*vsmImageData%height) then
     call vsmLogPrint("px", px)
     call vsmError("vsmRaySegment : exceed pixel number")
   end if

   pos = px*vsmRSBNum*vsmRSD
   do i=1, vsmRSBNum
     if(vsmImageData%rsdata(pos+1) < 1)then
       vsmImageData%rsdata(pos+1) = px+1
       exit
     else
       pos = pos + vsmRSD
     end if
   end do

   intin = int(in/vsmIntval%ray_interval/10.0)
   intout = int(out/vsmIntval%ray_interval/10.0)

!   vsmImageData%rsdata(pos + 2) = intout
#ifndef CTRANS
   vsmImageData%rsdata(pos + 2) = transfer(real(in), 1)
#else
  fin = real(in)
  call VSMTRANSFER4F2I(fin, intein)
  vsmImageData%rsdata(pos + 2) = intein
#endif
  vsmImageData%rsdata(pos + 3) = vsmColorToIntdata(color)
!   vsmImageData%rsdata(pos + 4) = transfer(color(4), 1)
  color(1:4) = 0.0

 end subroutine vsmRaySegment

  function vsmColorToIntdata(color)
    real, dimension(4), intent(in) :: color
    integer, dimension(4) :: icolor
    integer :: i
    integer :: vsmColorToIntdata
  
    icolor(1:4) = color(1:4) * 256
    do i=1, 4
      if(icolor(i) > 255) icolor(i) = 255
    end do

    vsmColorToIntdata = icolor(1)
    do i=2, 4
      vsmColorToIntdata=ishft(vsmColorToIntdata, 8)
      vsmColorToIntdata=ior(vsmColorToIntdata, icolor(i))
    end do
  end function vsmColorToIntdata

  function vsmColorToIntdata3(color)
    real, dimension(4), intent(in) :: color
    integer, dimension(4) :: icolor
    integer :: i
    integer :: vsmColorToIntdata3
  
    icolor(1:3) = color(1:3) * 256
    do i=1, 3
      if(icolor(i) > 255) icolor(i) = 255
    end do

    vsmColorToIntdata3 = icolor(1)
    do i=2, 3
      vsmColorToIntdata3=ishft(vsmColorToIntdata3, 8)
      vsmColorToIntdata3=ior(vsmColorToIntdata3, icolor(i))
    end do
  end function vsmColorToIntdata3

  subroutine vsmIntdata2Color(icolor, color)
    integer,intent(in) :: icolor
    real, dimension(4), intent(out) :: color
    integer, parameter :: imax = 255
    integer :: i, icol

    icol = icolor
    color(4) = iand(icol, imax)
    do i=3, 1, -1
      icol = ishft(icol, -8)
      color(i) = iand(icol, imax)
    end do
    
    color(1:4) = color(1:4)/256.0
  end subroutine vsmIntdata2Color

  subroutine vsmIntdata2Color3(icolor, color)
    integer,intent(in) :: icolor
    real, dimension(4), intent(inout) :: color
    integer, parameter :: imax = 255
    integer :: i, icol

    icol = icolor
    color(3) = iand(icol, imax)
    do i=2, 1, -1
      icol = ishft(icol, -8)
      color(i) = iand(icol, imax)
    end do
    
    color(1:3) = color(1:3)/256.0
  end subroutine vsmIntdata2Color3

 subroutine vsmClearRaySegmentR
   integer :: i
!$omp parallel do private(i)
   do i=1, vsmImageData%rimgsize*vsmRSBNum*vsmRSD, vsmRSD
     vsmImageData%rsdata_r(i) = -1
     vsmImageData%rsdata_r(i+1) = int_max
     vsmImageData%rsdata_r(i+2) = 0
  end do
 end subroutine vsmClearRaySegmentR

 subroutine vsmClearRaySegment
   integer :: i
!$omp parallel do private(i)
   do i=1, vsmImageData%rimgsize*vsmRSBNum*vsmRSD, vsmRSD
     vsmImageData%rsdata(i) = -1
     vsmImageData%rsdata(i+1) = int_max
     vsmImageData%rsdata(i+2) = 0
  end do
 end subroutine vsmClearRaySegment

 subroutine vsmCheckRaySegment
   integer :: i

!$omp parallel do private(i)
   do i=1, vsmImageData%rimgsize*vsmRSBNum*vsmRSD, vsmRSD
     if(vsmImageData%rsdata(i) > vsmImageData%rimgsize) then
        print *,"rsdata",vsmImageData%rsdata(i:i+2),vsmImageData%rimgsize,i,&
        vsmImageData%rimgsize*vsmRSBNum*vsmRSD
        call vsmError("check ray segment rs")
     end if

     if(vsmImageData%rsdata_r(i)  > vsmImageData%rimgsize) then
        print *,"rsdata",vsmImageData%rsdata_r(i:i+2),vsmImageData%rimgsize,i,&
        vsmImageData%rimgsize*vsmRSBNum*vsmRSD
        call vsmError("check ray segment rs_r")
     end if
  end do

 end subroutine vsmCheckRaySegment

 subroutine vsmMergeRaySegmentMP(stage, pxnum)
   integer, intent(in) :: stage, pxnum
   integer :: i, k, l, ele!,j
   integer :: p, p1, p2, r1, r2!, p0, err
   real :: dist1, dist2
   integer :: loopsize, loop, res, offset, lsize
   loopsize = pxnum
   loop = loopsize/vsmVecLen
   res = loopsize - loop*vsmVecLen

   ele = vsmRSBNum* (2**(stage-1))

!$omp parallel do private(l,lsize,i,offset,p,p1,p2,r1,r2,dist1,dist2,k)
   do l=1, loop+1
     if(l==loop+1) then
       lsize = res
     else
       lsize = vsmVecLen
     end if
     offset = (l-1)*vsmVecLen

   do i=1+offset, lsize+offset
     p = (i-1)*ele*vsmRSD*2
!     p0 = p
     p1 = (i-1)*ele*vsmRSD
     p2 = p1+pxnum*vsmRSD*ele
     r1 = 0; r2 = 0

     do while(r1 < ele .and. r2 < ele)
       if(vsmImageData%rsdata_r(p1+1) < 1 .or. vsmImageData%rsdata_r(p2+1) < 1) then
         exit
       end if

       if(vsmImageData%rsdata_r(p1+1) /= vsmImageData%rsdata_r(p2+1)) then
         print *,"p1, p2",vsmImageData%rsdata_r(p1+1), vsmImageData%rsdata_r(p2+1), i
         call vsmError("wrong pixel position")
       end if

       if(vsmImageData%rsdata_r(p1+1) > vsmImageData%rimgsize .or. &
           vsmImageData%rsdata_r(p2+1) > vsmImageData%rimgsize) then
         print *,"px1",vsmImageData%rsdata_r(p1+1:p1+3),r1,ele, p1
         print *,"px2",vsmImageData%rsdata_r(p2+1:p2+3),r2,ele, p2
         call vsmError("merge: px is strange")
       end if

#ifndef CTRANS
       dist1 = transfer(vsmImageData%rsdata_r(p1+2),1.0)
       dist2 = transfer(vsmImageData%rsdata_r(p2+2), 1.0)
#else       
       call VSMTRANSFER4I2F(vsmImageData%rsdata_r(p1+2), dist1)
       call VSMTRANSFER4I2F(vsmImageData%rsdata_r(p2+2), dist2)
#endif

       if(dist1 < dist2) then
        vsmImageData%rsdata(p+1:p+vsmRSD) = vsmImageData%rsdata_r(p1+1:p1+vsmRSD)
         r1 = r1 + 1; p1=p1+vsmRSD
       else
         vsmImageData%rsdata(p+1:p+vsmRSD) = vsmImageData%rsdata_r(p2+1:p2+vsmRSD)
         r2 = r2 + 1; p2=p2+vsmRSD
       end if
       p = p + vsmRSD
     end do

     if(vsmImageData%rsdata_r(p1+1) > 0) then
       do k=r1, ele-1
         if(vsmImageData%rsdata_r(p1+1) < 1) then
          exit
         end if
         vsmImageData%rsdata(p+1:p+vsmRSD) = vsmImageData%rsdata_r(p1+1:p1+vsmRSD)
         p1=p1+vsmRSD; p=p+vsmRSD
       end do
     end if

     if(vsmImageData%rsdata_r(p2+1) > 0) then
       do k=r2, ele-1
         if(vsmImageData%rsdata_r(p2+1) < 1) then
          exit
         end if
         vsmImageData%rsdata(p+1:p+vsmRSD) = vsmImageData%rsdata_r(p2+1:p2+vsmRSD)
         p2=p2+vsmRSD; p=p+vsmRSD
       end do
     end if
   end do
   end do
 end subroutine vsmMergeRaySegmentMP

 subroutine vsmMergeRaySegment(stage, pxnum)
   integer, intent(in) :: stage, pxnum
   integer :: i, k, r1, r2, ele
   integer :: p, p1, p2
   real :: dist1, dist2

   ele = vsmRSBNum* (2**(stage-1))
   do i=1, pxnum
     p = (i-1)*ele*vsmRSD*2
!     p0 = p
     p1 = (i-1)*ele*vsmRSD
     p2 = p1+pxnum*vsmRSD*ele
     r1 = 0; r2 = 0

     do while(r1 < ele .and. r2 < ele)
       if(vsmImageData%rsdata_r(p1+1) < 1 .or. vsmImageData%rsdata_r(p2+1) < 1) then
         exit
       end if

       if(vsmImageData%rsdata_r(p1+1) /= vsmImageData%rsdata_r(p2+1)) then
         print *,"p1, p2",vsmImageData%rsdata_r(p1+1), vsmImageData%rsdata_r(p2+1), i
         call vsmError("wrong pixel position")
       end if

       if(vsmImageData%rsdata_r(p1+1) > vsmImageData%rimgsize .or. &
           vsmImageData%rsdata_r(p2+1) > vsmImageData%rimgsize) then
         print *,"px1",vsmImageData%rsdata_r(p1+1:p1+3),r1,ele, p1
         print *,"px2",vsmImageData%rsdata_r(p2+1:p2+3),r2,ele, p2
         call vsmError("merge: px is strange")
       end if

!       if(transfer(vsmImageData%rsdata_r(p1+2),1.0) == transfer(vsmImageData%rsdata_r(p2+2), 1.0)) then
!         print *,"==",transfer(vsmImageData%rsdata_r(p1+2),1.0),vsmImageData%rsdata_r(p1+1)
!       end if
!       print *,"",transfer(vsmImageData%rsdata_r(p1+2),1.0), transfer(vsmImageData%rsdata_r(p2+2),1.0)
!       intd1 = vsmImageData%rsdata_r(p1+2)
!       intd2 = vsmImageData%rsdata_r(p2+2)
#ifndef CTRANS
!       dist1 = transfer(intd1,1.0)
!       dist2 = transfer(intd2, 1.0)
       dist1 = transfer(vsmImageData%rsdata_r(p1+2),1.0)
       dist2 = transfer(vsmImageData%rsdata_r(p2+2), 1.0)
#else 
!       call VSMTRANSFER4I2F(intd1, dist1)
!       call VSMTRANSFER4I2F(intd2, dist2)
       call VSMTRANSFER4I2F(vsmImageData%rsdata_r(p1+2), dist1)
       call VSMTRANSFER4I2F(vsmImageData%rsdata_r(p2+2), dist2)
#endif

       if(dist1 < dist2) then
         vsmImageData%rsdata(p+1:p+vsmRSD) = vsmImageData%rsdata_r(p1+1:p1+vsmRSD)
         r1 = r1 + 1; p1=p1+vsmRSD
       else
         vsmImageData%rsdata(p+1:p+vsmRSD) = vsmImageData%rsdata_r(p2+1:p2+vsmRSD)
         r2 = r2 + 1; p2=p2+vsmRSD
       end if
       p = p + vsmRSD
     end do

     if(vsmImageData%rsdata_r(p1+1) > 0) then
       do k=r1, ele-1
         if(vsmImageData%rsdata_r(p1+1) < 1) then
          exit
         end if
         vsmImageData%rsdata(p+1:p+vsmRSD) = vsmImageData%rsdata_r(p1+1:p1+vsmRSD)
         p1=p1+vsmRSD; p=p+vsmRSD
       end do
     end if

     if(vsmImageData%rsdata_r(p2+1) > 0) then
       do k=r2, ele-1
         if(vsmImageData%rsdata_r(p2+1) < 1) then
          exit
         end if
         vsmImageData%rsdata(p+1:p+vsmRSD) = vsmImageData%rsdata_r(p2+1:p2+vsmRSD)
         p2=p2+vsmRSD; p=p+vsmRSD
       end do
     end if

   end do
 end subroutine vsmMergeRaySegment

 subroutine vsmBlendRaySegmentF(stage)
   integer, intent(in) :: stage
   integer :: i, j, ele, rele!, size
   integer :: p1, p2, p, px, w
   integer :: dist
   real, dimension(4) :: col2, color!, col1
   real :: lumi_a, one_m_salpha, z

   call vsmLogPrint("raysegmentF stage=",stage)

   ele = vsmRSBNum* (2**(stage))
   do i=1, vsmImageData%rimgsize*vsmRSBNum/ele
     p = (i-1)*ele*vsmRSD
     p1 = p

     px = vsmImageData%rsdata(p1+1)
     if(px < 1) then
       cycle
     end if

     if(px > vsmImageData%rimgsize) then
        call vsmLogPrint("px = ", px)
        print *,"i, p1+1",i, p1+1
        do j=1, p1+3
           print *,"rsdata",j, vsmImageData%rsdata(j)
        end do
        call vsmError("blend ray segment: big px")
     end if

     rele = 0
     do j=1, ele
       if(vsmImageData%rsdata(p1+1) == px) then
         rele = rele + 1
         p1 = p1 + vsmRSD
       else
         exit
       end if
     end do

     p1 = p
     p2 = p + vsmRSD
     call vsmIntdata2Color(vsmImageData%rsdata(p1+3), color)
     dist = vsmImageData%rsdata(p1+2)
!     color(4) = transfer(vsmImageData%rsdata(p1+4),1.0)

     do j=1, rele-1
       call vsmIntdata2Color(vsmImageData%rsdata(p2+3), col2)
!       col2(4) = transfer(vsmImageData%rsdata(p2+4), 1.0)

       one_m_salpha = 1.0 - color(4)
       lumi_a = col2(4) * one_m_salpha
       color(1:3) = color(1:3) + col2(1:3)*one_m_salpha ! RGB
       color(4)   = color(4)   + lumi_a             ! A

       dist = vsmImageData%rsdata(p2+2)
       p2 = p2 + vsmRSD
     end do
     w = (px-1)*5
#ifndef CTRANS
     z = transfer(dist, 1.0)
#else
!     call vsmLogPrint("vsmtransfer4 i2f")
     call VSMTRANSFER4I2F(dist, z)
!     call vsmLogPrint("vsmtransfer4 i2f ends")
#endif
!     call vsmLogPrint("intmax= ",int_max)
!     call vsmLogPrint("image w+1:w+4 = color 1:4 ",px)
!     call vsmLogPrint("image rimgsize ",vsmimagedata%rimgsize)
     vsmImageData%image(w+1:w+4) = color(1:4)
!     call vsmLogPrint("image w+5 = color z")
     vsmImageData%image(w+5) = z
   end do
 end subroutine vsmBlendRaySegmentF

 subroutine vsmCopyRaySegment(stpix, pxnum, offset, stage)
   integer, intent(in) :: stpix, pxnum, offset, stage
   integer :: i, st

   st = (stpix-1)*vsmRSD*vsmRSBNum*(2**(stage-1))
   if(vsmDbg) then
     call vsmLogPrint("copy start",st)
     call vsmLogPrint("copy end",st+pxnum*vsmRSD*vsmRSBNum*(2**(stage-1)))
     call vsmLogPrint("copied offset",offset)
     call vsmLogPrint("copied end",offset+pxnum*vsmRSD*vsmRSBNum*(2**(stage-1)))
     call vsmLogPrint("pix num",vsmImageData%rimgsize*vsmRSD*vsmRSBNum)
   end if

   do i=1, pxnum*vsmRSD*vsmRSBNum*(2**(stage-1))
     vsmImageData%rsdata_r(i+offset) = vsmImageData%rsdata(st+i)
   end do
 end subroutine vsmCopyRaySegment

 subroutine vsmSetImageSize(w, h)
  integer, intent(in) :: w, h

  if(w < 0 .or. h < 0) call vsmError('image size is bizzare')
  vsmImageData%width = w
  vsmImageData%height = h
 end subroutine vsmSetImageSize

 subroutine vsmInitBGColor
  vsmImageData%bgcolor(1:3) = 0.0
  vsmImageData%bgc = .false.
 end subroutine vsmInitBGColor

 subroutine vsmInitImage
  integer :: allocstate
  integer :: comm, myrank, psize, pxl, rimgsize
  integer :: residue, allpix 
  call vsmLogPrint("vsmInitImage")
  call vsmGetMPIVars(comm, myrank, psize)

  allpix = vsmImageData%width*vsmImageData%height
  residue = mod(allpix, psize)
  pxl = allpix/psize

  if(residue == 0) then
    rimgsize = allpix
  else
    rimgsize = (pxl+1)*psize
    pxl = rimgsize/psize
  end if

  vsmImageData%pxlsize = pxl
  vsmImageData%rimgsize = rimgsize

  call vsmLogPrint("pxlsize",pxl)
  call vsmLogPrint("rimgsize",rimgsize)
  call vsmLogPrint("allpixsize",allpix)
  
  if(mod(rimgsize, 2) /= 0) rimgsize = rimgsize + 1

  allocate(vsmImageData%image(rimgsize*5), stat=allocstate) ! *5 -> rgbaz
  if(allocstate /= 0) then
   call vsmLogPrint("allocstate",allocstate)
   call vsmError("can't allocate image")
  end if

  allocate(vsmImageData%image_r(rimgsize*5), stat=allocstate) ! *5 -> rgbaz
  if(allocstate /= 0) then
   call vsmLogPrint("allocstate",allocstate)
   call vsmError("can't allocate image r")
  end if

  allocate(vsmImageData%rsdata(rimgsize*vsmRSBNum*vsmRSD), stat=allocstate) ! *5 -> rgbaz
  if(allocstate /= 0) then
   call vsmLogPrint("allocstate",allocstate)
   call vsmError("can't allocate rsdata")
  end if

  allocate(vsmImageData%rsdata_r(rimgsize*vsmRSBNum*vsmRSD), stat=allocstate) ! *5 -> rgbaz
  if(allocstate /= 0) then
   call vsmLogPrint("allocstate",allocstate)
   call vsmError("can't allocate rsdata_r")
  end if

  if(vsmImgKind == 0) then
!    allocate(imageUC(OFFSET+(3*vsmImageData%width + mod(vsmImageData%width,4))*vsmImageData%height))
    allocate(vsmImageData%imageUC(54+(3*vsmImageData%width + mod(vsmImageData%width,4))*vsmImageData%height))
  else if(vsmImgKind == 1) then
    allocate(vsmImageData%imageUC(3*vsmImageData%width*vsmImageData%height), stat = allocstate)
  else if(vsmImgKind == 2) then
    allocate(vsmImageData%imageUC(3*vsmImageData%width*vsmImageData%height), stat = allocstate)
  end if

  if(allocstate /= 0) then
   call vsmLogPrint("allocstate",allocstate)
   call vsmError("can't allocate ImageUC")
  end if

  if(vsmDbg) then
    call vsmLogPrint("width for allocation", vsmImageData%width)
    call vsmLogPrint("height for allocation", vsmImageData%height)
  end if
  call vsmClearImage

  call vsmClearRaySegment
  call vsmClearRaySegmentR

 end subroutine vsmInitImage

 subroutine vsmFinalizeImage
  deallocate(vsmImageData%image, vsmImageData%image_r)
  deallocate(vsmImageData%rsdata, vsmImageData%rsdata_r)
  deallocate(vsmImageData%imageUC)
 end subroutine vsmFinalizeImage

 subroutine vsmSetBGcolor(rgb)
  real, dimension(3), intent(in) :: rgb

  call vsmLogPrint("vsmSetBGcolor", rgb(1:3))
  vsmImageData%bgcolor(1:3) = rgb(1:3)
  vsmImageData%bgcolor(1) = vsmScaleF01(vsmImageData%bgcolor(1))
  vsmImageData%bgcolor(2) = vsmScaleF01(vsmImageData%bgcolor(2))
  vsmImageData%bgcolor(3) = vsmScaleF01(vsmImageData%bgcolor(3))
  if(vsmImageData%bgcolor(1) >=1.0/256 .or. vsmImageData%bgcolor(2) >= 1.0/256 .or. vsmImageData%bgcolor(3) >= 1.0/256) then
    vsmImageData%bgc = .true.
    call vsmLogPrint("bgc = true", vsmImageData%bgcolor)
  end if
 end subroutine vsmSetBGcolor

 subroutine set_bgcolor(rgb)
  real, dimension(3), intent(in) :: rgb

  vsmImageData%bgcolor(1:3) = rgb(1:3)
 end subroutine set_bgcolor

 subroutine vsmCalcDp(DP, cnum)
   integer, intent(in) :: cnum
   real(kind=vdp), intent(out) :: DP

   call vsmyyCalcDp(DP, cnum)

 end subroutine vsmCalcDp

 subroutine vsmyyCalcDP(DP, cam_num) ! 20200219
   integer, intent(in) :: cam_num
   real(kind=vdp), intent(out) :: DP
   real(kind=vdp) :: ray2dat
   real(kind=vdp), dimension(3) :: norm, pixel
   real(kind=vdp) :: dist2pxl
   integer :: i, j
   real(kind=vdp) :: min_ray2ldat
   real(kind=vdp), dimension(3) :: center, tcenter
   integer :: threadn, maxthread
   real(kind=vdp), dimension(0:vsmMaxOMP) :: tminray
   real(kind=vdp), dimension(3) :: origin, cend, cfront, cright, topleft, cup, cpos
   real(kind=vdp) :: fovy, screen_height, screen_width,near

     call vsmInitScreen(cam_num)
     if(.not.vsmCamera(cam_num)%pers) then
        ! Parallel Projection
!        DP = vsmImageData%dp
        return
     end if
     min_ray2ldat = vsmFaraway
     select case (vsmCoordData%YinYang)
        case (0)
      ! Now Yin
           cup(1:3) = vsmCamera(cam_num)%up(1:3)
           cright(1:3) = vsmCamera(cam_num)%right(1:3)
           cpos(1:3) = vsmCamera(cam_num)%pos(1:3)
           cfront(1:3) = vsmCamera(cam_num)%front(1:3)

        case (1)
      ! Now Yang
!         x_yy = -x(1)
!         y_yy = x(3)
!         z_yy = x(2)
           call vsmToYangCoord(vsmCamera(cam_num)%up, cup)
           call vsmToYangCoord(vsmCamera(cam_num)%right, cright)
           call vsmToYangCoord(vsmCamera(cam_num)%pos, cpos)
           call vsmToYangCoord(vsmCamera(cam_num)%front, cfront)

        case default
           call   vsmError("calcdp: Yin-Yang judge")
     end select
     near =  vsmCamera(cam_num)%near
     fovy = vsmCamera(cam_num)%fovy
     screen_height = 2.0_vdp*tan(vsmDeg2Rad(fovy/2.0_vdp))*near
     vsmImageData%dp = screen_height/dble(vsmImageData%height)
     screen_width  = vsmImageData%dp*dble(vsmImageData%width)

      topleft(1:3) = cpos(1:3) + near*cfront(1:3) + & 
        screen_height/2.0_vdp*cup(1:3) - screen_width/2.0_vdp*cright(1:3)
        
     call vsmAnglesToCoords(vsmLocalCoord%lorigin, origin)
     call vsmAnglesToCoords(vsmLocalCoord%lcend, cend)

    tminray = vsmFarAwayd10
    maxthread = 1
!$ maxthread = omp_get_max_threads()
    if (maxthread > vsmMaxOMP) call vsmError("The number of OpenMP's threads exceeds the limit.")
     ! Perspective Projection

!$omp parallel do private(i,j,pixel,norm,dist2pxl,ray2dat,threadn) schedule (dynamic, 1)
     do j=0, vsmImageData%height+1
       threadn = 0
!$    threadn = omp_get_thread_num()
       do i=0, vsmImageData%width+1
            pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*cright(1:3) - (j-1)*cup(1:3))
            norm(1:3) = pixel(1:3) - cpos(1:3)
            dist2pxl = sqrt(dot_product(norm, norm))
            call vsmNormVec(norm)

            ray2dat = vsmyyRayDist(pixel, norm, origin, cend)+ dist2pxl
!                            vsmRayDist(pixel, norm, vsmLocalCoord%lorigin, vsmLocalCoord%lcend) + dist2pxl
            if(ray2dat < tminray(threadn)) then
               tminray(threadn) = ray2dat
            end if
            if(ray2dat == dist2pxl) exit
       end do
     end do
     min_ray2ldat = minval(tminray(0:maxthread-1))
     if(min_ray2ldat < vsmFarAwayd10) then
       DP = min_ray2ldat*vsmImageData%dp/vsmCamera(cam_num)%near
     else
       tcenter = (vsmLocalCoord%lorigin(1:3) + vsmLocalCoord%lcend(1:3))/2.0_vdp
       call vsmAnglesToCoords(tcenter, center)
       norm = center - cpos
       ray2dat = sqrt(dot_product(norm, norm))
       DP = -ray2dat*vsmImageData%dp/vsmCamera(cam_num)%near
     end if

 end subroutine vsmyyCalcDP

 function vsmCheckPX(px)
   integer, dimension(2), intent(in) :: px
   logical :: vsmCheckPX

   if(px(1) < 1 .or. px(1) > vsmImageData%width) then
     vsmCheckPX = .false.
     return
   end if

   if(px(2) < 1 .or. px(2) > vsmImageData%height) then
     vsmCheckPX = .false.
     return
   end if
   vsmCheckPX = .true.
 end function vsmCheckPX

 !initialize screen. everytime called
 subroutine vsmInitScreen(cnum)
  integer, intent(in) :: cnum
  real(kind=vdp), dimension(3) :: campos, camfront, camup, camright ! camera's
  real(kind=vdp) :: near, fovy
  real(kind=vdp) :: screen_width, screen_height

!  call vsmLogPrint("------vsm Init Screen----------", cnum)
!  call vsmClearImage

  campos = vsmCamera(cnum)%pos
  camfront = vsmCamera(cnum)%front
  camup = vsmCamera(cnum)%up
  camright = vsmCamera(cnum)%right
  near = vsmCamera(cnum)%near

  if(vsmCamera(cnum)%pers) then
    fovy = vsmCamera(cnum)%fovy
    screen_height = 2.0_vdp*tan(vsmDeg2Rad(fovy/2.0_vdp))*near
    vsmImageData%dp = screen_height/dble(vsmImageData%height)
    screen_width  = vsmImageData%dp*dble(vsmImageData%width)
  else
    screen_width = vsmCamera(cnum)%width
    vsmImageData%dp = screen_width/dble(vsmImageData%width)
    screen_height = screen_width*dble(vsmImageData%height)/dble(vsmImageData%width)
  end if

  vsmImageData%topleft(1:3) = campos(1:3) + near*camfront(1:3) + & 
        screen_height/2.0_vdp*camup(1:3) - screen_width/2.0_vdp*camright(1:3)

!  call vsmLogPrint("topleft", vsmImageData%topleft)
!  call vsmLogPrint("dp", vsmImageData%dp)
 end subroutine vsmInitScreen

 !clean buffers
 subroutine vsmCleanImage
  integer :: i, wdth, rimgsize
  integer :: j,k,l,loop,res
  rimgsize = vsmImageData%rimgsize

!$omp parallel do private(i,wdth)
  do i=1, rimgsize
    wdth = (i-1)*5
    vsmImageData%image(wdth+1:wdth+4) = 0.0
    vsmImageData%image(wdth+5) = real(vsmFaraway)
    vsmImageData%image_r(wdth+1:wdth+4) = 0.0
    vsmImageData%image_r(wdth+5) = real(vsmFaraway)
  end do
 end subroutine vsmCleanImage

 subroutine vsmClearImage
   call vsmCleanImage
 end subroutine vsmClearImage

 subroutine vsmClearImage4Solid
   call vsmClearImage_buf4Solid
 end subroutine vsmClearImage4Solid

 subroutine vsmClearImage_buf4Solid
  integer :: i, wdth, wch
  integer :: j,k,l,loop,res
  integer :: ifaraway
  real :: ffaraway

#ifndef CTRANS
  ifaraway = transfer(vsmFarAway, 1)
#else
  ffaraway = real(vsmFarAway)
  call VSMTRANSFER4F2I(ffaraway, ifaraway)
#endif

  wch = vsmImageData%width*vsmImageData%height
!$omp parallel do private(i,wdth)
  do i=1, wch
    wdth = (i-1)*2
    vsmImageData%image(wdth+1) = 0.0
    vsmImageData%image(wdth+2) = real(vsmFaraway)

    vsmImageData%image(wdth+1+2*wch) = 0.0
    vsmImageData%image(wdth+2+2*wch) = real(vsmFaraway)

    vsmImageData%image_r(wdth+1) = 0.0
    vsmImageData%image_r(wdth+2) = real(vsmFaraway)

    vsmImageData%image_r(wdth+1+2*wch) = 0.0
    vsmImageData%image_r(wdth+2+2*wch) = real(vsmFaraway)

    vsmImageData%rsdata(wdth+1) = 0
    vsmImageData%rsdata(wdth+2) = ifaraway

    vsmImageData%rsdata(wdth+1+2*wch) = 0
    vsmImageData%rsdata(wdth+2+2*wch) = ifaraway

    vsmImageData%rsdata(wdth+1+4*wch) = 0
    vsmImageData%rsdata(wdth+2+4*wch) = ifaraway

    vsmImageData%rsdata(wdth+1+6*wch) = 0
    vsmImageData%rsdata(wdth+2+6*wch) = ifaraway

    vsmImageData%rsdata(wdth+1+8*wch) = 0
    vsmImageData%rsdata(wdth+2+8*wch) = ifaraway

    vsmImageData%rsdata(wdth+1+10*wch) = 0
    vsmImageData%rsdata(wdth+2+10*wch) = ifaraway

    vsmImageData%rsdata(wdth+1+12*wch) = 0
    vsmImageData%rsdata(wdth+2+12*wch) = ifaraway

    vsmImageData%rsdata(wdth+1+14*wch) = 0
    vsmImageData%rsdata(wdth+2+14*wch) = ifaraway

    vsmImageData%rsdata(wdth+1+16*wch) = 0
    vsmImageData%rsdata(wdth+2+16*wch) = ifaraway

    vsmImageData%rsdata_r(wdth+1) = 0
    vsmImageData%rsdata_r(wdth+2) = ifaraway

    vsmImageData%rsdata_r(wdth+1+2*wch) = 0
    vsmImageData%rsdata_r(wdth+2+2*wch) = ifaraway

    vsmImageData%rsdata_r(wdth+1+4*wch) = 0
    vsmImageData%rsdata_r(wdth+2+4*wch) = ifaraway

    vsmImageData%rsdata_r(wdth+1+6*wch) = 0
    vsmImageData%rsdata_r(wdth+2+6*wch) = ifaraway

    vsmImageData%rsdata_r(wdth+1+8*wch) = 0
    vsmImageData%rsdata_r(wdth+2+8*wch) = ifaraway

    vsmImageData%rsdata_r(wdth+1+10*wch) = 0
    vsmImageData%rsdata_r(wdth+2+10*wch) = ifaraway

    vsmImageData%rsdata_r(wdth+1+12*wch) = 0
    vsmImageData%rsdata_r(wdth+2+12*wch) = ifaraway

    vsmImageData%rsdata_r(wdth+1+14*wch) = 0
    vsmImageData%rsdata_r(wdth+2+14*wch) = ifaraway

    vsmImageData%rsdata_r(wdth+1+16*wch) = 0
    vsmImageData%rsdata_r(wdth+2+16*wch) = ifaraway

  end do

!$omp parallel do private(i)
  do i=4*wch+1, 5*wch
    vsmImageData%image(i) = 0.0
    vsmImageData%image_r(i) = real(vsmFaraway)
  end do

 end subroutine vsmClearImage_buf4Solid

 subroutine vsmClearRImage
   call vsmLogPrint("clearRimage ")
   call vsmClearRImage_buf
     call vsmLogPrint("clearRimage ends")
 end subroutine vsmClearRImage

 subroutine vsmClearRImage_buf
  integer :: i, wdth, rimgsize
  integer :: j,k,l,loop,res

  rimgsize = vsmImageData%rimgsize

!$omp parallel do private(i,wdth)
  do i=1, rimgsize
    wdth = (i-1)*5
    vsmImageData%image_r(wdth+1:wdth+4) = 0.0
    vsmImageData%image_r(wdth+5) = real(vsmFaraway)
!    vsmImageData%zbuf_r(i) = real(vsmFaraway)
  end do
 end subroutine vsmClearRImage_buf

 subroutine vsmClearImage_buf
  integer :: i, wdth, rimgsize
  integer :: j,k,l,loop,res
  rimgsize = vsmImageData%rimgsize

!$omp parallel do private(i,wdth)
  do i=1, rimgsize
    wdth = (i-1)*5
    vsmImageData%image(wdth+1:wdth+4) = 0.0
    vsmImageData%image(wdth+5) = real(vsmFaraway)
!    vsmImageData%zbuf(i) = real(vsmFaraway)
  end do
 end subroutine vsmClearImage_buf

 subroutine vsmMergeBuf4Solid
   integer :: i, wdt, wch, zwdt
   real :: fdis1, fdis2
   integer :: idis1, idis2
   integer :: j,k,l,loop,res
   real, parameter :: ffaraway = real(vsmFarAway)
   real :: dcolor
   real, dimension(3) :: color

   wch = vsmImageData%width*vsmImageData%height
   call vsmLogPrint("Merge")

!$omp parallel do private(i,wdt, zwdt)
   do i=1, wch
     wdt = (i-1)*2

     if(vsmImageData%image_r(wdt+2) > vsmImageData%image(wdt+2)) then
       vsmImageData%image_r(wdt+1:wdt+2) = vsmImageData%image(wdt+1:wdt+2)
     end if

     zwdt = wdt + 2*wch
     if(vsmImageData%image_r(wdt+2) > vsmImageData%image(zwdt+2)) then
       vsmImageData%image_r(wdt+1:wdt+2) = vsmImageData%image(zwdt+1:zwdt+2)
     end if

     if(vsmImageData%image_r(wdt+2) > vsmImageData%image_r(zwdt+2)) then
       vsmImageData%image_r(wdt+1:wdt+2) = vsmImageData%image_r(zwdt+1:zwdt+2)
     end if

     zwdt = i + 4*wch !(i-1) + 1 + 4*wch
     if(vsmImageData%image_r(wdt+2) > vsmImageData%image_r(zwdt)) then
       vsmImageData%image_r(wdt+1) = vsmImageData%image(zwdt)
       vsmImageData%image_r(wdt+2) = vsmImageData%image_r(zwdt)
     end if
   end do

!$omp parallel do private(i,wdt, zwdt, color, dcolor)
   do i=1, wch
     wdt = (i-1)*2
     zwdt = (i-1)*5

     if(vsmImageData%image_r(wdt+2) > ffaraway*0.5) then
       vsmImageData%image(zwdt+1:zwdt+4) = 0.0
       vsmImageData%image(zwdt+5) = ffaraway
     else
       dcolor = vsmImageData%image_r(wdt+1)
       call vsmUnpack1SP23SP(color, dcolor)
       vsmImageData%image(zwdt+1:zwdt+3) = color(1:3)
       vsmImageData%image(zwdt+4) = 1.0
       vsmImageData%image(zwdt+5) = vsmImageData%image_r(wdt+2)
     end if
   end do

  call vsmClearRImage_buf

!$omp parallel do private(i,zwdt)
   do i=wch+1, vsmImageData%rimgsize
     zwdt = (i-1)*5
     vsmImageData%image(zwdt+1:zwdt+4) = 0.0
     vsmImageData%image(zwdt+5) = ffaraway
   end do

!$omp parallel do private(i,wdt, zwdt, fdis1, fdis2, idis1, idis2)
   do i=1, wch
     wdt = (i-1)*2

     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata(wdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata(wdt+1:wdt+2)
     end if

     zwdt = wdt + 2*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 4*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 6*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 8*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 10*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 12*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 14*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 16*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata(zwdt+1:zwdt+2)
     end if


     zwdt = wdt + 2*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata_r(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata_r(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 4*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata_r(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata_r(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 6*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata_r(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata_r(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 8*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata_r(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata_r(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 10*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata_r(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata_r(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 12*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata_r(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata_r(zwdt+1:zwdt+2)
     end if

     zwdt = wdt + 14*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata_r(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata_r(zwdt+1:zwdt+2)
     end if
   
     zwdt = wdt + 16*wch
     idis1 = vsmImageData%rsdata_r(wdt+2); idis2 = vsmImageData%rsdata_r(zwdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0); fdis2 = transfer(idis2, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1); call VSMTRANSFER4I2F(idis2, fdis2)
#endif
     if(fdis1 > fdis2) then
       vsmImageData%rsdata_r(wdt+1:wdt+2) = vsmImageData%rsdata_r(zwdt+1:zwdt+2)
     end if
   end do


!$omp parallel do private(i,wdt, zwdt, color, dcolor, idis1, fdis1)
   do i=1, wch
     wdt = (i-1)*2
     zwdt = (i-1)*5

     idis1 = vsmImageData%rsdata_r(wdt+2)
#ifndef CTRANS
     fdis1 = transfer(idis1, 1.0)
#else
     call VSMTRANSFER4I2F(idis1, fdis1)
#endif

     if(vsmImageData%image(zwdt+5) > fdis1) then
       idis1 = vsmImageData%rsdata_r(wdt+1)
       call vsmUnpack1INT23SP(color, idis1)
       vsmImageData%image(zwdt+1:zwdt+3) = color(1:3)
       vsmImageData%image(zwdt+4) = 1.0
       vsmImageData%image(zwdt+5) = fdis1
     end if

   end do
  call vsmClearRImage_buf

 end subroutine vsmMergeBuf4Solid
 
  subroutine vsmAlphaBlending(st_pixel, ed_pixel)
   integer, intent(in) :: st_pixel, ed_pixel
   integer :: i, i1
   real :: vsmFarAwayF01
   integer :: wdt, wdt1
   real :: zbuf1, zbuf2
   real, dimension(4) :: color, col1, col2
   real :: one_m_salpha, lumi_a

   vsmFarAwayF01 = (real(vsmFaraway))*0.1
!$omp parallel do private(i,wdt,wdt1,i1,zbuf1,zbuf2,col1,col2,lumi_a,one_m_salpha,color)
   do i=st_pixel, ed_pixel
     wdt = (i-1)*5
     i1=i-st_pixel+1
     wdt1=(i1-1)*5

     zbuf1 = vsmImageData%image(wdt+5)
     zbuf2 = vsmImageData%image_r(wdt1+5)

     if(zbuf1 > vsmFarAwayF01 .and. zbuf2 > vsmFarAwayF01) cycle

     col1(1:4) = vsmImageData%image(wdt+1:wdt+4)
     col2(1:4) = vsmImageData%image_r(wdt1+1:wdt1+4)


     if(zbuf1 > zbuf2) then
       one_m_salpha = 1.0 - col2(4)
       lumi_a = col1(4) * one_m_salpha

       color(1:3) = col2(1:3) + col1(1:3)*one_m_salpha ! RGB
       color(4)   = col2(4)   + lumi_a             ! A

       vsmImageData%image(wdt+1:wdt+4) = color(1:4)
       vsmImageData%image(wdt+5) = zbuf2
     else
       one_m_salpha = 1.0 - col1(4)
       lumi_a = col2(4) * one_m_salpha

       color(1:3) = col1(1:3) + col2(1:3)*one_m_salpha ! RGB
       color(4)   = col1(4)   + lumi_a             ! A

       vsmImageData%image(wdt+1:wdt+4) = color(1:4)
       vsmImageData%image(wdt+5) = zbuf1
     end if
   end do
 end subroutine vsmAlphaBlending

 subroutine vsmSolidBlending(st_pixel, ed_pixel)
   integer, intent(in) :: st_pixel, ed_pixel
   integer :: i, i1, wdt, wdt1, loopsize, loop, offset, j, res

!$omp parallel do private(i,wdt,wdt1,i1)
   do i=st_pixel, ed_pixel
     wdt = (i-1)*5
     i1=i-st_pixel+1
     wdt1=(i1-1)*5
     if(vsmImageData%image(wdt+5) > vsmImageData%image_r(wdt1+5)) then
       vsmImageData%image(wdt+1:wdt+5) = vsmImageData%image_r(wdt1+1:wdt1+5)
     end if
   end do
 end subroutine vsmSolidBlending

  subroutine vsmBGBlending
   integer :: st_pixel, ed_pixel
   integer :: i, wdt!, i1, wdt1
   real, dimension(4) :: color, col1, col2
   real :: zbuf1, zbuf2
   real :: one_m_salpha, lumi_a
   
   call vsmLogPrint("BGCOLOR = ", vsmImageData%bgc)
   call vsmLogPrint("BGCOLOR", vsmImageData%bgcolor(1:3))
   if(.not.vsmImageData%BGC) return
   st_pixel = 1; ed_pixel = vsmImageData%rimgsize
   zbuf2 = real(vsmFaraway)*0.1
   col2(1:3) = vsmImageData%bgcolor(1:3); col2(4) = 1.0

!$omp parallel do private(i,wdt,zbuf1,col1,lumi_a,one_m_salpha,color)
   do i=st_pixel, ed_pixel
     wdt = (i-1)*5

     zbuf1 = vsmImageData%image(wdt+5)

!     if(zbuf1 > real(vsmFaraway)*0.1 .and. zbuf2 > real(vsmFaraway)*0.1) cycle

     col1(1:4) = vsmImageData%image(wdt+1:wdt+4)

     if(zbuf1 > zbuf2) then
       one_m_salpha = 1.0 - col2(4)
       lumi_a = col1(4) * one_m_salpha

       color(1:3) = col2(1:3) + col1(1:3)*one_m_salpha ! RGB
       color(4)   = col2(4)   + lumi_a             ! A

       vsmImageData%image(wdt+1:wdt+4) = color(1:4)
       vsmImageData%image(wdt+5) = zbuf2
     else
       one_m_salpha = 1.0 - col1(4)
       lumi_a = col2(4) * one_m_salpha

       color(1:3) = col1(1:3) + col2(1:3)*one_m_salpha ! RGB
       color(4)   = col1(4)   + lumi_a             ! A

       vsmImageData%image(wdt+1:wdt+4) = color(1:4)
       vsmImageData%image(wdt+5) = zbuf1
     end if
   end do
 end subroutine vsmBGBlending

 subroutine vsmMergeBuf
   integer :: i, wdt, rimgsize
   integer :: j,k,l,loop,res
   rimgsize = vsmImageData%rimgsize
!$omp parallel do private(i,wdt)
   do i=1, rimgsize
     wdt = (i-1)*5
     if(vsmImageData%image(wdt+5) > vsmImageData%image_r(wdt+5)) then
       vsmImageData%image(wdt+1:wdt+5) = vsmImageData%image_r(wdt+1:wdt+5)
     end if
   end do
  call vsmClearRImage_buf

 end subroutine vsmMergeBuf

 subroutine vsmWriteImage(file)
  character(*), intent(in) :: file
  character(128) :: filename

  call vsmLogPrint("write image")
  call vsmLogFlush

  if(vsmImgKind == 0) then
     filename = file//'.bmp'
     call vsmWriteImageBMP(filename)
  else if(vsmImgKind == 1) then
     filename = file//'.ppm'
     call vsmWriteImagePPM(filename)
  else if(vsmImgKind == 2) then
     filename = file//'.png'
     call vsmWriteImagePNG(filename)
  end if

  call vsmLogPrint("write image ends")
 end subroutine vsmWriteImage

 !save image as ppm
 subroutine vsmWriteImagePNG(filename)
  integer, parameter :: MaxC = 255 ! unsigned char -> 0 - 255
  integer :: i, j, k, l, m
  integer, dimension(3) :: rgb
  character(*), intent(in) :: filename
!  character(3*vsmImageData%width*vsmImageData%height) :: imageUC ! unsigned chara
!  integer :: allocstate
  
  call vsmLogPrint("write png image")
  call vsmLogFlush

!$omp parallel do private(i,j,k,l,m,rgb)
    do j=1, vsmImageData%height
      do i=1, vsmImageData%width
         ! real::rgba -> unsigned char :: rgb
         m = ((i-1)+(j-1)*vsmImageData%width)
         k = m*5
         rgb(1) = nint(256*vsmImageData%image(k+1))
         rgb(2) = nint(256*vsmImageData%image(k+2))
         rgb(3) = nint(256*vsmImageData%image(k+3))
         l = m*3
         vsmImageData%imageUC(l+1:l+1) = char(min(rgb(1), MaxC))
         vsmImageData%imageUC(l+2:l+2) = char(min(rgb(2), MaxC))
         vsmImageData%imageUC(l+3:l+3) = char(min(rgb(3), MaxC))
      end do
    end do

#ifdef CPNG
   call VSMWRITEPNG(filename, len_trim(filename), vsmImageData%width, &
           vsmImageData%height, vsmImageData%imageUC)
#else
   call vsmError("PNG format is not available. Modify the Makefile and recompile vismo.")
#endif
  call vsmLogPrint("write png image ends")
 end subroutine vsmWriteImagePNG

 subroutine vsmWriteImageBMP(filename)
  integer, parameter :: MaxC = 255 ! unsigned char -> 0 - 255
  integer, parameter :: OFFSET = 54
  integer :: i, j, k, l, m, ImgUnitNum, cwidth
  integer :: width, height, filesize
  integer, dimension(3) :: rgb
  character(*), intent(in) :: filename
  integer :: val = 255
  character :: val1, val2, val3, val4
!  character(OFFSET+(3*vsmImageData%width + mod(vsmImageData%width,4))*vsmImageData%height) :: imageUC ! unsigned chara
  call vsmLogPrint("write bmp image")
  call vsmLogFlush
#ifndef CBMPPPM
  width = vsmImageData%width
  height = vsmImageData%height
  filesize = (width*3+mod(width,4))*height+OFFSET
  ! BMP Header
     ! ---------File Header ---------
     vsmImageData%imageUC(1:2) = "BM"
     ! File size 4 bytes
     val1 = char(iand(val, filesize)); val2 = char(iand(val, ishft(filesize, -8)));
     val3 = char(iand(val, ishft(filesize, -16)));  val4 = char(iand(val, ishft(filesize, -24)))
     if(.not. bigend) then
         vsmImageData%imageUC(3:3) = val1;  vsmImageData%imageUC(4:4) = val2
         vsmImageData%imageUC(5:5) = val3;  vsmImageData%imageUC(6:6) = val4
     else
         vsmImageData%imageUC(3:3) = val1;  vsmImageData%imageUC(4:4) = val2
         vsmImageData%imageUC(5:5) = val3;  vsmImageData%imageUC(6:6) = val4
         !imageUC(3:3) = val4;  imageUC(4:4) = val3; imageUC(5:5) = val2;  imageUC(6:6) = val1     
     end if
     !imageUC(7:10) = char(0) ! reserved area
     do i=7, 10
        vsmImageData%imageUC(i:i) = char(0) ! reserved area
     end do
     ! offset to image data
     vsmImageData%imageUC(11:11) = char(54); vsmImageData%imageUC(12:12) = char(0)
     vsmImageData%imageUC(13:13) = char(0); vsmImageData%imageUC(14:14) = char(0)
     ! --------- Information Header ---------
     ! information header size
     vsmImageData%imageUC(15:15) = char(40); vsmImageData%imageUC(16:16) = char(0)
     vsmImageData%imageUC(17:17) = char(0); vsmImageData%imageUC(18:18) = char(0)
     ! width size
     val1 = char(iand(val, width)); val2 = char(iand(val, ishft(width, -8)));
     val3 = char(iand(val, ishft(width, -16)));  val4 = char(iand(val, ishft(width, -24)))
     if(.not. bigend) then
         vsmImageData%imageUC(19:19) = val1;  vsmImageData%imageUC(20:20) = val2
         vsmImageData%imageUC(21:21) = val3;  vsmImageData%imageUC(22:22) = val4
     else
         vsmImageData%imageUC(19:19) = val1;  vsmImageData%imageUC(20:20) = val2
         vsmImageData%imageUC(21:21) = val3;  vsmImageData%imageUC(22:22) = val4
!         imageUC(19:19) = val4;  imageUC(20:20) = val3; imageUC(21:21) = val2;  imageUC(22:22) = val1
     end if

     ! Height size
     val1 = char(iand(val, height)); val2 = char(iand(val, ishft(height, -8)));
     val3 = char(iand(val, ishft(height, -16)));  val4 = char(iand(val, ishft(height, -24)))
     if(.not. bigend) then
         vsmImageData%imageUC(23:23) = val1;  vsmImageData%imageUC(24:24) = val2
         vsmImageData%imageUC(25:25) = val3;  vsmImageData%imageUC(26:26) = val4
     else
         vsmImageData%imageUC(23:23) = val1;  vsmImageData%imageUC(24:24) = val2
         vsmImageData%imageUC(25:25) = val3;  vsmImageData%imageUC(26:26) = val4
!         imageUC(23:23) = val4;  imageUC(24:24) = val3; imageUC(25:25) = val2;  imageUC(26:26) = val1
     end if

     vsmImageData%imageUC(27:27) = char(1); vsmImageData%imageUC(28:28) = char(0); 
     vsmImageData%imageUC(29:29) = char(24); vsmImageData%imageUC(30:30) = char(0); 
     vsmImageData%imageUC(31:31) = char(0); vsmImageData%imageUC(32:32) = char(0); 
     vsmImageData%imageUC(33:33) = char(0); vsmImageData%imageUC(34:34) = char(0)
    !image data size
     vsmImageData%imageUC(35:35) = char(0); vsmImageData%imageUC(36:36) = char(0)
     vsmImageData%imageUC(37:37) = char(0); vsmImageData%imageUC(38:38) = char(0)

!     imageUC(39:54) = char(0)
     do i=39, 54
        vsmImageData%imageUC(i:i) = char(0) ! reserved area
     end do

    cwidth = 3*vsmImageData%width + mod(vsmImageData%width,4)

!$omp parallel do private(i,j,k,l,m,rgb)
    do j=1, vsmImageData%height
      do i=1, vsmImageData%width
         ! real::rgba -> unsigned char :: rgb
         m = ((i-1)+(j-1)*vsmImageData%width)
         k = m*5
         rgb(1) = nint(256*vsmImageData%image(k+1))
         rgb(2) = nint(256*vsmImageData%image(k+2))
         rgb(3) = nint(256*vsmImageData%image(k+3))
         l = 3*(i-1)+(height- j)*cwidth+OFFSET
         vsmImageData%imageUC(l+1:l+1) = char(min(rgb(3), MaxC)) ! not RGB but BGR
         vsmImageData%imageUC(l+2:l+2) = char(min(rgb(2), MaxC))
         vsmImageData%imageUC(l+3:l+3) = char(min(rgb(1), MaxC))
      end do
    end do

  ImgUnitNum = vsmSearchUnitNum("save bmp image")
  open(unit=ImgUnitNum, file=trim(filename), action='write', status='replace', access='stream')
  write(ImgUnitNum, '(A)', advance ='no') vsmImageData%imageUC(1:filesize)
  !call flush(vsmUnitNum2)
  close(ImgUnitNum)
#else
   call VSMWRITEBMP(filename, len_trim(filename), vsmImageData%width, &
           vsmImageData%height, vsmImageData%image)
#endif
  call vsmLogPrint("write bmp image ends")
  !print *, filesize
 end subroutine vsmWriteImageBMP

 !save image as ppm
 subroutine vsmWriteImagePPM(filename)
  integer, parameter :: MaxC = 255 ! unsigned char -> 0 - 255
  integer :: i, j, k, l, m, ImgUnitNum, width, height!, filesize
  integer, dimension(3) :: rgb
  character(*), intent(in) :: filename
!  character(3*vsmImageData%width*vsmImageData%height) :: imageUC ! unsigned chara

  width = vsmImageData%width; height = vsmImageData%height
  call vsmLogPrint("write image")
  call vsmLogFlush
!$omp parallel do private(i,j,k,l,m,rgb)
    do j=1, height
      do i=1, width
         ! real::rgba -> unsigned char :: rgb
         m = ((i-1)+(j-1)*vsmImageData%width)
         k = m*5
         rgb(1) = nint(256*vsmImageData%image(k+1))
         rgb(2) = nint(256*vsmImageData%image(k+2))
         rgb(3) = nint(256*vsmImageData%image(k+3))
         l = m*3
         vsmImageData%imageUC(l+1:l+1) = char(min(rgb(1), MaxC))
         vsmImageData%imageUC(l+2:l+2) = char(min(rgb(2), MaxC))
         vsmImageData%imageUC(l+3:l+3) = char(min(rgb(3), MaxC))
      end do
    end do

#ifndef CBMPPPM
  ImgUnitNum = vsmSearchUnitNum("save ppm image")
  open(unit=ImgUnitNum, file=trim(filename), action='write', status='replace', access='stream')
    write(imgUnitNum, '(A)') 'P6'
    write(imgUnitNum, '(I5, 1X, I5)') width, height
    write(imgUnitNum, '(I3)') MaxC
    write(ImgUnitNum, '(A)', advance ='no') vsmImageData%imageUC(1:3*width*height)
  !call flush(vsmUnitNum2)
  close(ImgUnitNum)
#else
   call VSMWRITEPPM(filename, len_trim(filename), vsmImageData%width, &
           vsmImageData%height, vsmImageData%imageUC)
#endif

  call vsmLogPrint("write ppm image ends")
 end subroutine vsmWriteImagePPM

 subroutine vsmPaintPixel4(px, col, dist)
  integer, dimension(2), intent(in) :: px
  real, dimension(:), intent(in) :: col
  real(kind=vdp), intent(in) :: dist
  integer :: wdth

  if(.not.vsmCheckPX(px)) call vsmError("Error :: vsmPaintPixel4")
  wdth = 5*((px(1)-1) + vsmImageData%width * (px(2)-1))
  vsmImageData%image(wdth+1:wdth+4) = col(1:4) ! RGBA
  vsmImageData%image(wdth+5) = real(dist)

 end subroutine vsmPaintPixel4

 subroutine vsmPaintPixel4R(px, col, dist)
  integer, dimension(2), intent(in) :: px
  real, dimension(:), intent(in) :: col
  real(kind=vdp), intent(in) :: dist
  integer :: wdth

  if(.not.vsmCheckPX(px)) call vsmError("Error :: vsmPaintPixel4")
  wdth = 5*((px(1)-1) + vsmImageData%width * (px(2)-1))
  vsmImageData%image_r(wdth+1:wdth+4) = col(1:4) ! RGBA
  vsmImageData%image_r(wdth+5) = real(dist)

 end subroutine vsmPaintPixel4R

 subroutine vsmPaintPixel3(px, col, distd)
  integer, dimension(2), intent(in) :: px
  real, dimension(:), intent(in) :: col
  real(kind=vdp), intent(in) :: distd
  real :: dist
  integer :: wdth

  if(.not.vsmCheckPX(px)) call vsmError("Error :: vsmPaintPixel3")

  dist = real(distd)
  wdth = 5*((px(1)-1) + vsmImageData%width * (px(2)-1))
  if(dist < vsmImageData%image(wdth+5)) then
    vsmImageData%image(wdth+1:wdth+3) = col(1:3) ! RGB
    vsmImageData%image(wdth+4)        = 1.0 ! A
    vsmImageData%image(wdth+5) = dist
  end if
 end subroutine vsmPaintPixel3

 subroutine vsmPaintPixelBlend(px, col, distd)
  integer, dimension(2), intent(in) :: px
  real, dimension(3), intent(in) :: col
  real(kind=vdp), intent(in) :: distd
  real, dimension(4) :: col1, color
  real :: dist, lumi_a, one_m_salpha
  integer :: wdth

  ! blend with solid for frame only
  if(.not.vsmCheckPX(px)) call vsmError("Error :: vsmPaintPixelBlend")

  dist = real(distd)
  wdth = 5*((px(1)-1) + vsmImageData%width * (px(2)-1))
  if(dist < vsmImageData%image(wdth+5)) then
    vsmImageData%image(wdth+1:wdth+3) = col(1:3) ! RGB
    vsmImageData%image(wdth+4)        = 1.0 ! A

!    col1(1)=dist; col1(2)=vsmImageData%image(wdth+5); col1(3)=0.0; col1(4)=0.0
    vsmImageData%image(wdth+5) = dist
!    call vsmLogPrint("blend frame",col1)
  else
    col1(1:4) = vsmImageData%image(wdth+1:wdth+4)

    one_m_salpha = 1.0 - col1(4)
    !lumi_a = col(4) * one_m_salpha
    lumi_a = one_m_salpha ! 1.0*one_m_salpha

    color(1:3) = col1(1:3) + col(1:3)*one_m_salpha ! RGB
    color(4)   = col1(4)   + lumi_a             ! A

    vsmImageData%image(wdth+1:wdth+4) = color(1:4)
!    vsmImageData%image(wdth+5) = zbuf2
  end if
 end subroutine vsmPaintPixelBlend

 subroutine vsmPaintPixel3s(px, col, dist)
  integer, dimension(2), intent(in) :: px
  real, dimension(:), intent(in) :: col
  real(kind=vdp), intent(in) :: dist
  integer :: tnum

  tnum = 0

  call vsmLogPrint("3s is called")

!$ tnum=omp_get_thread_num()

  if( mod(tnum,2) == 0) then
     call vsmPaintPixel3B0(px, col, dist)
     !print *,"3B0",tnum
  else
     call vsmPaintPixel3B1(px, col, dist)
     !print *,"3B1",tnum
  end if

 end subroutine vsmPaintPixel3s

#ifdef OMPNEST
 subroutine vsmPaintPixel3ms(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer :: tnum
  tnum = 0

!$ tnum=omp_get_thread_num()

  select case(tnum)
  case(:-1)
    call vsmError("PaintPixel3ms less than 0")
  case(0)
    call vsmPaintPixel3mB0(pxbuf, ps, pe, pxwl, pxhl, width)
  case(1)
    call vsmPaintPixel3mB1(pxbuf, ps, pe, pxwl, pxhl, width)
  case(2)
    call vsmPaintPixel3mB2(pxbuf, ps, pe, pxwl, pxhl, width)
  case(3)
    call vsmPaintPixel3mB3(pxbuf, ps, pe, pxwl, pxhl, width)
  case(4)
    call vsmPaintPixel3mB4(pxbuf, ps, pe, pxwl, pxhl, width)

  case(5:13)
    call vsmPaintPixel3mB5R(pxbuf, ps, pe, pxwl, pxhl, width, tnum)

  case(14:22)
    call vsmPaintPixel3mB5L(pxbuf, ps, pe, pxwl, pxhl, width, tnum)

  case(23:)
    call vsmError("PaintPixel3ms more than 6")
  end select

 end subroutine vsmPaintPixel3ms

 subroutine vsmPaintPixel3mB0(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i!, wch
  real :: dcolor
  integer :: tnuml2

  tnuml2 = vsmGetOMPNestL2()

!$omp parallel do private(i) num_threads(tnuml2)
  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb0")
  end do

!$omp parallel do private(i,wdth,zwdth,dcolor) num_threads(tnuml2)
  do i=1, pe
    wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1))
    zwdth = wdth+2
    if(pxbuf(4,i) < vsmImageData%image(zwdth)) then
      call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
      vsmImageData%image(wdth+1) = dcolor ! RGB
      vsmImageData%image(zwdth) = pxbuf(4,i)
    end if
  end do

 end subroutine vsmPaintPixel3mB0


 subroutine vsmPaintPixel3mB1(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  real :: dcolor
  integer :: tnuml2

  tnuml2 = vsmGetOMPNestL2()

!$omp parallel do private(i) num_threads(tnuml2)
  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb1")
  end do

  wch = vsmImageData%width*vsmImageData%height

!$omp parallel do private(i,wdth,zwdth,dcolor) num_threads(tnuml2)
  do i=1, pe
    wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + 2*wch
    zwdth = wdth+2
    if(pxbuf(4,i) < vsmImageData%image(zwdth)) then
      call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
      vsmImageData%image(wdth+1) = dcolor ! RGB
      vsmImageData%image(zwdth) = pxbuf(4,i)
    end if
  end do

 end subroutine vsmPaintPixel3mB1


 subroutine vsmPaintPixel3mB2(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i
  real :: dcolor
  integer :: tnuml2

  tnuml2 = vsmGetOMPNestL2()

!$omp parallel do private(i) num_threads(tnuml2)
  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb2")
  end do

!$omp parallel do private(i,wdth,zwdth,dcolor) num_threads(tnuml2)
  do i=1, pe
    wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1))
    zwdth = wdth+2
    if(pxbuf(4,i) < vsmImageData%image_r(zwdth)) then
      call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
      vsmImageData%image_r(wdth+1) = dcolor ! RGB
      vsmImageData%image_r(zwdth) = pxbuf(4,i)
    end if
  end do

 end subroutine vsmPaintPixel3mB2

 subroutine vsmPaintPixel3mB3(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  real :: dcolor
  integer :: tnuml2

  tnuml2 = vsmGetOMPNestL2()

!$omp parallel do private(i) num_threads(tnuml2)
  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb3")
  end do

  wch = vsmImageData%width*vsmImageData%height

!$omp parallel do private(i,wdth,zwdth,dcolor) num_threads(tnuml2)
  do i=1, pe
    wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + 2*wch
    zwdth = wdth+2
    if(pxbuf(4,i) < vsmImageData%image_r(zwdth)) then
      call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
      vsmImageData%image_r(wdth+1) = dcolor ! RGB
      vsmImageData%image_r(zwdth) = pxbuf(4,i)
    end if
  end do

 end subroutine vsmPaintPixel3mB3

 subroutine vsmPaintPixel3mB4(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  real :: dcolor
  integer :: tnuml2

  tnuml2 = vsmGetOMPNestL2()

!$omp parallel do private(i) num_threads(tnuml2)
  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb4")
  end do

  wch = vsmImageData%width*vsmImageData%height

!$omp parallel do private(i,wdth,zwdth,dcolor) num_threads(tnuml2)
  do i=1, pe
    wdth = (px(1,i)-1) + vsmImageData%width * (px(2,i)-1) + 4*wch
    zwdth = wdth+1
    if(pxbuf(4,i) < vsmImageData%image_r(zwdth)) then
      call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
      vsmImageData%image(zwdth) = dcolor ! RGB
      vsmImageData%image_r(zwdth) = pxbuf(4,i)
    end if
  end do

 end subroutine vsmPaintPixel3mB4

 subroutine vsmPaintPixel3mB5R(pxbuf, ps, pe, pxwl, pxhl, width, tnum)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width, tnum
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  integer :: dcolor
  integer :: idis
  real :: fdis
  integer :: tnuml2

  if(tnum < 5 .or. tnum > 13) call vsmError("pixel3mb5r")
  tnuml2 = vsmGetOMPNestL2()

!$omp parallel do private(i) num_threads(tnuml2)
  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb5")
  end do

  wch = vsmImageData%width*vsmImageData%height

!$omp parallel do private(i,idis,fdis,wdth,zwdth,dcolor) num_threads(tnuml2)
  do i=1, pe
    wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + wch*(tnum-5)*2
    zwdth = wdth+2
    idis = vsmImageData%rsdata_r(zwdth)
#ifndef CTRANS
    fdis = transfer(idis, 1.0)
#else
     call VSMTRANSFER4I2F(idis, fdis)
#endif

    if(pxbuf(4,i) < fdis) then
      call vsmPack3SP21INT(pxbuf(1:3, i), dcolor)
      vsmImageData%rsdata_r(wdth+1) = dcolor ! RGB
#ifndef CTRANS
      idis = transfer(pxbuf(4,i), 1)
#else
     call VSMTRANSFER4F2I(pxbuf(4,i), idis)
#endif
      vsmImageData%rsdata_r(zwdth) = idis
    end if
  end do

 end subroutine vsmPaintPixel3mB5R

 subroutine vsmPaintPixel3mB5L(pxbuf, ps, pe, pxwl, pxhl, width, tnum)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width, tnum
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  integer :: dcolor
  integer :: idis
  real :: fdis
  integer :: tnuml2

  if(tnum < 14 .or. tnum > 22) call vsmError("pixel3mb5l")
  tnuml2 = vsmGetOMPNestL2()

!$omp parallel do private(i) num_threads(tnuml2)
  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb5")
  end do

  wch = vsmImageData%width*vsmImageData%height

!$omp parallel do private(i,idis,fdis,wdth,zwdth,dcolor) num_threads(tnuml2)
  do i=1, pe
    wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + wch*(tnum-14)*2
    zwdth = wdth+2
    idis = vsmImageData%rsdata(zwdth)
#ifndef CTRANS
    fdis = transfer(idis, 1.0)
#else
     call VSMTRANSFER4I2F(idis, fdis)
#endif

    if(pxbuf(4,i) < fdis) then
      call vsmPack3SP21INT(pxbuf(1:3, i), dcolor)
      vsmImageData%rsdata(wdth+1) = dcolor ! RGB
#ifndef CTRANS
      idis = transfer(pxbuf(4,i), 1)
#else
     call VSMTRANSFER4F2I(pxbuf(4,i), idis)
#endif
      vsmImageData%rsdata(zwdth) = idis
    end if
  end do

 end subroutine vsmPaintPixel3mB5L
#else
 subroutine vsmPaintPixel3ms(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer :: tnum, omp_get_thread_num!,num
  tnum = 0

!$ tnum=omp_get_thread_num()
tnum = mod(tnum, vsmMaxOMP4Solid)

  select case(tnum)
  case(:-1)
    call vsmError("PaintPixel3ms less than 0")
  case(0)
    call vsmPaintPixel3mB0(pxbuf, ps, pe, pxwl, pxhl, width)
  case(1)
    call vsmPaintPixel3mB1(pxbuf, ps, pe, pxwl, pxhl, width)
  case(2)
    call vsmPaintPixel3mB2(pxbuf, ps, pe, pxwl, pxhl, width)
  case(3)
    call vsmPaintPixel3mB3(pxbuf, ps, pe, pxwl, pxhl, width)
  case(4)
    call vsmPaintPixel3mB4(pxbuf, ps, pe, pxwl, pxhl, width)

  case(5:13)
    call vsmPaintPixel3mB5R(pxbuf, ps, pe, pxwl, pxhl, width, tnum)

  case(14:22)
    call vsmPaintPixel3mB5L(pxbuf, ps, pe, pxwl, pxhl, width, tnum)

  case(23:)
    call vsmError("PaintPixel3ms more than 6")
  end select

 end subroutine vsmPaintPixel3ms

 subroutine vsmPaintPixel3mB0(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i!, wch
  real :: dcolor
  integer :: onum
  onum = 1
!$ onum = omp_get_max_threads()

  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb0")
  end do

  if(onum <= vsmMaxOMP4Solid) then
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1))
      zwdth = wdth+2
      if(pxbuf(4,i) < vsmImageData%image(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image(wdth+1) = dcolor ! RGB
        vsmImageData%image(zwdth) = pxbuf(4,i)
      end if
    end do
  else
!$omp critical (vismo_pp0)
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1))
      zwdth = wdth+2
      if(pxbuf(4,i) < vsmImageData%image(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image(wdth+1) = dcolor ! RGB
        vsmImageData%image(zwdth) = pxbuf(4,i)
      end if
    end do
!$omp end critical (vismo_pp0)
  end if

 end subroutine vsmPaintPixel3mB0


 subroutine vsmPaintPixel3mB1(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  real :: dcolor
  integer :: onum
  onum = 1
!$ onum = omp_get_max_threads()

  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb1")
  end do

  wch = vsmImageData%width*vsmImageData%height

  if(onum <= vsmMaxOMP4Solid+1) then
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + 2*wch
      zwdth = wdth+2
      if(pxbuf(4,i) < vsmImageData%image(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image(wdth+1) = dcolor ! RGB
        vsmImageData%image(zwdth) = pxbuf(4,i)
      end if
    end do
  else
!$omp critical (vismo_pp1)
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + 2*wch
      zwdth = wdth+2
      if(pxbuf(4,i) < vsmImageData%image(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image(wdth+1) = dcolor ! RGB
        vsmImageData%image(zwdth) = pxbuf(4,i)
      end if
    end do
!$omp end critical (vismo_pp1)
  end if

 end subroutine vsmPaintPixel3mB1


 subroutine vsmPaintPixel3mB2(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i!, wch
  real :: dcolor
  integer :: onum
  onum = 1
!$ onum = omp_get_max_threads()

  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb2")
  end do

  if(onum <= vsmMaxOMP4Solid+2) then
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1))
      zwdth = wdth+2
      if(pxbuf(4,i) < vsmImageData%image_r(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image_r(wdth+1) = dcolor ! RGB
        vsmImageData%image_r(zwdth) = pxbuf(4,i)
      end if
    end do
  else
!$omp critical (vismo_pp2)
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1))
      zwdth = wdth+2
      if(pxbuf(4,i) < vsmImageData%image_r(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image_r(wdth+1) = dcolor ! RGB
        vsmImageData%image_r(zwdth) = pxbuf(4,i)
      end if
    end do
!$omp end critical (vismo_pp2)
  end if
 end subroutine vsmPaintPixel3mB2

 subroutine vsmPaintPixel3mB3(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  real :: dcolor
  integer :: onum
  onum = 1
!$ onum = omp_get_max_threads()

  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb3")
  end do

  wch = vsmImageData%width*vsmImageData%height

  if(onum <= vsmMaxOMP4Solid+3) then
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + 2*wch
      zwdth = wdth+2
      if(pxbuf(4,i) < vsmImageData%image_r(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image_r(wdth+1) = dcolor ! RGB
        vsmImageData%image_r(zwdth) = pxbuf(4,i)
      end if
    end do
  else
!$omp critical (vismo_pp3)
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + 2*wch
      zwdth = wdth+2
      if(pxbuf(4,i) < vsmImageData%image_r(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image_r(wdth+1) = dcolor ! RGB
        vsmImageData%image_r(zwdth) = pxbuf(4,i)
      end if
    end do
!$omp end critical (vismo_pp3)
  end if

 end subroutine vsmPaintPixel3mB3


 subroutine vsmPaintPixel3mB4(pxbuf, ps, pe, pxwl, pxhl, width)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  real :: dcolor
  integer :: onum
  onum = 1
!$ onum = omp_get_max_threads()

  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb4")
  end do

  wch = vsmImageData%width*vsmImageData%height

  if(onum <= vsmMaxOMP4Solid+4) then
    do i=1, pe
      wdth = (px(1,i)-1) + vsmImageData%width * (px(2,i)-1) + 4*wch
      zwdth = wdth+1
      if(pxbuf(4,i) < vsmImageData%image_r(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image(zwdth) = dcolor ! RGB
        vsmImageData%image_r(zwdth) = pxbuf(4,i)
      end if
    end do
  else
!$omp critical (vismo_pp4)
    do i=1, pe
      wdth = (px(1,i)-1) + vsmImageData%width * (px(2,i)-1) + 4*wch
      zwdth = wdth+1
      if(pxbuf(4,i) < vsmImageData%image_r(zwdth)) then
        call vsmPack3SP21SP(pxbuf(1:3, i), dcolor)
        vsmImageData%image(zwdth) = dcolor ! RGB
        vsmImageData%image_r(zwdth) = pxbuf(4,i)
      end if
    end do
!$omp end critical (vismo_pp4)
  end if

 end subroutine vsmPaintPixel3mB4

 subroutine vsmPaintPixel3mB5R(pxbuf, ps, pe, pxwl, pxhl, width, tnum)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width, tnum
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  integer :: dcolor
  integer :: idis
  real :: fdis
  integer :: onum
  onum = 1
!$ onum = omp_get_max_threads()

  if(tnum < 5 .or. tnum > 13) call vsmError("pixel3mb5r")
  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb5")
  end do

  wch = vsmImageData%width*vsmImageData%height

  if(onum <= vsmMaxOMP4Solid+5) then
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + wch*(tnum-5)*2
      zwdth = wdth+2
      idis = vsmImageData%rsdata_r(zwdth)
#ifndef CTRANS
      fdis = transfer(idis, 1.0)
#else
      call VSMTRANSFER4I2F(idis, fdis)
#endif

      if(pxbuf(4,i) < fdis) then
        call vsmPack3SP21INT(pxbuf(1:3, i), dcolor)
        vsmImageData%rsdata_r(wdth+1) = dcolor ! RGB
#ifndef CTRANS
        idis = transfer(pxbuf(4,i), 1)
#else
        call VSMTRANSFER4F2I(pxbuf(4,i), idis)
#endif
        vsmImageData%rsdata_r(zwdth) = idis
      end if
    end do
  else
!$omp critical (vismo_pp5R)
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + wch*(tnum-5)*2
      zwdth = wdth+2
      idis = vsmImageData%rsdata_r(zwdth)
#ifndef CTRANS
      fdis = transfer(idis, 1.0)
#else
      call VSMTRANSFER4I2F(idis, fdis)
#endif

      if(pxbuf(4,i) < fdis) then
        call vsmPack3SP21INT(pxbuf(1:3, i), dcolor)
        vsmImageData%rsdata_r(wdth+1) = dcolor ! RGB
#ifndef CTRANS
        idis = transfer(pxbuf(4,i), 1)
#else
        call VSMTRANSFER4F2I(pxbuf(4,i), idis)
#endif
        vsmImageData%rsdata_r(zwdth) = idis
      end if
    end do
!$omp end critical (vismo_pp5R)
  end if

 end subroutine vsmPaintPixel3mB5R

 subroutine vsmPaintPixel3mB5L(pxbuf, ps, pe, pxwl, pxhl, width, tnum)
  real, dimension(4, vsmPbLen), intent(in) :: pxbuf
  integer, intent(in) :: ps, pe, pxwl, pxhl, width, tnum
  integer, dimension(2, vsmPbLen) :: px
  integer :: wdth, zwdth, i, wch
  integer :: dcolor
  integer :: idis
  real :: fdis
  integer :: onum
  onum = 1
!$ onum = omp_get_max_threads()

  if(tnum < 14 .or. tnum > 22) call vsmError("pixel3mb5l")
  do i=0, pe-1
    px(1, i+1) = pxwl + mod(i+ps-1, width)
    px(2, i+1) = pxhl +(i+ps-1)/width
    if(.not.vsmCheckPX(px(1:2,i+1))) call vsmError("vsmPaintPixel3mb5")
  end do

  wch = vsmImageData%width*vsmImageData%height

  if(onum <= vsmMaxOMP4Solid+14) then
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + wch*(tnum-14)*2
      zwdth = wdth+2
      idis = vsmImageData%rsdata(zwdth)
#ifndef CTRANS
      fdis = transfer(idis, 1.0)
#else
      call VSMTRANSFER4I2F(idis, fdis)
#endif

      if(pxbuf(4,i) < fdis) then
        call vsmPack3SP21INT(pxbuf(1:3, i), dcolor)
        vsmImageData%rsdata(wdth+1) = dcolor ! RGB
#ifndef CTRANS
        idis = transfer(pxbuf(4,i), 1)
#else
        call VSMTRANSFER4F2I(pxbuf(4,i), idis)
#endif
        vsmImageData%rsdata(zwdth) = idis
      end if
    end do
  else
!$omp critical (vismo_pp5L)
    do i=1, pe
      wdth = 2*((px(1,i)-1) + vsmImageData%width * (px(2,i)-1)) + wch*(tnum-14)*2
      zwdth = wdth+2
      idis = vsmImageData%rsdata(zwdth)
#ifndef CTRANS
      fdis = transfer(idis, 1.0)
#else
      call VSMTRANSFER4I2F(idis, fdis)
#endif

      if(pxbuf(4,i) < fdis) then
        call vsmPack3SP21INT(pxbuf(1:3, i), dcolor)
        vsmImageData%rsdata(wdth+1) = dcolor ! RGB
#ifndef CTRANS
        idis = transfer(pxbuf(4,i), 1)
#else
        call VSMTRANSFER4F2I(pxbuf(4,i), idis)
#endif
        vsmImageData%rsdata(zwdth) = idis
      end if
    end do
!$omp end critical (vismo_pp5L)
  end if

 end subroutine vsmPaintPixel3mB5L

#endif

 subroutine vsmPaintPixel3B1(px, col, distd)
  integer, dimension(2), intent(in) :: px
  real, dimension(:), intent(in) :: col
  real(kind=vdp), intent(in) :: distd
  integer :: wdth, zwdth
  real :: dist

  dist = real(distd)

  if(.not.vsmCheckPX(px)) call vsmError("vsmPaintPixel3s")
  
  wdth  = 5*((px(1)-1) + vsmImageData%width * (px(2)-1))
  zwdth = wdth + 5

  if(dist >= vsmImageData%image_r(zwdth)) return
  if(dist >= vsmImageData%image(zwdth)) return

!$omp critical (vsm_solidB)
    if(dist < vsmImageData%image_r(zwdth)) then
      vsmImageData%image_r(zwdth) = dist
      vsmImageData%image_r(wdth+1:wdth+3) = col(1:3) ! RGB
      vsmImageData%image_r(wdth+4)         = 1.0 ! A
    end if
!$omp end critical (vsm_solidB)

 end subroutine vsmPaintPixel3B1


 subroutine vsmPaintPixel3B0(px, col, distd)
  integer, dimension(2), intent(in) :: px
  real, dimension(:), intent(in) :: col
  real(kind=vdp), intent(in) :: distd
  integer :: wdth, zwdth
  real :: dist

  dist = real(distd)

  if(.not.vsmCheckPX(px)) call vsmError("vsmPaintPixel3s")
  
  wdth = 5*((px(1)-1) + vsmImageData%width * (px(2)-1))
  zwdth = wdth+5

  if(dist >= vsmImageData%image(zwdth)) return
  if(dist >= vsmImageData%image_r(zwdth)) return

!$omp critical (vsm_solidA)
    if(dist < vsmImageData%image(zwdth)) then
      vsmImageData%image(zwdth) = dist
      vsmImageData%image(wdth+1:wdth+3) = col(1:3) ! RGB
      vsmImageData%image(wdth+4)         = 1.0 ! A
    end if
!$omp end critical (vsm_solidA)

 end subroutine vsmPaintPixel3B0

 subroutine vsmClearPixel(px)
  integer, dimension(2), intent(in) :: px
  integer :: wdth

  if(.not.vsmCheckPX(px)) call vsmError("vsmClearPixel")

  wdth = 5*((px(1)-1) + vsmImageData%width * (px(2)-1))
  vsmImageData%image(wdth+1:wdth+4) = 0.0 ! RGB
  vsmImageData%image(wdth+5) = real(vsmFarAway)
 end subroutine vsmClearPixel

 subroutine vsmCheckImage
  integer :: i, j, p, er
  integer, dimension(2) :: px

  call vsmLogPrint("check image")
!$omp parallel do private(i,j,p,px,er)
  do j=1, vsmImageData%height
    do i=1, vsmImageData%width
      er = 0
      p = (vsmImageData%width * (j-1) + (i-1))*5
      px(1) = i; px(2)=j
      if(vsmImageData%image(p+1) < 0.0 .or. vsmImageData%image(p+2) < 0.0 .or.&
        vsmImageData%image(p+3) < 0.0 .or. vsmImageData%image(p+4) < 0.0 .or.&
        vsmImageData%image(p+4) > 1.0) then
        call vsmLogPrint("color4 is strange ",px)
        call vsmLogPrint("col=", vsmImageData%image(p+1:p+4))
        er = 1
      end if

      if(vsmImageData%image(p+5) < 0.0 .or. vsmImageData%image(p+5) > real(vsmFarAway)) then
        call vsmLogPrint("Z is strange ",px)
        call vsmLogPrint("Z=",vsmImageData%image(p+5))
        er = 1
      end if

      if(er /= 0) then
        call vsmError("image data is wrong: CheckImage")
      end if
    end do
  end do

  call vsmLogPrint("check image ends")
 end subroutine

 subroutine vsmCheckImageL(stpixel, edpixel)
  integer, intent(in) :: stpixel, edpixel
  integer :: i, er, p

  call vsmLogPrint("check image L")

!$omp parallel do private(p,er)
    do i=stpixel, edpixel
      er=0
      p = (i-1)*5
      if(vsmImageData%image(p+1) < 0.0 .or. vsmImageData%image(p+2) < 0.0 .or.&
        vsmImageData%image(p+3) < 0.0 .or. vsmImageData%image(p+4) < 0.0 .or.&
        vsmImageData%image(p+4) > 1.0) then
        call vsmLogPrint("color4 is strange ",p)
        call vsmLogPrint("col=", vsmImageData%image(p+1:p+4))
        er = 1
      end if

      if(vsmImageData%image(p+5) < 0.0 .or. vsmImageData%image(p+5) > real(vsmFarAway)) then
        call vsmLogPrint("Z is strange ",p)
        call vsmLogPrint("Z=",vsmImageData%image(p+5))
        er = 1
      end if

      if(er /= 0) then
        call vsmError("image data is wrong: CheckImageL")
      end if
    end do
  call vsmLogPrint("check image L ends")
 end subroutine

 subroutine vsmGetPixelColor4(px, color)
   integer, dimension(2), intent(in) :: px
   real, dimension(4), intent(out) :: color
   integer :: wdth

   if(.not.vsmCheckPX(px)) call vsmError("Error :: vsmGetPixelColor4")
   wdth = 5*((px(1)-1) + vsmImageData%width * (px(2)-1))
   color(1:4) = vsmImageData%image(wdth+1:wdth+4)
 end subroutine
 
 function vsmGetZbuf(px)
   integer, dimension(2), intent(in) :: px
   real(kind=vdp) :: vsmGetZbuf
   integer :: wdth

   if(.not.vsmCheckPX(px)) call vsmError("vsmGetZbuf")
   wdth = (px(1)-1 + vsmImageData%width * (px(2)-1))*5
   vsmGetZbuf = dble(vsmImageData%image(wdth+5))
 end function vsmGetZbuf

end module vsmImage


