module vsmOctree4iso
  use vsmConst
  use vsmFunction
  use vsmCoordinates
  use vsmColormap
  use vsmData
  use vsmParam

  implicit none
  private

type vismo_OctNode4iso
  integer, dimension(3) :: vertex
  integer, dimension(3) :: OctDatSize
  !morton number  
  integer :: parent
  integer :: moi
  integer, dimension(0:7) :: children
  integer :: MyNumber ! morton number

  integer :: level ! level of octree
  logical :: skip
  real(kind=vdp) :: txmin, tymin, tzmin, txmax, tymax, tzmax
  integer, dimension(3) :: SkipVertexMin, SkipVertexMax
  real(kind=vdp), dimension(3) ::  SkipMin, SkipMax
 end type vismo_OctNode4iso

 type vismo_OctCommon4iso
  logical :: uniform
  integer :: maxLevel
  integer :: vis_num, Row
  real(kind=vdp) :: xmin, ymin, zmin
  real(kind=vdp), dimension(:), allocatable :: coordx, coordy, coordz
  logical, dimension(3) :: vis_on ! 1:isosurface, 2:volume rendering, 3:slices
 end type vismo_OctCommon4iso

 type(vismo_OctNode4iso), dimension(0:vsmMaxOctNum), save :: vsmOctant4iso
 type(vismo_OctCommon4iso), save :: vsmOctComDat4iso

 public vsmInitOctree4iso, vsmFinalizeOctree4iso, vsmOctDivide4iso, vsmOctPrepare4iso, vsmRayJump4iso
 public vsmSetOctComDat4iso, vsmUpdateOctree4iso, vsmyyRayJump4iso

 contains

 subroutine vsmInitOctree4iso
  integer :: i

  do i=0, vsmMaxOctNum-1
    vsmOctant4iso(i)%children(0:7) = -1

    vsmOctant4iso(i)%parent = -1
    vsmOctant4iso(i)%MyNumber = -1
    vsmOctant4iso(i)%skip = .false.

    vsmOctant4iso(i)%vertex(1:3) = -1
    vsmOctant4iso(i)%SkipVertexMin(1:3) = -1
    vsmOctant4iso(i)%SkipVertexMax(1:3) = -1
    vsmOctant4iso(i)%SkipMin(1:3) = -1.0d+10
    vsmOctant4iso(i)%SkipMax(1:3) = -1.0d+10

    vsmOctant4iso(i)%txmin = 1.0d+10;  vsmOctant4iso(i)%txmax = -1.0d+10
    vsmOctant4iso(i)%tymin = 1.0d+10;  vsmOctant4iso(i)%tymax = -1.0d+10
    vsmOctant4iso(i)%tzmin = 1.0d+10;  vsmOctant4iso(i)%tzmax = -1.0d+10
  end do

  vsmOctComDat4iso%vis_num      = 0
  vsmOctComDat4iso%uniform      = .true.
  vsmOctComDat4iso%maxLevel     = vsmMaxOctLevel - 1
  vsmOctComDat4iso%vis_on(1:3)  = .false.

  vsmOctComDat4iso%xmin = 1.0d+10
  vsmOctComDat4iso%ymin = 1.0d+10
  vsmOctComDat4iso%zmin = 1.0d+10

  vsmOctComDat4iso%Row  = 0

  call vsmLogPrint("Octree: vsmOctComDat%maxLevel", vsmOctComDat4iso%maxLevel)
  call vsmLogPrint("Octree Initialized")
 end subroutine vsmInitOctree4iso

 subroutine vsmSetOctComDat4iso(vn)
   integer, intent(in) :: vn

   vsmOctComDat4iso%vis_num = vn

   if(vsmVisParam(vn)%iso_num > 0 ) then
     vsmOctComDat4iso%vis_on(1) = .true.
   else
     vsmOctComDat4iso%vis_on(1) = .false.
   end if

   if(vsmVisParam(vn)%volren_num > 0 ) then
     vsmOctComDat4iso%vis_on(2) = .true.
   else
     vsmOctComDat4iso%vis_on(2) = .false.
   end if

   if(vsmVisParam(vn)%slice_num > 0 ) then
     vsmOctComDat4iso%vis_on(3) = .true.
   else
     vsmOctComDat4iso%vis_on(3) = .false.
   end if

 end subroutine vsmSetOctComDat4iso

 subroutine vsmFinalizeOctree4iso
   if(allocated(vsmOctComDat4iso%coordx)) deallocate(vsmOctComDat4iso%coordx)
   if(allocated(vsmOctComDat4iso%coordy)) deallocate(vsmOctComDat4iso%coordy)
   if(allocated(vsmOctComDat4iso%coordz)) deallocate(vsmOctComDat4iso%coordz)
 end subroutine vsmFinalizeOctree4iso

 subroutine vsmRayJump4iso(pos, dirc, ray_, r_skip)
  real(kind=vdp), dimension(3), intent(in) :: pos, dirc
  real(kind=vdp), intent(out) :: ray_
  logical, intent(out) :: r_skip
  integer :: i, nd
  integer :: div_num, integ_m!, count_
  integer :: intx, inty, intz
  integer :: mnumray

  !call vsmLogPrint("Octree's vsmRayJump is called")
  intx = -1; inty = -1; intz = -1
  r_skip = .false.

  if(.not.vsmIsInsidelWOE(pos)) then
    call vsmLogPrint("RayJump: out of data:", pos)
    call vsmError("RayJump: input pos is strange")
  end if

  div_num = vsmOctComDat4iso%Row
  !div_num = ishft(div_num, vsmOctComDat4iso%maxLevel)

!  integ_m = 1; count_ = 1
!  do i=1, vsmOctComDat4iso%maxLevel-1
!    count_ = count_ * 8
!    integ_m = integ_m + count_
!  end do

  if(vsmOctComDat4iso%coordx(div_num) == pos(1)) then
    intx = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat4iso%coordx(i) <= pos(1)) then
        intx = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat4iso%coordy(div_num) == pos(2)) then
    inty = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat4iso%coordy(i) <= pos(2)) then
        inty = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat4iso%coordz(div_num) == pos(3)) then
    intz = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat4iso%coordz(i) <= pos(3)) then
        intz = i
        exit
      end if
    end do
  end if

  if(intx == -1 .or. inty == -1 .or. intz == -1) then
    if(vsmDbg) then
      do i=0, div_num
        call vsmLogPrint("Octree:i=", i)
        call vsmLogPrint("Octree:coordx ", vsmOctComDat4iso%coordx(i))
        call vsmLogPrint("Octree:coordy ", vsmOctComDat4iso%coordy(i))
        call vsmLogPrint("Octree:coordz ", vsmOctComDat4iso%coordz(i))
      end do
    end if
    call vsmError("RayJump: intx,y or z is negative")
  end if

  if(vsmOctComDat4iso%maxLevel > 0) then
    integ_m = vsmOctNum(vsmOctComDat4iso%maxLevel-1)
  else
    integ_m = 0
  end if
  mnumray =  vsmGetMortonNum4iso(intx, inty, intz)
  nd = mnumray+integ_m

  if(vsmMaxOctNum <= nd ) call vsmError("OctNum: vsmRayJump")
  call vsmRaySkip4iso(pos, dirc, nd, ray_, r_skip)

  !call vsmLogPrint("Octree's vsmRayJump ended")
 end subroutine vsmRayJump4iso

 subroutine vsmyyRayJump4iso(pos, spos, dirc, ray_, r_skip)
  real(kind=vdp), dimension(3), intent(in) :: pos, dirc, spos
  real(kind=vdp), intent(out) :: ray_
  logical, intent(out) :: r_skip
  integer :: i, nd
  integer :: div_num, integ_m!, count_
  integer :: intx, inty, intz
  integer :: mnumray

  !call vsmLogPrint("Octree's vsmRayJump is called")
  intx = -1; inty = -1; intz = -1
  r_skip = .false.

  if(.not.vsmIsInsidelWOE(spos)) then
    call vsmLogPrint("RayJump: out of data:", pos)
    call vsmError("RayJump: input pos is strange")
  end if

  div_num = vsmOctComDat4iso%Row
  !div_num = ishft(div_num, vsmOctComDat4iso%maxLevel)

!  integ_m = 1; count_ = 1
!  do i=1, vsmOctComDat4iso%maxLevel-1
!    count_ = count_ * 8
!    integ_m = integ_m + count_
!  end do

  if(vsmOctComDat4iso%coordx(div_num) == spos(1)) then
    intx = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat4iso%coordx(i) <= spos(1)) then
        intx = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat4iso%coordy(div_num) == spos(2)) then
    inty = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat4iso%coordy(i) <= spos(2)) then
        inty = i
        exit
      end if
    end do
  end if

  if(vsmOctComDat4iso%coordz(div_num) == spos(3)) then
    intz = div_num-1 
  else
    do i=div_num-1, 0, -1
      if(vsmOctComDat4iso%coordz(i) <= spos(3)) then
        intz = i
        exit
      end if
    end do
  end if

  if(intx == -1 .or. inty == -1 .or. intz == -1) then
    if(vsmDbg) then
      do i=0, div_num
        call vsmLogPrint("Octree:i=", i)
        call vsmLogPrint("Octree:coordx ", vsmOctComDat4iso%coordx(i))
        call vsmLogPrint("Octree:coordy ", vsmOctComDat4iso%coordy(i))
        call vsmLogPrint("Octree:coordz ", vsmOctComDat4iso%coordz(i))
      end do
    end if
    call vsmError("RayJump: intx,y or z is negative")
  end if

  if(vsmOctComDat4iso%maxLevel > 0) then
    integ_m = vsmOctNum(vsmOctComDat4iso%maxLevel-1)
  else
    integ_m = 0
  end if
  mnumray =  vsmGetMortonNum4iso(intx, inty, intz)
  nd = mnumray+integ_m

  if(vsmMaxOctNum <= nd ) call vsmError("OctNum: vsmRayJump")
  call vsmyyRaySkip4iso(pos, spos, dirc, nd, ray_, r_skip)

  !call vsmLogPrint("Octree's vsmRayJump ended")
 end subroutine vsmyyRayJump4iso

 subroutine vsmRaySkipVoxel4iso(i0, pos, dirc, ray_, r_skip)
  integer, dimension(3), intent(in) :: i0
  real(kind=vdp), dimension(3), intent(in) :: pos
  real(kind=vdp), dimension(3), intent(in) :: dirc
  real(kind=vdp), intent(out) :: ray_
  logical, intent(out) :: r_skip

  integer :: i, j, k, ip1,jp1,kp1,im1,jm1,km1
  real(kind=vdp), dimension(6) :: a
  real(kind=vdp), dimension(3) :: cv_min, cv_max
  integer :: vis_num
  integer :: n, nscal, l
  real(kind=vdp) :: iso_level
  real(kind=vdp) :: max_val, min_val, b
  real(kind=vdp), dimension(8) :: dat
!  real, dimension(8) :: alpha
  real :: maxalpha
  real(kind=vdp), dimension(3, 8) :: tpos

  vis_num = vsmOctComDat4iso%vis_num
  i=i0(1); j = i0(2); k=i0(3)
  r_skip = .false.

  ip1 = i+1; jp1 = j+1; kp1 = k+1
  im1=i-1; jm1=j-1; km1=k-1

  if(.not.vsmCheckLCoord(i, j, k)) call vsmError("Error:: vsmRaySkipVoxel")
  if(.not.vsmCheckLCoord(ip1, jp1, kp1)) call vsmError("Error:: vsmRaySkipVoxel2")

!  if(i == node%OctDatSize(1)) i = node%OctDatSize(1) - 1
!  if(j == node%OctDatSize(2)) j = node%OctDatSize(2) - 1
!  if(k == node%OctDatSize(3)) k = node%OctDatSize(3) - 1

!  if(i < 1 .or. i > node%OctDatSize(1) - 1) call vsmError("RaySkip voxel : i is strange")
!  if(j < 1 .or. j > node%OctDatSize(2) - 1) call vsmError("RaySkip voxel : j is strange")
!  if(k < 1 .or. k > node%OctDatSize(3) - 1) call vsmError("RaySkip voxel : k is strange")

  if(vsmOctComDat4iso%vis_on(1)) then
    do n=1, vsmVisParam(vis_num)%iso_num
      nscal = vsmVisParam(vis_num)%isoparam(n)%nscal
      iso_level = vsmVisParam(vis_num)%isoparam(n)%iso_level

      dat(1) = vsmScalData(nscal)%scal(i,   j,   k)
      dat(2) = vsmScalData(nscal)%scal(ip1, j,   k)
      dat(3) = vsmScalData(nscal)%scal(i,   jp1, k)
      dat(4) = vsmScalData(nscal)%scal(ip1, jp1, k)
      dat(5) = vsmScalData(nscal)%scal(i,   j,   kp1)
      dat(6) = vsmScalData(nscal)%scal(ip1, j,   kp1)
      dat(7) = vsmScalData(nscal)%scal(i,   jp1, kp1)
      dat(8) = vsmScalData(nscal)%scal(ip1, jp1, kp1)

      max_val = maxval(dat); min_val = minval(dat)

      if(min_val <= iso_level .and. max_val >= iso_level) then
        r_skip = .false. 
        return
      end if
    end do
  end if

 if(vsmOctComDat4iso%vis_on(2)) then
   do n=1, vsmVisParam(vis_num)%volren_num
     nscal = vsmVisParam(vis_num)%volparam(n)%nscal

      dat(1) = vsmScalData(nscal)%scal(i,   j,   k)
      dat(2) = vsmScalData(nscal)%scal(ip1, j,   k)
      dat(3) = vsmScalData(nscal)%scal(i,   jp1, k)
      dat(4) = vsmScalData(nscal)%scal(ip1, jp1, k)
      dat(5) = vsmScalData(nscal)%scal(i,   j,   kp1)
      dat(6) = vsmScalData(nscal)%scal(ip1, j,   kp1)
      dat(7) = vsmScalData(nscal)%scal(i,   jp1, kp1)
      dat(8) = vsmScalData(nscal)%scal(ip1, jp1, kp1)

     maxalpha = vsmGetMaxAlpha(dat, nscal)

     if(maxalpha > vsmSkipAlpha) then
        r_skip = .false. 
        return
     end if
   end do
 end if

 if(vsmOctComDat4iso%vis_on(3)) then
   do n=1, vsmVisParam(vis_num)%slice_num
     if(vsmOctComDat4iso%uniform) then
       tpos(1,1) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(im1)
       tpos(2,1) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(jm1)
       tpos(3,1) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(km1)

       tpos(1,2) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(i  )
       tpos(2,2) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(jm1)
       tpos(3,2) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(km1)

       tpos(1,3) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(im1)
       tpos(2,3) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(j  )
       tpos(3,3) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(km1)

       tpos(1,4) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(i  )
       tpos(2,4) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(j  )
       tpos(3,4) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(km1)

       tpos(1,5) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(im1)
       tpos(2,5) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(jm1)
       tpos(3,5) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(k)

       tpos(1,6) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(i  )
       tpos(2,6) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(jm1)
       tpos(3,6) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(k)

       tpos(1,7) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(im1)
       tpos(2,7) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(j  )
       tpos(3,7) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(k)

       tpos(1,8) = vsmLocalCoord%origin(1)+vsmLocalCoord%dx(1)*(i  )
       tpos(2,8) = vsmLocalCoord%origin(2)+vsmLocalCoord%dx(2)*(j  )
       tpos(3,8) = vsmLocalCoord%origin(3)+vsmLocalCoord%dx(3)*(k)
     else
       tpos(1,1) = vsmLocalCoord%coordx(i)
       tpos(2,1) = vsmLocalCoord%coordy(j)
       tpos(3,1) = vsmLocalCoord%coordz(k)

       tpos(1,2) = vsmLocalCoord%coordx(ip1)
       tpos(2,2) = vsmLocalCoord%coordy(j)
       tpos(3,2) = vsmLocalCoord%coordz(k)

       tpos(1,3) = vsmLocalCoord%coordx(i)
       tpos(2,3) = vsmLocalCoord%coordy(jp1)
       tpos(3,3) = vsmLocalCoord%coordz(k)

       tpos(1,4) = vsmLocalCoord%coordx(ip1)
       tpos(2,4) = vsmLocalCoord%coordy(jp1)
       tpos(3,4) = vsmLocalCoord%coordz(k)

       tpos(1,5) = vsmLocalCoord%coordx(i)
       tpos(2,5) = vsmLocalCoord%coordy(j)
       tpos(3,5) = vsmLocalCoord%coordz(kp1)

       tpos(1,6) = vsmLocalCoord%coordx(ip1)
       tpos(2,6) = vsmLocalCoord%coordy(j)
       tpos(3,6) = vsmLocalCoord%coordz(kp1)

       tpos(1,7) = vsmLocalCoord%coordx(i)
       tpos(2,7) = vsmLocalCoord%coordy(jp1)
       tpos(3,7) = vsmLocalCoord%coordz(kp1)

       tpos(1,8) = vsmLocalCoord%coordx(ip1)
       tpos(2,8) = vsmLocalCoord%coordy(jp1)
       tpos(3,8) = vsmLocalCoord%coordz(kp1)
     end if

     do l=1, 8
       dat(l) = vsmIsAboveSlice(vis_num, n, tpos(1:3,l))
     end do

     min_val =  minval(dat)
     max_val =  maxval(dat)

     if(min_val*max_val <= 0.0_vdp) then
       r_skip = .false. 
       return
     end if

   end do
 end if

 if(vsmOctComDat4iso%uniform) then
   cv_min(1) = vsmLocalCoord%origin(1) + im1*vsmLocalCoord%dx(1)
   cv_max(1) = vsmLocalCoord%origin(1) + i*vsmLocalCoord%dx(1)

   cv_min(2) = vsmLocalCoord%origin(2) + jm1*vsmLocalCoord%dx(2)
   cv_max(2) = vsmLocalCoord%origin(2) + j*vsmLocalCoord%dx(2)

   cv_min(3) = vsmLocalCoord%origin(3) + km1*vsmLocalCoord%dx(3)
   cv_max(3) = vsmLocalCoord%origin(3) + k*vsmLocalCoord%dx(3)
 else
   cv_min(1) = vsmLocalCoord%coordx(i)
   cv_max(1) = vsmLocalCoord%coordx(ip1)

   cv_min(2) = vsmLocalCoord%coordy(j)
   cv_max(2) = vsmLocalCoord%coordy(jp1)

   cv_min(3) = vsmLocalCoord%coordz(k)
   cv_max(3) = vsmLocalCoord%coordz(kp1)
 end if

 if( pos(1) > cv_max(1) .or. pos(1) < cv_min(1) .or. & 
     pos(2) > cv_max(2) .or. pos(2) < cv_min(2) .or. &
     pos(3) > cv_max(3) .or. pos(3) < cv_min(3)) then
    call vsmLogPrint("Octree: Ray fail cvmin: ",cv_min)
    call vsmLogPrint("Octree: Ray fail cvmax: ",cv_max)
    call vsmLogPrint("Octree: Ray fail pos: ",pos)
    r_skip = .false.
    return
 end if

 a = -1.0_vdp

 if(dirc(1) /= 0.0_vdp) then
   a(1) = (cv_min(1)-pos(1))/dirc(1)
   a(2) = (cv_max(1)-pos(1))/dirc(1)
 end if
  
 if(dirc(2) /= 0.0_vdp) then
   a(3) = (cv_min(2)-pos(2))/dirc(2)
   a(4) = (cv_max(2)-pos(2))/dirc(2)
 end if

 if(dirc(3) /= 0.0_vdp) then
   a(5) = (cv_min(3)-pos(3))/dirc(3)
   a(6) = (cv_max(3)-pos(3))/dirc(3)
 end if

  where(a < 0.0_vdp)
    a = vsmFaraway
  end where

  b = minval(a)
  ray_ = vsmIntval%ray_interval * int(b/vsmIntval%ray_interval)

  if(int(b/vsmIntval%ray_interval) > 0 .and. ray_ == vsmIntval%ray_interval * b/vsmIntval%ray_interval ) then
     ray_ = vsmIntval%ray_interval * (int(b/vsmIntval%ray_interval)-1)
   end if

  r_skip = .true.
 end subroutine vsmRaySkipVoxel4iso

 recursive subroutine vsmUpdateOctree4iso(node)
   integer, intent(in) :: node
   logical :: r_skip
   integer :: i

!   call vsmLogPrint("Octree: Update Octree is called")
   vsmOctant4iso(node)%skip = .false.

   if(vsmOctant4iso(vsmOctant4iso(node)%parent)%skip) then
     !the parent of root is root itself. skip is always false
     vsmOctant4iso(node)%skip = .true.
     call vsmGetSkipVertex4iso(node, r_skip)

       if(vsmOctant4iso(node)%moi == 0) then
         call vsmLogPrint("Octree: in node%parent%skip root ", vsmOctant4iso(node)%skip)
       end if

   else if(SurveyData4iso(node) == 0) then
     vsmOctant4iso(node)%skip = .true.
     vsmOctant4iso(node)%SkipVertexMin(1:3) = vsmOctant4iso(node)%vertex(1:3)
     vsmOctant4iso(node)%SkipVertexMax(1:3) = vsmOctant4iso(node)%vertex(1:3) + vsmOctant4iso(node)%OctDatSize(1:3)-1

     if(vsmOctant4iso(node)%moi == 0) then
       call vsmLogPrint("Octree: in SurveyData root ", vsmOctant4iso(node)%skip)
     end if

   end if

   if(vsmOctant4iso(node)%moi == 0) call vsmLogPrint("Octree: root ", vsmOctant4iso(node)%skip)

   if(vsmOctant4iso(node)%skip) then

     if(vsmOctComDat4iso%uniform) then
       vsmOctant4iso(node)%SkipMin(1) = vsmOctComDat4iso%xmin + vsmLocalCoord%dx(1)*(vsmOctant4iso(node)%SkipVertexMin(1)-1)
       vsmOctant4iso(node)%SkipMax(1) = vsmOctComDat4iso%xmin + vsmLocalCoord%dx(1)*(vsmOctant4iso(node)%SkipVertexMax(1)-1)
       vsmOctant4iso(node)%SkipMin(2) = vsmOctComDat4iso%ymin + vsmLocalCoord%dx(2)*(vsmOctant4iso(node)%SkipVertexMin(2)-1)
       vsmOctant4iso(node)%SkipMax(2) = vsmOctComDat4iso%ymin + vsmLocalCoord%dx(2)*(vsmOctant4iso(node)%SkipVertexMax(2)-1)
       vsmOctant4iso(node)%SkipMin(3) = vsmOctComDat4iso%zmin + vsmLocalCoord%dx(3)*(vsmOctant4iso(node)%SkipVertexMin(3)-1)
       vsmOctant4iso(node)%SkipMax(3) = vsmOctComDat4iso%zmin + vsmLocalCoord%dx(3)*(vsmOctant4iso(node)%SkipVertexMax(3)-1)
     else
       vsmOctant4iso(node)%SkipMin(1) = vsmLocalCoord%coordx(vsmOctant4iso(node)%SkipVertexMin(1))
       vsmOctant4iso(node)%SkipMax(1) = vsmLocalCoord%coordx(vsmOctant4iso(node)%SkipVertexMax(1))
       vsmOctant4iso(node)%SkipMin(2) = vsmLocalCoord%coordy(vsmOctant4iso(node)%SkipVertexMin(2))
       vsmOctant4iso(node)%SkipMax(2) = vsmLocalCoord%coordy(vsmOctant4iso(node)%SkipVertexMax(2))
       vsmOctant4iso(node)%SkipMin(3) = vsmLocalCoord%coordz(vsmOctant4iso(node)%SkipVertexMin(3))
       vsmOctant4iso(node)%SkipMax(3) = vsmLocalCoord%coordz(vsmOctant4iso(node)%SkipVertexMax(3))
     end if

  else
    vsmOctant4iso(node)%SkipMin(1:3) = -1.0d+10
    vsmOctant4iso(node)%SkipMax(1:3) = -1.0d+10
  end if 
 
  if(vsmOctant4iso(node)%level == vsmOctComDat4iso%maxLevel) then
    !call vsmLogPrint("Level == maxLevel: vsmOctant4iso(node)%level", vsmOctant4iso(node)%level)
    !call vsmLogPrint("Level == maxLevel: vsmOctComDat4iso%maxLevel", vsmOctComDat4iso%maxLevel)
    !call vsmLogPrint("Octree: Update Octree ends", vsmOctant4iso(node)%moi)
    return
  end if

  do i=0, 7
    call vsmUpdateOctree4iso(vsmOctant4iso(node)%children(i))
  end do

  !call vsmLogPrint("Octree: Update Octree ends", node%moi)
 end subroutine vsmUpdateOctree4iso

 subroutine vsmOctPrepare4iso
  integer, dimension(3) :: vtx

  call vsmSetGrid4iso
  vtx(1:3) = 1
  call vsmSetVertex4iso(vtx, vsmLocalCoord%nx, 0)
 end subroutine vsmOctPrepare4iso

 recursive subroutine vsmOctDivide4iso(node)
  integer, intent(in) :: node
  logical :: root

  integer :: n, nx, ny, nz, b
  integer :: i, allocstat
!  logical :: r_skip
  integer, dimension(3) :: hsize, hsizep1
  integer, dimension(3, 0:7) :: child_vertex, child_size
  integer :: next_level
  integer :: Mnum
  integer :: integ_num!, tinteg_num

  if(node == 0) then
    root = .true.
  else
    root = .false.
  end if

  if(root) then
    call vsmLogPrint("vsmOctDivide is called")
   end if

  call vsmCalcMinMax4iso(node)

  LEVEL0 : if(root) then
    vsmOctant4iso(node)%level  = 0
    vsmOctant4iso(node)%parent = 0
    nx = vsmOctant4iso(node)%OctDatSize(1) - 1 ! num of control volume
    ny = vsmOctant4iso(node)%OctDatSize(2) - 1
    nz = vsmOctant4iso(node)%OctDatSize(3) - 1
    n  = vsmMaxOctLevel

    do i=1, vsmMaxOctLevel
      nx = nx/2; ny = ny/2; nz = nz/2
   
      if(vsmDbg) then
        call vsmLogPrint("nx = ",nx)
        call vsmLogPrint("ny = ",ny)
        call vsmLogPrint("nz = ",nz)
      end if

      if(nx < 1 .or. ny < 1 .or. nz < 1) exit
   
      n=n-1

      if(nx == 1 .or. ny == 1 .or. nz == 1) exit

    end do

    vsmOctComDat4iso%maxLevel = vsmMaxOctLevel - n
    call vsmLogPrint("Octree: MaxLevel = ",vsmOctComDat4iso%maxLevel)

    b = 1
    b = ishft(b, vsmOctComDat4iso%maxLevel)
  
    call vsmLogPrint("Octree: Row = ",b)
    vsmOctComDat4iso%Row = b
    allocate(vsmOctComDat4iso%coordx(0:b), vsmOctComDat4iso%coordy(0:b), vsmOctComDat4iso%coordz(0:b), stat = allocstat)
    if(allocstat /= 0) call vsmError('cant allocate octree coord')

    vsmOctant4iso(node)%moi = 0
    vsmOctant4iso(node)%MyNumber = 0

  end if LEVEL0
 
  if(vsmOctant4iso(node)%level == vsmOctComDat4iso%maxLevel) then
    ! call vsmLogPrint("vsmOctant4iso(node)level==vsmOctComDat4iso%maxLevel:", vsmOctant4iso(node)%level)
    return
  end if

  hsize(1:3) = vsmOctant4iso(node)%OctDatSize(1:3)/2
  hsizep1(1:3) = hsize(1:3) + 1;
  
! child 0
  child_vertex(1:3, 0) = vsmOctant4iso(node)%Vertex(1:3)
  child_size(1:3, 0) = hsizep1(1:3)

! child 1
  child_vertex(1,1) = vsmOctant4iso(node)%Vertex(1) + hsize(1)
  child_vertex(2,1) = vsmOctant4iso(node)%Vertex(2)
  child_vertex(3,1) = vsmOctant4iso(node)%Vertex(3)
  child_size(1,1) = vsmOctant4iso(node)%OctDatSize(1) - hsize(1)
  child_size(2,1) = hsizep1(2)
  child_size(3,1) = hsizep1(3)

! child 2
  child_vertex(1,2) = vsmOctant4iso(node)%Vertex(1)
  child_vertex(2,2) = vsmOctant4iso(node)%Vertex(2) + hsize(2)
  child_vertex(3,2) = vsmOctant4iso(node)%Vertex(3)
  child_size(1,2) = hsizep1(1)
  child_size(2,2) = vsmOctant4iso(node)%OctDatSize(2) - hsize(2)
  child_size(3,2) = hsizep1(3)

! child 3
  child_vertex(1,3) = vsmOctant4iso(node)%Vertex(1) + hsize(1)
  child_vertex(2,3) = vsmOctant4iso(node)%Vertex(2) + hsize(2)
  child_vertex(3,3) = vsmOctant4iso(node)%Vertex(3)
  child_size(1,3) = vsmOctant4iso(node)%OctDatSize(1) - hsize(1)
  child_size(2,3) = vsmOctant4iso(node)%OctDatSize(2) - hsize(2)
  child_size(3,3) = hsizep1(3)

! child 4
  child_vertex(1, 4) = vsmOctant4iso(node)%Vertex(1)
  child_vertex(2, 4) = vsmOctant4iso(node)%Vertex(2)
  child_vertex(3, 4) = vsmOctant4iso(node)%Vertex(3) + hsize(3)
  child_size(1, 4) = hsizep1(1)
  child_size(2, 4) = hsizep1(2)
  child_size(3, 4) = vsmOctant4iso(node)%OctDatSize(3) - hsize(3)

! child 5
  child_vertex(1, 5) = vsmOctant4iso(node)%Vertex(1) + hsize(1)
  child_vertex(2, 5) = vsmOctant4iso(node)%Vertex(2)
  child_vertex(3, 5) = vsmOctant4iso(node)%Vertex(3) + hsize(3)
  child_size(1, 5) = vsmOctant4iso(node)%OctDatSize(1) - hsize(1)
  child_size(2, 5) = hsizep1(2)
  child_size(3, 5) = vsmOctant4iso(node)%OctDatSize(3) - hsize(3)

! child 6
  child_vertex(1, 6) = vsmOctant4iso(node)%Vertex(1)
  child_vertex(2, 6) = vsmOctant4iso(node)%Vertex(2) + hsize(2)
  child_vertex(3, 6) = vsmOctant4iso(node)%Vertex(3) + hsize(3)
  child_size(1, 6) = hsizep1(1)
  child_size(2, 6) = vsmOctant4iso(node)%OctDatSize(2) - hsize(2)
  child_size(3, 6) = vsmOctant4iso(node)%OctDatSize(3) - hsize(3)

! child 7
  child_vertex(1, 7) = vsmOctant4iso(node)%Vertex(1) + hsize(1)
  child_vertex(2, 7) = vsmOctant4iso(node)%Vertex(2) + hsize(2)
  child_vertex(3, 7) = vsmOctant4iso(node)%Vertex(3) + hsize(3)
  child_size(1, 7) = vsmOctant4iso(node)%OctDatSize(1) - hsize(1)
  child_size(2, 7) = vsmOctant4iso(node)%OctDatSize(2) - hsize(2)
  child_size(3, 7) = vsmOctant4iso(node)%OctDatSize(3) - hsize(3)

  next_level = vsmOctant4iso(node)%level + 1
  Mnum = vsmOctant4iso(node)%MyNumber*8

!  integ_num = 1; tinteg_num = 1
!  do i=1, vsmOctant4iso(node)%level
!    tinteg_num = tinteg_num * 8
!    integ_num = integ_num + tinteg_num
!  end do

  integ_num = vsmOctNum(vsmOctant4iso(node)%level)

  do i=0, 7
    vsmOctant4iso(node)%children(i) = integ_num + Mnum + i
    vsmOctant4iso(vsmOctant4iso(node)%children(i))%level = next_level
    vsmOctant4iso(vsmOctant4iso(node)%children(i))%MyNumber = Mnum+i
  end do

  !call vsmLogPrint("moi : ", vsmOctant4iso(node)%moi)
  !call vsmLogPrint("children's numbers: ", vsmOctant4iso(node)%children)
 
  do i=0, 7
    if(vsmOctant4iso(node)%children(i) >= vsmMaxOctNum) then
      call vsmLogPrint("vsmOctant4iso(node)%children(i) >= vsmMaxOctNum:", vsmOctant4iso(node)%children(i))
      call vsmError("vsmOctant4iso(node) number is strange")
    end if
    vsmOctant4iso(vsmOctant4iso(node)%children(i))%moi = vsmOctant4iso(node)%children(i)
    vsmOctant4iso(vsmOctant4iso(node)%children(i))%parent = vsmOctant4iso(node)%moi
    call vsmSetVertex4iso(child_vertex(1:3, i), child_size(1:3, i), vsmOctant4iso(node)%children(i))
    !call vsmLogPrint("call vsmOctDivide: ", vsmOctant4iso(node)%children(i))
    call vsmOctDivide4iso(vsmOctant4iso(node)%children(i))
  end do

  if(vsmOctant4iso(node)%level == 0) then
    call vsmLogPrint("call vsmCalcCoord")
    call vsmCalcCoord4iso
  end if

  !call vsmLogPrint("vsmOctDivide ends:", vsmOctant4iso(node)%moi)
 end subroutine vsmOctDivide4iso

 subroutine vsmSetGrid4iso
  vsmOctComDat4iso%uniform = vsmLocalCoord%uniform
  vsmOctComDat4iso%xmin = vsmLocalCoord%origin(1)
  vsmOctComDat4iso%ymin = vsmLocalCoord%origin(2)
  vsmOctComDat4iso%zmin = vsmLocalCoord%origin(3)
 end subroutine vsmSetGrid4iso

 subroutine vsmSetVertex4iso(vx, sz, node)
   integer, dimension(3), intent(in) :: vx, sz
   integer, intent(in) :: node

   vsmOctant4iso(node)%vertex(1:3) = vx(1:3)
   vsmOctant4iso(node)%OctDatSize(1:3) = sz(1:3)
   !call vsmLogPrint("Octree:sz", sz)
 end subroutine vsmSetVertex4iso

 function SurveyData4iso(node)
  integer, intent(in) :: node
  integer ::  SurveyData4iso

  SurveyData4iso = yySurveyData4iso(node)

 end function SurveyData4iso

 function yySurveyData4iso(node)
  integer, intent(in) :: node
  integer ::  yySurveyData4iso
  integer :: i, j, k, nscal, n, vis_num, ip1,jp1,kp1
  real(kind=vdp) :: iso_level, max_val, min_val
  real(kind=vdp), dimension(8) :: dat
!  real(kind=vdp), dimension(3) :: pos, spos
!  real :: alpha
!  integer :: above, below
!  logical :: anset

  !call vsmLogPrint("Octree: survey data is called, level = ",node%level)
  vis_num = vsmOctComDat4iso%vis_num

  !call vsmLogPrint("vsnum",vis_num)
  if(vis_num < 1) call vsmError("survayData visnum")

  yySurveyData4iso = 0
  if(vsmOctComDat4iso%vis_on(1)) then ! isosurface
    if(vsmOctant4iso(node)%Vertex(1)<1 .or. vsmOctant4iso(node)%Vertex(2)<1 .or. &
       vsmOctant4iso(node)%Vertex(3)<1) call vsmError("i or j or k<1")

    if(vsmOctant4iso(node)%Vertex(1)+vsmOctant4iso(node)%OctDatSize(1)-1 > vsmLocalCoord%nx(1) &
      .or. vsmOctant4iso(node)%Vertex(2)+vsmOctant4iso(node)%OctDatSize(2)-1 > vsmLocalCoord%nx(2) &
      .or. vsmOctant4iso(node)%Vertex(3)+vsmOctant4iso(node)%OctDatSize(3)-1 > vsmLocalCoord%nx(3)) then
       call vsmError("i or j or k too big")
    end if

    do n=1, vsmVisParam(vis_num)%iso_num
      nscal = vsmVisParam(vis_num)%isoparam(n)%nscal
      iso_level = vsmVisParam(vis_num)%isoparam(n)%iso_level
      !print *,"nscal iso_level", nscal, iso_level
      do k=vsmOctant4iso(node)%Vertex(3), vsmOctant4iso(node)%Vertex(3)+vsmOctant4iso(node)%OctDatSize(3)-2
!$omp parallel do private(i,j,ip1,jp1,kp1,dat,max_val,min_val)
        do j=vsmOctant4iso(node)%Vertex(2), vsmOctant4iso(node)%Vertex(2)+vsmOctant4iso(node)%OctDatSize(2)-2
          do i=vsmOctant4iso(node)%Vertex(1), vsmOctant4iso(node)%Vertex(1)+vsmOctant4iso(node)%OctDatSize(1)-2
            if(yySurveyData4iso == 0) then
               ip1 = i+1; jp1 = j+1; kp1 = k+1
               dat(1) = vsmScalData(nscal)%scal(i,   j,   k)
               dat(2) = vsmScalData(nscal)%scal(ip1, j,   k)
               dat(3) = vsmScalData(nscal)%scal(i,   jp1, k)
               dat(4) = vsmScalData(nscal)%scal(ip1, jp1, k)
               dat(5) = vsmScalData(nscal)%scal(i,   j,   kp1)
               dat(6) = vsmScalData(nscal)%scal(ip1, j,   kp1)
               dat(7) = vsmScalData(nscal)%scal(i,   jp1, kp1)
               dat(8) = vsmScalData(nscal)%scal(ip1, jp1, kp1)

               max_val = maxval(dat)
               min_val = minval(dat)

               if(min_val <= iso_level .and. max_val >= iso_level) then
                  yySurveyData4iso = 1
               end if
            end if

          end do !i
        end do !j
        if(yySurveyData4iso > 0) then
           call vsmLogPrint("SurveyData4Iso: Isosurface")
           exit
         end if
      end do !k
      if(yySurveyData4iso > 0) then
         call vsmLogPrint("SurveyData4Iso: Isosurface")
         exit
      end if
    end do
   !call vsmLogPrint("survey data for isosurface ends")
  end if

  !call vsmLogPrint("Octree: survey data ends")
 end function yySurveyData4iso


 subroutine vsmGetGMinMax4iso(gmin, gmax, node)
  real(kind=vdp), dimension(3), intent(out) :: gmin, gmax
  integer, intent(in) :: node

  gmin(1) = vsmOctant4iso(node)%txmin; gmin(2) = vsmOctant4iso(node)%tymin; gmin(3) = vsmOctant4iso(node)%tzmin
  gmax(1) = vsmOctant4iso(node)%txmax; gmax(2) = vsmOctant4iso(node)%tymax; gmax(3) = vsmOctant4iso(node)%tzmax

 end subroutine vsmGetGMinMax4iso

 subroutine vsmCalcMinMax4iso(node)
   integer, intent(in) :: node
   real(kind=vdp), dimension(3) :: temp

   if(vsmOctComDat4iso%uniform) then
      vsmOctant4iso(node)%txmin = vsmOctComDat4iso%xmin + vsmLocalCoord%dx(1)*(vsmOctant4iso(node)%Vertex(1)-1)
      vsmOctant4iso(node)%txmax = vsmOctant4iso(node)%txmin + vsmLocalCoord%dx(1)*(vsmOctant4iso(node)%OctDatSize(1)-1)

      vsmOctant4iso(node)%tymin = vsmOctComDat4iso%ymin + vsmLocalCoord%dx(2)*(vsmOctant4iso(node)%Vertex(2)-1)
      vsmOctant4iso(node)%tymax = vsmOctant4iso(node)%tymin + vsmLocalCoord%dx(2)*(vsmOctant4iso(node)%OctDatSize(2)-1)

      vsmOctant4iso(node)%tzmin = vsmOctComDat4iso%zmin + vsmLocalCoord%dx(3)*(vsmOctant4iso(node)%Vertex(3)-1)
      vsmOctant4iso(node)%tzmax = vsmOctant4iso(node)%tzmin + vsmLocalCoord%dx(3)*(vsmOctant4iso(node)%OctDatSize(3)-1)

      temp(1) = vsmOctant4iso(node)%txmin; temp(2) = vsmOctant4iso(node)%tymin; temp(3) = vsmOctant4iso(node)%tzmin;
      if(.not.vsmisinsidelwo(temp)) then
         call vsmLogPrint("calc min max1:", temp)
      end if

      temp(1) = vsmOctant4iso(node)%txmax; temp(2) = vsmOctant4iso(node)%tymax; temp(3) = vsmOctant4iso(node)%tzmax;
      if(.not.vsmisinsidelwo(temp)) then
         call vsmLogPrint("calc min max2:", temp)
      end if
    else
      vsmOctant4iso(node)%txmin = vsmLocalCoord%coordx(vsmOctant4iso(node)%Vertex(1))
      vsmOctant4iso(node)%txmax = vsmLocalCoord%coordx(vsmOctant4iso(node)%Vertex(1)+vsmOctant4iso(node)%OctDatSize(1)-1)

      vsmOctant4iso(node)%tymin = vsmLocalCoord%coordy(vsmOctant4iso(node)%Vertex(2))
      vsmOctant4iso(node)%tymax = vsmLocalCoord%coordy(vsmOctant4iso(node)%Vertex(2)+vsmOctant4iso(node)%OctDatSize(2)-1)

      vsmOctant4iso(node)%tzmin = vsmLocalCoord%coordz(vsmOctant4iso(node)%Vertex(3))
      vsmOctant4iso(node)%tzmax = vsmLocalCoord%coordz(vsmOctant4iso(node)%Vertex(3)+vsmOctant4iso(node)%OctDatSize(3)-1)
   end if
 end subroutine vsmCalcMinMax4iso

 function vsmGetMortonNum4iso(x, y, z)
   integer, intent(in) :: x, y, z
   integer :: vsmGetMortonNum4iso
   integer ::  n, ax, ay, az, bx, by, bz, b, i, di
   !call vsmLogPrint("maxlevel",vsmOctComDat4iso%maxLevel)
   n=1 
   bx=0; by=0; bz=0
   do i=0, vsmOctComDat4iso%maxLevel-1
     ax = iand(x, n)
     ay = iand(y, n)
     az = iand(z, n)

     !ax = ishft(ax, 3*i  )
     !ay = ishft(ay, 3*i+1)
     !az = ishft(az, 3*i+2)

     di = 2*i

     ax = ishft(ax, di  )
     ay = ishft(ay, di+1)
     az = ishft(az, di+2)

     bx = ior(bx, ax)
     by = ior(by, ay)
     bz = ior(bz, az)

     n = ishft(n, 1)
   end do

   b = 0
   b = ior(b, bx)
   b = ior(b, by)
   vsmGetMortonNum4iso = ior(b, bz)!b = ior(b, bz)

!   vsmGetMortonNum4iso = b
   if(vsmGetMortonNum4iso < 0 .or. 8**vsmOctComDat4iso%maxLevel <= vsmGetMortonNum4iso) then
     call vsmLogPrint("Morton num is strange:", vsmGetMortonNum4iso)
     call vsmError("ends")
   end if
 end function vsmGetMortonNum4iso


 subroutine vsmGetSkipVertex4iso(node, r_skip)
   !integer, dimension(3), intent(out) :: minvtx, maxvtx
   !type(vismo_octnode), intent(in) :: node
   integer, intent(in) :: node
   logical, intent(out) :: r_skip
   integer :: parent

   if(vsmOctant4iso(node)%skip) then
     parent = vsmOctant4iso(node)%parent
     vsmOctant4iso(node)%SkipVertexMin(1:3) = vsmOctant4iso(parent)%SkipVertexMin(1:3)
     vsmOctant4iso(node)%SkipVertexMax(1:3) = vsmOctant4iso(parent)%SkipVertexMax(1:3)
     r_skip = .true.
   else
     r_skip = .false.
   end if
 end subroutine vsmGetSkipVertex4iso

 subroutine vsmRaySkip4iso(pos, dirc, node, ray_, r_skip)
   real(kind=vdp), dimension(3), intent(in) :: pos
   real(kind=vdp), dimension(3), intent(in) :: dirc
   integer, intent(in) :: node
   real(kind=vdp), intent(out) :: ray_
   logical, intent(inout) :: r_skip

   real(kind=vdp), dimension(6) :: a
   real(kind=vdp) :: b

   ray_ = 0.0_vdp; 
   r_skip = .false.;
   a=-1.0_vdp; 
   b=0.0_vdp

   if(.not. vsmOctant4iso(node)%skip) then
     r_skip = .false.
     return
   end if

   if(.not. vsmIsInsideNode4iso(pos, node)) then
     r_skip = .false.
     return
   end if

   if(dirc(1) /= 0.0_vdp) then
     a(1) = (vsmOctant4iso(node)%SkipMin(1)-pos(1))/dirc(1)
     a(2) = (vsmOctant4iso(node)%SkipMax(1)-pos(1))/dirc(1)
   end if
  
   if(dirc(2) /= 0.0_vdp) then
     a(3) = (vsmOctant4iso(node)%SkipMin(2)-pos(2))/dirc(2)
     a(4) = (vsmOctant4iso(node)%SkipMax(2)-pos(2))/dirc(2)
   end if

   if(dirc(3) /= 0.0_vdp) then
     a(5) = (vsmOctant4iso(node)%SkipMin(3)-pos(3))/dirc(3)
     a(6) = (vsmOctant4iso(node)%SkipMax(3)-pos(3))/dirc(3)
   end if

   where(a < 0.0_vdp)
     a = vsmFaraway
   end where

   b = minval(a)
   ray_ = vsmIntval%ray_interval * int(b/vsmIntval%ray_interval)
   if(int(b/vsmIntval%ray_interval) > 0 .and. ray_ == vsmIntval%ray_interval * b/vsmIntval%ray_interval ) then
     ray_ = vsmIntval%ray_interval * (int(b/vsmIntval%ray_interval)-1)
!   if(int(b/vsmIntval%ray_interval) > 0) then
!     ray_ = vsmIntval%ray_interval * (int(b/vsmIntval%ray_interval)-0.1_vdp)
   end if
   r_skip = .true. 
   if(ray_ < 0.0_vdp) then
     if(vsmDbg) then
       call vsmLogPrint("a=",a)
       call vsmLogPrint("minval(a)=",minval(a))
      !call vsmLogPrint("b=",b)
      !call vsmLogPrint("intval=",vsmIntval%ray_interval)
      !call vsmLogPrint("real=",b/vsmIntval%ray_interval)
      !call vsmLogPrint("int=",int(b/vsmIntval%ray_interval))
       call vsmLogPrint("skipmin",vsmOctant4iso(node)%skipmin)
       call vsmLogPrint("skipmax",vsmOctant4iso(node)%skipmax)
     end if
     call vsmError("ray_ is negative")
   end if
 end subroutine vsmRaySkip4iso

 subroutine vsmyyRaySkip4iso(pos, spos, dirc, node, ray_, r_skip)
   real(kind=vdp), dimension(3), intent(in) :: pos, spos
   real(kind=vdp), dimension(3), intent(in) :: dirc
   integer, intent(in) :: node
   real(kind=vdp), intent(out) :: ray_
   logical, intent(inout) :: r_skip
   real(kind=vdp), dimension(3) :: rtpmin, rtpmax!, x
   real(kind=vdp) :: b

   ray_ = 0.0_vdp
   r_skip = .false.

   if(.not. vsmOctant4iso(node)%skip) then
     r_skip = .false.
     return
   end if

   if(.not. vsmIsInsideNode4iso(spos, node)) then
     r_skip = .false.
     return
   end if

   rtpmax(1) = vsmOctant4iso(node)%SkipMax(1)
   rtpmin(1) = vsmOctant4iso(node)%SkipMin(1)
   rtpmax(2) = vsmOctant4iso(node)%SkipMax(2)
   rtpmin(2) = vsmOctant4iso(node)%SkipMin(2)
   rtpmax(3) = vsmOctant4iso(node)%SkipMax(3)
   rtpmin(3) = vsmOctant4iso(node)%SkipMin(3)

!   call vsmAnglesToCoords(pos, x)
!   b = vsmyyRayDist(x, dirc, rtpmin, rtpmax)
   b = vsmyyRayDistInside(pos, spos, dirc, rtpmin, rtpmax)
   ray_ = vsmIntval%ray_interval * int(b/vsmIntval%ray_interval)
   !call vsmLogPrint("ray1=",ray_)
   !call vsmLogPrint("ray2=",vsmIntval%ray_interval * b/vsmIntval%ray_interval)
!   if(int(b/vsmIntval%ray_interval) > 0 .and. ray_ == vsmIntval%ray_interval * b/vsmIntval%ray_interval ) then
   if(int(b/vsmIntval%ray_interval) > 0 .and. ray_ == vsmIntval%ray_interval * b/vsmIntval%ray_interval) then
     ray_ = vsmIntval%ray_interval * (int(b/vsmIntval%ray_interval)-1)
     r_skip = .true. 
   else if(int(b/vsmIntval%ray_interval) == 0) then
     r_skip = .false.
   end if
   !call vsmLogPrint("final ray1=",ray_)
!   print *,"skip",b,ray_,rtpmin,rtpmax
   if(ray_ < 0.0_vdp) then
     if(vsmDbg) then
      ! call vsmLogPrint("a=",a)
      ! call vsmLogPrint("minval(a)=",minval(a))
       !call vsmLogPrint("b=",b)
       !call vsmLogPrint("intval=",vsmIntval%ray_interval)
       !call vsmLogPrint("real=",b/vsmIntval%ray_interval)
       !call vsmLogPrint("int=",int(b/vsmIntval%ray_interval))
       call vsmLogPrint("skipmin",vsmOctant4iso(node)%skipmin)
       call vsmLogPrint("skipmax",vsmOctant4iso(node)%skipmax)
     end if
     call vsmError("ray_ is negative")
   end if
 end subroutine vsmyyRaySkip4iso


 function vsmIsInsideNode4iso(pos, node) !is_rayin
  logical :: vsmIsInsideNode4iso
  integer, intent(in) :: node
  real (kind=vdp), dimension(3), intent(in) :: pos

  if(pos(1) >= vsmOctant4iso(node)%txmax .or. pos(1) < vsmOctant4iso(node)%txmin .or. pos(2) >= vsmOctant4iso(node)%tymax .or. &
  pos(2) < vsmOctant4iso(node)%tymin .or. pos(3) >= vsmOctant4iso(node)%tzmax .or. pos(3) < vsmOctant4iso(node)%tzmin) then
    vsmIsInsideNode4iso = .false.
  else 
    vsmIsInsideNode4iso = .true.
  end if
 end function vsmIsInsideNode4iso


 subroutine vsmCalcCoord4iso
  integer :: numRow, num_sv!, tsv
  integer :: i, nd
  integer :: bx, by, bz
  real(kind=vdp), dimension(3) :: gmin, gmax

  call vsmLogPrint("vsmCalcCoord is called")
  if(vsmOctComDat4iso%maxLevel == 0) then 
    vsmOctComDat4iso%coordx(0) = vsmOctant4iso(0)%txmin
    vsmOctComDat4iso%coordy(0) = vsmOctant4iso(0)%tymin
    vsmOctComDat4iso%coordz(0) = vsmOctant4iso(0)%tzmin

    vsmOctComDat4iso%coordx(1) = vsmOctant4iso(0)%txmax
    vsmOctComDat4iso%coordy(1) = vsmOctant4iso(0)%tymax
    vsmOctComDat4iso%coordz(1) = vsmOctant4iso(0)%tzmax
    return
  end if
  
!  num_sv = 1; tsv = 1
!  do i=1, vsmOctComDat4iso%maxLevel-1
!   tsv = ishft(tsv, 3)
!   num_sv = num_sv + tsv
!  end do

  if(vsmOctComDat4iso%maxLevel > 0) then
    num_sv = vsmOctNum(vsmOctComDat4iso%maxLevel-1)
  else
    num_sv = 0
  end if

  call vsmLogPrint("numsv: ",num_sv)
!  num_sv = num_sv - 1

  numRow = vsmOctComDat4iso%Row !ishft(a, vsmOctComDat4iso%maxLevel)
  call vsmLogPrint("numRow: ",numRow)


  vsmOctComDat4iso%coordx(0) = vsmOctant4iso(0)%txmin
  vsmOctComDat4iso%coordy(0) = vsmOctant4iso(0)%tymin
  vsmOctComDat4iso%coordz(0) = vsmOctant4iso(0)%tzmin

  do i=0, numRow - 1

   bx = vsmGetMortonNum4iso(i, 0, 0)
   nd = num_sv+bx
   call vsmGetGMinMax4iso(gmin, gmax, nd)
   vsmOctComDat4iso%coordx(i+1) = gmax(1)

   by = vsmGetMortonNum4iso(0, i, 0)
   nd = num_sv+by
   call vsmGetGMinMax4iso(gmin, gmax, nd)
   vsmOctComDat4iso%coordy(i+1) = gmax(2)
   
   bz = vsmGetMortonNum4iso(0, 0, i)
   nd = num_sv+bz
   call vsmGetGMinMax4iso(gmin, gmax, nd)
   vsmOctComDat4iso%coordz(i+1) = gmax(3)

   if(vsmDbg) then
     call vsmLogPrint("Row ", i)
     call vsmLogPrint("bx:", bx)
     call vsmLogPrint("gmax(1):", gmax(1))
     call vsmLogPrint("by:", by)
     call vsmLogPrint("gmax(2):", gmax(2))
     call vsmLogPrint("bz:", bz)
     call vsmLogPrint("gmax(3):", gmax(3))
   end if
  end do

  call vsmLogPrint("vsmCalcCoord ends")
 end subroutine vsmCalcCoord4iso
 
end module

