module vsmRayCaster
    use vsmConst
    use vsmFunction
    use vsmData
    use vsmCoordinates
    use vsmColormap
    use vsmParam
    use vsmCameraData
    use vsmLightData
    use vsmDepthQ
    use vsmImage
    use vsmIntpol
    use vsmOctree
    use vsmOctree4Iso
    use vsmBTree

    implicit none
    private

    type vismo_cslice
        integer :: num
        real(kind=vdp) :: val
        integer :: type
    end type vismo_cslice

    type vismo_rcastdata
     integer, dimension(:), allocatable :: state
     real(kind=vdp), dimension(:, :), allocatable :: rpos
     real(kind=vdp), dimension(:,:,:), allocatable :: hit
     real(kind=vdp), dimension(:,:),allocatable :: swaparray!vsmImageData%width*vsmImageData%height, 4
     real(kind=vdp), dimension(:,:),allocatable :: norm, pixel, normo, pixelo!vsmImageData%width*vsmImageData%height, 3
     real(kind=vdp), dimension(:),allocatable :: ray2ldat, ray2wdat, ray2dat!vsmImageData%width*vsmImageData%height
     real(kind=vdp), dimension(:),allocatable :: dist2pxl, ray, prevray !vsmImageData%width*vsmImageData%height
     real(kind=vdp), dimension(:,:,:),allocatable :: vsmyyRayCastDataHit!3, vsmImageData%width, vsmImageData%height
     real(kind=vdp), dimension(:,:),allocatable :: vsmyyRayCastDataRPos!vsmImageData%width*vsmImageData%height, 3
     integer, dimension(:,:),allocatable :: px, pxo!vsmImageData%width*vsmImageData%height, 2
     integer, dimension(:),allocatable :: cont, N, cont2, cont3!vsmImageData%width*vsmImageData%height
    end type vismo_rcastdata

   public vsmFinalizeyyRaycast, vsmInityyRaycast

   public vsmScalarVisualization

    type(vismo_rcastdata), save :: yyraycasdat
    contains

 subroutine vsmInityyRaycast(width, height)
   integer, intent(in) :: width, height
   integer :: allocstate

  call vsmLogPrint("vsmInityyraycast")
  allocate(yyraycasdat%state(width*height), stat=allocstate) ! *5 -> rgbaz
  if(allocstate /= 0) then
   call vsmLogPrint("allocstate",allocstate)
   call vsmError("can't allocate state")
  end if

  allocate(yyraycasdat%rpos(width*height, 3), stat=allocstate) ! *5 -> rgbaz
  if(allocstate /= 0) then
   call vsmLogPrint("allocstate",allocstate)
   call vsmError("can't allocate rpos")
  end if

  allocate(yyraycasdat%hit(3, width, height), stat=allocstate) ! *5 -> rgbaz
  if(allocstate /= 0) then
   call vsmLogPrint("allocstate",allocstate)
   call vsmError("can't allocate hit")
  end if

 end subroutine vsmInityyRaycast

 subroutine vsmFinalizeyyRaycast
   deallocate(yyraycasdat%state)
   deallocate(yyraycasdat%hit, yyraycasdat%rpos)
 end subroutine vsmFinalizeyyRaycast

 subroutine vsmyyGradScal(i, j, k, nscal, vec)
   integer, intent(in) :: i, j, k, nscal
   real(kind=vdp), dimension(3), intent(out) :: vec
   real(kind=vdp) :: dfd1, dfd2, dfd3, dis_r, dis_t, dis_p
   real(kind=vdp) :: radius, sin_theta
#ifndef FLOAT
   real(kind=vdp), dimension(:,:,:), pointer :: data_pointer
#else
   real, dimension(:,:,:), pointer :: data_pointer
#endif
   integer, dimension(3) :: n
   real(kind=vdp), dimension(3) :: delta, minc

   if(.not.vsmCheckLCoord(i,j,k)) call vsmError("coord is strange:: vsmyyGradScal")

   n(1) = vsmLocalCoord%nx(1)
   n(2) = vsmLocalCoord%nx(2)
   n(3) = vsmLocalCoord%nx(3)

   delta(1) = vsmLocalCoord%dx(1)
   delta(2) = vsmLocalCoord%dx(2)
   delta(3) = vsmLocalCoord%dx(3)

   minc(1) = vsmLocalCoord%origin(1)
   minc(2) = vsmLocalCoord%origin(2)
   minc(3) = vsmLocalCoord%origin(3)

   data_pointer => vsmScalData(nscal)%scal

   !********** r **********
   if(i == 1) then
      dfd1 = data_pointer(2, j, k) - data_pointer(1, j, k)
      dis_r = delta(1)
      radius = minc(1)
   else if(i == n(1)) then
      dfd1 = data_pointer(n(1), j, k) - data_pointer(n(1)-1,j,k)
      dis_r = delta(1)
      radius = delta(1) *(n(1) -1) + minc(1)
   else 
      dfd1 = data_pointer(i+1,j,k)- data_pointer(i-1,j,k)
      dis_r = 2.0_vdp * delta(1)
      radius = delta(1) * (i-1) + minc(1)
   end if

   dfd1 = dfd1/dis_r

   !********** theta **********
   if(j == 1) then
      dfd2 = data_pointer(i,2,k)- data_pointer(i,1,k)
      dis_t = delta(2)
      sin_theta = sin(minc(2))
   else if(j == n(2)) then
      dfd2 = data_pointer(i,n(2),k) - data_pointer(i,n(2)-1,k)
      dis_t = delta(2)
      sin_theta = sin((n(2)-1) * delta(2) + minc(2))
   else
      dfd2 = data_pointer(i,j+1,k) - data_pointer(i,j-1,k)
      dis_t = 2.0_vdp * delta(2)
      sin_theta = sin((j-1) * delta(2) + minc(2))
   end if

   dfd2 = dfd2/dis_t/radius

   !********** phi;; Phi is periodic *********
   if(k == 1) then
      dfd3 = data_pointer(i,j,2) - data_pointer(i,j,1)
      dis_p = delta(3)
   else if(k == n(3)) then
      dfd3 = data_pointer(i,j,n(3)) - data_pointer(i,j,n(3)-1)
      dis_p = delta(3)
   else
      dfd3 = data_pointer(i,j,k+1) - data_pointer(i,j,k-1)
      dis_p = 2.0_vdp * delta(3)
   end if

   dfd3 = dfd3/dis_p/radius/sin_theta

   vec(1) = -dfd1
   vec(2) = -dfd2
   vec(3) = -dfd3

   if(isnan(vec(1)) .or. isnan(vec(2)) .or. isnan(vec(3))) then
     call vsmLogPrint("vec1 in grad",vec)
     call vsmLogPrint("delta", delta)
     call vsmLogPrint("i, j, k in grad",i)
     call vsmLogPrint("i, j, k in grad",j)
     call vsmLogPrint("i, j, k in grad",k)
     call vsmError("vec is NAN")
   end if

 end subroutine vsmyyGradScal

 subroutine vsmCopyNormvec(dx, dy, dz, norm_vect, temp_vect, calcd)
   integer, intent(in) :: dx, dy, dz
   real(kind=vdp), dimension(3, 8), intent(inout) :: norm_vect, temp_vect
   integer, dimension(8), intent(inout) :: calcd
   integer :: i

  select case (dx)
    case (-1) ! dx

      select case (dy)
    case (-1) ! dy

        select case (dz)
    case (-1)
      temp_vect(1:3, 8) = norm_vect(1:3, 1)
      calcd(8) = 0

    case (0)
      temp_vect(1:3, 4) = norm_vect(1:3, 1)
      calcd(4) = 0;
      temp_vect(1:3, 8) = norm_vect(1:3, 5)
      calcd(8) = 0

    case (1)
      temp_vect(1:3, 4) = norm_vect(1:3, 5)
      calcd(4) = 0;

    case default
      call vsmError ('failed to copy normal vectors')
          end select

   case (0) ! dy

   select case(dz)
     case (-1)
       temp_vect(1:3, 6) = norm_vect(1:3, 1)
       calcd(6) = 0
       temp_vect(1:3, 8) = norm_vect(1:3, 3)
       calcd(8) = 0

     case (0)
       temp_vect(1:3, 2) = norm_vect(1:3, 1)
       calcd(2) = 0
       temp_vect(1:3, 4) = norm_vect(1:3, 3)
       calcd(4) = 0
       temp_vect(1:3, 6) = norm_vect(1:3, 5)
       calcd(6) = 0
       temp_vect(1:3, 8) = norm_vect(1:3, 7)
       calcd(8) = 0

     case (1)
       temp_vect(1:3, 2) = norm_vect(1:3, 5)
       calcd(2) = 0
       temp_vect(1:3, 4) = norm_vect(1:3, 7)
       calcd(4) = 0

     case default
       call vsmError ('failed to copy normal vectors')
          end select

   case (1) ! dy

   select case (dz)
     case (-1)
       temp_vect(1:3, 6) = norm_vect(1:3, 3)
       calcd(6) = 0

     case (0)
       temp_vect(1:3, 2) = norm_vect(1:3, 3)
       calcd(2) = 0
       temp_vect(1:3, 6) = norm_vect(1:3, 7)
       calcd(6) = 0

     case (1)
       temp_vect(1:3, 2) = norm_vect(1:3, 7)
       calcd(2) = 0

     case default
       call vsmError ('failed to copy normal vectors')

     end select

      case default ! dy
         call vsmError ('failed to copy normal vectors')

    end select

    case (0) ! dx

      select case (dy)
    case (-1) ! dy

   select case(dz)
     case (-1)
       temp_vect(1:3, 7) = norm_vect(1:3, 1)
       calcd(7) = 0
       temp_vect(1:3, 8) = norm_vect(1:3, 2)
       calcd(8) = 0

     case (0)
       temp_vect(1:3, 3) = norm_vect(1:3, 1)
       calcd(3) = 0
       temp_vect(1:3, 4) = norm_vect(1:3, 2)
       calcd(4) = 0
       temp_vect(1:3, 7) = norm_vect(1:3, 5)
       calcd(7) = 0
       temp_vect(1:3, 8) = norm_vect(1:3, 6)
       calcd(8) = 0

     case (1)
       temp_vect(1:3, 3) = norm_vect(1:3, 5)
       calcd(3) = 0
       temp_vect(1:3, 4) = norm_vect(1:3, 6)
       calcd(4) = 0

     case default
       call vsmError ('failed to copy normal vectors')

   end select

 case (0) ! dy

   select case (dz)
     case (-1)
       temp_vect(1:3, 5) = norm_vect(1:3, 1)
       calcd(5) = 0;
       temp_vect(1:3, 6) = norm_vect(1:3, 2)
       calcd(6) = 0;
       temp_vect(1:3, 7) = norm_vect(1:3, 3)
       calcd(7) = 0;
       temp_vect(1:3, 8) = norm_vect(1:3, 4)
       calcd(8) = 0;

     case (1)
       temp_vect(1:3, 1) = norm_vect(1:3, 5)
       calcd(1) = 0;
       temp_vect(1:3, 2) = norm_vect(1:3, 6)
       calcd(2) = 0;
       temp_vect(1:3, 3) = norm_vect(1:3, 7)
       calcd(3) = 0;
       temp_vect(1:3, 4) = norm_vect(1:3, 8)
       calcd(4) = 0;

     case default
       call vsmError ('failed to copy normal vectors')

           end select

 case (1)! dy

   select case(dz)
     case (-1)
       temp_vect(1:3, 5) = norm_vect(1:3, 3)
       calcd(5) = 0;
       temp_vect(1:3, 6) = norm_vect(1:3, 4)
       calcd(6) = 0;

     case (0)
       temp_vect(1:3, 1) = norm_vect(1:3, 3)
       calcd(1) = 0;
       temp_vect(1:3, 2) = norm_vect(1:3, 4)
       calcd(2) = 0;
       temp_vect(1:3, 5) = norm_vect(1:3, 7)
       calcd(5) = 0;
       temp_vect(1:3, 6) = norm_vect(1:3, 8)
       calcd(6) = 0;

     case (1)
       temp_vect(1:3, 1) = norm_vect(1:3, 7)
       calcd(1) = 0;
       temp_vect(1:3, 2) = norm_vect(1:3, 8)
       calcd(2) = 0;

     case default
       call vsmError ('failed to copy normal vectors')

   end select

        case default
    call vsmError ('failed to copy normal vectors')

       end select


    case (1)
      select case(dy)
    case (-1) ! dy

   select case (dz)
     case (-1)
       temp_vect(1:3, 7) = norm_vect(1:3, 2)
       calcd(7) = 0

     case (0)
       temp_vect(1:3, 3) = norm_vect(1:3, 2)
       calcd(3) = 0
       temp_vect(1:3, 7) = norm_vect(1:3, 6)
       calcd(7) = 0

     case (1)
       temp_vect(1:3, 3) = norm_vect(1:3, 6)
       calcd(3) = 0

            case default
              call vsmError ('failed to copy normal vectors')

          end select

 case (0) ! dy

   select case (dz)
     case (-1)
       temp_vect(1:3, 5) = norm_vect(1:3, 2)
       calcd(5) = 0
       temp_vect(1:3, 7) = norm_vect(1:3, 4)
       calcd(7) = 0

     case (0)
       temp_vect(1:3, 1) = norm_vect(1:3, 2)
       calcd(1) = 0
       temp_vect(1:3, 3) = norm_vect(1:3, 4)
       calcd(3) = 0
       temp_vect(1:3, 5) = norm_vect(1:3, 6)
       calcd(5) = 0
       temp_vect(1:3, 7) = norm_vect(1:3, 8)
       calcd(7) = 0

     case (1)
       temp_vect(1:3, 1) = norm_vect(1:3, 6)
       calcd(1) = 0
       temp_vect(1:3, 3) = norm_vect(1:3, 8)
       calcd(3) = 0

            case default
              call vsmError ('failed to copy normal vectors')

         end select

 case (1)

   select case(dz)
     case (-1)
       temp_vect(1:3, 5) = norm_vect(1:3, 4)
       calcd(5) = 0

     case (0)
       temp_vect(1:3, 1) = norm_vect(1:3, 4)
       calcd(1) = 0
       temp_vect(1:3, 5) = norm_vect(1:3, 8)
       calcd(5) = 0

     case (1)
       temp_vect(1:3, 1) = norm_vect(1:3, 8)
       calcd(1) = 0

            case default
              call vsmError ('failed to copy normal vectors')

          end select

         case default
            call vsmError ('failed to copy normal vectors')

       end select

     case default
           call vsmError ('failed to copy normal vectors')

    end select
    do i=1, 8
      if(calcd(i) == 0) then
        if(isnan(temp_vect(1,i)) .or. isnan(temp_vect(2,i)) .or. isnan(temp_vect(3,i))) then
     	   print *,"vsmCopyNormvec ",temp_vect(1:3,i)
   	     print *,"dx dy dz i=",dx,dy,dz,i
         call vsmError("temp_vect is NAN")
        end if
      end if
   end do
 end subroutine vsmCopyNormvec

 subroutine vsmGradScal(i, j, k, nscal, vec)
   integer, intent(in) :: i, j, k, nscal
   real(kind=vdp), dimension(3), intent(out) :: vec
   real(kind=vdp) :: dis, x1, x2

   if(.not.vsmCheckLCoord(i,j,k)) call vsmError("coord is strange:: vsmGradScal")

   if (i == vsmLocalCoord%nx(1)) then
      x1 = vsmScalData(nscal)%scal(i, j, k);
      x2 = vsmScalData(nscal)%scal(i-1, j, k);

      if(vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(1)
      else
        dis = vsmLocalCoord%coordx(i) - vsmLocalCoord%coordx(i-1)
      end if
   else if(i == 1) then
      x1 = vsmScalData(nscal)%scal(2, j, k)
      x2 = vsmScalData(nscal)%scal(1, j, k)

      if (vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(1)
      else
        dis = vsmLocalCoord%coordx(2) - vsmLocalCoord%coordx(1)
      end if
   else
      x1 = vsmScalData(nscal)%scal(i+1, j, k)
      x2 = vsmScalData(nscal)%scal(i-1, j, k)

      if (vsmLocalCoord%uniform) then
        dis = 2.0_vdp*vsmLocalCoord%dx(1)
      else
        dis = vsmLocalCoord%coordx(i+1) - vsmLocalCoord%coordx(i-1)
      end if
   end if

   vec(1) = (x2 - x1)/ dis

   if(j == vsmLocalCoord%nx(2)) then
      x1 = vsmScalData(nscal)%scal(i, j, k)
      x2 = vsmScalData(nscal)%scal(i, j-1, k)

      if(vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(2)
      else
        dis = vsmLocalCoord%coordy(j) - vsmLocalCoord%coordy(j-1)
      end if
   else if(j == 1) then
      x1 = vsmScalData(nscal)%scal(i, 2, k)
      x2 = vsmScalData(nscal)%scal(i, 1, k)

      if(vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(2)
      else
        dis = vsmLocalCoord%coordy(2) - vsmLocalCoord%coordy(1)
      end if
   else
      x1 = vsmScalData(nscal)%scal(i, j+1, k)
      x2 = vsmScalData(nscal)%scal(i, j-1, k)

      if (vsmLocalCoord%uniform) then
        dis = 2.0_vdp*vsmLocalCoord%dx(2)
      else
        dis = vsmLocalCoord%coordy(j+1) - vsmLocalCoord%coordy(j-1)
      end if
   end if

   vec(2) = (x2 - x1) / dis

   if(k == vsmLocalCoord%nx(3)) then
      x1 = vsmScalData(nscal)%scal(i, j, k)
      x2 = vsmScalData(nscal)%scal(i, j, k-1)

      if(vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(3)
      else
        dis = vsmLocalCoord%coordz(k) - vsmLocalCoord%coordz(k-1)
      end if
   else if(k == 1) then
      x1 = vsmScalData(nscal)%scal(i, j, 2)
      x2 = vsmScalData(nscal)%scal(i, j, 1)

      if(vsmLocalCoord%uniform) then
        dis = vsmLocalCoord%dx(3)
      else
        dis = vsmLocalCoord%coordz(2) - vsmLocalCoord%coordz(1)
      end if
   else
      x1 = vsmScalData(nscal)%scal(i, j, k+1)
      x2 = vsmScalData(nscal)%scal(i, j, k-1)

      if (vsmLocalCoord%uniform) then
        dis = 2.0_vdp*vsmLocalCoord%dx(3)
      else
        dis = vsmLocalCoord%coordz(k+1) - vsmLocalCoord%coordz(k-1)
      end if
   end if

   vec(3) = (x2 - x1) / dis

   if(isnan(vec(1)) .or. isnan(vec(2)) .or. isnan(vec(3))) then
     call vsmLogPrint("vec in grad scal",vec)
     call vsmLogPrint("delta", vsmLocalCoord%dx)
     call vsmLogPrint("i in grad",i)
     call vsmLogPrint("j in grad",j)
     call vsmLogPrint("k in grad",k)
     call vsmError("vec is NAN")
   end if

 end subroutine vsmGradScal

 subroutine vsmNormScalar(ci, nci, nscal, norm)
   integer, dimension(3), intent(in) :: ci
   integer, dimension(3), intent(inout) :: nci
   integer, intent(in) :: nscal
   real(kind=vdp), dimension(3, 8), intent(inout) :: norm
   integer :: l
   integer, dimension(3) :: dxyz
   integer, dimension(8) :: calcd
   real(kind=vdp), dimension(3, 8) :: temp_norm
   !call vsmLogPrint("normscalar ci=",ci)
   !call vsmLogPrint("normscalar nci=",nci)

   if (ci(1) /= nci(1) .or. ci(2) /= nci(2) .or. ci(3) /= nci(3)) then

     dxyz(1:3) = ci(1:3) - nci(1:3)
     calcd(1:8) = 1
     if (maxval(abs(dxyz)) < 2) then
       call vsmCopyNormvec (dxyz(1), dxyz(2), dxyz(3), norm, temp_norm, calcd)
     end if

!     call vsmLogPrint("dx dy dz = ",dxyz)
!     call vsmLogPrint("calcd = ",calcd)

     do l = 1, 8
       if (calcd(l) == 1) then
         select case (l)
           case (1)
             call vsmGradScal(ci(1),   ci(2),   ci(3), nscal, temp_norm(1:3,1))
           case (2)
             call vsmGradScal(ci(1)+1, ci(2),   ci(3), nscal, temp_norm(1:3,2))
           case (3)
             call vsmGradScal(ci(1),   ci(2)+1, ci(3), nscal, temp_norm(1:3,3))
           case (4)
             call vsmGradScal(ci(1)+1, ci(2)+1, ci(3), nscal, temp_norm(1:3,4))
           case (5)
             call vsmGradScal(ci(1),   ci(2),   ci(3)+1, nscal, temp_norm(1:3,5))
           case (6)
             call vsmGradScal(ci(1)+1, ci(2),   ci(3)+1, nscal, temp_norm(1:3,6))
           case (7)
             call vsmGradScal(ci(1),   ci(2)+1, ci(3)+1, nscal, temp_norm(1:3,7))
           case (8)
             call vsmGradScal(ci(1)+1, ci(2)+1, ci(3)+1, nscal, temp_norm(1:3,8))
         end select
       end if
     norm(1:3, l) = temp_norm(1:3, l)

     if(isnan(norm(1, l)) .or. isnan(norm(2, l)) .or. isnan(norm(3, l))) then
       call vsmLogPrint("norm in normscalar",norm(:, l))
       call vsmLogPrint("ci in normscalar", ci)
       call vsmError("NAN exist in norm scalar")
     end if

     end do

    nci(1:3) = ci(1:3)
!    call vsmLogPrint("normscalar ci=",ci)
!    call vsmLogPrint("normscalar nci=",nci)
   end if
 end subroutine vsmNormScalar

 subroutine vsmyyNormScalar(ci, nci, nscal, norm)
   integer, dimension(3), intent(in) :: ci
   integer, dimension(3), intent(inout) :: nci
   integer, intent(in) :: nscal
   real(kind=vdp), dimension(3, 8), intent(inout) :: norm
   integer :: l
   integer, dimension(3) :: dxyz
   integer, dimension(8) :: calcd
   real(kind=vdp), dimension(3, 8) :: temp_norm
   !call vsmLogPrint("normscalar ci=",ci)
   !call vsmLogPrint("normscalar nci=",nci)

   if(ci(1) /= nci(1) .or. ci(2) /= nci(2) .or. ci(3) /= nci(3)) then

     dxyz(1:3) = ci(1:3) - nci(1:3)
     calcd(1:8) = 1
     if (maxval(abs(dxyz)) < 2) then
       call vsmCopyNormvec (dxyz(1), dxyz(2), dxyz(3), norm, temp_norm, calcd)
     end if
!     call vsmLogPrint("dx dy dz = ",dxyz)
!     call vsmLogPrint("calcd = ",calcd)

     do l = 1, 8
       if(calcd(l) == 1) then
         select case (l)
           case (1)
             call vsmyyGradScal(ci(1),   ci(2),   ci(3), nscal, temp_norm(1:3,1))
           case (2)
             call vsmyyGradScal(ci(1)+1, ci(2),   ci(3), nscal, temp_norm(1:3,2))
           case (3)
             call vsmyyGradScal(ci(1),   ci(2)+1, ci(3), nscal, temp_norm(1:3,3))
           case (4)
             call vsmyyGradScal(ci(1)+1, ci(2)+1, ci(3), nscal, temp_norm(1:3,4))
           case (5)
             call vsmyyGradScal(ci(1),   ci(2),   ci(3)+1, nscal, temp_norm(1:3,5))
           case (6)
             call vsmyyGradScal(ci(1)+1, ci(2),   ci(3)+1, nscal, temp_norm(1:3,6))
           case (7)
             call vsmyyGradScal(ci(1),   ci(2)+1, ci(3)+1, nscal, temp_norm(1:3,7))
           case (8)
             call vsmyyGradScal(ci(1)+1, ci(2)+1, ci(3)+1, nscal, temp_norm(1:3,8))
         end select
       end if
     norm(1:3, l) = temp_norm(1:3, l)

     if(isnan(norm(1, l)) .or. isnan(norm(2, l)) .or. isnan(norm(3, l))) then
       call vsmLogPrint("norm in normscalar",norm(:, l))
       call vsmLogPrint("ci in normscalar", ci)
       call vsmError("NAN exist in yy norm scalar")
     end if

     end do

    nci(1:3) = ci(1:3)
!    call vsmLogPrint("normscalar ci=",ci)
!    call vsmLogPrint("normscalar nci=",nci)
   end if
 end subroutine vsmyyNormScalar

 subroutine vsmyySlicerIM(pixel, norm, ray, dist2pxl, vis_num, cam_num, px, cpos, cfront)
   ! inputed variables
   real(kind=vdp), dimension(3), intent(in) :: pixel, norm, cpos, cfront
   real(kind=vdp), intent(in) :: ray, dist2pxl
   integer, intent(in) :: vis_num, cam_num
   integer, dimension(2), intent(in) :: px

   ! variables
   type(vismo_cslice), dimension(2*vsmMaxSlice) :: cslidat
   integer :: i, j, sl_c, crs_points, snum
   integer :: nscal
   integer, dimension(3) :: ci
   real, dimension(4) :: color, color_ ! RGBA
 !  real :: one_m_salpha
   real, dimension(3) :: lumi
   real :: amb, dif, spec, shin, alpha
   real(kind=vdp), dimension(3) :: ray_pos, spos
   real(kind=vdp) :: ray_, inpro, radius
   real(kind=vdp) :: v1, v2
   real(kind=vdp), dimension(3) :: dx
   real(kind=vdp), dimension(2) :: dist
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp), dimension(8) :: grid_val
   real(kind=vdp) :: scal_val
   real, dimension(3) :: normal_vect
   real(kind=vdp), dimension(4) :: eq
   real(kind=vdp) :: fdistance

   !initialization
   call vsmSetZeroF(color); call vsmSetZeroF(color_)
   crs_points = 0
   ray_ = ray
   ci(1:3) = -1
   inpro = dot_product(norm, cfront)
   color(1:4) = 0.0

   do i = 1, vsmVisParam(vis_num)%slice_num

      if(vsmVisParam(vis_num)%sliceparam(i)%type == 0) then
        if(vsmCoordData%YinYang == vsmYang) then
          eq(1) = -vsmVisParam(vis_num)%sliceparam(i)%sc(1)
          eq(2) = vsmVisParam(vis_num)%sliceparam(i)%sc(3)
          eq(3) = vsmVisParam(vis_num)%sliceparam(i)%sc(2)
          eq(4) = vsmVisParam(vis_num)%sliceparam(i)%sc(4)
        else
          eq(1:4) = vsmVisParam(vis_num)%sliceparam(i)%sc(1:4)
        end if
        v1 = eq(1)* norm(1) + eq(2)* norm(2) + eq(3)* norm(3)

        if(abs(v1) < 1.0D-15) cycle

        v2 = -(eq(1)* pixel(1) + eq(2)* pixel(2) + eq(3)* pixel(3) + eq(4))
        ray_ = v2 / v1

        if(ray_ <= 0.0_vdp) cycle !Near

        crs_points = crs_points + 1
        cslidat(crs_points)%num = i
        cslidat(crs_points)%val = ray_
        cslidat(crs_points)%type = 0

      else if(vsmVisParam(vis_num)%sliceparam(i)%type == 1) then
        radius = vsmVisParam(vis_num)%sliceparam(i)%sc(1)
        call vsmyyRayDistSphere4Rslice(pixel, norm, radius, dist)
        do j=1, 2
          if(dist(j) <= 0.0_vdp) cycle !Near
          ray_ = dist(j)
          crs_points = crs_points + 1
          cslidat(crs_points)%num = i
          cslidat(crs_points)%val = ray_
          cslidat(crs_points)%type = 1
        end do
      else
        call vsmError("slice type is strange")
      end if

   end do

   do sl_c=1, crs_points
     snum = cslidat(sl_c)%num
     nscal = vsmVisParam(vis_num)%sliceparam(snum)%nscal

     ray_ = cslidat(sl_c)%val
     fdistance = ray_ + dist2pxl

     if(fdistance*inpro >= vsmCamera(cam_num)%far) cycle

     ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

     call vsmCoordsToAngles(ray_pos, spos)
     if(.not.vsmIsInsidel(spos)) cycle
!     if(vsmIsAnotherSet(ray_pos, spos)) cycle ! never semi-transparent
!     call vsmCoordsToAngles(ray_pos, spos)
     call vsmFindCoord(spos, dx, ci)
     ! for interpolation
     call vsmSetFrac(dx, vx)
     call vsmSetScalGridVal(nscal, ci, grid_val)

     ! calculate each scalar value at current point
     scal_val = vsmGetScalVal(vx, grid_val)

     ! color slices
     call vsmGetSliceColor(vis_num, snum, alpha, amb, dif, spec, shin)

     if(cslidat(sl_c)%type == 0) then
        if(vsmCoordData%YinYang == vsmYang) then
          normal_vect(1) = -vsmVisParam(vis_num)%sliceparam(snum)%sc(1)
          normal_vect(2) = vsmVisParam(vis_num)%sliceparam(snum)%sc(3)
          normal_vect(3) = vsmVisParam(vis_num)%sliceparam(snum)%sc(2)
        else
          normal_vect(1:3) = vsmVisParam(vis_num)%sliceparam(snum)%sc(1:3)
        end if
     else if(cslidat(sl_c)%type == 1) then
       normal_vect = ray_pos
     end if

     call vsmNormVecF(normal_vect)

      if(vsmCamera(cam_num)%pers) then
        call vsmCalcColor(amb, dif, spec, shin, cpos, &
            ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
      else
        call vsmCalcColorOrtho(amb, dif, spec, shin, cfront, &
            ray_pos, normal_vect, vsmVisParam(vis_num)%light_used,  lumi)
      end if

     call vsmSetZeroF(color_)
     call vsmGetSColor(scal_val, nscal, color_)

     color(1:3) = lumi(1:3)*color_(1:3) ! RGB
     color(4)   = 1.0        ! A

     call vsmPaintPixel3(px, color, fdistance)
     ! end of color slices
   end do

 end subroutine vsmyySlicerIM

 subroutine vsmyyVolumeRendererIM(pixel, norm, rn_, ray2ldat, ray2dat, dist2pxl, vis_method, vis_num, cam_num, px, &
   state, rpos, cpos, cfront)
    ! inputed variables
   real(kind=vdp), dimension(3), intent(in) :: pixel, norm,cpos, cfront
   real(kind=vdp), intent(in) :: ray2ldat, dist2pxl, ray2dat
   logical, dimension(3), intent(in) :: vis_method
   integer, intent(in) :: vis_num, cam_num, rn_
   integer, dimension(2), intent(in) :: px
   logical :: z_on
   integer, dimension(:), intent(inout) :: state
   real(kind=vdp), dimension(3), intent(inout) :: rpos

   ! variables
   integer :: rn, trn
   integer :: integ_loop, i, vol_c
   integer :: nscal, NumScal
   integer, dimension(3) :: ci, prev_ci
   integer, dimension(3, vsmMaxScalar) :: nci
   real, dimension(4) :: color ! RGBA
   real :: one_m_salpha
   real, dimension(4) :: color_
   real, dimension(3) :: amb, dif, spec, lumi
   real :: shin, lumi_a
   real(kind=vdp), dimension(3) :: ray_pos, prev_ray_pos, spos
   real(kind=vdp) :: rintval, ray_, inpro, zdis, pray_
   real(kind=vdp), dimension(3) :: dx, nvectrtp
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp) :: rayBack

   real(kind=vdp), dimension(3, 8, vsmMaxScalar) :: scalnorm

   real(kind=vdp), dimension(8) :: prev_vx

   real(kind=vdp), dimension(8, vsmMaxScalar) :: grid_val
   real(kind=vdp), dimension(vsmMaxScalar) :: scal_val
   real, dimension(3) :: normal_vect
   real :: dq_factor
   logical :: first_in, inset
   real(kind=vdp) :: fdistance, first_inset!, firstin_ray
   !octree
   logical :: skip, skipcalc, endFlag, InsideL, insidesphere
   integer :: comm, myrank, psize

   if(.not.vis_method(3)) call vsmError("vsm yy Volume RenderIM is called but method(3) is false")

   call vsmGetMPIVars(comm, myrank, psize)

   !initialization
   endFlag = .false.
   inset = .false.
   insidesphere = .true.

   NumScal = vsmDataInfo%NumOfScal
   prev_ci(1:3) = -1
   nci(1:3, 1:vsmMaxScalar) = -2 ! different from prev_ci values
   call vsmSetZeroF(color)

   prev_vx=0.0_vdp

   first_in = .false.
   zdis = vsmGetZbuf(px)

   ray_pos(1:3) = pixel(1:3) + norm(1:3) * (zdis - dist2pxl)
   call vsmCoordsToAngles(ray_pos, spos)
   if(vsmIsInsidel(spos)) then
     if(vsmIsAnotherSet(ray_pos, spos)) then
       z_on = .false.
     else
       z_on = .true.
     end if
   else
     z_on = .false.
   end if


   rintval = vsmIntval%ray_interval
   rn = rn_

   ray_ = rintval*rn

   pray_ = ray2ldat
   prev_ray_pos(1:3) = pixel(1:3) + norm(1:3) * pray_

   first_inset = ray_
   inpro = dot_product(norm, cfront)
   ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

   dx = 0.0_vdp

   call vsmCoordsToAngles(ray_pos, spos)

   if(.not.vsmIsInsidel(spos)) then
     rn = rn + 1
     ray_ = rintval*rn
     ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
     call vsmCoordsToAngles(ray_pos, spos)
   end if

   if(.not.vsmIsInsidel(spos)) then
     state(px(1) + (px(2)-1)*vsmImageData%width) = 2
     rpos(1:3) = ray_pos(1:3)
     return
   else
     if(first_in .eqv. .false.) then
       first_in = .true.
     end if
   end if

   call vsmCoordsToAngles(prev_ray_pos, spos)
   if(.not.vsmIsInsidelWO(spos)) then
      rayBack = ray2ldat
      do i=1, 100
         prev_ray_pos(1:3) = pixel(1:3) + norm(1:3) * rayBack
         call vsmCoordsToAngles(prev_ray_pos, spos)
         if(vsmIsInsidelWO(spos)) then
             !call vsmLogPrint("Modified(prev ray pos) ", prev_ray_pos)
             exit
         end if
         rayBack = rayBack + rintval*0.01_vdp
      end do

      call vsmCoordsToAngles(prev_ray_pos, spos)
      if(.not.vsmIsInsidelWO(spos)) then
         prev_ray_pos(1:3) = ray_pos(1:3)
         !call vsmLogPrint("Modified2(prev ray pos) ", prev_ray_pos)
      end if
   end if

   call vsmCoordsToAngles(prev_ray_pos, spos)
   call vsmFindCoord(spos, dx, prev_ci)
   call vsmSetFrac(dx, prev_vx)
   do i=1, NumScal
     call vsmSetScalGridVal(i, prev_ci, grid_val(1:8, i))
   end do

!   call vsmLogPrint("integloop starts", px)
   do integ_loop = 1, vsmMaxRayCount

     fdistance = ray_ + dist2pxl
     if(fdistance*inpro >= vsmCamera(cam_num)%far) then
        state(px(1) + (px(2) -1) *vsmImageData%width) = 0
        exit
     end if

     ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

     call vsmCoordsToAngles(ray_pos, spos)
     InsideL = vsmIsInsidel(spos)

     if(.not. InsideL) then
       endFlag = .true.
       insidesphere = vsmIsInsideSphere(ray_pos)
       if(inset) then
         call vsmRaySegment(first_inset+dist2pxl, ray_ +dist2pxl-rintval, color, px)
         inset = .false.
       end if

       rpos(1:3) = ray_pos(1:3)
       !call vsmLogPrint("prev_ray_pos ", prev_ray_pos)
       call vsmCoordsToAngles(prev_ray_pos, spos)
       rayBack = vsmyyRayDistInside(prev_ray_pos, spos, norm, vsmLocalCoord%lorigin, vsmLocalCoord%lcend)
       ray_pos(1:3) = prev_ray_pos(1:3) + norm(1:3) * rayBack

       call vsmCoordsToAngles(ray_pos, spos)
       InsideL = vsmIsInsidel(spos)
       if(.not.InsideL) then
         do i=1, 100
            rayBack = rayBack - rintval*0.01_vdp
            ray_pos(1:3) = prev_ray_pos(1:3) + norm(1:3) * rayBack
            call vsmCoordsToAngles(ray_pos, spos)
            if(vsmIsInsidel(spos)) then
              !print *,"modify ray_pos", ray_pos
              exit
            end if
         end do
         call vsmCoordsToAngles(ray_pos, spos)
         InsideL = vsmIsInsidel(spos)
         if(.not.InsideL) then
            ray_pos(1:3) = prev_ray_pos(1:3)
           ! call vsmError("ray cannot enter the data : raycaster")
         end if
       end if
     else
       if(inset) then
         if(vsmIsAnotherSet(ray_pos, spos)) then
           call vsmRaySegment(first_inset + dist2pxl, ray_+dist2pxl, color, px)
           inset = .false.
         end if
       else
         if(.not. vsmIsAnotherSet(ray_pos, spos)) then
           first_inset = ray_
           inset = .true.
         end if
       end if
     end if

     call vsmCoordsToAngles(ray_pos, spos)
     if(.not.vsmIsInsidelWO(spos)) then
       rpos(1:3) = ray_pos(1:3)
       insidesphere = vsmIsInsideSphere(ray_pos)
       if(.not.insidesphere) then
         state(px(1) + (px(2)-1)*vsmImageData%width) = 0
       else
         state(px(1) + (px(2)-1)*vsmImageData%width) = 2
       end if
       exit
     end if

     if(ray_ + dist2pxl >= zdis) then
       state(px(1) + (px(2)-1)*vsmImageData%width) = 0
       rpos(1:3) = ray_pos(1:3)
       call vsmGetPixelColor4(px, color_)
       one_m_salpha = 1.0 - color(4)
       lumi_a = color_(4) * one_m_salpha
       color(1:3) = color(1:3) + color_(1:3)*lumi_a ! RGB
       color(4)   = color(4)   + lumi_a           ! A
       if(inset) then
         fdistance = first_inset + dist2pxl
       else
         fdistance = zdis
       end if
       call vsmRaySegment(fdistance, zdis, color, px)
       return
     end if

     ! for interpolation
     call vsmCoordsToAngles(ray_pos, spos)
     call vsmFindCoord(spos, dx, ci)
     call vsmSetFrac(dx, vx)

     if(ci(1)/=prev_ci(1) .or. ci(2)/=prev_ci(2) .or. ci(3)/=prev_ci(3) .or. integ_loop==1) then
       do i=1, NumScal
         call vsmSetScalGridVal(i, ci, grid_val(1:8, i))
       end do
       skipcalc = .true.
     else
       skipcalc = .false.
     end if

     ! calculate each scalar value at current point
     do i=1, NumScal
       scal_val(i) = vsmGetScalVal(vx, grid_val(1:8, i))
     end do

     if(endFlag) then
       state(px(1)+(px(2)-1)*vsmImageData%width) = 2
       exit
     end if
     ! volume rendering
    ! call vsmCoordsToAngles(ray_pos, spos)
     if(.not.endFlag .and. color(4) < vsmErtValue .and. InsideL .and. inset) then !the last is for pc
       do vol_c=1, vsmVisParam(vis_num)%volren_num
         nscal = vsmVisParam(vis_num)%volparam(vol_c)%nscal
         call vsmGetVolColor(vis_num, vol_c, amb(1), dif(1), spec(1), shin)
         ! calculating normal vector
         if(ci(1) /= nci(1, nscal) .or. ci(2) /= nci(2, nscal) .or. ci(3) /= nci(3, nscal)) then
           call vsmyyNormScalar(ci, nci(1:3, nscal), nscal, scalnorm(1:3, 1:8, nscal))
         end if
         call vsmGetIntpolVectd(scalnorm(1:3, 1:8, nscal), vx, nvectrtp)

         normal_vect(1) = sin(spos(2)) * cos(spos(3))*nvectrtp(1) + cos(spos(2)) * cos(spos(3))*nvectrtp(2) &
              -sin(spos(3))*nvectrtp(3)
         normal_vect(2) =sin(spos(2)) * sin(spos(3))*nvectrtp(1) + cos(spos(2)) * sin(spos(3))*nvectrtp(2) &
             + cos(spos(3))*nvectrtp(3)
         normal_vect(3) = cos(spos(2))*nvectrtp(1) -sin(spos(2))*nvectrtp(2)
         call vsmNormVecF(normal_vect)

         if(vsmCamera(cam_num)%pers) then
           call vsmCalcColor(amb(1), dif(1), spec(1), shin, cpos, &
               ray_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
         else
           call vsmCalcColorOrtho(amb(1), dif(1), spec(1), shin, cfront, &
               ray_pos, normal_vect, vsmVisParam(vis_num)%light_used,  lumi)
         end if

         call vsmSetZeroF(color_)
         call vsmGetSColor(scal_val(nscal), nscal, color_)
         if(color_(4) <= vsmSkipAlpha) cycle

         lumi(1:3) = lumi(1:3)*color_(1:3)
         if(vsmVisParam(vis_num)%depthq) then
           dq_factor = vsmDepthQFactor(fdistance)
           lumi(1:3) = dq_factor*lumi(1:3)
         end if

         ! recalc alpha for short intval
         color_(4) = 1.0 - (1.0-color_(4))**vsmRayStride
         one_m_salpha = 1.0 - color(4)
         lumi_a = color_(4) * one_m_salpha
         color(1:3) = color(1:3) + lumi(1:3)*lumi_a ! RGB
         color(4)   = color(4)   + lumi_a           ! A
       end do

       if(color(4) >= vsmErtValue) then
         state(px(1)+(px(2)-1)*vsmImageData%width) = 0
         rpos(1:3) = ray_pos(1:3)
         exit
       end if
     end if
     ! end of volume rendering

!     prev_ci(1:3) = ci(1:3)

     ! ------------- skip some space ----------
!     skipcalc = .false.
     if(skipcalc) then
!     if(.false.) then
!       call vsmLogPrint("skipcalc in")
       skip = .false.
       ! skip a super voxel
!       call vsmyyRayJump(ray_pos, spos, norm, skip_intval, skip)
        call vsmyyRayJumpI(ray_pos, spos, norm, trn, skip)
!       skip = .false.
       if(skip) then
         rn = rn + trn
         ray_ = rintval*rn
         ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
         call vsmCoordsToAngles(ray_pos, spos)

         if(.not.vsmIsInsidelWO(spos)) then
            rn = rn - 1
            ray_ = rintval*rn
            ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
            call vsmCoordsToAngles(ray_pos, spos)
         end if

         if(.not.vsmIsInsidelWO(spos)) then
            call vsmLogPrint("by skip space, ray went out of local data")
            call vsmFindCoord(ray_pos, dx, ci)
            call vsmLogPrint("out pos", ray_pos)
            call vsmLogPrint("out ci", ci)
            call vsmError("by skip, ray went out of local data")
         end if

         if(.not.vsmIsInsidel(spos)) then
            if(z_on .and. ray_+dist2pxl >= zdis) then
              call vsmGetPixelColor4(px, color_)
              one_m_salpha = 1.0 - color(4)
              lumi_a = color_(4) * one_m_salpha
              color(1:3) = color(1:3) + color_(1:3)*lumi_a ! RGB
              color(4)   = color(4)   + lumi_a           ! A
            end if
            rpos(1:3) = ray_pos(1:3)
            insidesphere = vsmIsInsideSphere(ray_pos)
            if(.not.insidesphere) then
              state(px(1) + (px(2)-1)*vsmImageData%width) = 0
            else
              state(px(1) + (px(2)-1)*vsmImageData%width) = 2
            end if
            exit
         end if

         ! for interpolation
         call vsmFindCoord(spos, dx, ci)
         call vsmSetFrac(dx, vx)
!         print *,"new ci",ci

         do i=1, NumScal
           call vsmSetScalGridVal(i, ci, grid_val(1:8, i))
         end do

         ! calculate each scalar value at current point
         do i=1, NumScal
           scal_val(i) = vsmGetScalVal(vx, grid_val(1:8, i))
         end do

       end if
     end if
     !-----------------------------------------
     prev_ci(1:3) = ci(1:3)
     prev_ray_pos(1:3) = ray_pos(1:3)
     rn = rn + 1
     ray_ = rintval*rn

!     prev_scal_val(1:NumScal) = scal_val(1:NumScal)
     prev_vx = vx
   end do

   if(inset) then
      fdistance = first_inset + dist2pxl
      ray_ = ray_ + dist2pxl
      call vsmRaySegment(fdistance, ray_, color, px)
   end if
 end subroutine vsmyyVolumeRendererIM

 subroutine vsmyyIsoRendererIM(pixel, norm, rn_, ray2ldat, ray2dat, dist2pxl, vis_method, vis_num, cam_num, px, &
   state, rpos, cpos, cfront)
    ! inputed variables
   real(kind=vdp), dimension(3), intent(in) :: pixel, norm,cpos, cfront
   real(kind=vdp), intent(in) :: ray2ldat, dist2pxl, ray2dat
   logical, dimension(3), intent(in) :: vis_method
   integer, intent(in) :: vis_num, cam_num, rn_
   integer, dimension(2), intent(in) :: px
   integer, dimension(:), intent(inout) :: state
   real(kind=vdp), dimension(3), intent(inout) :: rpos

   ! variables
   integer :: rn
   integer :: integ_loop, i, iso_c
   integer :: nscal, NumScal
   integer, dimension(3) :: ci, prev_ci
   integer, dimension(3, vsmMaxScalar) :: nci
   real, dimension(4) :: color ! RGBA
   real, dimension(3) :: amb, dif, spec, lumi
   real :: shin, alpha
   real(kind=vdp), dimension(3) :: ray_pos, prev_ray_pos, spos
   real(kind=vdp) :: rintval, ray_, skip_intval, inpro, zdis, pray_
   real(kind=vdp), dimension(3) :: dx, nvectrtp
   real(kind=vdp), dimension(8) :: vx
   real(kind=vdp) :: judge1, judge2, rayBack

   integer :: iso_s
   integer, dimension(3) :: prev_isoci, isoci
   real(kind=vdp), dimension(3) :: isodx, iso_pos, iso_pos_f, iso_pos_b, isospos!, planec
   real(kind=vdp) :: judgeI,judgeIf,isoval,isovalf,isovalb!, judgeIb,isolength, seg1,seg2
   real(kind=vdp), dimension(8) :: isovx, iso_grid_val
   real(kind=vdp), dimension(3, 8, vsmMaxScalar) :: scalnorm

   real(kind=vdp), dimension(8) :: prev_vx
   real(kind=vdp) :: ray_ratio

   real(kind=vdp), dimension(8, vsmMaxScalar) :: grid_val
   real(kind=vdp), dimension(vsmMaxScalar) :: scal_val, prev_scal_val
   real, dimension(3) :: normal_vect
   logical :: first_in!, inset
   real(kind=vdp) :: fdistance
   !octree
   logical :: skip, skipcalc, endFlag, InsideL, insidesphere!, isoinsidel
   integer :: comm, myrank, psize

   if(.not. vis_method(2)) call vsmError("vsmyyisorendererIM is called but method(2) is false")

 !  call vsmLogPrint("renderer top state = ", state(px(1)+(px(2)-1)*vsmImageData%width))

   call vsmGetMPIVars(comm, myrank, psize)

   !initialization
   endFlag = .false.
   insidesphere = .true.

   NumScal = vsmDataInfo%NumOfScal
   prev_ci(1:3) = -1
   nci = -2 ! different from prev_ci values
   call vsmSetZeroF(color)

   prev_vx=0.0_vdp

   first_in = .false.
   zdis = vsmGetZbuf(px)

   rintval = vsmIntval%ray_interval
   rn = rn_
   if(rn >= 0) then
     ray_ = rintval*rn
   else
     ray_ = ray2ldat
   end if
   pray_ = ray2ldat
   prev_ray_pos(1:3) = pixel(1:3) + norm(1:3) * pray_

   inpro = dot_product(norm, cfront)
   ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

   dx = 0.0_vdp

   call vsmCoordsToAngles(ray_pos, spos)

   do i=1, 100
      if(vsmIsInsidelWOE(spos)) then
        exit
      end if
      ray_ = ray_ + rintval*0.01_vdp
      ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_
      call vsmCoordsToAngles(ray_pos, spos)
   end do

   call vsmCoordsToAngles(ray_pos, spos)
   if(.not.vsmIsInsidelWOE(spos)) then
       state(px(1) + (px(2)-1)*vsmImageData%width) = 2
       rpos(1:3) = ray_pos(1:3)
       return
   else
       if(first_in .eqv. .false.) then
         first_in = .true.
       end if
   end if

   prev_ray_pos(1:3) = ray_pos(1:3)

   call vsmCoordsToAngles(prev_ray_pos, spos)
   call vsmFindCoord(spos, dx, prev_ci)
   call vsmSetFrac(dx, prev_vx)
   do i=1, NumScal
     call vsmSetScalGridVal(i, prev_ci, grid_val(1:8, i))
   end do
   do i=1, NumScal
     prev_scal_val(i) = vsmGetScalVal(prev_vx, grid_val(1:8, i))
   end do

!   call vsmLogPrint("integloop starts", px)
!   call vsmLogPrint("integloop starts raypos=", ray_pos)
!   call vsmCoordsToAngles(ray_pos, spos)
!   call vsmLogPrint("ISINSIDELWO",vsmIsInsidelWO(spos))
   do integ_loop = 1, vsmMaxRayCount
!     call vsmLogPrint("integloop top", integ_loop)
     fdistance = ray_ + dist2pxl

     if(fdistance*inpro >= vsmCamera(cam_num)%far) then
        state(px(1) + (px(2) -1) *vsmImageData%width) = 0
        exit
     end if

     ray_pos(1:3) = pixel(1:3) + norm(1:3) * ray_

     call vsmCoordsToAngles(ray_pos, spos)
     InsideL = vsmIsInsidelWOE(spos)

     if(.not. InsideL) then
       endFlag = .true.
       insidesphere = vsmIsInsideSphere(ray_pos)
       rpos(1:3) = ray_pos(1:3)
       !call vsmLogPrint("prev_ray_pos ", prev_ray_pos)
       call vsmCoordsToAngles(prev_ray_pos, spos)
       rayBack = vsmyyRayDistInside(prev_ray_pos, spos, norm, vsmLocalCoord%origin, vsmLocalCoord%cend)
       ray_pos(1:3) = prev_ray_pos(1:3) + norm(1:3) * rayBack

       call vsmCoordsToAngles(ray_pos, spos)
       InsideL = vsmIsInsidelWOE(spos)
       if(.not.InsideL) then
         do i=1, 100
            rayBack = rayBack - rintval*0.01_vdp
            ray_pos(1:3) = prev_ray_pos(1:3) + norm(1:3) * rayBack
            call vsmCoordsToAngles(ray_pos, spos)
            if(vsmIsInsidelWOE(spos)) then
              !print *,"modify ray_pos", ray_pos
              exit
            end if
         end do
         call vsmCoordsToAngles(ray_pos, spos)
         InsideL = vsmIsInsidelWOE(spos)
         if(.not.InsideL) then
            ray_pos(1:3) = prev_ray_pos(1:3)
           ! call vsmError("ray cannot enter the data : raycaster")
         end if
       end if
     end if

     call vsmCoordsToAngles(ray_pos, spos)
     if(.not.vsmIsInsidelWOE(spos)) then
       rpos(1:3) = ray_pos(1:3)
       insidesphere = vsmIsInsideSphere(ray_pos)
       if(.not.insidesphere) then
         state(px(1) + (px(2)-1)*vsmImageData%width) = 0
       else
         state(px(1) + (px(2)-1)*vsmImageData%width) = 2
       end if
       exit
     end if

     if(ray_ + dist2pxl >= zdis) then
       state(px(1) + (px(2)-1)*vsmImageData%width) = 0
       rpos(1:3) = ray_pos(1:3)
       return
     end if

     ! for interpolation
     call vsmCoordsToAngles(ray_pos, spos)
     call vsmFindCoord(spos, dx, ci)
     call vsmSetFrac(dx, vx)

     if(ci(1)/=prev_ci(1) .or. ci(2)/=prev_ci(2) .or. ci(3)/=prev_ci(3) .or. integ_loop==1) then
       do i=1, NumScal
         call vsmSetScalGridVal(i, ci, grid_val(1:8, i))
       end do
       skipcalc = .true.
     else
       skipcalc = .false.
     end if

     ! calculate each scalar value at current point
     do i=1, NumScal
       scal_val(i) = vsmGetScalVal(vx, grid_val(1:8, i))
     end do

     ! isosurface
!       call vsmLogPrint("iso loop starts")
       do iso_c = 1, vsmVisParam(vis_num)%iso_num
!         call vsmLogPrint("iso loop iso_c", iso_c)
         nscal = vsmVisParam(vis_num)%isoparam(iso_c)%nscal
         judge1 = vsmIsAboveIsoL(vis_num, iso_c, scal_val(nscal))
         judge2 = vsmIsAboveIsoL(vis_num, iso_c, prev_scal_val(nscal))

         if(judge1* judge2 > 0.0_vdp) then
           cycle
         else if(judge1 == 0.0_vdp .and. judge2 == 0.0_vdp) then
           cycle
         else
           if(judge1* judge2 < 0.0_vdp) then

             ray_ratio = abs(judge1)/abs(scal_val(nscal) - prev_scal_val(nscal)) ! always positive or 0.0
            ! for safty
             ray_ratio = vsmScaleD01(ray_ratio)

             iso_pos_b(1:3) = prev_ray_pos(1:3)
!             judgeIb = judge2
             isovalb = prev_scal_val(nscal)

             iso_pos_f(1:3) = ray_pos(1:3)
             judgeIf = judge1
             isovalf = scal_val(nscal)

             iso_pos(1:3) = iso_pos_f(1:3) + (iso_pos_b(1:3) - iso_pos_f(1:3))*ray_ratio
             prev_isoci(1:3) = ci(1:3)
             iso_grid_val(1:8) = grid_val(1:8, nscal)

             do iso_s =1, 3
                call vsmCoordsToAngles(iso_pos, isospos)
                call vsmFindCoord(isospos, isodx, isoci)
                call vsmSetFrac(isodx, isovx)
                if(prev_isoci(1)/=isoci(1) .or. prev_isoci(2)/=isoci(2) .or. prev_isoci(3)/=isoci(3)) then
                    call vsmSetScalGridVal(nscal, isoci, iso_grid_val(1:8))
                end if

                isoval = vsmGetScalVal(isovx, iso_grid_val(1:8))
                judgeI = vsmIsAboveIsoL(vis_num, iso_c, isoval)

                if(judgeI*judgeIf < 0.0_vdp) then
                   iso_pos_b(1:3) = iso_pos(1:3)
!                   judgeIb = judgeI
                   isovalb = isoval
                else if(judgeI == 0.0_vdp) then
                  exit
                else
                   iso_pos_f(1:3) = iso_pos(1:3)
                   judgeIf = judgeI
                   isovalf = isoval
                end if
                ray_ratio = abs(judgeIf)/abs(isovalf - isovalb) ! always positive or 0.0
                ! for safty
                ray_ratio = vsmScaleD01(ray_ratio)

                iso_pos(1:3) = iso_pos_f(1:3) + (iso_pos_b(1:3) - iso_pos_f(1:3))*ray_ratio
                prev_isoci(1:3) = isoci(1:3)
             end do

           else if(judge1 == 0.0_vdp .and.  judge2 /= 0.0_vdp) then
             iso_pos(1:3) = ray_pos(1:3)
           else if(judge1 /= 0.0_vdp .and.  judge2 == 0.0_vdp) then
             iso_pos(1:3) = prev_ray_pos(1:3)
           end if

           call vsmCoordsToAngles(iso_pos, isospos)

           call vsmFindCoord(isospos, isodx, isoci)
           call vsmSetFrac(isodx, isovx)
           
            if(vsmIsAnotherSet(iso_pos, isospos) .and. (isoci(2) == 1 .or. isoci(3) ==1 .or. &
                      isoci(2) == vsmLocalCoord%nx(2)-1 .or. isoci(3) ==vsmLocalCoord%nx(3)-1)) then
                state(px(1)+(px(2)-1)*vsmImageData%width) = 0
                rpos(1:3) = ray_pos(1:3)
                return
            end if

           if(prev_isoci(1)/=isoci(1) .or. prev_isoci(2)/=isoci(2) .or. prev_isoci(3)/=isoci(3)) then
               call vsmSetScalGridVal(nscal, isoci, iso_grid_val(1:8))
           end if

           call vsmyyNormScalar(isoci, nci(1:3, nscal), nscal, scalnorm(1:3, 1:8, nscal))
           call vsmGetIntpolVectd(scalnorm(1:3, 1:8, nscal), isovx, nvectrtp)

           normal_vect(1) = sin(isospos(2)) * cos(isospos(3))*nvectrtp(1) + cos(isospos(2)) * cos(isospos(3))*nvectrtp(2) &
              - sin(isospos(3))*nvectrtp(3)
           normal_vect(2) =sin(isospos(2)) * sin(isospos(3))*nvectrtp(1) + cos(isospos(2)) * sin(isospos(3))*nvectrtp(2) &
             + cos(isospos(3))*nvectrtp(3)
           normal_vect(3) = cos(isospos(2))*nvectrtp(1) -sin(isospos(2))*nvectrtp(2)

           call vsmNormVecF(normal_vect)

           call vsmGetIsoColor(vis_num, iso_c, alpha, amb, dif, spec, shin)
           call vsmSetZerof(lumi)

           if(vsmCamera(cam_num)%pers) then
             call vsmCalcColorv(amb, dif, spec, shin, cpos, &
                 iso_pos, normal_vect, vsmVisParam(vis_num)%light_used, lumi)
             ray_ratio = sqrt(dot_product(cpos - iso_pos, cpos - iso_pos))
           else
             call vsmCalcColorvOrtho(amb, dif, spec, shin, cfront, &
                 iso_pos, normal_vect, vsmVisParam(vis_num)%light_used,  lumi)
              ray_ratio = dot_product(iso_pos - cpos, cfront)
           end if

           color(1:3) = lumi(1:3) ! RGB
           color(4)   = 1.0           ! A

           state(px(1)+(px(2)-1)*vsmImageData%width) = 0
           rpos(1:3) = ray_pos(1:3)
           call vsmPaintPixel3(px, color, ray_ratio)
!           call vsmLogPrint("iso color = ", color)
           return
         end if
       end do
!       end if
!     call vsmLogPrint("end of isoloop state = ", state(px(1)+(px(2)-1)*vsmImageData%width))
!     call vsmLogPrint("end of isoloop flag = ", endFlag)
     ! end of isosurface

     if(endFlag) then
       state(px(1)+(px(2)-1)*vsmImageData%width) = 2
       exit
     end if
     prev_ci(1:3) = ci(1:3)

     ! ------------- skip some space ----------
     if(skipcalc) then
!     if(.false.) then
!       call vsmLogPrint("skipcalc in")
       skip = .false.
       ! skip a super voxel
!        call vsmLogPrint("isoskip ", integ_loop)
        call vsmyyRayJump4iso(ray_pos, spos, norm, skip_intval, skip)

       if(skip) then
!         call vsmLogPrint("b4 skip raypos ", ray_pos)
         ray_ = ray_ + skip_intval
         ray_pos(1:3) = ray_pos(1:3) + skip_intval* norm(1:3)
!         call vsmLogPrint("after ray skip ", ray_pos)
         call vsmCoordsToAngles(ray_pos, spos)
         !print *,"skip",skip_intval

         if(.not.vsmIsInsidelWOE(spos)) then
            call vsmLogPrint("by skip space, ray went out of local data ", integ_loop)
            call vsmFindCoord(spos, dx, ci)
            call vsmLogPrint("out pos", ray_pos)
            call vsmLogPrint("out ci", ci)
            call vsmError("by skip, ray went out of local data")
         end if

         ! for interpolation
         call vsmFindCoord(spos, dx, ci)
         call vsmSetFrac(dx, vx)

         do i=1, NumScal
           call vsmSetScalGridVal(i, ci, grid_val(1:8, i))
         end do
         prev_ci(1:3) = ci(1:3)

         ! calculate each scalar value at current point
         do i=1, NumScal
           scal_val(i) = vsmGetScalVal(vx, grid_val(1:8, i))
         end do

         prev_scal_val(1:NumScal) = scal_val(1:NumScal)
         ray_ = ray_ + rintval

         prev_ray_pos(1:3) = ray_pos(1:3)
         prev_vx = vx
!         call vsmLogPrint("end of skip")
         cycle
       end if

     end if
     !-----------------------------------------

     prev_ray_pos(1:3) = ray_pos(1:3)
     ray_ = ray_ + rintval
     prev_scal_val(1:NumScal) = scal_val(1:NumScal)
     prev_vx = vx
!     call vsmLogPrint("end os integ loop")
   end do
 end subroutine vsmyyIsoRendererIM

 subroutine vsmScalarVisualization(vis_num, cam_num, blend, dp)
     integer, intent(in) :: vis_num, cam_num
     logical, intent(in) :: blend
     real(kind=vdp), intent(in) :: dp

     call vsmyyScalarVisualization(vis_num, cam_num, blend, dp)

 end  subroutine vsmScalarVisualization

 subroutine vsmyyScalarVisualization(vis_num, cam_num, blend, dp)
     integer, intent(in) :: vis_num, cam_num
     logical, intent(in) :: blend
     real(kind=vdp), intent(in) :: dp
     integer :: i, j
     real(kind=vdp), dimension(3) :: norm, pixel, topleft
     real(kind=vdp), dimension(3) :: cright, cup, cfront, cpos
     real(kind=vdp) :: dist2pxl, ray2ldat, ray2wdat, ray2dat
     real(kind=vdp), dimension(3) :: pos
     real(kind=vdp) :: zdis
     real, dimension(4) :: color
     integer :: step, N
     integer, dimension(2) :: px
     logical, dimension(3) :: v_method, rmv_method
     integer :: comm, myrank, psize
     

     call vsmGetMPIVars(comm, myrank, psize)

     if(vsmVisParam(vis_num)%slice_num > 0) then
         v_method(1) = .true.; rmv_method(1) = .true. 
     else
         v_method(1) = .false.; rmv_method(1) = .false.
     end if

     if(vsmVisParam(vis_num)%iso_num > 0) then
         v_method(2) = .true.; rmv_method(2) = .true.
     else
         v_method(2) = .false.; rmv_method(2) = .false.
     end if

     if(vsmVisParam(vis_num)%volren_num > 0) then
         v_method(3) = .true.; rmv_method(3) = .true.
     else
         v_method(3) = .false.; rmv_method(3) = .false.
     end if

     call vsmLogPrint("entering scalar visualization")

     select case (vsmCoordData%YinYang)
        case (0)
      ! Now Yin
           topleft(1:3) = vsmImageData%topleft(1:3)
           cup(1:3) = vsmCamera(cam_num)%up(1:3)
           cright(1:3) = vsmCamera(cam_num)%right(1:3)
           cpos(1:3) = vsmCamera(cam_num)%pos(1:3)
           cfront(1:3) = vsmCamera(cam_num)%front(1:3)

        case (1)
      ! Now Yang
!         x_yy = -x(1)
!         y_yy = x(3)
!         z_yy = x(2)
           call vsmToYangCoord(vsmImageData%topleft, topleft)
           call vsmToYangCoord(vsmCamera(cam_num)%up, cup)
           call vsmToYangCoord(vsmCamera(cam_num)%right, cright)
           call vsmToYangCoord(vsmCamera(cam_num)%pos, cpos)
           call vsmToYangCoord(vsmCamera(cam_num)%front, cfront)

        case default
           call   vsmError("coords to angles: Yin-Yang judge")
     end select

     topleft(1:3) = topleft(1:3) + vsmImageData%dp/2.0_vdp * (-cup(1:3) + cright(1:3))

! if(myrank == 2 .or. myrank == 3 .or. myrank == 4 .or. myrank == 6) print *,"scalar visualization",myrank

!$omp parallel do private(i,j,pixel, norm, dist2pxl, px, ray2ldat, ray2wdat, ray2dat) schedule(dynamic, 1)
     do j=1, vsmImageData%height
         do i=1, vsmImageData%width
             ! calculate normal vactor
             pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*cright(1:3) - (j-1)*cup(1:3))

             if(vsmCamera(cam_num)%pers) then
             ! Perspective Projection
                 norm(1:3) = pixel(1:3) - cpos(1:3)
                 dist2pxl = sqrt(dot_product(norm, norm))
                 call vsmNormVec(norm)
             else
             ! Parallel Projection
                 norm(1:3) = cfront(1:3)
                 dist2pxl = vsmCamera(cam_num)%near
             end if

             px(1) = i; px(2) = j
             ray2dat = vsmyyRayDistSphere(pixel, norm)

             if(ray2dat > vsmFarAwayd10) then
                yyraycasdat%hit(1,i,j) = vsmFarAway
                yyraycasdat%hit(2,i,j) = vsmFarAway
                yyraycasdat%hit(3,i,j) = vsmFarAway
                yyraycasdat%state(i+vsmImageData%width*(j-1)) = 0
                cycle
             end if

             yyraycasdat%hit(1,i,j) = ray2dat
             ray2wdat = vsmyyRayDist(pixel, norm, vsmLocalCoord%origin, vsmLocalCoord%cend)
             ray2ldat = vsmyyRayDist(pixel, norm, vsmLocalCoord%lorigin, vsmLocalCoord%lcend)

             if(ray2ldat > vsmFarAwayd10 .and. ray2wdat > vsmFarAwayd10) then
             ! ray does not hit the local data
                yyraycasdat%hit(2,i,j) = vsmFarAway
                yyraycasdat%hit(3,i,j) = vsmFarAway
                yyraycasdat%state(i+vsmImageData%width*(j-1)) = 0
                cycle
             else
                yyraycasdat%hit(2,i,j) = ray2wdat
                yyraycasdat%hit(3,i,j) = ray2ldat
!                yyraycasdat%state(i+vsmImageData%width*(j-1)) = 1
             end if
         end do
     end do

     if(v_method(1)) then
         call vsmLogPrint("yy slicer")
!$omp parallel do private(i,j,pixel, norm, dist2pxl, px, ray2ldat, ray2dat) schedule(dynamic, 1)
         do j=1, vsmImageData%height
             do i=1, vsmImageData%width
                 if(yyraycasdat%hit(2,i,j) > vsmFarAwayd10) cycle
                 ! calculate normal vactor
                 pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*cright(1:3) - (j-1)*cup(1:3))

                 if(vsmCamera(cam_num)%pers) then
                 ! Perspective Projection
                     norm(1:3) = pixel(1:3) - cpos(1:3)
                     dist2pxl = sqrt(dot_product(norm, norm))
                     call vsmNormVec(norm)
                 else
                 ! Parallel Projection
                     norm(1:3) = cfront(1:3)
                     dist2pxl = vsmCamera(cam_num)%near
                 end if

                px(1) = i; px(2) = j
                ray2dat = yyraycasdat%hit(1,i,j);  ray2ldat = yyraycasdat%hit(2,i,j)

                call vsmyySlicerIM(pixel, norm, ray2ldat, dist2pxl, vis_num, cam_num, px, cpos, cfront)

            end do
        end do
        rmv_method(1) = .false.
     end if


     if(rmv_method(2)) then
         call vsmLogPrint("yy isosurface")

!$omp parallel do private(i,j,pixel, norm, dist2pxl, px, ray2ldat, ray2dat, step, N) firstprivate(myrank) schedule(dynamic, 1)
         do j=1, vsmImageData%height
             do i=1, vsmImageData%width
                 if(yyraycasdat%hit(2,i,j) > vsmFarAwayd10) cycle
                 ! calculate normal vactor
                 pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*cright(1:3) - (j-1)*cup(1:3))

                 if(vsmCamera(cam_num)%pers) then
                     ! Perspective Projection
                     norm(1:3) = pixel(1:3) - cpos(1:3)
                     dist2pxl = sqrt(dot_product(norm, norm))
                     call vsmNormVec(norm)
                 else
                     ! Parallel Projection
                     norm(1:3) = cfront(1:3)
                     dist2pxl = vsmCamera(cam_num)%near
                 end if

                 px(1) = i; px(2) = j
                 ray2ldat = yyraycasdat%hit(2,i,j)
                 yyraycasdat%state(i+vsmImageData%width*(j-1)) = 1

                 do
!                     call vsmLogPrint("yyscalvis start do state = ", yyraycasdat%state(i+(j-1)*vsmImageData%width))
!                     call vsmLogPrint("yyscalvis start do px = ", px)
!                     call vsmLogPrint("yyscalvis start do ray2ldat = ", ray2ldat)
                     ray2dat = yyraycasdat%hit(1, i, j) 
                     ! only isosurface
!                     ray = ray2ldat! + 1.0D-10
                     N = -1

                          yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3) = 0.0_vdp
                         call vsmyyisoRendererIM(pixel, norm, N, ray2ldat, ray2dat, dist2pxl, rmv_method, vis_num, cam_num, &
                            px, yyraycasdat%state, yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3), &
                              cpos, cfront)

!                   call vsmLogPrint("yyscalvis state = ", yyraycasdat%state(i+(j-1)*vsmImageData%width))
!                   call vsmLogPrint("yyscalvis rpos = ", yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3))
!                   call vsmLogPrint("yyscalvis rpos dist from center = ", sqrt(dot_product(yyraycasdat%rpos(i+&
!                   (j-1)*vsmImageData%width, 1:3), yyraycasdat%rpos(i+&
!                   (j-1)*vsmImageData%width, 1:3))))
!                   call vsmCoordsToAngles(&
!                      yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3), spos)
!                   call vsmLogPrint("yyscalvis insidelwo = ", vsmIsInsidelWO(spos))

                   if(yyraycasdat%state(i+(j-1)*vsmImageData%width) /= 2) exit

                   if(vsmCamera(cam_num)%pers) then
                     ray2ldat = vsmyyRayDist(yyraycasdat%rpos(i+(j-1)*vsmImageData%width,1:3), norm,&
                         vsmLocalCoord%origin, vsmLocalCoord%cend)
 !                    call vsmLogPrint("relative ray2ldat = ", ray2ldat)

                     if(ray2ldat > vsmFarawayd10) exit
                     ray2ldat = ray2ldat + sqrt(dot_product(yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3) - pixel, &
                        yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3) - pixel))
                   else
                     ray2ldat = vsmyyRayDist(yyraycasdat%rpos(i+(j-1)*vsmImageData%width,1:3), norm, &
                         vsmLocalCoord%origin, vsmLocalCoord%cend)
                     if(ray2ldat > vsmFarawayd10) exit
                     ray2ldat = ray2ldat + dot_product(yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3) - pixel, cfront)
                   end if
  !                 call vsmLogPrint("yyscalvis end do state = ", yyraycasdat%state(i+(j-1)*vsmImageData%width))
                 end do
              end do
            end do
          end if

          if(vsmVisParam(vis_num)%depthq) then
             call vsmDepthCueingImage
          end if

     if(rmv_method(3)) then
         call vsmLogPrint("yy volume rendering")
         call vsmBinaryTreeIM(cam_num, vis_num, 0)

         call vsmClearRImage
!$omp parallel do private(i,j,pixel, norm, dist2pxl, px, ray2ldat, ray2dat, color, zdis, pos) schedule(dynamic, 1)
         do j=1, vsmImageData%height
             do i=1, vsmImageData%width
                 px(1) = i; px(2) = j
                 if(yyraycasdat%hit(1,i,j) > vsmFarAwayd10) then
                    zdis = vsmGetZbuf(px)
                    if(zdis < vsmFarAwayd10) then
                       call vsmGetPixelColor4(px, color)
                       call vsmPaintPixel4R(px, color, zdis)
                       call vsmClearPixel(px)
                    end if
                    cycle
                 end if
                 ! calculate normal vactor
                 pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*vsmCamera(cam_num)%right(1:3) &
                      - (j-1)*vsmCamera(cam_num)%up(1:3))

                 if(vsmCamera(cam_num)%pers) then
                 ! Perspective Projection
                     norm(1:3) = pixel(1:3) - vsmCamera(cam_num)%pos(1:3)
                     dist2pxl = sqrt(dot_product(norm, norm))
                     call vsmNormVec(norm)
                 else
                 ! Parallel Projection
                     norm(1:3) = vsmCamera(cam_num)%front(1:3)
                     dist2pxl = vsmCamera(cam_num)%near
                 end if

                 zdis = vsmGetZbuf(px)
                 if(zdis < vsmFarAwayd10) then
                    pos(1:3) = pixel(1:3) + norm(1:3)*(zdis - dist2pxl)
                    if(.not.vsmIsInsideSphere(pos)) then
                      call vsmGetPixelColor4(px, color)
                      call vsmPaintPixel4R(px, color, zdis)
                      call vsmClearPixel(px)
                    end if
                 end if

            end do
        end do


!$omp parallel do private(i,j,pixel,norm,dist2pxl,px,ray2ldat,ray2dat,step,N) firstprivate(myrank) schedule(dynamic, 1)
         do j=1, vsmImageData%height
             do i=1, vsmImageData%width
                 px(1) = i; px(2) = j
                 if(yyraycasdat%hit(3,i,j) > vsmFarAwayd10) then
                    call vsmClearPixel(px)
                    cycle
                 end if
                 ! calculate normal vactor
                 pixel(1:3)  = topleft(1:3) + vsmImageData%dp*((i-1)*cright(1:3) - (j-1)*cup(1:3))

                 if(vsmCamera(cam_num)%pers) then
                     ! Perspective Projection
                     norm(1:3) = pixel(1:3) - cpos(1:3)
                     dist2pxl = sqrt(dot_product(norm, norm))
                     call vsmNormVec(norm)
                 else
                     ! Parallel Projection
                     norm(1:3) = cfront(1:3)
                     dist2pxl = vsmCamera(cam_num)%near
                 end if

                 ray2ldat = yyraycasdat%hit(3,i,j)
                 yyraycasdat%state(i+vsmImageData%width*(j-1)) = 1

                 do
                   ray2dat = yyraycasdat%hit(1, i, j) 

                   step = int(ray2ldat/vsmIntval%ray_interval)
                   if(step >= 0) then
!                       ray = vsmIntval%ray_interval*step! + 1.0D-10
                       N=step
                   else
!                       ray = ray2ldat! + 1.0D-10
                       N = 0
                   end if

                          yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3) = 0.0_vdp
                         call vsmyyVolumeRendererIM(pixel, norm, N, ray2ldat, ray2dat, dist2pxl, rmv_method, vis_num, cam_num, &
                            px, yyraycasdat%state, yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3), &
                              cpos, cfront)

                   if(yyraycasdat%state(i+(j-1)*vsmImageData%width) /= 2) exit

                   if(vsmCamera(cam_num)%pers) then
                     ray2ldat = vsmyyRayDist(yyraycasdat%rpos(i+(j-1)*vsmImageData%width,1:3), norm,&
                         vsmLocalCoord%lorigin, vsmLocalCoord%lcend)

                     if(ray2ldat > vsmFarawayd10) exit
                     ray2ldat = ray2ldat + sqrt(dot_product(yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3) - pixel, &
                        yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3) - pixel))
                   else
                     ray2ldat = vsmyyRayDist(yyraycasdat%rpos(i+(j-1)*vsmImageData%width,1:3), norm, &
                         vsmLocalCoord%origin, vsmLocalCoord%cend)
                     if(ray2ldat > vsmFarawayd10) exit
                     ray2ldat = ray2ldat + dot_product(yyraycasdat%rpos(i+(j-1)*vsmImageData%width, 1:3) - pixel, cfront)
                   end if
                 end do

                 call vsmClearPixel(px)
              end do
            end do
        end if

   call vsmLogPrint("scalar visualization ends")

 end subroutine vsmyyScalarVisualization

end module vsmRayCaster

