#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include "vismo.h"

const int RDSIZE = 200;
const int THSIZE = 190/2+2;
const int PHSIZE = 280/2+2; /* (1, 2,...., 60, 61, 62), (60, 61, 62, ..., 121)  60, 61, 62 are overlapped*/
const  int mpisize = 8;
const  int maxmpisize = 64;
 
double *scalar, *scalar2;
double  *vectx, *vecty, *vectz;
double origin[3], dx[3];
int n[3];
double loccorner[3], rad[3];
int ln[3], rln[3], lx0[3];

void makeCrdData(double *origin, int *n, double *dx, double *loccorner, int *ln, int *rln, int *lx0, int myrank){
   int i;
   n[0] = RDSIZE; n[1] = (THSIZE-2)*2; n[2] = (PHSIZE-2)*2;

   dx[0] = (100.0 - 50.0)/(n[0]-1);
   dx[1] = (137.5 - 42.5)/(n[1]-1);
   dx[2] =  (140.0-(-140.0))/(n[2]-1);
   
   origin[0] = 50.0;
   origin[1] = 42.5;
   origin[2] = -140.0;


   /* local data*/
   switch(myrank){
   case 0:
     for (i=0;i<3;i++) loccorner[i] = origin[i];
     rln[0] = RDSIZE; rln[1] = THSIZE-2; rln[2] = PHSIZE-2;
     for (i=0;i<3;i++) lx0[i] = 1;
     ln[0]  = RDSIZE; ln[1] = THSIZE; ln[2]  = PHSIZE;
     break;

   case 1:
     loccorner[0] = origin[0];
     loccorner[1] = origin[1]+dx[1]*(THSIZE-4);
     loccorner[2] = origin[2];
     rln[0] = RDSIZE; rln[1] = THSIZE-1; rln[2] = PHSIZE-2;
     for (i=0;i<3;i++) lx0[i] = 1;
     lx0[1] = 2;
     ln[0]  = RDSIZE; ln[1] = THSIZE; ln[2]  = PHSIZE;
     break;

   case 2:
     loccorner[0] = origin[0];
     loccorner[1] = origin[1];
     loccorner[2] = origin[2]+dx[2]*(PHSIZE-4);

     rln[0] = RDSIZE; rln[1] = THSIZE-2; rln[2] = PHSIZE-1;
     for (i=0;i<3;i++) lx0[i] = 1;
     lx0[2] = 2;
     ln[0]  = RDSIZE; ln[1] = THSIZE; ln[2]  = PHSIZE;
     break;

   case 3:
     loccorner[0] = origin[0];
     loccorner[1] = origin[1]+dx[1]*(THSIZE-4);
     loccorner[2] = origin[2]+dx[2]*(PHSIZE-4);
     rln[0] = RDSIZE; rln[1] = THSIZE-1; rln[2] = PHSIZE-1;
     for (i=0;i<3;i++) lx0[i] = 2;
     lx0[0] = 1;
     ln[0]  = RDSIZE; ln[1] = THSIZE; ln[2]  = PHSIZE;
     break;

   case 4:
     for (i=0;i<3;i++) loccorner[i] = origin[i];
     rln[0] = RDSIZE; rln[1] = THSIZE-2; rln[2] = PHSIZE-2;
     for (i=0;i<3;i++) lx0[i] = 1;
     ln[0]  = RDSIZE; ln[1] = THSIZE; ln[2]  = PHSIZE;
     break;

   case 5:
     loccorner[0] = origin[0];
     loccorner[1] = origin[1]+dx[1]*(THSIZE-4);
     loccorner[2] = origin[2];
     rln[0] = RDSIZE; rln[1] = THSIZE-1; rln[2] = PHSIZE-2;
     for (i=0;i<3;i++) lx0[i] = 1;
     lx0[1] = 2;
     ln[0]  = RDSIZE; ln[1] = THSIZE; ln[2]  = PHSIZE;
     break;

   case 6:
     for (i=0;i<3;i++) loccorner[i] = origin[i];
     loccorner[2] = origin[2]+dx[2]*(PHSIZE-4);

     rln[0] = RDSIZE; rln[1] = THSIZE-2; rln[2] = PHSIZE-1;
     for (i=0;i<3;i++) lx0[i] = 1;
     lx0[2] = 2;
     ln[0]  = RDSIZE; ln[1] = THSIZE; ln[2]  = PHSIZE;
     break;

   case 7:
     loccorner[0] = origin[0];
     loccorner[1] = origin[1]+dx[1]*(THSIZE-4);
     loccorner[2] = origin[2]+dx[2]*(PHSIZE-4);
     rln[0] = RDSIZE; rln[1] = THSIZE-1; rln[2] = PHSIZE-1;
     for (i=0;i<3;i++) lx0[i] = 2;
     lx0[0] = 1;
     ln[0]  = RDSIZE; ln[1] = THSIZE; ln[2]  = PHSIZE;
     break;
     
     default:
         printf("myrank is strange (%d) \n", myrank);
         exit(1);
         break;
   }
}

 void LoadFldData(double *scal, double *vectx, double *vecty, double *vectz, int myrank){
   FILE *fp;
   int tmp;

   switch (myrank){
     case 0 :
       if ((fp = fopen("../yydata/vorz_n0.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(scal, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velx_n0.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectx, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/vely_n0.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vecty, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velz_n0.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectz, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);
       break;

   case 1:
       if ((fp = fopen("../yydata/vorz_n1.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(scal, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velx_n1.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectx, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/vely_n1.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vecty, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velz_n1.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectz, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);
       break;

   case 2:
       if ((fp = fopen("../yydata/vorz_n2.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(scal, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velx_n2.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectx, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/vely_n2.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vecty, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velz_n2.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectz, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);
       break;

   case 3:
       if ((fp = fopen("../yydata/vorz_n3.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(scal, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velx_n3.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectx, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/vely_n3.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vecty, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velz_n3.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectz, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);
       break;

   case 4:
       if ((fp = fopen("../yydata/vorz_e0.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(scal, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velx_e0.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectx, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/vely_e0.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vecty, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velz_e0.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectz, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);
       break;

   case 5:
       if ((fp = fopen("../yydata/vorz_e1.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(scal, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velx_e1.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectx, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/vely_e1.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vecty, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velz_e1.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectz, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);
       break;

   case 6:
       if ((fp = fopen("../yydata/vorz_e2.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(scal, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velx_e2.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectx, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/vely_e2.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vecty, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velz_e2.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectz, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);
       break;

   case 7:
       if ((fp = fopen("../yydata/vorz_e3.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(scal, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velx_e3.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectx, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/vely_e3.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vecty, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);

       if ((fp = fopen("../yydata/velz_e3.dat", "rb")) == NULL) {
        printf("file open error\n");
        exit(1);
       }
       fread(&tmp, sizeof(int), 1, fp);
       fread(vectz, sizeof(double), RDSIZE*THSIZE*PHSIZE, fp);
       fclose(fp);
       break;

    default:
        printf("rank is strange (5d)\n", myrank);
        exit(1);
       break;
   }

 }

int main(int argc, char **argv){
   int myrank, psize;
   int step;
   int gsize[3];

   double  tb, te, t0, t1, t;
   double tbmin, temax;
   scalar = (double *)malloc(sizeof(double)*RDSIZE*THSIZE*PHSIZE);
   vectx = (double *)malloc(sizeof(double)*RDSIZE*THSIZE*PHSIZE);
   vecty = (double *)malloc(sizeof(double)*RDSIZE*THSIZE*PHSIZE);
   vectz = (double *)malloc(sizeof(double)*RDSIZE*THSIZE*PHSIZE);

   MPI_Init(&argc,&argv); 
   MPI_Comm_size(MPI_COMM_WORLD,&psize); 
   MPI_Comm_rank(MPI_COMM_WORLD,&myrank); 
   
/*!$   call omp_set_num_threads(1)*/

   if(mpisize != psize) {
      printf("mpisize /= psize %d %d \n", mpisize, psize);
      MPI_Finalize();
      exit(1);
   }

   if(myrank == 0) printf("vismo with C code\n");
   if(myrank == 0) printf("making data\n");

   t0 = MPI_Wtime();
   /* --------- Making Data --------- */
   makeCrdData(origin, n, dx, loccorner, ln, rln, lx0, myrank);
   LoadFldData(scalar, vectx, vecty, vectz, myrank);
   /* ------ End of Making Data ----- */
   t1 = MPI_Wtime();

   MPI_Barrier(MPI_COMM_WORLD);
   tb = MPI_Wtime();

     if(myrank == 0) printf("vismo Init\n");
     vismo__init(MPI_COMM_WORLD, myrank, psize, "configtest.vsm");

     if(myrank == 0) printf("vismo SetData\n");
     gsize[0] = RDSIZE; gsize[1]=THSIZE; gsize[2]=PHSIZE;
     vismo__addscalarv(scalar, gsize);
     vismo__addvectorv(vectx, vecty, vectz, gsize);

     rad[0] = 50.0; rad[1] = 100.0; rad[2] = (rad[1] - rad[0])/(RDSIZE-1); /*rad1:min, rad2:max, rad3: dr*/

    if(myrank < 4) {
     if(myrank == 0) printf("vismo SetCoord\n");
       vismo__initcoords(vismo__Yin, rad); /* Uniform = .true., Rect = .false.*/
    } else {
       vismo__initcoords(vismo__Yang, rad); /* Uniform = .true., Rect = .false.*/
    }

     vismo__setunicoord(n, origin, dx, vismo__Radian);
     vismo__setlocalunicoord(ln, lx0, rln, loccorner, dx, vismo__Radian);

     if(myrank == 0) printf("vismo PrepareVis\n");
     vismo__preparevis();
   /* < computation loop>*/
     if(myrank == 0) printf("vismo Visualization\n");
     step = 0;
     vismo__visualization(step);/* ! arg is timestep (int)*/
   /* <end of computation loop>*/

     if(myrank == 0) printf("vismo Finalize\n");
     vismo__finalize();

   MPI_Barrier(MPI_COMM_WORLD);
   te = MPI_Wtime();

   if(myrank == 0){
     printf("total time: %f\n", te-tb);
     /* print *,"making data time", t1-t0*/
   }

   free(scalar); free(vectx); free(vecty); free(vectz);

   MPI_Finalize();
}

