program yytest8
   use mpi
   use vismo
 
   implicit none
   integer, parameter :: RDSIZE = 200
   integer, parameter :: THSIZE = 190/2+2
   integer, parameter :: PHSIZE = 280/2+2 ! (1, 2,...., 60, 61, 62), (60, 61, 62, ..., 121)  60, 61, 62 are overlapped
   integer, parameter :: dp = selected_real_kind(14)

   integer :: myrank, ierr, psize
   integer, parameter :: mpisize = 8
   integer, parameter :: maxmpisize = 64

   real(kind=dp), allocatable, dimension(:, :, :), target :: scalar, scalar2
   real(kind=dp), allocatable, dimension(:, :, :), target :: vectx, vecty, vectz
   real(kind=dp), dimension(3) :: origin, dx
   integer, dimension(3) :: n
   real(kind=dp), dimension(3) :: locorigin
   real(kind=dp), dimension(3) :: rad
   integer, dimension(3) :: ln, rln, lx0

   integer :: i

   logical :: vismo_on
   real(kind=dp) :: tb, te, t0, t1, t
   real(kind=dp) :: tbmin, temax


   allocate(scalar(RDSIZE, THSIZE, PHSIZE))
   allocate(vectx(RDSIZE, THSIZE, PHSIZE), vecty(RDSIZE, THSIZE, PHSIZE), vectz(RDSIZE, THSIZE, PHSIZE))
   call mpi_init(ierr)
   call mpi_comm_size(mpi_comm_world, psize, ierr)
   call mpi_comm_rank(mpi_comm_world, myrank, ierr)
   
!$   call omp_set_num_threads(1)

   vismo_on = .true.

   if(mpisize /= psize) then
      print *, "mpisize /= psize ", mpisize, psize
      call mpi_finalize(ierr)
      stop
   end if

   if(myrank == 0) print *, "making data"

   t0 = MPI_Wtime()
   ! --------- Making Data ---------
   call makeCrdData(origin, n, dx, locorigin, ln, rln, lx0, myrank)
   call LoadFldData(scalar, vectx, vecty, vectz, myrank)
   ! ------ End of Making Data -----
   t1 = MPI_Wtime()

   call mpi_barrier(mpi_comm_world, ierr)
   tb = MPI_Wtime()
   if(vismo_on) then
     if(myrank == 0) print *, "vismo Init"
     call vismo__init(mpi_comm_world, myrank, psize, "configtest.vsm")

     if(myrank == 0) print *, "vismo SetData"
     call vismo__addScalar(scalar, RDSIZE, THSIZE, PHSIZE)
     call vismo__addVector(vectx, vecty, vectz, RDSIZE, THSIZE, PHSIZE)
     rad(1) = 50.0_dp; rad(2) = 100.0_dp; rad(3) = (rad(2) - rad(1))/(RDSIZE-1) !rad1:min, rad2:max, rad3: dr

    if(myrank < 4) then
       if(myrank == 0) print *, "vismo SetCoord"
       call vismo__initCoords(vismo__Yin, rad) ! Uniform = .true., Rect = .false.
    else
       call vismo__initCoords(vismo__Yang, rad) ! Uniform = .true., Rect = .false.
    end if
!     print *,"",myrank, n,origin,dx,ln,lx0,rln,locorigin

     call vismo__setUniCoord(n, origin, dx)
     call vismo__setLocalUniCoord(ln, lx0, rln, locorigin, dx)

     if(myrank == 0) print *, "vismo PrepareVis"
     call vismo__prepareVis
   ! < computation loop>
     if(myrank == 0) print *, "vismo Visualization"
     call vismo__visualization(0) ! arg is timestep (integer)
   ! <end of computation loop>

     if(myrank == 0) print *, "vismo Finalize"
     call vismo__finalize
   end if ! vismo_on

   call mpi_barrier(mpi_comm_world, ierr)
   te = MPI_Wtime()
!   call mpi_reduce(te, temax, 1, mpi_double_precision, mpi_max, 0, mpi_comm_world, ierr)
!   call mpi_reduce(tb, tbmin, 1, mpi_double_precision, mpi_min, 0, mpi_comm_world, ierr)

   if(myrank == 0) then
     print *,"total time:", te-tb
     !print *,"making data time", t1-t0
   end if

   deallocate(scalar, vectx, vecty, vectz)!
   !deallocate(scalar)
   call mpi_finalize(ierr)

 contains


 subroutine makeCrdData(origin, n, dx, locorigin, ln, rln, lx0, myrank)
   real(kind=dp), dimension(3), intent(out) :: origin, dx, locorigin
   integer, dimension(3), intent(out) :: n, ln, rln, lx0
   integer, intent(in) :: myrank

   ! whole data
   n(1) = RDSIZE; n(2) = (THSIZE-2)*2; n(3) = (PHSIZE-2)*2

   dx(1) = (100.0_dp - 50.0_dp)/(n(1)-1)
   dx(2) = (137.5_dp - 42.5_dp)/(n(2)-1)
   dx(3) =  (140.0_dp-(-140.0_dp))/(n(3)-1)
   
   origin(1) = 50.0_dp
   origin(2) = 42.5_dp
   origin(3) = -140.0_dp


   ! local data
   select case (myrank)
   case (0)
     locorigin(1:3) = origin(1:3)
     rln(1) = RDSIZE; rln(2) = THSIZE-2; rln(3) = PHSIZE-2
     lx0(1:3) = 1
     ln(1)  = RDSIZE; ln(2) = THSIZE; ln(3)  = PHSIZE

   case (1)
     locorigin(1) = origin(1)
     locorigin(2) = origin(2)+dx(2)*(THSIZE-4)
     locorigin(3) = origin(3)
     rln(1) = RDSIZE; rln(2) = THSIZE-1; rln(3) = PHSIZE-2
     lx0(1:3) = 1
     lx0(2) = 2
     ln(1)  = RDSIZE; ln(2) = THSIZE; ln(3)  = PHSIZE

   case (2)
     locorigin(1:2) = origin(1:2)
     locorigin(3) = origin(3)+dx(3)*(PHSIZE-4)

     rln(1) = RDSIZE; rln(2) = THSIZE-2; rln(3) = PHSIZE-1
     lx0(1:2) = 1
     lx0(3) = 2
     ln(1)  = RDSIZE; ln(2) = THSIZE; ln(3)  = PHSIZE

   case (3)
     locorigin(1) = origin(1)
     locorigin(2) = origin(2)+dx(2)*(THSIZE-4)
     locorigin(3) = origin(3)+dx(3)*(PHSIZE-4)
     rln(1) = RDSIZE; rln(2) = THSIZE-1; rln(3) = PHSIZE-1
     lx0(1) = 1
     lx0(2:3) = 2
     ln(1)  = RDSIZE; ln(2) = THSIZE; ln(3)  = PHSIZE

   case (4)
     locorigin(1:3) = origin(1:3)
     rln(1) = RDSIZE; rln(2) = THSIZE-2; rln(3) = PHSIZE-2
     lx0(1:3) = 1
     ln(1)  = RDSIZE; ln(2) = THSIZE; ln(3)  = PHSIZE

   case (5)
     locorigin(1) = origin(1)
     locorigin(2) = origin(2)+dx(2)*(THSIZE-4)
     locorigin(3) = origin(3)
     rln(1) = RDSIZE; rln(2) = THSIZE-1; rln(3) = PHSIZE-2
     lx0(1:3) = 1
     lx0(2) = 2
     ln(1)  = RDSIZE; ln(2) = THSIZE; ln(3)  = PHSIZE

   case (6)
     locorigin(1:2) = origin(1:2)
     locorigin(3) = origin(3)+dx(3)*(PHSIZE-4)

     rln(1) = RDSIZE; rln(2) = THSIZE-2; rln(3) = PHSIZE-1
     lx0(1:2) = 1
     lx0(3) = 2
     ln(1)  = RDSIZE; ln(2) = THSIZE; ln(3)  = PHSIZE

   case (7)
     locorigin(1) = origin(1)
     locorigin(2) = origin(2)+dx(2)*(THSIZE-4)
     locorigin(3) = origin(3)+dx(3)*(PHSIZE-4)
     rln(1) = RDSIZE; rln(2) = THSIZE-1; rln(3) = PHSIZE-1
     lx0(1) = 1
     lx0(2:3) = 2
     ln(1)  = RDSIZE; ln(2) = THSIZE; ln(3)  = PHSIZE
   end select
   print *,"",myrank, locorigin(2), locorigin(2)+(THSIZE-1)*dx(2), locorigin(3), locorigin(3)+(PHSIZE-1)*dx(3)
 end subroutine makeCrdData

 logical function isnan(v)
   real(kind=dp) :: v
   isnan = .false.
   if( v.ne.v ) then
       isnan = .true.
   endif
 end function isnan

 subroutine LoadFldData(scal, vectx, vecty, vectz, myrank)
   real(kind=dp), dimension(RDSIZE, THSIZE, PHSIZE), intent(inout) :: scal
   real(kind=dp), dimension(RDSIZE, THSIZE, PHSIZE), intent(inout) :: vectx, vecty, vectz
   integer, intent(in) :: myrank
   integer :: unit
   integer :: i, j, k, num
   real(kind=dp) :: val

   unit = 18
   select case (myrank)
   case (0)
    open(unit,file='../yydata/vorz_n0.dat', form='unformatted', status='old')
    read(unit) scal
    close(unit)

    open(unit,file='../yydata/velx_n0.dat', form='unformatted', status='old')
    read(unit) vectx
    close(unit)

    open(unit,file='../yydata/vely_n0.dat', form='unformatted', status='old')
    read(unit) vecty
    close(unit)

    open(unit,file='../yydata/velz_n0.dat', form='unformatted', status='old')
    read(unit) vectz
    close(unit)

   case (1)
    open(unit,file='../yydata/vorz_n1.dat', form='unformatted', status='old')
    read(unit) scal
    close(unit)

    open(unit,file='../yydata/velx_n1.dat', form='unformatted', status='old')
    read(unit) vectx
    close(unit)

    open(unit,file='../yydata/vely_n1.dat', form='unformatted', status='old')
    read(unit) vecty
    close(unit)

    open(unit,file='../yydata/velz_n1.dat', form='unformatted', status='old')
    read(unit) vectz
    close(unit)

   case (2)
    open(unit,file='../yydata/vorz_n2.dat', form='unformatted', status='old')
    read(unit) scal
    close(unit)

    open(unit,file='../yydata/velx_n2.dat', form='unformatted', status='old')
    read(unit) vectx
    close(unit)

    open(unit,file='../yydata/vely_n2.dat', form='unformatted', status='old')
    read(unit) vecty
    close(unit)

    open(unit,file='../yydata/velz_n2.dat', form='unformatted', status='old')
    read(unit) vectz
    close(unit)

   case (3)
    open(unit,file='../yydata/vorz_n3.dat', form='unformatted', status='old')
    read(unit) scal
    close(unit)

    open(unit,file='../yydata/velx_n3.dat', form='unformatted', status='old')
    read(unit) vectx
    close(unit)

    open(unit,file='../yydata/vely_n3.dat', form='unformatted', status='old')
    read(unit) vecty
    close(unit)

    open(unit,file='../yydata/velz_n3.dat', form='unformatted', status='old')
    read(unit) vectz
    close(unit)

   case (4)
    open(unit,file='../yydata/vorz_e0.dat', form='unformatted', status='old')
    read(unit) scal
    close(unit)

    open(unit,file='../yydata/velx_e0.dat', form='unformatted', status='old')
    read(unit) vectx
    close(unit)

    open(unit,file='../yydata/vely_e0.dat', form='unformatted', status='old')
    read(unit) vecty
    close(unit)

    open(unit,file='../yydata/velz_e0.dat', form='unformatted', status='old')
    read(unit) vectz
    close(unit)

   case (5)
    open(unit,file='../yydata/vorz_e1.dat', form='unformatted', status='old')
    read(unit) scal
    close(unit)

    open(unit,file='../yydata/velx_e1.dat', form='unformatted', status='old')
    read(unit) vectx
    close(unit)

    open(unit,file='../yydata/vely_e1.dat', form='unformatted', status='old')
    read(unit) vecty
    close(unit)

    open(unit,file='../yydata/velz_e1.dat', form='unformatted', status='old')
    read(unit) vectz
    close(unit)

   case (6)
    open(unit,file='../yydata/vorz_e2.dat', form='unformatted', status='old')
    read(unit) scal
    close(unit)

    open(unit,file='../yydata/velx_e2.dat', form='unformatted', status='old')
    read(unit) vectx
    close(unit)

    open(unit,file='../yydata/vely_e2.dat', form='unformatted', status='old')
    read(unit) vecty
    close(unit)

    open(unit,file='../yydata/velz_e2.dat', form='unformatted', status='old')
    read(unit) vectz
    close(unit)

   case (7)
    open(unit,file='../yydata/vorz_e3.dat', form='unformatted', status='old')
    read(unit) scal
    close(unit)

    open(unit,file='../yydata/velx_e3.dat', form='unformatted', status='old')
    read(unit) vectx
    close(unit)

    open(unit,file='../yydata/vely_e3.dat', form='unformatted', status='old')
    read(unit) vecty
    close(unit)

    open(unit,file='../yydata/velz_e3.dat', form='unformatted', status='old')
    read(unit) vectz
    close(unit)

   end select

 end subroutine LoadFldData

end program yytest8

