module vismo
#ifndef CLANG
 use mpi
#endif
 use vsmConst
 use vsmFunction
 use vsmParam
 use vsmParamReader
 use vsmParamBroadcast
 use vsmData
 use vsmCoordinates
 use vsmOctree
 use vsmOctree4Iso
 use vsmRayCaster
 use vsmArrows
 use vsmStreamLine
 use vsmFrame
 use vsmBSwap
 use vsmBTree
 use vsmColorMap
 use vsmCameraData
 use vsmLightData
 use vsmDepthQ
 use vsmImage
!$ use omp_lib !changed
 implicit none
 private

 integer, parameter :: FNSIZE=256
 character(FNSIZE), save :: vsmFileName

 integer, parameter, public :: vismo__Yin = 0
 integer, parameter, public :: vismo__Yang = 1

interface vsmAddScalar
  module procedure vsmAddScalarS, vsmAddScalarV
end interface vsmAddScalar

interface vsmAddVector
  module procedure vsmAddVectorS, vsmAddVectorV
end interface vsmAddVector

interface vismo__addScalar
  module procedure vsmAddScalarS, vsmAddScalarV
end interface vismo__addScalar

interface vismo__addVector
  module procedure vsmAddVectorS, vsmAddVectorV
end interface vismo__addVector

 public vsmInit, vsmAddScalar, vsmAddVector, vsmInitCoords, vsmSetUniCoord
 public vsmSetLocalUniCoord, vsmPrepareVis
 public vsmVisualization, vsmFinalize
 
 public vismo__init, vismo__addScalar, vismo__addVector, vismo__initCoords
 public vismo__setUniCoordron, vismo__setLocalUniCoordron
 public vismo__setUniCoordroff, vismo__setLocalUniCoordroff
 public vismo__prepareVis, vismo__visualization, vismo__finalize
 public vismo__init4c, vismo__initCoords4c, vismo__visualization4c, vismo__visualization4cv
 public vismo__addScalarS, vismo__addScalarV, vismo__addVectorS, vismo__addVectorV

 contains

 ! Initialization of vismo
 subroutine vismo__init4c(rank, psize, vsmfile)
   integer, intent(in) :: rank, psize
   integer :: comm, ierr
   character(*), intent(in) :: vsmfile
   comm = 0
   call vismo__init(comm, rank, psize, vsmfile)
 end subroutine vismo__init4c

 subroutine vsmInit(comm, rank, psize, vsmfile)
   integer, intent(in) :: comm, rank, psize
   character(*), intent(in) :: vsmfile
   call vismo__init(comm, rank, psize, vsmfile)
 end subroutine vsmInit

 subroutine vismo__init(comm, rank, psize, vsmfile)
   integer, intent(in) :: comm, rank, psize
   character(*), intent(in) :: vsmfile
   character(4) :: ascrank
   integer :: onum
   logical :: fexist

   onum = 1
   write(ascrank, '(I4.4)') rank 

   vsmFileName = vsmfile

   call vsmLogStart('vsmlog'//ascrank//'.txt')
   call vsmLogPrint("vsmInit is called")

   ! initialization of parameters
!$ onum = omp_get_max_threads() !changed
   call vsmInitVisParam(comm, rank, psize, onum)
   call vsmLogPrint("MPI Processes:", psize)
   call vsmLogPrint("OpenMP threads:", onum) !changed
   ! initialization of variables
   call vsmInitColmap(vsmMaxScalar, vsmMaxVector)
   call vsmInitCamera
   call vsmInitBSwap(rank, psize)
   call vsmInitBTree(rank, psize)
   call vsmInitData
   call vsmInitOctree
   call vsmInitOctree4Iso
   call vsmInitLight
   call vsmInitBGColor
   call vsmInitFrame

   ! reading and broadcasting parameters
   call vsmInitParamReader
   if (rank == 0) then

     call vsmInvestigateFiles(trim(vsmFileName), fexist)
     if(.not.fexist) then
       call vsmError("Could not find all the config files. be careful about INCLUDE")
     end if

     call vsmReadParam(trim(vsmFileName), .true.)
   end if

   call vsmParamBcast

   call vsmLogPrint("vsmInit ends")
 end subroutine vismo__init

 ! ---- Setting data
 subroutine vsmAddScalarS(scalar, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOAT
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: scalar
#else
   real, dimension(nx,ny,nz), target, intent(in) :: scalar
#endif
   call vsmLogPrint("vsmAddScalarS is called")
   call vsmAddScalData(scalar, nx, ny, nz)
   call vsmLogPrint("vsmAddScalarS ends")
 end subroutine vsmAddScalarS

 subroutine vismo__addScalarS(scalar, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOAT
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: scalar
#else
   real, dimension(nx,ny,nz), target, intent(in) :: scalar
#endif
   call vsmLogPrint("vsmAddScalarS is called")
   call vsmAddScalData(scalar, nx, ny, nz)
   call vsmLogPrint("vsmAddScalarS ends")
 end subroutine vismo__addScalarS

 subroutine vsmAddScalarV(scalar, nx)
   integer, dimension(3), intent(in) :: nx
#ifndef FLOAT
   real(kind=vdp), dimension(nx(1),nx(2),nx(3)), target, intent(in) :: scalar
#else
   real, dimension(nx(1),nx(2),nx(3)), target, intent(in) :: scalar
#endif
   call vsmLogPrint("vsmAddScalarV is called")
   call vsmAddScalData(scalar, nx(1), nx(2), nx(3))
   call vsmLogPrint("vsmAddScalarV ends")
 end subroutine vsmAddScalarV

 subroutine vismo__addScalarV(scalar, nx)
   integer, dimension(3), intent(in) :: nx
#ifndef FLOAT
   real(kind=vdp), dimension(nx(1),nx(2),nx(3)), target, intent(in) :: scalar
#else
   real, dimension(nx(1),nx(2),nx(3)), target, intent(in) :: scalar
#endif
   call vsmLogPrint("vsmAddScalarV is called")
   call vsmAddScalData(scalar, nx(1), nx(2), nx(3))
   call vsmLogPrint("vsmAddScalarV ends")
 end subroutine vismo__addScalarV

 subroutine vsmAddVectorS(vect_x, vect_y, vect_z, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOAT
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#else
   real, dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#endif
   call vsmLogPrint("vsmAddVectorS is called")
   call vsmAddVectData(vect_x, vect_y, vect_z, nx, ny, nz)
   call vsmLogPrint("vsmAddVectorS ends")
 end subroutine vsmAddVectorS

 subroutine vismo__addVectorS(vect_x, vect_y, vect_z, nx, ny, nz)
   integer, intent(in) :: nx, ny, nz
#ifndef FLOAT
   real(kind=vdp), dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#else
   real, dimension(nx,ny,nz), target, intent(in) :: vect_x, vect_y, vect_z
#endif
   call vsmLogPrint("vsmAddVectorS is called")
   call vsmAddVectData(vect_x, vect_y, vect_z, nx, ny, nz)
   call vsmLogPrint("vsmAddVectorS ends")
 end subroutine vismo__addVectorS

 subroutine vsmAddVectorV(vect_x, vect_y, vect_z, nx)
   integer, dimension(3), intent(in) :: nx
#ifndef FLOAT
   real(kind=vdp), dimension(nx(1),nx(2),nx(3)), target, intent(in) :: vect_x, vect_y, vect_z
#else
   real, dimension(nx(1),nx(2),nx(3)), target, intent(in) :: vect_x, vect_y, vect_z
#endif
   call vsmLogPrint("vsmAddVectorV is called")
   call vsmAddVectData(vect_x, vect_y, vect_z, nx(1), nx(2), nx(3))
   call vsmLogPrint("vsmAddVectorV ends")
 end subroutine vsmAddVectorV

 subroutine vismo__addVectorV(vect_x, vect_y, vect_z, nx)
   integer, dimension(3), intent(in) :: nx
#ifndef FLOAT
   real(kind=vdp), dimension(nx(1),nx(2),nx(3)), target, intent(in) :: vect_x, vect_y, vect_z
#else
   real, dimension(nx(1),nx(2),nx(3)), target, intent(in) :: vect_x, vect_y, vect_z
#endif
   call vsmLogPrint("vsmAddVectorV is called")
   call vsmAddVectData(vect_x, vect_y, vect_z, nx(1), nx(2), nx(3))
   call vsmLogPrint("vsmAddVectorV ends")
 end subroutine vismo__addVectorV
 ! ---- End of data

 ! ---- Setting coordinates ----
 subroutine vismo__initCoords4c(yyz, rad)
  integer, intent(in) :: yyz
  real(kind=vdp), dimension(3), intent(in) :: rad !Min, Max
  if (yyz == 0) then
    call vismo__initCoords(vismo__Yin, rad)
  else if (yyz == 1) then
    call vismo__initCoords(vismo__Yang, rad)
  end if
 end subroutine vismo__initCoords4c

 subroutine vismo__initCoords(yyz, rad, YYZM)
!  logical, intent(in) :: uniform
  integer, intent(in) :: yyz
  real(kind=vdp), dimension(3), intent(in) :: rad !Min, Max
  logical, intent(in), optional :: YYZM
  logical :: YYZMode
  logical :: uniform

  uniform = .true.
  call vsmLogPrint("vsmSetInitCoords is called")
   if(present(YYZM)) then
      YYZMode = YYZM
   else
      YYZMode = .false.
   end if

  call vsmLogPrint("YYZ = ", yyz)
  call vsmInitCoord(uniform, yyz, rad, YYZMode)
  call vsmLogPrint("vsmSetInitCoords ends")
 end subroutine vismo__initCoords

 subroutine vsmInitCoords(yyz, rad, YYZM)
!  logical, intent(in) :: uniform
  integer, intent(in) :: yyz
  real(kind=vdp), dimension(3), intent(in) :: rad !Min, Max
  logical, intent(in), optional :: YYZM

  call vismo__initCoords(yyz, rad, YYZM)
 end subroutine vsmInitCoords

 ! << Uniform>>
 subroutine vismo__setUniCoordron(nx, origin, dx)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  call vismo__setUniCoord(nx, origin, dx, .true.)
 end subroutine vismo__setUniCoordron

 subroutine vismo__setUniCoordroff(nx, origin, dx)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  call vismo__setUniCoord(nx, origin, dx, .false.)
 end subroutine vismo__setUniCoordroff

 subroutine vismo__setUniCoord(nx, origin, dx, radon)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  logical, intent(in), optional :: radon
  call vsmSetUniCoord(nx, origin, dx, radon)
 end subroutine vismo__setUniCoord

 subroutine vsmSetUniCoord(nx, origin, dx, radon)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  logical, intent(in), optional :: radon
  call vsmLogPrint("vsmSetUniCoord is called")
  if(.not.present(radon)) then
    call vsmSetUniform(nx, origin, dx)
  else
    call vsmSetUniform(nx, origin, dx, radon)
  end if
  call vsmLogPrint("vsmSetUniCoord ends")
 end subroutine vsmSetUniCoord

 subroutine vsmSetLocalUniCoord(nx, lx0, lnx, origin, dx, radon)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  logical, optional :: radon
  call vismo__setLocalUniCoord(nx, lx0, lnx, origin, dx, radon)
 end subroutine vsmSetLocalUniCoord

 subroutine vismo__setLocalUniCoordron(nx, lx0, lnx, origin, dx)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  call vismo__SetLocalUniCoord(nx, lx0, lnx, origin, dx, .true.)
 end subroutine vismo__setLocalUniCoordron

 subroutine vismo__setLocalUniCoordroff(nx, lx0, lnx, origin, dx)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  call vismo__SetLocalUniCoord(nx, lx0, lnx, origin, dx, .false.)
 end subroutine vismo__setLocalUniCoordroff

 subroutine vismo__setLocalUniCoord(nx, lx0, lnx, origin, dx, radon)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(3), intent(in) :: origin
  real(kind=vdp), dimension(3), intent(in) :: dx
  logical, optional :: radon
  call vsmLogPrint("vsmSetLocalUniCoord is called")
  if(.not.present(radon)) then
    call vsmSetLUniform(nx, lx0, lnx, origin, dx)
  else
    call vsmSetLUniform(nx, lx0, lnx, origin, dx, radon)
  end if

  call vsmLogPrint("vsmSetLocalUniCoord ends")
 end subroutine vismo__setLocalUniCoord

 ! << Rect >>
 subroutine vsmSetRectCoord(nx, coordx, coordy, coordz)
  integer, dimension(3), intent(in) :: nx
  real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
  real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
  real(kind=vdp), dimension(nx(3)), intent(in) :: coordz

  call vsmLogPrint("vsmSetRectCoord is called")
  call vsmSetRect(nx, coordx, coordy, coordz)
  call vsmLogPrint("vsmSetRectCoord ends")
 end subroutine vsmSetRectCoord

 subroutine vsmSetLocalRectCoord(nx, lx0, lnx, coordx, coordy, coordz)
  integer, dimension(3), intent(in) :: nx, lx0, lnx
  real(kind=vdp), dimension(nx(1)), intent(in) :: coordx
  real(kind=vdp), dimension(nx(2)), intent(in) :: coordy
  real(kind=vdp), dimension(nx(3)), intent(in) :: coordz

  call vsmLogPrint("vsmSetLocalRectCoord is called")
  call vsmSetLRect(nx, lx0, lnx, coordx, coordy, coordz)
  call vsmLogPrint("vsmSetLocalRectCoord ends")
 end subroutine vsmSetLocalRectCoord

 ! --- End of coordinates

 ! Preparing visualization
 ! should be called after both coord and data settings
 subroutine vsmPrepareVis
   call vismo__prepareVis
 end subroutine vsmPrepareVis

 subroutine vismo__prepareVis
!   logical :: root
   integer :: i, calcdpflag
   call vsmLogPrint("vsmPrepareVis is called")
   call vsmInitImage ! 2014/12/2
   call vsmOctPrepare
   call vsmOctDivide(0)
   call vsmOctPrepare4Iso
   call vsmOctDivide4Iso(0)
!   root = .true.
!   call vsmOctDivide(vsmOctant(0), root)

   call vsmArrowsPrepare
   call vsmCheckParam
   call vsmPrepareCoord
   call vsmPrepareLight(vsmCoordData%YinYang)
   calcdpflag = 0
   do i=1, vsmNumVisKind
     if(vsmVisParam(i)%sline_num > 0 .or. vsmVisParam(i)%arrow_num > 0) then
       calcdpflag=calcdpflag+1
     end if 
   end do
   if(calcdpflag > 0) then
     do i=1, vsmCameraNum
       call vsmCalcDP(vsmCamera(i)%dp, i)
     end do
   end if
   call vsmLogPrint("vsmPrepareVis end")
 end subroutine vismo__prepareVis

 ! Visualization
 subroutine vsmVectVis(vnum, cnum, dp)
  integer, intent(in) :: vnum, cnum
  real(kind=vdp), intent(in) :: dp

  call vsmLogPrint("------ vsmVectVis is called")

      call vsmDrawStreamLinesIM(vnum, cnum, dp)
      call vsmDrawArrowsIM(vnum, cnum)

  call vsmLogPrint("------ vsmVectVis ends")
 end subroutine vsmVectVis

 subroutine vsmScalVis(vnum, cnum, bon, dp)
  integer, intent(in) :: vnum, cnum
  logical, intent(in) :: bon
  real(kind=vdp), intent(in) :: dp
  call vsmLogPrint("------ vsmScalVis is called")
  call vsmInityyRaycast(vsmImageData%width, vsmImageData%height)

  call vsmScalarVisualization(vnum, cnum, bon, dp)

  call vsmFinalizeyyRaycast
  call vsmLogPrint("------ vsmScalVis ends")
 end subroutine vsmScalVis

 subroutine vsmIsBlendOn(visnum, bon)
   integer, intent(in)  :: visnum
   logical, intent(out) :: bon
   real :: alpha
   integer :: i

   bon = .false.

   ! volume rendering
   if(vsmVisParam(visnum)%volren_num > 0) then
     bon = .true.
   end if

   ! isosurface
   do i=1, vsmVisParam(visnum)%iso_num
     alpha = 1.0
     call vsmGetIsoAlpha(visnum, i, alpha)
     if(alpha < vsmErtValue) then
       bon = .true.
       vsmVisParam(visnum)%iso_solid = .false.
       exit
     end if
   end do

   ! slice
   do i=1, vsmVisParam(visnum)%slice_num
     alpha = 1.0
     call vsmGetSliceAlpha(visnum, i, alpha)
     if(alpha < vsmErtValue) then
       bon = .true.
       vsmVisParam(visnum)%slice_solid = .false.
       exit
     end if
   end do
 end subroutine vsmIsBlendOn

 subroutine vismo__visualization4c(time)
   integer, intent(in) :: time
   call vismo__visualization(time)
 end subroutine vismo__visualization4c

 subroutine vismo__visualization4cv(time, vex)
   integer, intent(in) :: time
   integer, dimension(vsmMaxVisKind), intent(in) :: vex
   logical, dimension(vsmMaxVisKind) :: visex
   integer :: i
   visex = .false.
   do i=1, vsmNumViskind
     if(vex(i) == 1) then
       visex(i) = .true.
     end if
   end do
   call vismo__visualization(time, visex)
 end subroutine vismo__visualization4cv

 subroutine vsmVisualization(time, visex)
   integer, intent(in) :: time
   logical, dimension(vsmMaxVisKind), optional, intent(in) :: visex
   call vismo__visualization(time, visex)
 end  subroutine vsmVisualization

 subroutine vismo__visualization(time, visex)
   integer, intent(in) :: time
   character(5) :: viskind, camnum 
   character(4) :: rank 
   character(9) :: timestep
   integer :: i, j
   integer :: comm, myrank, psize!, ierr
   integer :: UnNum, iostatus

   logical :: blend, nest_on
   logical, save :: first = .true.
   logical, dimension(vsmMaxVisKind), optional, intent(in) :: visex
   logical, dimension(vsmMaxVisKind) :: visexcute
   integer :: onum

   real(kind=vdp) :: dp
   !print *,"pcmode", vsmPcMode
   call vsmLogPrint("vsmVisualization is called")
   call vsmGetMPIVars(comm, myrank, psize)
   nest_on = .true.
   if(present(visex)) then
      visexcute(1:vsmMaxVisKind) = visex(1:vsmMaxVisKind)
   else
     visexcute(1:vsmMaxVisKind) = .true.
   end if
    onum = 1
!$ onum = omp_get_max_threads() !changed

   if(first) then
     first = .false.
     if (myrank == 0) then
        UnNum = vsmSearchUnitNum("DataInfo")
        open(UnNum, file=trim(vsmDataInfoFile), action='write', &
            iostat=iostatus)

        if(iostatus > 0) then
           call vsmError("cant open data info file")
        end if
        
        write(UnNum, '(1A)', advance='no') 'COORDMIN '
        do i=1, size(vsmCoord%origin)
           write(UnNum, '(1F16.10,1X)', advance='no') vsmCoord%origin(i)
        end do
        write(UnNum, *) ''
        
        write(UnNum, '(1A)', advance='no') 'COORDMAX '
        do i=1, size(vsmCoord%cend)
           write(UnNum, '(1F16.10,1X)', advance='no') vsmCoord%cend(i)
        end do
        write(UnNum, *) ''

        write(UnNum, '(1A,1F16.10)') "RAYINTV ", vsmIntval%ray_interval
        write(UnNum, '(1A,1F16.10)') "SLINEDH ", vsmIntval%dh

        write(UnNum, '(1A,1I5)') "SCALNUM ", vsmDataInfo%NumOfScal
        write(UnNum, '(1A,1I5)') "VECTNUM ", vsmDataInfo%NumOfVect

        close(UnNum)

      end if

   end if
   ! updating octree for scalar visualization
   call vsmint2asc9(time, timestep)

   if(vsmDbg) then
     call vsmLogPrint("viskindnum:", vsmNumVisKind)
     call vsmLogPrint("cameranum:", vsmCameraNum)
   end if

   do j=1, vsmNumViskind
     if(.not.visexcute(j)) cycle
     blend = .false.

     call vsmLogPrint("viskindloop:", j)

     call vsmIsBlendOn(j, blend)

     call vsmint2asc5(j, viskind)
        call vsmLogPrint("octree processing: volume rendering")
        call vsmSetOctComDat(j)
        call vsmUpdateOctree(0)

        call vsmLogPrint("octree processing: isosurface")
     call vsmSetOctComDat4Iso(j)
     call vsmUpdateOctree4Iso(0)

!     call vsmUpdateOctree(vsmOctant(0))

     do i=1, vsmCameraNum
!       if(vsmVisParam(j)%volren_num > 0) then
!         call vsmLogPrint("clear raysegment")

!         call vsmClearRaySegment
!         call vsmClearRaySegmentR

!       end if
       call vsmint2asc5(i, camnum)
       call vsmLogPrint("cameraloop:", i)
       ! initialization of camera and image
       call vsmInitScreen(i)

!       call vsmCalcDP(dp, i)
       if(.not.vsmCamera(i)%pers) then
         dp = vsmImageData%dp
       else
         dp = vsmCamera(i)%dp
       end if

       !print *, "b4 vector", myrank
       if(vsmVisParam(j)%sline_num > 0 .or. vsmVisParam(j)%arrow_num > 0) then

         call vsmClearImage4Solid
#ifdef OMPNEST
!$         nest = omp_get_nested()
!$         if(.not.nest) call omp_set_nested(nest_on)
#endif
         ! vector visualization
         call vsmVectVis(j, i, dp)
         ! merge local buffer
#ifdef OMPNEST
!$         call omp_set_nested(nest)
#endif
         call vsmMergeBuf4Solid

!          print *,"vect"
       else
         call vsmClearImage
       end if

       if(vsmVisParam(j)%frameparam%insphere .and. myrank == 0) call vsmDrawInnerSphere(i, j)

       if(vsmVisParam(j)%frameparam%inspframe) call vsmDrawSphereFramePara(i,.false.,  0, myrank, psize)
       if(vsmVisParam(j)%frameparam%outspframe) call vsmDrawSphereFramePara(i,.false.,  1, myrank, psize)
       if(vsmVisParam(j)%frameparam%inyyframe) call vsmDrawYinYangFrameAPara(i, .false., 0, 2, myrank, psize)
       if(vsmVisParam(j)%frameparam%outyyframe) call vsmDrawYinYangFrameAPara(i, .false., 1, 2, myrank, psize)
       if(vsmVisParam(j)%frameparam%egyyframe)  call vsmDrawYinYangFrameBPara(i, .false., 2, myrank, psize)

      if(blend .or. (vsmVisParam(j)%slice_num == 0 .and. vsmVisParam(j)%iso_num == 0)) then
        if(vsmBSwapParam%bswap) then
         call vsmBinarySwapIM(.true., i, j, time)
        else
         call vsmBinarySwap4SolidIM(i, j, time)
        end if
      end if

      if(vsmVisParam(j)%slice_num == 0 .and. vsmVisParam(j)%iso_num == 0) then
!        print *,"vect ptcl depth", .not. vsmVisParam(j)%slice_solid,  .not. vsmVisParam(j)%iso_solid
        if(vsmVisParam(j)%depthq) call vsmDepthCueingImage
      end if

      if(vsmVisParam(j)%slice_num > 0 .or. vsmVisParam(j)%iso_num > 0 .or. vsmVisParam(j)%volren_num > 0) then
       if(vsmVisParam(j)%volren_num > 0) then
         call vsmLogPrint("clear raysegment")

         call vsmClearRaySegment
         call vsmClearRaySegmentR

       end if

         ! scalar visualization
        call vsmScalVis(j, i, blend, dp)
        !print *, "scal vis has ended", myrank

        if(blend) then
           if(vsmBSwapParam%bswap) then
             call vsmBinarySwapIMRS(i, j, time)
           else
             call vsmError("num of mpi processes must be a power of two when volume rendering is used.")
           end if
           call vsmAlphaBlending(1, vsmImageData%rimgsize)
        end if
      end if

      if(.not.blend) then
!         call vsmMergeImg1Img2
           if(vsmDbg) then
             call vsmint2asc4(myrank, rank)
             call vsmWriteImage(vsmImageFile//'r'//rank//'t'//timestep//'am.ppm')
           end if

            call vsmBinaryTreeIM(i, j, time)

      end if

       ! saving image
      if(myrank == 0) then
        !call vsmDrawFrame(i, blend)
        call vsmBGBlending
        call vsmWriteImage(vsmImageFile//'v'//viskind//'c'//camnum//'t'//timestep)
      end if
     end do
   end do
   call vsmLogPrint("vsmVisualization ends")
 end subroutine vismo__visualization

 ! Finalizing vismo
 subroutine vsmFinalize
   call vismo__finalize
 end  subroutine vsmFinalize
 
 subroutine vismo__finalize
   call vsmLogPrint("vsmFinalize is called")
   call vsmFinalizeImage
   call vsmFinalizeBSwap
   call vsmFinalizeBTree
   call vsmFinalizeColmap
   call vsmFinalizeCoord
   call vsmFinalizeOctree
   call vsmFinalizeOctree4Iso
   call vsmLogPrint("vsmFinalize ends")
   call vsmLogEnd
 end subroutine vismo__finalize

end module vismo

